(*
   Include File for IndexNotation,...
 *)

(*************************** for Indices ****************************)
(*** Local Constants ***)
normalIndices       = {"a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o"}
dummyIndices        = {"p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"}
lenDummy            = Length[dummyIndices]
dummyPreStr         = "$"

greekIndices        = {"\[Alpha]", "\[Beta]", "\[Gamma]", "\[Delta]", "\[Mu]", "\[Nu]", "\[Rho]",
                       "\[Sigma]", "\[Epsilon]", "\[Kappa]", "\[Lambda]", "\[Tau]"}
greekDummyIndices   = {"\[Chi]", "\[Psi]", "\[Phi]", "\[Xi]", "\[Eta]", "\[Theta]", "\[Omega]",
                       "\[Pi]", "\[Zeta]", "\[Iota]"}
greekLenDummy       = Length[greekDummyIndices]
greekDummyPreStr    = "@"

capitalIndices      = {"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O"}
capitalDummyIndices = {"P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z"}
capitalLenDummy     = Length[capitalDummyIndices]
capitalDummyPreStr  = "#"

barDummyPreStr      = "!"
dotDummyPreStr      = "~"
hatDummyPreStr      = "%"
oneIndices          = "0"    (* for one-dimensional space *)

allIndices          = Join[normalIndices, greekIndices, capitalIndices, {oneIndices}]
allIndices          = Join[Map[StringJoin["l", #]&, allIndices], Map[StringJoin["u", #]&, allIndices]]

iIndexClassTable    = {ComponentIndexQ, NormalIndexQ, GreekIndexQ, CapitalIndexQ,
                       BarIndexQ, DotIndexQ, HatIndexQ, OneIndexQ}

(*** Local Variables ***)
pointerDummy        = 1  (* for getting an unique dummy index *)
greekPointerDummy   = 1
capitalPointerDummy = 1
barPointerDummy     = 1
dotPointerDummy     = 1
hatPointerDummy     = 1

(*** Exported Functions ***)
(* BarIndexQ tag *)
BarIndexQ[a_]  := indexClassQ[a] === BarIndexQ
BarIndexQ[___] := False

(* CapitalIndexQ tag *)
CapitalIndexQ[a_]  := indexClassQ[a] === CapitalIndexQ
CapitalIndexQ[___] := False

(* ComponentIndexQ tag *)
ComponentIndexQ[0]         := False
ComponentIndexQ[a_Integer] := True
ComponentIndexQ[___]       := False

(* DnIndexQ tag *)
DnIndexQ[a_Integer, testQ_:(True&)] := testQ[a] && a < 0
DnIndexQ[a_Symbol,  testQ_:(True&)] := testQ[a] && DnIndexQ[ToString[a]]
DnIndexQ[a_String]                  := StringLength[a] > 1 && StringMatchQ[a, "l*"]
DnIndexQ[a_String,  testQ_]         := testQ[a] && DnIndexQ[a]
DnIndexQ[___]                       := False

(* DotIndexQ tag *)
DotIndexQ[a_]  := indexClassQ[a] === DotIndexQ
DotIndexQ[___] := False

(* Dummy tag *)
(* return unique dummy down-index for each class *)
Dummy[BarIndexQ]     := "l" <> barDummyPreStr     <> ToString[barPointerDummy++]
Dummy[CapitalIndexQ] := "l" <> capitalDummyPreStr <> ToString[capitalPointerDummy++]
Dummy[DotIndexQ]     := "l" <> dotDummyPreStr     <> ToString[dotPointerDummy++]
Dummy[GreekIndexQ]   := "l" <> greekDummyPreStr   <> ToString[greekPointerDummy++]
Dummy[HatIndexQ]     := "l" <> hatDummyPreStr     <> ToString[hatPointerDummy++]
Dummy[NormalIndexQ]  := "l" <> dummyPreStr        <> ToString[pointerDummy++]
Dummy[OneIndexQ]     := "l" <> oneIndices
Dummy[]              := Dummy[mainClassQL[][[1]]]

(* DummyIndexQ tag *)
DummyIndexQ[a_String] := True
DummyIndexQ[___]      := False

(* DummyPair tag *)
(* return unique dummy pair: {dn, up} for each class *)
DummyPair[BarIndexQ]     := {"l" <> barDummyPreStr     <> ToString[barPointerDummy],
                             "u" <> barDummyPreStr     <> ToString[barPointerDummy++]}
DummyPair[CapitalIndexQ] := {"l" <> capitalDummyPreStr <> ToString[capitalPointerDummy],
                             "u" <> capitalDummyPreStr <> ToString[capitalPointerDummy++]}
DummyPair[DotIndexQ]     := {"l" <> dotDummyPreStr     <> ToString[dotPointerDummy],
                             "u" <> dotDummyPreStr     <> ToString[dotPointerDummy++]}
DummyPair[GreekIndexQ]   := {"l" <> greekDummyPreStr   <> ToString[greekPointerDummy],
                             "u" <> greekDummyPreStr   <> ToString[greekPointerDummy++]}
DummyPair[HatIndexQ]     := {"l" <> hatDummyPreStr     <> ToString[hatPointerDummy],
                             "u" <> hatDummyPreStr     <> ToString[hatPointerDummy++]}
DummyPair[NormalIndexQ]  := {"l" <> dummyPreStr        <> ToString[pointerDummy],
                             "u" <> dummyPreStr        <> ToString[pointerDummy++]}
DummyPair[OneIndexQ]     := {"l" <> oneIndices,
                             "u" <> oneIndices}
DummyPair[]              := DummyPair[mainClassQL[][[1]]]

(* FlipIndex tag *)
(* dn <-> up *)
FlipIndex[a_Integer,         testQ_:(True&)] := -a /; testQ[a]
FlipIndex[a_Symbol?DnIndexQ, testQ_:(True&)] := ToExpression[ "u" <> StringDrop[ToString[a], 1] ] /; testQ[a]
FlipIndex[a_Symbol?UpIndexQ, testQ_:(True&)] := ToExpression[ "l" <> StringDrop[ToString[a], 1] ] /; testQ[a]
FlipIndex[a_String?DnIndexQ, testQ_:(True&)] := "u" <> StringDrop[a, 1] /; testQ[a]
FlipIndex[a_String?UpIndexQ, testQ_:(True&)] := "l" <> StringDrop[a, 1] /; testQ[a]
FlipIndex[a_,                ___]            := a

(* GetPairs tag *)
(* return contracted pairs: {{dn,up},..} from indexL satisfying testQ *)
GetPairs[{}, ___]                     := {}
GetPairs[{_}, ___]                    := {}
GetPairs[indexL_List, testQ_:(True&)] :=
   Module[ {idxL = Select[indexL, testQ]},
      (* choose tensorial, remove duplicates, to Dn indices, sort, and then split *)
      idxL = Split @ Sort @ (ToDnIndex /@ (Union @ Select[idxL, TensorialIndexQ]));
      idxL = Select[idxL, (Length[#] === 2)&];  (* choose pairs *)
      Map[{#[[1]], ToUpIndex @ #[[2]]}&, idxL]  (* to {dn, up} *)
   ]
GetPairs[___]                         := (Message[GetPairs::usage]; {})

(* GreekIndexQ tag *)
GreekIndexQ[a_]  := indexClassQ[a] === GreekIndexQ
GreekIndexQ[___] := False

(* HatIndexQ tag *)
HatIndexQ[a_]  := indexClassQ[a] === HatIndexQ
HatIndexQ[___] := False

(* IndexOrderedQ tag *)
IndexOrderedQ[a_?AtomQ, b_?AtomQ] := OrderedQ[forOrderingIndex /@ {a, b}]
IndexOrderedQ[aL_List]            := OrderedQ[forOrderingIndex /@ aL]
IndexOrderedQ[aL_List, bL_List]   := OrderedQ[{forOrderingIndex /@ aL, forOrderingIndex /@ bL}]
IndexOrderedQ[___]                := (Message[IndexOrderedQ::usage]; True)

(* NormalIndexQ tag *)
NormalIndexQ[a_]  := indexClassQ[a] === NormalIndexQ
NormalIndexQ[___] := False

(* OneIndexQ tag *)
OneIndexQ[a_]  := indexClassQ[a] === OneIndexQ
OneIndexQ[___] := False

(* PairQ tag *)
(* is a contracted pair satisfying testQ? *)
PairQ[a_Integer,         __]                                := False
PairQ[_,                 a_Integer,         ___]            := False
PairQ[a_Symbol,          b_Symbol,          testQ_:(True&)] := testQ[a] && strPairQ[ToString[a], ToString[b]]
PairQ[a_String,          b_String,          testQ_:(True&)] := testQ[a] && strPairQ[a, b]
PairQ[pairs:({_, _}..)]                                     := PairQ[pairs, True&]
PairQ[pairs:({_, _}..),                     testQ_]         :=
   AllQ @ Map[PairQ[#[[1]], #[[2]], testQ]&, {pairs}]
PairQ[___]                                                  := False

   (* where *)
   strPairQ["l" <> oneIndices, _]                 := False
   strPairQ["u" <> oneIndices, _]                 := False
   strPairQ[_,                 "l" <> oneIndices] := False
   strPairQ[_,                 "u" <> oneIndices] := False
   strPairQ[a_String,          b_String]          :=
      If [StringLength[a] > 1 && StringLength[b] > 1 && StringDrop[a, 1] === StringDrop[b, 1],
         True,
      (* else *)
         False
      ] /; (StringMatchQ[a, "l*"] && StringMatchQ[b, "u*"]) || (StringMatchQ[a, "u*"] && StringMatchQ[b, "l*"])
   strPairQ[_,                 _]                 := False

(* RegularIndexQ tag *)
RegularIndexQ[a_Symbol] := True
RegularIndexQ[___]      := False

(* TensorialIndexQ tag *)
(* is a non-(component or one-dimensional) tensorial index? *)
TensorialIndexQ[a_?ComponentIndexQ, ___]            := False
TensorialIndexQ[a_?OneIndexQ,       ___]            := False
TensorialIndexQ[a_,                 testQ_:(True&)] := testQ[a] && indexClassQ[a] =!= errorIndexQ
TensorialIndexQ[___]                                := False

(* ToDnIndex tag *)
ToDnIndex[a_Integer?Positive, testQ_:(True&)] := -a /; testQ[a]
ToDnIndex[a_Symbol?UpIndexQ,  testQ_:(True&)] := ToExpression["l" <> StringDrop[ToString[a], 1]] /; testQ[a]
ToDnIndex[a_String?UpIndexQ,  testQ_:(True&)] := "l" <> StringDrop[a, 1] /; testQ[a]
ToDnIndex[a_,                 ___]            := a  (* if not UpIndexQ *)

(* ToUpIndex tag *)
ToUpIndex[a_Integer?Negative, testQ_:(True&)] := -a /; testQ[a]
ToUpIndex[a_Symbol?DnIndexQ,  testQ_:(True&)] := ToExpression["u" <> StringDrop[ToString[a], 1]] /; testQ[a]
ToUpIndex[a_String?DnIndexQ,  testQ_:(True&)] := "u" <> StringDrop[a, 1] /; testQ[a]
ToUpIndex[a_,                 ___]            := a  (* if not DnIndexQ *)

(* UpIndexQ tag *)
UpIndexQ[a_Integer, testQ_:(True&)] := testQ[a] && a > 0
UpIndexQ[a_Symbol,  testQ_:(True&)] := testQ[a] && UpIndexQ[ToString[a]]
UpIndexQ[a_String]                  := StringLength[a] > 1 && StringMatchQ[a, "u*"]
UpIndexQ[a_String,  testQ_]         := testQ[a] && UpIndexQ[a]
UpIndexQ[___]                       := False

(*** Local Functions ***)
(* checkIndices tag *)
checkIndices[indexL_List, withMsg_:True] :=
(* are there duplicate regular indices in indexL? *)
   Module[ {idxL},
      (* remove component, one-dimensional indices, and dummy indices *)
      idxL = Select[Select[indexL, TensorialIndexQ], RegularIndexQ];

      (* duplicated indices? *)
      If [Length[Union[idxL]] =!= Length[idxL],
         If [withMsg, Message[Msg::err, "duplicated indices:", idxL, "", ""]];
         False,
      (* else *)
         True
      ]
   ]

(* errorIndexQ tag *)
errorIndexQ[a_] := (Message[Msg::err, "illegal index", a, "", ""]; a)

(* forOrderingIndex tag *)
forOrderingIndex[aIndex_Integer] := Abs[aIndex]
forOrderingIndex[aIndex_String]  := forOrderingIndexAux[aIndex]
forOrderingIndex[aIndex_]        := aIndex

   (* where *)
   (* for neglecting ordering of dummy indices: due to automatic ResetDummies in Tensor package *)
   forOrderingIndexAux[dummyIndex_?DnIndexQ] := lzz
   forOrderingIndexAux[dummyIndex_?UpIndexQ] := uzz
   forOrderingIndexAux[dummyIndex_]          := dummyIndex

(* forPrintIndex tag *)
(* for pretty printing an index, return {expr for each index-type, string of spaces} *)
forPrintIndex[a_Integer] := Module[ {n = Abs[a]}, If [n > 9, {UnderBar[n], "  "}, {n, " "}]]
forPrintIndex[a_Symbol]  :=
   Switch [indexClassQ[a],
      HatIndexQ, {OverHat @ StringDrop[ToString[a], 2], " "},
      BarIndexQ, {OverBar @ StringDrop[ToString[a], 2], " "},
      DotIndexQ, {OverDot @ StringDrop[ToString[a], 2], " "},
      _,         {StringDrop[ToString[a], 1],           " "}
   ]
forPrintIndex["l" <> oneIndices] := {oneIndices, " "}
forPrintIndex["u" <> oneIndices] := {oneIndices, " "}
forPrintIndex[a_String]  :=
   Module[ {pos},
      pos = ToExpression @ StringDrop[a, 2];
      If [ !IntegerQ[pos], Return[{a, " "}] ];  (* error *)

      Switch [indexClassQ[a],
         NormalIndexQ,  If [ pos <= lenDummy,        Return[{dummyIndices[[pos]], " "}] ],
         GreekIndexQ,   If [ pos <= greekLenDummy,   Return[{greekDummyIndices[[pos]], " "}] ],
         CapitalIndexQ, If [ pos <= capitalLenDummy, Return[{capitalDummyIndices[[pos]], " "}] ],
         HatIndexQ,     If [ pos <= lenDummy,        Return[{OverHat @ dummyIndices[[pos]], " "}] ],
         BarIndexQ,     If [ pos <= lenDummy,        Return[{OverBar @ dummyIndices[[pos]], " "}] ],
         DotIndexQ,     If [ pos <= lenDummy,        Return[{OverDot @ dummyIndices[[pos]], " "}] ],
         OneIndexQ,     Return[{StringDrop[a, 1], " "}],
         _,             Return[{a, " "}]  (* error *)
      ];

      (* if pos > LenDummy *)
      {StringDrop[a, 1], StringJoin @@ Table[" ", {StringLength[a] - 1}]}  (* adjust spaces *)
   ]

(* indexClassQ tag *)
(* return iClassQ from a index *)
indexClassQ[a_Integer]         := ComponentIndexQ
indexClassQ[a_Symbol]          :=
   Module[ {str = ToString[a]},
      If [ StringLength[str] < 2, Return[errorIndexQ] ];

      (* check dn/up format: if not "l*" or "u*" *)
      If [ !StringMatchQ[str, "l*"] && !StringMatchQ[str, "u*"], Return[errorIndexQ] ];

      str = StringDrop[str, 1];  (* drop dn/up flag *)

      If [ MemberQ[normalIndices, str], Return[NormalIndexQ] ];
      If [ MemberQ[greekIndices, str],   Return[GreekIndexQ] ];
      If [ MemberQ[capitalIndices, str], Return[CapitalIndexQ] ];
      If [StringLength[str] === 2 && MemberQ[normalIndices, StringDrop[str,1]],
         Switch [StringTake[str, 1],
            "h", Return[HatIndexQ],
            "b", Return[BarIndexQ],
            "d", Return[DotIndexQ]
         ]
      ];
      If [ str === oneIndices, Return[OneIndexQ] ];

      errorIndexQ
   ]
indexClassQ["l" <> oneIndices] := OneIndexQ
indexClassQ["u" <> oneIndices] := OneIndexQ
indexClassQ[a_String]          :=
   Switch [StringTake[a, {2}],
      dummyPreStr,        NormalIndexQ,
      greekDummyPreStr,   GreekIndexQ,
      capitalDummyPreStr, CapitalIndexQ,
      hatDummyPreStr,     HatIndexQ,
      barDummyPreStr,     BarIndexQ,
      dotDummyPreStr,     DotIndexQ,
      _,                  errorIndexQ
   ] /; StringLength[a] > 2

(* upupdndnIndexQ tag *)
upupdndnIndexQ[aIndexL_] := SameQ @ (Sequence @@ Map[UpIndexQ, aIndexL])
(* are indices all dn all up or? *)
(********************************************************************)

(************************ for SetIndexClass *************************)
(*** Local Variables ***)
iClassTable = {{NormalIndexQ}, {ComponentIndexQ}}  (* default index class *)

(*** Exported Functions ***)
(* SetIndexClass tag *)
(* set Main/Sub index classes *)
SetIndexClass[iMain_Symbol, iSubL_:{ComponentIndexQ}]       := SetIndexClass[{iMain}, iSubL]
SetIndexClass[iMainL:{_Symbol..}, iSubL_:{ComponentIndexQ}] :=
   Module[ {},
      If [Intersection[iMainL, iSubL] =!= {},
         Message[Msg::err, "invalid main-sub relation between", iMainL, "and", iSubL]; Return[]
      ];

      If [ !AllQ[checkIndexClass /@ Join[iMainL, iSubL]], Return[] ];
      iClassTable = {iMainL, iSubL};
   ]
SetIndexClass[___]                                         := Message[SetIndexClass::usage]

   (* checkIndexClass tag *)
   checkIndexClass[iClassQ_] :=
      If [!MemberQ[iIndexClassTable, iClassQ],
         Message[Msg::err, "illegal index class:", iClassQ, "", ""];
         False,
      (* else *)
         True
      ]

(* MainIndexQ tag *)
MainIndexQ[aIndex_] := MemberQ[mainClassQL[], indexClassQ[aIndex]]
MainIndexQ[___]     := False

(* SubIndexQ tag *)
SubIndexQ[aIndex_] := MemberQ[subClassQL[], indexClassQ[aIndex]]
SubIndexQ[___]     := False

(* ValidIndexQ tag *)
(* check validity of an index. return False without a message for an invalid index. *)
ValidIndexQ[aIndex_] := checkIndexWith[aIndex, False]
ValidIndexQ[___]     := False

(* ValidIndicesQ tag *)
(* check validity of indices. return False without a message for invalid indices. *)
ValidIndicesQ[indexL_List] := checkIndicesWith[indexL, False]
ValidIndicesQ[___]         := False

(*** Local Functions ***)
(* checkIndexWith tag *)
(* check consistency of an index. *)
checkIndexWith[0,       withMsg_:True] := (
      If [ withMsg, Message[Msg::err, "not allowed component index", 0, "", ""] ];
      False
   )
checkIndexWith[aIndex_, withMsg_:True] :=
   If [!MemberQ[Join[mainClassQL[], subClassQL[]], indexClassQ[aIndex]],
      If [ withMsg, Message[Msg::err, "invalid index", aIndex, "", ""] ];
      False,
   (* else *)
      True
   ]

(* checkIndicesWith tag *)
checkIndicesWith[indexL_List, withMsg_:True] :=
(* check validity of an index list. return False with a corresponding message when withMsg == True. *)
   Module[ {},
      If [ !AllQ[indexL, checkIndexWith[#, withMsg]&], Return[False] ];

      checkIndices[indexL, withMsg]
   ]

(* mainClassQL tag *)
mainClassQL[] := iClassTable[[1]]

(* subClassQL tag *)
subClassQL[] := iClassTable[[2]]
(********************************************************************)

(*************************** for Objects ****************************)
(*** Exported Functions ***)
(* ClearName tag *)
(* clear an object name for redefining. *)
ClearName[oName_Symbol?DefinedNameQ] :=
   Module[ {},
      If [MemberQ[iPredefinedNameTable, oName],
         Message[Msg::err, "Pre-defined object", oName, "can't be removed!", ""]; Return[]
      ];
      Unprotect[oName];
      Remove[oName];
   ]
ClearName[oNameL:{_Symbol..}]        := Module[ {}, Map[ClearName, oNameL]; ] /; AllQ[oNameL, DefinedNameQ]
ClearName[___]                       := Message[ClearName::usage]

(* DefinedNameQ tag *)
(* is oName defined as an object name? *)
DefinedNameQ[oName_Symbol] := ValueQ[iNameTable[oName]]
DefinedNameQ[___]          := False

(*** Local Functions ***)
(* getNamePrtStr tag *)
getNamePrtStr[oName_] := iNameTable[oName] [[1]]

(* getNameRank tag *)
getNameRank[oName_] := iNameTable[oName] [[2]]

(* getNameSymmetries tag *)
getNameSymmetries[oName_] := iNameTable[oName] [[3]]

(* getNameType tag *)
getNameType[oName_] := iNameTable[oName] [[4,1]]
(********************************************************************)

(************************ for DefineTensor **************************)
(*** Exported Functions ***)
(* DefinedTensorQ tag *)
DefinedTensorQ[oName_Symbol?DefinedNameQ] := getNameType[oName] === DefinedTensorQ
DefinedTensorQ[___]                       := False

(* DefineTensor tag *)
DefineTensor[oName_Symbol, permW_String]                                :=
   defineTensor[oName, ToString[oName], permW, False]
DefineTensor[oName_Symbol, prtStr_String, permW_String]                 :=
   defineTensor[oName, prtStr, permW, False]
DefineTensor[oNameL:{(_Symbol)..}, permW_String]                        := (
      Thread[defineTensor[oNameL, Map[ToString, oNameL], permW, False], List, 2];
   )
DefineTensor[oNameL:{(_Symbol)..}, prtStrL:{(_String)..}, permW_String] := (
      If [Length[oNameL] =!= Length[prtStrL],
         Message[Msg::err, "incompatible arguments:", oNameL, "and", prtStrL]; Return[]
      ];
      Thread[defineTensor[oNameL, prtStrL, permW, False], List, 2];
   )
DefineTensor[___] := Message[DefineTensor::usage]

(* tag PermWeight *)
(* return full permmutation-weights in string format *)
PermWeight[permS_String?permWeightQ] :=
   Module[ {rc},
      rc = makePermWeightGroup[permS];
      If [ rc === Null, Return[] ];

      toPermWeight[rc]
   ]
PermWeight[___]                      := Message[PermWeight::usage]

(* SetAntisymmetric tag *)
SetAntisymmetric[oName_Symbol?DefinedTensorQ] := setSymmetric[oName, True]

(* SetSymmetric tag *)
SetSymmetric[oName_Symbol?DefinedTensorQ] := setSymmetric[oName, False]

   (* where *)
   setSymmetric[oName_, isAnti:(True | False)] :=
      Module[ {nRank = getNameRank[oName], symL, wL, permW},
         If [nRank > 1,
            symL = Permutations @ Range[nRank];  (* long evaluation for nRank > 7 due to Permutations *)
            If [ isAnti, wL = Map[Signature[#]&, symL], wL = Table[1, {Length[symL]}] ];
            permW = Transpose[{symL, wL}];

            (* update symW *)
            iNameTable[oName] = ReplacePart[iNameTable[oName], Sort[permW], 3];
         ]
      ]

(* Symmetries tag *)
(* return full symmetries of oName *)
Symmetries[oName_Symbol?DefinedTensorQ] := toPermWeight @ getNameSymmetries[oName]

(*** Local Functions ***)
(* defineTensor tag *)
defineTensor[oName_Symbol, prtStr_String, permW_String, isInternalQ_] :=
   Module[ {},
      If [!isInternalQ,  (* check when external call *)
         If [MemberQ[iPredefinedNameTable, oName],
            Message[Msg::err, oName, "is pre-defined.", "", ""]; Return[]
         ];

         If [MemberQ[allIndices, ToString[oName]],
            Message[Msg::err, oName, "is used as an index name!", "", ""]; Return[]
         ]
      ];

      If [ !putTensorTable[oName, prtStr, permW], Return[] ];

      Format[oName[args___]] := printTensor[oName, prtStr, {args}];  (* Pretty Printing *)
      oName[args__] := 0 /; FreeQ[{args}, Pattern | Blank] && vanishingObjectQ[oName, {args}];
   ]

   (* where *)
   printTensor[oName_, prtStr_, aIndexL_] :=  (* Pretty Printing an indical object *)
      Module[ {dnS = "", upS = "", rc, spaces},
         Map[If [DnIndexQ[#], {rc, spaces} = forPrintIndex[#];
                              dnS = SequenceForm[dnS, rc];     upS = SequenceForm[upS, spaces],
             (* else *)       {rc, spaces} = forPrintIndex[#];
                              dnS = SequenceForm[dnS, spaces]; upS = SequenceForm[upS, rc]
             ]&, aIndexL];
         Subsuperscript[ToExpression[prtStr], dnS, upS]
      ]

(* makePermWeightGroup tag *)
makePermWeightGroup[""]                       := {}  (* scalar *)
makePermWeightGroup[permS_String?permWeightQ] :=
   Module[ {permW, rcL},
      permW = Union @ toPermWeight[permS];  (* convert and remove duplicate permW *)

      If [permWeightQ[permW] === False,
         Message[Msg::err, "illegal permWeight(s):", permS, "", ""]; Return[],
      (* else *)
         rcL = makePermWeightGroupAux[permW]
      ];

      If [permWeightQ[rcL] === False,  (* check consistency *)
         Message[Msg::err, "illegal permWeight(s):", permS, "", ""]; Return[]
      ];

      rcL
   ]

   makePermWeightGroupAux[permS_String] := makePermWeightGroupAux @ Union @ toPermWeight[permS]
   makePermWeightGroupAux[permW_List]   :=
      Module[ {orgPermL, orgWL, permL, wL, nPermL, aPerm, aW, i, j},
         permL = Transpose[permW]; orgPermL = permL[[1]]; orgWL = permL[[2]];

         permL = orgPermL; wL = orgWL; nPermL = Length[permL];
         While [True,
            For [i = 1, i <= Length[orgPermL], i++,
               For [j = 1, j <= nPermL, j++,
                  aPerm = Map[orgPermL[[i]] [[#]]&, permL[[j]]]; aW = orgWL[[i]] * wL[[j]];
                  If [ !MemberQ[permL, aPerm], permL = Join[permL, {aPerm}]; wL = Join[wL, {aW}] ]
               ]
            ];
            If [ nPermL === Length[permL], Break[] ];  (* closing group *)
            nPermL = Length[permL]
         ];
         Transpose[{permL, wL}]
      ]

(* permWeightQ tag *)
permWeightQ[""]           :=  True  (* scalar *)
permWeightQ[permS_String] :=  (* has valid characters in permS *)
   Module[ {strL = Characters[permS]},
      strL = Union @ Select[strL, (StringPosition[" +-", #] === {})&];  (* drop ' ', '+', '-' *)

      (* are normal characters? *)
      If [ MemberQ[Map[Position[normalIndices, #]&, strL], {}], Return[False] ];

      True
   ]
permWeightQ[{}] := True  (* scalar *)
permWeightQ[permW:{{{_Integer..}, -1|1}..}] :=  (* is valid permWeight? *)
   Module[ {aL, sortedL},
      aL = Union[permW]; (* remove duplicate perms *)
      sortedL = Range @ Length @ aL[[1,1]];  (* {1,2,..} *)

      (* check each onePermWeight *)
      If [ MemberQ[Map[onePermWeightQ[#, sortedL]&, aL], False], Return[False] ];

      (* is same perm but different weights? *)
      aL = Map[#[[1]]&, aL];  (* perms without weight *)
      If [ Length @ aL =!= Length @ Union[aL], Return[False] ];

      True
   ]
permWeightQ[___] := False

   (* where *)
   onePermWeightQ[aPerm:{{_Integer..}, -1|1}, sortedL_List] :=  (* {{1,3,2,4}, 1} or {{1,3,2,4}, -1} *)
      Module[ {},
         (* is a combination of ditinct numbers? *)
         If [ Sort @ aPerm[[1]] =!= sortedL, Return[False] ];

         (* is {1,2,...} but antysymmetry? *)
         If [ aPerm === {sortedL, -1}, False, True ]
      ]
   onePermWeightQ[___] := False

(* putTensorTable tag *)
putTensorTable[oName_, prtStr_String, ""] := (  (* scalar *)
      iNameTable[oName] = {prtStr, 0, {}, {DefinedTensorQ}};
      True
   )
putTensorTable[oName_, prtStr_String, permS_String] :=
   (* return False for error(s) *)
   Module[ {permL},
      permL = makePermWeightGroup[permS];
      If [ permL === Null, Return[False] ];  (* illegal permWeight *)

      iNameTable[oName] = {prtStr, Length @ permL[[1,1]], Sort[permL], {DefinedTensorQ}};
      True
   ]

(* toPermWeight tag *)
toPermWeight[""]           :=  {}  (* scalar *)
toPermWeight[permS_String] :=      (* permS_String -> permW_List *)
   Module[ {strL, posL, nRank},
      strL = Select[Characters[permS], (# =!= " ")&];  (* neglect spaces *)

      (* strL = {sign,...} *)
      If [ strL[[1]] =!= "-" && strL[[1]] =!= "+", strL = Join[{"+"}, strL] ];
      strL = strL /. {"+" -> 1, "-" -> -1};

      (* get rank *)
      posL = Flatten @ Position[strL, 1 | -1];
      If [ Length[posL] === 1, nRank = Length[strL] - 1, nRank = posL[[2]] - posL[[1]] - 1 ];

      (* string indices -> positions of normal indices *)
      strL = Map[If[Head[#] =!= String, #, Position[normalIndices, #] [[1,1]]]&, strL];

      (* strL -> permW *)
      strL = Partition[strL, nRank + 1];
      Map[{Rest[#], First[#]}&, strL]
   ]
toPermWeight[{}]         := ""                                             (* scalar *)
toPermWeight[permW_List] := StringJoin @@ Map[toOnePermWeight[#]&, permW]  (* permW -> permS *)

   (* where *)
   toOnePermWeight[onePermW_List] :=  (* onePermW -> onePermS *)
      Module[ {str},
         str = StringJoin @@ Map[normalIndices[[#]]&, onePermW[[1]]];
         If [ onePermW[[2]] === -1, str = "-" <> str, str = "+" <> str ]
      ]

(* vanishingObjectQ tag *)
(* is a vanishing object by symmetries? *)
vanishingObjectQ[oName_?DefinedTensorQ, aIndexL_] := (* F_a^a -> 0 and F_{11} -> 0 but F_1^1 \neq 0 *)
   Module[ {indexL, antiSymL},
      (* neglect up/dn except component, one-dimensional, and sub indices *)
      indexL = Map[If [!TensorialIndexQ[#] || SubIndexQ[#], #, ToDnIndex[#]]&, aIndexL];

      (* anti-symmetric list *)
      antiSymL = Select[getNameSymmetries[oName], (#[[2]] === -1)&];
      If [ MemberQ[Map[indexL[[#[[1]]]]&, antiSymL], indexL], True, False ]
   ] /; getNameRank[oName] === Length[aIndexL]  (* for protecting invalid number of input indices *)
(********************************************************************)

(******************** for Define Operators **************************)
(*** Exported Functions ***)
(* DefinedOperatorQ tag *)
(* is opName defined as an operator name? *)
DefinedOperatorQ[opName_Symbol?DefinedNameQ] := getNameType[opName] === DefinedOperatorQ
DefinedOperatorQ[___]                        := False

(* DefineOperator tag *)
DefineOperator[opName_Symbol, prtForm_, opType_Symbol]              :=
   defineOperator[opName, prtForm, opType, False]
DefineOperator[opNameL:{(_Symbol)..}, prtFormL_List, opType_Symbol] := (
      If [Length[opNameL] =!= Length[prtFormL],
         Message[Msg::err, "incompatible arguments:", opNameL, "and", prtFormL]; Return[]
      ];
      Thread[defineOperator[opNameL, prtFormL, opType, False], List, 2];
   )
DefineOperator[___]                                                 := Message[DefineOperator::usage]

(*** Local Functions ***)
(* defineOperator tag *)
defineOperator[opName_Symbol, prtForm_, opType_Symbol, isInternalQ_] :=
   Module[ {arg, expr},
      If [!isInternalQ,  (* check when external call *)
         If [MemberQ[iPredefinedNameTable, opName],
            Message[Msg::err, opName, "is predefined.", "", ""]; Return[]
         ];

         If [MemberQ[allIndices, ToString[opName]],
            Message[Msg::err, "operator name", opName, "is used as an index name!", ""]; Return[]
         ];

         (* check operator type *)
         If [!MemberQ[{CD, LD, XD, XP, TP}, opType],
            Message[Msg::err, "illegal operator type:", opType, "", ""]; Return[]
         ]
      ];

      iNameTable[opName] = {prtForm, 0, {}, {DefinedOperatorQ, opType}};
      Format[opName[arg_, expr___]] := printOperator[opName, opType, prtForm, arg, expr];

      (* for TP-type operator only: for speed up pattern matching in Tensor package *)
      If [opType === TP,
         opName[expr_]          := expr /; FreeQ[expr, Pattern | Blank];
         opName[expr_Times, n_] := opName[List @@ expr, n] /; FreeQ[{expr, n}, Pattern | Blank];
         opName[expr_, n_]      := Power[expr, n] /; FreeQ[{expr, n}, Pattern | Blank] && Head[expr] =!= List;
      ];
   ]

   (* where *)
   printOperator[opName_, opType_, prtForm_, arg_, expr___] :=
      Switch [opType,
         CD, If [DnIndexQ[arg],
                SequenceForm[Subscript[prtForm, forPrintIndex[arg][[1]]], expr],
             (* else *)
                SequenceForm[Superscript[prtForm, forPrintIndex[arg][[1]]], expr]
             ],
         LD, If [DefinedTensorQ[arg],
                SequenceForm[Subscript[prtForm, getNamePrtStr[arg]], expr],
             (* else *)
                SequenceForm[Subscript[prtForm, arg], expr]
             ],
         XD, SequenceForm[prtForm, arg, expr],
         XP, FormatOperatorXP[opName, arg, expr],  (* must define for each XP-type operator *)
         TP, SequenceForm[Power[HoldForm[Evaluate[Times @@ arg]], expr]],  (* TP[{...}, n] *)
         _,  Message[Msg::dbg, printOperator, opName, "invalid operator type", opType]; Return[]
      ]

(* getOpType tag *)
getOpType[opName_] := iNameTable[opName][[4,2]]
(********************************************************************)

(************************ for IndicalAtom ***************************)
(***** Exported Functions *****)
(* ExpandIAtom tag *)
ExpandIAtom[expr_]   :=
   Module[ {rcExpr},
      (* wraping $FORWRAPIATOM for expanding indical expr *)
      rcExpr = WrapIAtom[expr, $FOREXPANDIATOM];

      Expand[rcExpr, $FOREXPANDIATOM] /. $FOREXPANDIATOM -> Identity
   ]
ExpandIAtom[args___] := args

(* ExtractIAtom tag *)
(* extract indical objects from aTerm *)
ExtractIAtom[aTerm_] :=
   Module[ {rc = splitTerm[aTerm] [[2]]},
      If [rc === 1,
         {},
      (* else *)
         If [ Head[rc] === Times, List @@ rc, {rc} ]
      ]
   ]
ExtractIAtom[___]    := Message[ExtractIAtom::usage]

(* FreeIAtomQ tag *)
FreeIAtomQ[expr_]  :=
   Module[ {rcExpr},
      (* wraping $FORWRAPIATOM *)
      rcExpr = WrapIAtom[expr, $FORFREEIATOM];

      FreeQ[rcExpr, $FORFREEIATOM]
   ]
FreeIAtomQ[args___] := True

(* IAtomQ tag *)
IAtomQ[(oName_?DefinedNameQ)[___]] := True
IAtomQ[___]                        := False

(* OrderingIndies tag *)
(* sorting indices using symmetries in expr *)
OrderingIndices[expr_]   := forEachIAtom[expr, orderingIndicesObject]
OrderingIndices[expr___] := expr

(* WrapIAtom tag *)
(* wraping indical atoms in expr with wrapSymb for some operations *)
WrapIAtom[expr_, wrapSymb_Symbol] := forEachIAtom[expr, wrap, wrapSymb]
WrapIAtom[___]                    := Message[WrapIAtom::usage]

   (* where *)
   wrap[expr_, wrapSymb_] := wrapSymb[expr]

(***** Local Functions *****)
(* forEachIAtom tag *)
(* visit every indical atoms in expr and then act actTo on them  *)
forEachIAtom[expr_Plus,                       actTo_, args___] := Map[forEachIAtom[#, actTo, args]&, expr]
forEachIAtom[expr_Times,                      actTo_, args___] := Map[forEachIAtom[#, actTo, args]&, expr]
forEachIAtom[expr_Power,                      actTo_, args___] := Map[forEachIAtom[#, actTo, args]&, expr]
forEachIAtom[expr_Equal,                      actTo_, args___] := Map[forEachIAtom[#, actTo, args]&, expr]
forEachIAtom[expr:(oName_?DefinedNameQ)[___], actTo_, args___] := actTo[expr, args]
forEachIAtom[expr_,                           _,      ___]     := expr

(* forEachTerm tag *)
(* visit each term in expr and then act actTo on them *)
(* For proper operatopn, ExpandIndicalAtom must be acted to expr prior to call this *)
forEachTerm[expr_Plus,  actTo_, args___] := Map[forEachTerm[#, actTo, args]&, expr]
forEachTerm[expr_Equal, actTo_, args___] := Map[forEachTerm[#, actTo, args]&, expr]
forEachTerm[expr_,      actTo_, args___] := actTo[expr, args]

(* orderingIndicesObject tag *)
(* ordering indices using symmetries of oName. *)
orderingIndicesObject[(oName_?DefinedTensorQ)[aIndices___]] := orderingIndicesObjectAux[oName[aIndices]]

   (* where *)
   orderingIndicesObjectAux[expr:oName_[aIndices___]] :=
      Module[ {nLen = Length[{aIndices}], allSymL, tmpF, symL, iL, indexWL, minPos},
         (* for protecting invalid number of input indices *)
         If [ getNameRank[oName] =!= nLen, Return[expr] ];
         If [ nLen < 2, Return[expr] ];

         allSymL = getNameSymmetries[oName];
         tmpF[symL_, iL_] := {{aIndices}[[ symL[[1]] ]], iL[[1]]};
         indexWL = Sort[MapIndexed[tmpF, allSymL], IndexOrderedQ[#1[[1]], #2[[1]]]&];
         minPos = indexWL[[1,2]];

         allSymL[[minPos, 2]] * expr[[ allSymL[[minPos,1]] ]]
      ] /; FreeQ[{oName, aIndices}, Pattern | Blank]

orderingIndicesObject[(opName_?DefinedOperatorQ)[arg_, expr___]] :=
   Switch [getOpType[opName],
      CD, opName[arg, orderingIndicesObject[expr]],
      LD, opName[arg, orderingIndicesObject[expr]],
      XD, opName[orderingIndicesObject[arg], expr],
      XP, opName[Sequence @@ (orderingIndicesObject /@ {arg, expr})],
      TP, opName[Sequence @@ (orderingIndicesObject /@ {arg, expr})],  (* no effect on arg *)
      _,  opName[arg, expr]  (* internal error *)
   ] /; FreeQ[{opName, arg, epxr}, Pattern | Blank]

(* splitTerm tag *)
(* return {non-indical, indical} of aTerm: Apply[Times, {non-indical, indical}] === aTerm *)
(* For proper operatopn, ExpandIAtom must be acted prior to calling this *)
(* Power of an indical atom is non-indical *)
splitTerm[aTerm:(oName_?DefinedNameQ)[___]] := {1, aTerm}
splitTerm[aTerm_Times]                      := Times @@ Map[splitTerm, List @@ aTerm]
splitTerm[expr_, ___]                       := {expr, 1}
(********************************************************************)

(*********************** for GetIndices *****************************)
(***** Exported Functions *****)
(* GetFreeIndices tag *)
(* return Union of free tensorial indices satisfying testQ in expr. *)
GetFreeIndices[expr_, testQ_:(True&)]     :=
   Module[ {freeIndexL},
      freeIndexL = indicesSatisfying[expr, {ValidIndexQ, TensorialIndexQ}];
      Select[Complement[freeIndexL, Flatten @ GetPairs[freeIndexL]], testQ]
   ]
GetFreeIndices[___]                       := Message[GetFreeIndices::usage]

(* GetIndices tag *)
(* return Union of indices satisfying testQ in expr. *)
GetIndices[expr_, testQ_:(True&)] := indicesSatisfying[expr, {ValidIndexQ, testQ}]
GetIndices[___]                   := Message[GetIndices::usage]

(* GetMainIndices tag *)
(* return Union of main-indices satisfying testQ in expr *)
GetMainIndices[expr_, testQ_:(True&)] := indicesSatisfying[expr, {MainIndexQ, testQ}]
GetMainIndices[___]                   := Message[GetMainIndices::usage]

(* GetSubIndices tag *)
(* return Union of sub-indices satisfying testQ in expr *)
GetSubIndices[expr_, testQ_:(True&)] := indicesSatisfying[expr, {SubIndexQ, testQ}]
GetSubIndices[___]                   := Message[GetSubIndices::usage]

(***** Local Functions *****)
(* getOriginalIndicesTerm tag *)
(* return indices from a term, where the indices do not sorted *)
getOriginalIndicesTerm[expr:(oName_?DefinedNameQ)[___], testQ_:(True&)] :=
   getOriginalIndicesTermFor[expr, testQ]
getOriginalIndicesTerm[aTerm_Times, testQ_:(True&)]                     :=
   Flatten @ Map[getOriginalIndicesTerm[#, testQ]&, List @@ aTerm]
getOriginalIndicesTerm[aPower_Power, testQ_:(True&)]                    :=
   Flatten @ Map[getOriginalIndicesTerm[#, testQ]&, List @@ aPower]
getOriginalIndicesTerm[___]                                             := {}

   (* where *)
   getOriginalIndicesTermFor[(oName_?DefinedTensorQ)[aIndices__], testQ_]       :=
      Select[{aIndices}, testQ]
   getOriginalIndicesTermFor[(opName_?DefinedOperatorQ)[arg_, expr___], testQ_] :=
      Switch [getOpType[opName],
         CD, If [testQ[arg],
                Join[{arg}, getOriginalIndicesTerm[expr, testQ]],
             (* else *)
                getOriginalIndicesTerm[expr, testQ]
             ],
         LD, getOriginalIndicesTerm[expr, testQ],
         XD, getOriginalIndicesTerm[arg, testQ],
         XP, Flatten @ Map[getOriginalIndicesTerm[#, testQ]&, {arg, expr}],
         TP, Flatten @ Map[getOriginalIndicesTerm[#, testQ]&, {Times @@ arg, expr}],
         _,  Message[Msg::dbg, getOriginalIndicesTerm, opName, "invalid operator type", getOpType[opName]]; {}
      ]

(* indicesSatisfying tag *)
(* return sorted indices satisfying testQL in expr *)
indicesSatisfying[expr:(oName_?DefinedNameQ)[___], testQL_:{True&}]   := indicesSatisfyingFor[expr, testQL]
indicesSatisfying[expr_Plus, testQL_:{True&}]                         :=
   Union @ Flatten @ Map[indicesSatisfying[#, testQL]&, List @@ expr]
indicesSatisfying[expr_Times, testQL_:{True&}]                        :=
   Union @ Flatten @ Map[indicesSatisfying[#, testQL]&, List @@ expr]
indicesSatisfying[expr_Power, testQL_:{True&}]                        :=
   Union @ Flatten @ Map[indicesSatisfying[#, testQL]&, List @@ expr]
indicesSatisfying[expr_Equal, testQL_:{True&}]                        :=
   Union @ Flatten @ Map[indicesSatisfying[#, testQL]&, List @@ expr]
indicesSatisfying[___]                                                := {}

   (* where *)
   indicesSatisfyingFor[(oName_?DefinedTensorQ)[aIndices__], testQL_]       :=
      Flatten @ (Intersection @@ Map[Select[{aIndices}, #]&, testQL])
   indicesSatisfyingFor[(opName_?DefinedOperatorQ)[arg_, expr___], testQL_] :=
      Switch [getOpType[opName],
         CD, If [AllQ[Map[#[arg]&, testQL]],
                Union @ Join[{arg}, indicesSatisfying[expr, testQL]],
             (* else *)
                indicesSatisfying[expr, testQL]
             ],
         LD, indicesSatisfying[expr, testQL],
         XD, indicesSatisfying[arg, testQL],          (* expr === Null *)
         XP, Union @ Flatten @ Map[indicesSatisfying[#, testQL]&, {arg, expr}],
         TP, Union @ Flatten @ Map[indicesSatisfying[#, testQL]&, {Times @@ arg, expr}],
         _,  Message[Msg::dbg, indicesSatisfying, opName, "invalid operator type", getOpType[opName]]; {}
      ]
(********************************************************************)

(************************* for Dummies ******************************)
(***** Exported Functions *****)
(* Dum tag *)
(* Dum contracted pairs which are not already dummed *)
Dum[expr_] := forEachTerm[ExpandIAtom[expr], dumTerm, RegularIndexQ]
Dum[___]   := Message[Dum::usage]

(* DumFresh tag *)
(* fresh Dum all contracted indices *)
DumFresh[expr_] := forEachTerm[ExpandIAtom[expr], dumTerm, TensorialIndexQ]
DumFresh[___]   := Message[DumFresh::usage]

(* ResetDummies tag *)
(* re-labelling dummy indices in expr *)
ResetDummies[expr_]   := forEachTerm[ExpandIAtom[expr], resetDummiesTerm]
ResetDummies[expr___] := expr

(* SplitDum tag *)
(* Split dummies with no checking of main-sub relation *)
SplitDum[rangeL:(({_, _, __}?VectorQ)..), expr_] :=
   Module[ {},
      If [!AllQ[Flatten[{rangeL}], DnIndexQ],
         Message[Msg::err, {rangeL}, "contains non-down indices", "", ""]; Return[expr]
      ];
      forEachTerm[ExpandIAtom[expr], sumDumSymbolicTerm, {rangeL}]
   ]
SplitDum[___]   := Message[SplitDum::usage]

(* SumDum tag *)
(* numeric SumDum *)
SumDum[{lower_Integer?Positive, upper_Integer?Positive}, expr_] :=
   If [lower < upper,
      forEachTerm[ExpandIAtom[expr], sumDumNumericTerm, {lower, upper}],
   (* else *)
      forEachTerm[ExpandIAtom[expr], sumDumNumericTerm, {upper, lower}]
   ]

   (* where *)
   sumDumNumericTerm[aTerm_, {lower_, upper_}] :=
      Module[ {ordTerm, oTerm},
         {ordTerm, oTerm} = splitTerm[aTerm];

         ordTerm * sumDumNumericAux[oTerm, {lower, upper}]
      ]

      (* where *)
      sumDumNumericAux[oTerm_, {lower_, upper_}] :=
         Module[ {pairL, rcL, i, j},
            pairL = GetPairs @ GetMainIndices[oTerm];
            If [ pairL === {}, Return[oTerm] ];

            rcL = oTerm;
            For [i = 1, i <= Length[pairL], i++,
               (* prepare for Table generation *)
               rcL = rcL /. {pairL[[i,1]] -> -j}; rcL = rcL /. {pairL[[i,2]] -> j};

               rcL = Table[rcL, {j, lower, upper}]  (* dummy indices -> component indices *)
            ];

            If [rcL === oTerm,
               oTerm,  (* not summed *)
            (* else *)
               Plus @@ Flatten[rcL, Length[pairL] - 1]
            ]
         ]

(* symbolic SumDum with checking of main-sub relation *)
SumDum[rangeL:(({_, _, __}?VectorQ)..), expr_] := sumDum[{rangeL}, expr] /; Head[expr] =!= List

   (* where *)
   sumDum[rangeL:{{_, _, __}..}, expr_] :=
      Module[ {i},
         If [!AllQ[Flatten[rangeL], DnIndexQ],  (* are all indices in rangeL dn? *)
            Message[Msg::err, rangeL, "contains non-down indices", "", ""]; Return[expr]
         ];

         (* check consistency of main-sub relation *)
         For [i = 1, i <= Length[rangeL], i++,
            If [!MainIndexQ[rangeL[[i,1]]],
               Message[Msg::err, "not main index", rangeL[[i,1]], "", ""];
               Return[expr]
            ];
            If [!AllQ[ Map[SubIndexQ, Drop[rangeL[[i]], 1]] ],
               Message[Msg::err, Drop[rangeL[[i]], 1], "contains non-sub indices", "", ""];
               Return[expr]
            ]
         ];

         forEachTerm[ExpandIAtom[expr], sumDumSymbolicTerm, rangeL]
      ]

(***** Local Functions *****)
(* dumTerm tag *)
dumTerm[aTerm_, selectQ_] :=
   Module[ {ordTerm, oTerm, indexL, iClassQL},
      {ordTerm, oTerm} = splitTerm[aTerm];  (* NOT for Power of scalars *)
      indexL = GetIndices[oTerm, selectQ];  (* indices satisfying selectQ *)
      If [ indexL === {}, Return[aTerm] ];

      (* choose tensorial-indexQ *)
      iClassQL = Complement[Join[mainClassQL[], subClassQL[]], {ComponentIndexQ, OneIndexQ}];

      ordTerm * Fold[dumTermAux[#1, indexL, #2]&, oTerm, iClassQL]
   ]

   (* where *)
   dumTermAux[oTerm_, {}]                := oTerm
   dumTermAux[oTerm_, {_}]               := oTerm
   dumTermAux[oTerm_, indexL_, iClassQ_] :=
      Module[ {pairL, dummyL, i},
         pairL = GetPairs @ Select[indexL, iClassQ];
         If [ pairL === {}, Return[oTerm] ];

         dummyL = Table[DummyPair[iClassQ], {i, Length[pairL]}];
         oTerm /. MapThread[Rule, {Flatten[pairL, 1], Flatten[dummyL, 1]}]
      ]

(* resetDummiesTerm tag *)
resetDummiesTerm[aTerm_] :=
(* reset dummy indices in aTerm *)
   Module[ {ordTerm, oTerm, dummyL, iClassQL},
      {ordTerm, oTerm} = splitTerm[aTerm];  (* NOT for Power of scalars. Use TP operator *)
      dummyL = Select[getOriginalIndicesTerm[oTerm], DummyIndexQ];
      If [ dummyL === {}, Return[aTerm] ];

      (* choose tensorial-indexQ *)
      iClassQL = Complement[Join[mainClassQL[], subClassQL[]], {ComponentIndexQ, OneIndexQ}];

      ordTerm * Fold[resetDummiesTermAux, oTerm, Map[Select[dummyL, #]&, iClassQL]]
   ]

   (* where *)
   resetDummiesTermAux[oTerm_, {}]      := oTerm
   resetDummiesTermAux[oTerm_, dummyL_] :=
      Module[ {splitL, posL, reOrderRule, newDummyL},
         (* split "l$num" or "u$num$ into {"l$", num} or {"u$", num} *)
         splitL = Map[{StringTake[#, 2], ToExpression[StringDrop[#, 2]]}&, dummyL];

         (* list of positions in DummyIndices *)
         posL = UnionList @ Map[#[[2]] &, splitL];

         (* re-numbering in accending order *)
         reOrderRule = MapThread[Rule, {posL, Range @ Length[posL]}];

         newDummyL = Map[{#[[1]], #[[2]] /. reOrderRule}&, splitL];   (* re-ordering *)
         newDummyL = Map[(#[[1]] <> ToString[#[[2]]]) &, newDummyL];  (* combine *)
         oTerm /. MapThread[Rule, {dummyL, newDummyL}]
      ]

(* sumDumSymbolicTerm tag *)
sumDumSymbolicTerm[aTerm_, rangeL_] :=
   Module[ {ordTerm, oTerm},
      {ordTerm, oTerm} = splitTerm[aTerm];

      ordTerm * sumDumSymbolicAux[oTerm, rangeL]
   ]

   (* where *)
   sumDumSymbolicAux[oTerm_, rangeL_] :=
      Module[ {indexL, newRangeL = {}, aIndex, rcTerm, aRangeL, ruleL, rcL, i},
         indexL = GetMainIndices[oTerm, TensorialIndexQ];

         (* get rangeL of dummy indices *)
         For [i = 1, i <= Length[rangeL], i++,
            aIndex = rangeL[[i,1]];
            If [Position[indexL, aIndex] === {},
               Message[Msg::warn, "no index or sub-index:", aIndex, "in", indexL]; Continue[],
            (* else *)
               If [Position[indexL, FlipIndex[aIndex]] == {},
                  Message[Msg::warn, "non-dummy or sub-index:", aIndex, "in", indexL]; Continue[]
               ]
            ];
            newRangeL = Join[newRangeL, {rangeL[[i]]}]
         ];
         If [newRangeL == {}, Return[oTerm]];

         (* substitution *)
         rcL = oTerm;
         For [i = 1, i <= Length[newRangeL], i++,
            (* {{la, ua}, {-1,1}, {li, ui},...} for {la, -1, li,...} *)
            aRangeL = Map[{#, FlipIndex[#]}&, newRangeL[[i]]];

            ruleL = {};
            For [j = 2, j <= Length[aRangeL], j++,
               ruleL = Join[ruleL, { {aRangeL[[1,1]] -> aRangeL[[j,1]],
                                      aRangeL[[1,2]] -> aRangeL[[j,2]]} }]
            ];
            rcL = Map[(rcL /. #)&, ruleL]
         ];

         If [rcL === oTerm,
            oTerm,  (* not summed *)
         (* else *)
            Plus @@ Flatten[rcL, Length[newRangeL] - 1]
         ]
      ]
(********************************************************************)

(************************ for Symmetrize ****************************)
(* Antisymmetrize tag *)
Antisymmetrize[indexL_List, expr_] :=
   Module[ {},
      (* check consistency of indexL*)
      If [!ValidIndicesQ[indexL],
         Message[Msg::err, "incompatible indices", indexL, "", ""];
         Return[expr]
      ];

      forEachTerm[ExpandIAtom[expr], symmetrizeTerm, indexL, True]
   ]
Antisymmetrize[___]                := Message[Antisymmetrize::usage]

(* Symmetrize tag *)
Symmetrize[symIndexL_List, expr_] :=
   Module[ {},
      (* check consistency of symIndexL*)
      If [!ValidIndicesQ[symIndexL],
         Message[Msg::err, "incompatible indices", symIndexL, "", ""];
         Return[expr]
      ];

      forEachTerm[ExpandIAtom[expr], symmetrizeTerm, symIndexL, False]
   ]
Symmetrize[___]                   := Message[Symmetrize::usage]

   (* where *)
   symmetrizeTerm[aTerm_, symIndexL_, antiQ_] :=
      Module[ {ordTerm, oTerm},
         {ordTerm, oTerm} = splitTerm[aTerm];
         ordTerm * symmetrizeTermAux[oTerm, symIndexL, antiQ, GetIndices[oTerm]]
      ]

      (* where *)
      symmetrizeTermAux[oTerm_, symIndexL_, antiQ_, indexL_] :=
         Module[ {permL, aPerm, reOrderRule, rc = 0, i},
            If [!AllQ[symIndexL, (MemberQ[indexL, #])&],  (* if no symIndex in oTerm *)
               Message[Msg::err, symIndexL, "is not subset of", indexL, ""]; Return[oTerm]
            ];

            (* all possible permutations of symIndexL: may be long *)
            permL = Permutations[symIndexL];

            For [i = 1, i <= Length[permL], i++,
               aPerm = permL[[i]];
               reOrderRule = MapThread[Rule, {symIndexL, aPerm}];
               If [antiQ === True, (* Antisymmetrize *)
                  rc += Signature[aPerm] * (oTerm /. reOrderRule),
               (* else *) (* Symmetrize *)
                  rc += oTerm /. reOrderRule
               ]
            ];
            rc / Length[symIndexL]!
         ]
(********************************************************************)

(************************ for ApplyRules ****************************)
(***** Exported Functions *****)
(* ApplyRules tag *)
ApplyRules[expr_, oRule:RuleDelayed[_, _Condition]]        := ApplyRules[expr, {oRule}]
ApplyRules[expr_, oRules:{(RuleDelayed[_, _Condition])..}] := applyObjectRules[expr, oRules]
ApplyRules[a___]                                           := Message[ApplyRules::usage]

(* ObjectRule tag *)
(* return a rule for input of ApplyRules *)
SetAttributes[ObjectRule, HoldAll]
ObjectRule[lhs_, rhs_, cond_:True] := objectRule[lhs, Evaluate[Dum[rhs]], cond]
ObjectRule[___]                    := (Message[ObjectRule::usage]; {})

(* ObjectSet tag *)
(* set lhs to rhs if cond hold, where dummy indices in rhs are preperly treated. *)
SetAttributes[ObjectSet, HoldAll]
ObjectSet[lhs_, rhs_, cond_:True] := objectSet[lhs, Evaluate[Dum[rhs]], cond]
ObjectSet[___]                    := Message[ObjectSet::usage]

(***** Local Functions *****)
(* applyObjectRules tag *)
applyObjectRules[expr_, oRules:{(RuleDelayed[_, _Condition])..}] :=
(* need to proper ordering of oRules due to the sequential substitutions *)
   Module[ {rcExpr, posL, powExpr, subExpr, rcSubExpr, rhs, aRule, i, j},
      rcExpr = Dum[expr];
      For [i = 1, i <= Length[oRules], i++,
         posL = {}; powExpr = {};
         If [noIndexedObjectQ[oRules[[i,1]]],  (* consider scalar expr *)
            posL = Position[rcExpr, Power[_, _Integer]];
            If [posL =!= {},
               powExpr = Extract[rcExpr, posL];
               If [!FreeQ[powExpr, oRules[[i,1]]],  (* scalar power expr *)
                  subExpr = Table[powExpr[[1,1]], {i, powExpr[[1,2]]}]  (* {f^2} -> {f, f} *)
               ]
            ]
         ];
         If [posL === {},  (* non-(scalar and Power) expr *)
            posL = Position[rcExpr, oRules[[i,1]]];
            subExpr = Extract[rcExpr, posL]
         ];

         rcSubExpr = subExpr;
         For [j = 1, j <= Length[rcSubExpr], j++,
            (* get expr with unique dummy indices in oRule[[i]] *)
            rhs = DumFresh @ oRules[[i,2,1]];

            (* apply the rule with unique dummies to subExpr *)
            aRule = ReplacePart[oRules[[i]], rhs, {2,1}];
            rcSubExpr = ReplacePart[rcSubExpr, rcSubExpr[[j]] /. aRule, j]
         ];

         (* replace the subExpr with rcSubExpr *)
         If [powExpr =!= {},  (* scalar power expr *)
            rcExpr = ReplacePart[rcExpr, Times @@ rcSubExpr, posL],
         (* else *)
            rcExpr = rcExpr /. MapThread[Rule, {subExpr, rcSubExpr}];
         ]
      ];
      rcExpr
   ]

   (* where *)
   (* has no indices? *)
   noIndexedObjectQ[expr_] :=
      Module[ {indices = getOriginalIndicesTerm[expr]},
         If [!FreeQ[indices, Pattern | Blank],  (* with la_,... *)
            False,
         (* else *)  (* has no tensorial indices? *)
            Select[indices, TensorialIndexQ] === {}
         ]
      ]
   noIndexedObjectQ[___]   := False

(* objectRule tag *)
SetAttributes[objectRule, HoldAll]
objectRule[lhs_, rhs_, cond_] := RuleDelayed[lhs, DumFresh[rhs] /; cond]

(* objectSet tag *)
SetAttributes[objectSet, HoldAll]
objectSet[lhs_, rhs_, cond_] := SetDelayed[lhs, DumFresh[rhs] /; cond]  (* dumFresh for each call of lhs *)
(********************************************************************)

(************************ for SyntaxCheck ***************************)
(* SyntaxCheck tag *)
(* return False for an error with appropriate messages. return itself for no-errors *)
SyntaxCheck[aPlus_Plus] :=
   Module[ {rcExpr = ExpandIAtom[aPlus], freeIndexL1, freeIndexL2, i},
      If [ syntaxCheckTerm[rcExpr[[1]]] === False, Return[False] ];

      (* get free indices of 1st term *)
      freeIndexL1 = GetFreeIndices @ rcExpr[[1]];

      For [i = 2, i <= Length[rcExpr], i++,
         If [ syntaxCheckTerm[rcExpr[[i]]] === False, Return[False] ];

         (* same free indices? *)
         freeIndexL2 = GetFreeIndices @ rcExpr[[i]];
         If [freeIndexL1 =!= freeIndexL2,
            Message[Msg::err, "incompatible free indices:", freeIndexL1, "and", freeIndexL2];
            Return[False]
         ]
      ];

      aPlus  (* no error *)
   ] /; FreeQ[aPlus, Pattern | Blank]
SyntaxCheck[aTerm_]     :=
    Module[ {expr = ExpandIAtom[aTerm]},
        Switch[ Head[expr],
            Plus,  Return[SyntaxCheck[expr]],
            _,     Return[syntaxCheckTerm[expr]]
        ];
    ] /; FreeQ[aTerm, Pattern | Blank]
SyntaxCheck[expr_]      := expr
SyntaxCheck[___]        := Message[SyntaxCheck::usage]

   (* where *)
   syntaxCheckTerm[aTerm_] :=
      Module[ {ordTerm, oTerm, pos},
         {ordTerm, oTerm} = splitTerm[aTerm];

         (* check Power expr *)
         If [!FreeIAtomQ[ordTerm],
            If [!FreeQ[syntaxCheckPower /@ Extract[ordTerm, Position[ordTerm, Power[_,_]]], False],
               Return[False]
            ]
         ];

         If [ oTerm === 1, Return[aTerm] ];

         (* check objects *)
         oTerm = forEachIAtom[oTerm, syntaxCheckObject];
         If [ !FreeQ[oTerm, False], Return[False] ];  (* there are errors *)

         If [ !checkIndices[getOriginalIndicesTerm[oTerm, ValidIndexQ]], Return[False] ];

         aTerm  (* no-errors *)
      ]

      (* where *)
      syntaxCheckPower[Power[a_, b_]] :=
         If [GetFreeIndices[a] =!= {},
            Message[Msg::err, "non-scalar expression in Power:", a, "", ""]; False,
         (* else *)
            If [GetFreeIndices[b] =!= {},
               Message[Msg::err, "non-scalar expression in Power:", b, "", ""]; False,
            (* else *)
               Power[a,b]  (* no error? *)
            ]
         ]
      syntaxCheckPower[expr_]         := expr

      syntaxCheckObject[(oName_?DefinedTensorQ)[]]                 :=
         If [getNameRank[oName] =!= 0,
            Message[Msg::err, "invalid number of indices for", oName, "", ""]; False,
         (* else *)
            oName[]  (* no errors *)
         ]
      syntaxCheckObject[(oName_?DefinedTensorQ)[aIndices__]]       := (
            (* check # of indices *)
            If [getNameRank[oName] =!= Length[{aIndices}],
               Message[Msg::err, {aIndices}, "has invalid number of indices for", oName, ""]; Return[False]
            ];

            (* check consistency of indices *)
            If [ !checkIndicesWith[{aIndices}], Return[False] ];

            oName[aIndices]  (* no errors *)
         )
      syntaxCheckObject[(opName_?DefinedOperatorQ)[arg_, expr___]] :=
         Switch[ getOpType[opName],
            CD, If [ SyntaxCheck[expr] === False, Return[False] ];
                If [ !checkIndexWith[arg], Return[False] ];
                If [ !checkIndices[Join[{arg}, GetIndices[expr]]], Return[False] ];
                opName[arg, expr],
            LD, If [ SyntaxCheck[expr] === False, Return[False] ];
                If [ !checkFirstArgLD[opName, arg], Return[False] ];
                opName[arg, expr],
            XD, If [ SyntaxCheck[arg] === False, Return[False] ];
                opName[arg, expr],
            XP, If [ MemberQ[SyntaxCheck /@ {arg, expr}, False], Return[False] ];
                If [!checkIndices[Flatten @ (GetIndices /@ {arg, expr})],
                   Return[False]
                ];
                opName[arg, expr],
            TP, If [Length[{arg, expr}] =!= 2,
                   Message[Msg::err, "two arguments are requied.", "", "", ""]; Return[False]
                ];
                If [ MemberQ[SyntaxCheck /@ {Times @@ arg, expr}, False], Return[False] ];
                If [GetFreeIndices[Times @@ arg] =!= {},
                   Message[Msg::err, "non-scalar expression", arg, "in power of tensor", ""]; Return[False],
                (* else *)
                   If [GetFreeIndices[expr] =!= {},
                      Message[Msg::err, "non-scalar expression:", expr, "in power of tensor", ""]; Return[False]
                   ]
                ];
                opName[arg, expr],
            _,  Message[Msg::err, "invalid operator type", getOpType[opName], "", ""]; Return[False]
         ]

         (* where *)
         (* check 1st argument of LD-type operator: Need to re-define for a specific LD-type operator *)
         checkFirstArgLD[opName_, arg_] :=  (* check LD-type operator *)
            If [!DefinedTensorQ[arg],
               Message[Msg::err, "non-tensorial argument", arg, "for an operator", opName];
               False,
            (* else *)
               True
            ]
(********************************************************************)

(************************ for Kaluza-Klein **************************)
(* tag SumDumKK *)
(* SumDum for Kaluza-Klein mode *)
SumDumKK[expr_] :=
   Module[ {subQL},
      (* get sub-indexQ's and remove ComponentIndexQ *)
      subQL = Complement[subClassQL[], {ComponentIndexQ}];

      If [Length[subQL] < 2,
         Message[Msg::err, "Kaluza-Klein mode needs, at least, two sub-index classes.", "", "", ""];
         expr,
      (* else *)
         forEachTerm[ExpandIAtom[expr], sumDumKK, subQL]
      ]
   ]
SumDumKK[___]   := Message[SumDumKK::usage]

   (* where *)
   sumDumKK[aTerm_, subQL_] :=
      Module[ {ordTerm, oTerm},
         {ordTerm, oTerm} = splitTerm[aTerm];

         ordTerm * sumDumKKAux[oTerm, subQL]
      ]

      (* where *)
      sumDumKKAux[oTerm_, subQL_] :=
         Module[ {rcTerm, mainIndexL, rangeL, i},
            (* Dum *)
            rcTerm = dumTerm[oTerm, RegularIndexQ];

            (* get main dummy indices*)
            mainIndexL = Select[GetMainIndices[rcTerm, DummyIndexQ], DnIndexQ];

            (* generate rangeL for calling sumDumSymbolicAux *)
            rangeL = {};
            For [i = 1, i <= Length[mainIndexL], i++,
               rangeL = Join[rangeL, {Join[{mainIndexL[[i]]}, Map[Dummy, subQL]]}]
            ];

            sumDumSymbolicAux[rcTerm, rangeL]
         ]

(* TableKK tag *)
(* Table for Kaluza-Klein mode *)
SetAttributes[TableKK, HoldAll]
TableKK[expr_, indexL_List] :=
   Module[ {subQL},
      (* get sub-indexQ's and remove ComponentIndexQ: KEEP ordering *)
      subQL = Flatten[subClassQL[] /. ComponentIndexQ -> {}];

      If [Length[subQL] < 2,
         Message[Msg::err, "Kaluza-Klein mode needs, at least, two sub-index classes.", "", "", ""];
         expr,
      (* else *)
         tableKK[expr, indexL, subQL]
      ]
   ] /; IAtomQ[expr] || Head[expr] === Times
TableKK[___]                       := Message[TableKK::usage]

   (* where *)
   tableKK[aTerm_, {},      subQL_] := aTerm
   tableKK[aTerm_, indexL_, subQL_] :=
      Module[ {mainIndexL, ordTerm, oTerm, newIndexL = {}},
         {ordTerm, oTerm} = splitTerm[aTerm];

         (* check consistency of indexL with main indices in oTerm *)
         mainIndexL = GetMainIndices[oTerm];
         For [i = 1, i <= Length[indexL], i++,
            If [!MemberQ[mainIndexL, indexL[[i]]],
               Message[Msg::warn, "no index or sub-index:", indexL[[i]], "in", mainIndexL]; Continue[]
            ];
            newIndexL = Join[newIndexL, {indexL[[i]]}]
         ];
         If [ newIndexL === {}, Return[aTerm] ];

         ordTerm * tableKKAux[oTerm, indexL, subQL, Length[subQL]]
      ]

      (* where *)
      tableKKAux[oTerm_, indexL_, subQL_, nSubQL_] :=
         Module[ {rc = oTerm, i},
            For [i = 1, i <= Length[indexL], i++,
               (* wrap $FORTABLEKK for avoiding automatic listable of subQL *)
               rc = tableKKAuxAux[rc, indexL[[i]], $FORTABLEKK[subQL, nSubQL]]
            ];
            rc
         ]

         (* where *)
         SetAttributes[tableKKAuxAux, Listable]
         tableKKAuxAux[oTerm_, mainIndex_, $FORTABLEKK[subQL_, nSubQL_]] :=
            Module[ {subFreeIndexL, tDummy},
               subFreeIndexL = GetFreeIndices[oTerm, SubIndexQ];

               (* dummy index for Table *)
               tDummy = Unique["m"];

               Table[ tableKKelement[oTerm, mainIndex, subFreeIndexL, subQL[[tDummy]]], {tDummy, nSubQL} ]
            ]

            (* where *)
            tableKKelement[oTerm_, mainIndex_, subFreeIndexL_, subQ_] :=
               Module[ {subIndex},
                  subIndex = indexFromClass[StringTake[ToString @ mainIndex, 1], subFreeIndexL, subQ];

                  oTerm /. mainIndex -> subIndex
            ]

            (* where *)
            indexFromClass[updnStr_String, subFreeIndexL_, iClass_] :=
            (* for getting an index which does not conflict with indices in subFreeIndexL *)
               Module[ {strIndexL, extraStr = ""},
                  Switch [iClass,
                     NormalIndexQ,
                        strIndexL = Complement[normalIndices,
                                               Map[StringDrop[#, 1]&,
                                                   ToString /@ Select[subFreeIndexL, NormalIndexQ]]],
                     GreekIndexQ,
                        strIndexL = Complement[greekIndices,
                                               Map[StringDrop[#, 1]&,
                                                   ToString /@ Select[subFreeIndexL, GreekIndexQ]]],
                     CapitalIndexQ,
                        strIndexL = Complement[capitalIndices,
                                               Map[StringDrop[#, 1]&,
                                                   ToString /@ Select[subFreeIndexL, CapitalIndexQ]]],
                     BarIndexQ,
                        strIndexL = Complement[normalIndices,
                                               Map[StringDrop[#, 1]&,
                                                   ToString /@ Select[subFreeIndexL, NormalIndexQ]]];
                        extraStr = "b",
                     DotIndexQ,
                        strIndexL = Complement[normalIndices,
                                               Map[StringDrop[#, 1]&,
                                                   ToString /@ Select[subFreeIndexL, NormalIndexQ]]];
                        extraStr = "d",
                     HatIndexQ,
                        strIndexL = Complement[normalIndices,
                                               Map[StringDrop[#, 1]&,
                                                   ToString /@ Select[subFreeIndexL, NormalIndexQ]]];
                        extraStr = "h",
                     OneIndexQ,
                        strIndexL = {oneIndices},
                     _,
                        Message[Msg::dbg, indexFromClass, "invalid index class", iClass, ""];
                        Return[ ToExpression[updnStr <> "a"] ]
                  ];
                  ToExpression[updnStr <> extraStr <> strIndexL[[1]]]
               ]
(********************************************************************)

(*************************** for Dnup *******************************)
(***** Exported Functions *****)
(* Dnup tag *)
(* dn and up contracted pairs, where covOpL are considered *)
Dnup[expr_, covOpL_:defaultCDTable] := forEachTerm[ExpandIAtom[expr], dnupTerm, covOpL]
Dnup[___]                           := Message[Dnup::usage]

(***** Local Functions *****)
(* canDnupQ tag *)
canDnupQ[(oName_?DefinedTensorQ)[___], _, _]                                := True   (* T[ua,la] *)
canDnupQ[(opName_?DefinedOperatorQ)[args__], covOpL_, {dnIndex_, upIndex_}] :=        (* CD[ua, T[la,lb] *)
   Module[ {flatL, nLen, dnPos, upPos},
      flatL = flattenObject[opName[args]];  (* CD[a, BD[b, T[indices]]] -> {{CD, a}, {BD, b}, T, {indices}} *)
      opLen = Length[flatL] - 2;

      dnPos = Position[flatL, dnIndex]; If [ dnPos === {}, Return[False] ];
      upPos = Position[flatL, upIndex]; If [ upPos === {}, Return[False] ];
      If [dnPos[[1,1]] <= opLen, dnPos = dnPos[[1,1]], dnPos = opLen + 1];
      If [upPos[[1,1]] <= opLen, upPos = upPos[[1,1]], upPos = opLen + 1];

      If [ dnPos > opLen && upPos > opLen, Return[True] ];    (* inside contraction *)
      If [ dnPos < upPos, {dnPos, upPos} = {upPos, dnPos} ];  (* swap *)

      (* get opNames upPos through dnPos - 1. Are they all contained in covOpL? *)
      AllQ[Map[MemberQ[covOpL, #]&,
               Map[#[[1]]&,
                   Take[flatL, {upPos, dnPos - 1}]]]]
   ]
canDnupQ[(opName_?DefinedOperatorQ)[args__], covOpL_, aIndex_?AtomQ]        :=
(* CD[la, T[lb, lc]] T[ub,ud] *)
   Module[ {flatL, pos, opLen},
      flatL = flattenObject[opName[args]];
      pos = Position[flatL, aIndex]; If [ pos === {}, Return[False] ];
      opLen = Length[flatL] - 2;
      If [pos[[1,1]] <= opLen, pos = pos[[1,1]], pos = opLen + 1];

      (* get opNames before pos - 1. Are they all contained in covOpL *)
      AllQ[Map[MemberQ[covOpL, #]&,
               Map[#[[1]]&,
                   Take[flatL, {1, pos - 1}]]]]
   ]

(* dnupTerm tag *)
dnupTerm[term_, covOpL_] :=
   Module[ {ordTerm, oTerm},
      {ordTerm, oTerm} = splitTerm[term];

      ordTerm * dnupTermAux[oTerm, covOpL]
   ]

   (* where *)
   dnupTermAux[expr:(oName_?DefinedNameQ)[___], covOpL_] := dnupObject[expr, covOpL]
   dnupTermAux[oTerm_,                          covOpL_] :=
      Module[ {pairL, rc, dnPos, upPos, i},
         pairL = GetPairs @ GetIndices[oTerm];  (* contracted pairs in the format {dn, up} *)
         If [ pairL === {}, Return[oTerm] ];

         rc = oTerm;
         For [i = 1, i <= Length[pairL], i++,
            dnPos = Position[oTerm, pairL[[i,1]]]; upPos = Position[oTerm, pairL[[i,2]]];
            If [dnPos[[1,1]] === upPos[[1,1]],  (* inside contraction of an object *)
               If [canDnupQ[oTerm[[dnPos[[1,1]]]], covOpL, pairL[[i]]],
                  (* if up-dn pair *)
                  If [Position[getOriginalIndicesTerm[oTerm[[dnPos[[1,1]]]]], pairL[[i,1]]] [[1,1]]
                      > Position[getOriginalIndicesTerm[oTerm[[dnPos[[1,1]]]]], pairL[[i,2]]] [[1,1]],
                     rc = rc /. MapThread[Rule, {pairL[[i]], FlipIndex /@ pairL[[i]]}]
                  ]
               ],
            (* else *)
               If [dnPos[[1,1]] > upPos[[1,1]], (* if upObject * dnObject *)
                  If [canDnupQ[Part[oTerm, {dnPos[[1,1]]}], covOpL, pairL[[i,1]]]  \
                      && canDnupQ[Part[oTerm, {upPos[[1,1]]}], covOpL, pairL[[i,2]]],
                     rc = rc /. MapThread[Rule, {pairL[[i]], FlipIndex /@ pairL[[i]]}]
                  ]
               ]
            ]
         ];
         rc
      ]

      (* where *)
      dnupObject[obj_, covOpL_] :=
         Module[ {pairL, rc, i},
            (* contracted pairs in the format {dn, up} *)
            pairL = GetPairs @ GetIndices[obj];
            If [ pairL === {}, Return[obj] ];

            rc = obj;
            For [i = 1, i <= Length[pairL], i++,
               If [canDnupQ[obj, covOpL, pairL[[i]]],
                  (* if up-dn pair *)
                  If [Position[getOriginalIndicesTerm[obj], pairL[[i,1]]] [[1,1]]
                      > Position[getOriginalIndicesTerm[obj], pairL[[i,2]]] [[1,1]],
                     rc = rc /. MapThread[Rule, {pairL[[i]], FlipIndex /@ pairL[[i]]}]
                  ]
               ]
            ];
            rc
         ]

(* flattenObject tag *)
(* flatten an indical object: CD[a, BD[b, T[...]]] -> {{CD, a}, {BD, b}, T, ...} *)
flattenObject[(oName_?DefinedTensorQ)[args___]]           := {oName, {args}}
flattenObject[(opName_?DefinedOperatorQ)[args_, expr___]] :=
   Switch [getOpType[opName],
      CD, Join[{{opName, args}}, flattenObject[expr]],
      LD, Join[{{opName, args}}, flattenObject[expr]],
      XD, Join[{{opName, {}}},   flattenObject[args]],
      XP, {{opName, args, expr}, opName, {}},           (* STRANGE? *)
      TP, {{opName, Times @@ args, expr}, opName, {}}   (* STRANGE? *)
   ]
(********************************************************************)

(************************** for Absorb ******************************)
(***** Exported Functions *****)
(* Absorb tag *)
Absorb[expr_, aMetric_Symbol?DefinedTensorQ, covOpL_:defaultCDTable] :=
   Module[ {rcExpr},
      (* is aMetric a metric-type tensor? *)
      If [getNameRank[aMetric] =!= 2,  (* need not be symmetric, e.g. in Spinor calculus! *)
         Message[Msg::err, "not 2nd rank metric:", aMetric, "", ""]; Return[]
      ];

      If [!AllQ[covOpL, DefinedOperatorQ],
         Message[Msg::err, "not defined operator(s)", covOpL, "", ""]; Return[]
      ];

      absorb[expr, aMetric, covOpL]
    ]
Absorb[___]                                                          := Message[Absorb::usage]

(***** Local Functions *****)
(* absorb tag *)
absorb[expr_, aMetric_, covOpL_] :=
   Module[ {rcExpr},
      rcExpr = WrapIAtom[expr, tpListToTimes];                        (* TP[_List, _] -> $FORTP[_Times, _] *)
      rcExpr = ExpandIAtom[rcExpr] //. absorbRuleWith[aMetric, covOpL];
      rcExpr /. { $FORTP[arg_Times, n_, tpName_] :> tpName[List @@ arg, n],
                  $FORTP[arg_,      n_, tpName_] :> tpName[arg, n] }  (* TP[_Times, _] -> $FORTP[_List, _] *)
   ]

(* absorbRuleWith tag *)
absorbRuleWith[aMetric_, covOpL_] :=  (* aMetric is covariantly constant under opNames in covOpL *)
{
   (* g_{ab} T^b *)
   aMetric[a_, b_] * oName_[pre___, c_, post___]                        :>
      oName[pre, a, post] /; canPairQ[oName, c, b],
   (* g_{ab} T^a *)
   aMetric[a_, b_] * oName_[pre___, c_, post___]                        :>
      oName[pre, b, post] /; canPairQ[oName, c, a],
   (* g_{ab} \pd^b T *)
   aMetric[a_, b_] * opName_[c_, expr_]                                 :>
      opName[a, expr] /; canPairQ[opName, c, b],
   (* g_{ab} \pd^a T *)
   aMetric[a_, b_] * opName_[c_, expr_]                                 :>
      opName[b, expr] /; canPairQ[opName, c, a],
   (* g_{ab} XD[A^a] *)
   aMetric[a_, b_] * opName_[expr_]                                     :>
      opName[aMetric[a, b] * expr //. absorbRuleWith[aMetric, covOpL]]  \
      /; DefinedOperatorQ[opName] && getOpType[opName] === XD && canAbsorbQ[expr, covOpL, {a, b}],
   aMetric[a_, b_] * opName_[pre___, expr_, post___]                    :>
      opName[pre, aMetric[a, b] * expr //. absorbRuleWith[aMetric, covOpL], post]  \
      /; DefinedOperatorQ[opName] && getOpType[opName] === XP && canAbsorbQ[expr, covOpL, {a, b}],
   (* g_{ab} \del^c T^b *)
   Times[preT___, aMetric[a_, b_], mT___, opName_[c_, expr_], postT___] :>
      Times[preT, mT, opName[c, aMetric[a, b] * expr //. absorbRuleWith[aMetric, covOpL]], postT]  \
      /; DefinedOperatorQ[opName] && MemberQ[covOpL, opName]  \
         && !canPairQ[opName, c, a] && !canPairQ[opName, c, b] && canAbsorbQ[expr, covOpL, {a, b}]
}

   (* where *)
   canPairQ[oName_?DefinedTensorQ,    c_, a_?TensorialIndexQ] := True /; PairQ[a,c]
   canPairQ[opName_?DefinedOperatorQ, c_, a_?TensorialIndexQ] := True /; getOpType[opName] === CD && PairQ[a,c]

   (* is raising or lowering possible for indices a and b *)
   canAbsorbQ[(oName_?DefinedTensorQ)[args___],     _,       {a_, b_}] :=
      If [Or @@ (Flatten @ Outer[canPairQ[oName, #1, #2]&, {args}, {a, b}]),
         True,
      (* else *)
         False
      ]
   canAbsorbQ[(opName_?DefinedOperatorQ)[c_, _],    _,       {a_, b_}] :=
      True /; canPairQ[opName, c, a] || canPairQ[opName, c, b]
   canAbsorbQ[(opName_?DefinedOperatorQ)[__],       covOpL_, {a_, b_}] :=
      False /; !MemberQ[covOpL, opName]
   canAbsorbQ[(opName_?DefinedOperatorQ)[_, expr_], covOpL_, {a_, b_}] :=
      canAbsorbQ[expr, covOpL, {a, b}] /; MemberQ[covOpL, opName]
   canAbsorbQ[(opName_?DefinedOperatorQ)[expr_],    covOpL_, {a_, b_}] :=
      canAbsorbQ[expr, covOpL, {a, b}] /; MemberQ[covOpL, opName] && getOpType[opName] === XD
   canAbsorbQ[expr_, covOpL_, {a_, b_}]                                := False

(* tpListToTimes tag *)
(* List to Times in TP-type for proper operation *)
tpListToTimes[(opName_?DefinedOperatorQ)[arg_List, n_]] :=
   If [getOpType[opName] === TP,
      $FORTP[Times @@ arg, n, opName],
   (* else *)
      opName[arg, n]
   ]
tpListToTimes[expr__]                                   := expr
(********************************************************************)

(**************************** for Can *******************************)
(***** Exported Functions *****)
(* Can tag *)
(* canonicalize indical objects in expr *)
Can[expr_, covOpL_:defaultCDTable] := canExpr[expr, covOpL] /; FreeQ[expr, Pattern | Blank]
Can[expr__]                        := expr  (* for Sequence input *)

(***** Local Functions *****)
(* canExpr tag *)
canExpr[expr_, covOpL_List] :=
   Module[ {rcExpr = Dum[ExpandIAtom[expr]]},
      If [iFlagTable[SyntaxCheckFlag],
         If [ SyntaxCheck[rcExpr] === False, Return[rcExpr] ]  (* check consistency for indices *)
      ];

      (* String to Expr of OneIndex for simplication *)
      rcExpr = symToStrOneIndex[rcExpr];

      (* re-labelling dummy indices and then sorting indices *)
      forEachIAtom[forEachTerm[rcExpr, canTerm, covOpL], orderingIndicesObject]
   ]
canExpr[expr_, _]           := expr

   (* where *)
   symToStrOneIndex[expr_] := expr  (* See IndexNotation *)

   canTerm[term_, covOpL_] := dnupTerm[resetDummiesTerm[term], covOpL]
(********************************************************************)
