(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     45610,       1240]*)
(*NotebookOutlinePosition[     46378,       1267]*)
(*  CellTagsIndexPosition[     46334,       1263]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["2-D SpaceTime (Coordinate Basis)", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {t, x};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(metric\  = \ {{\(-E\^\(2  \[Rho][t, x]\)\), 0}, {0, 
          E\^\(2  \[Rho][t, x]\)}}; \ MatrixForm[metric]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(-\[ExponentialE]\^\(2\ \[Rho][t, x]\)\), "0"},
            {"0", \(\[ExponentialE]\^\(2\ \[Rho][t, x]\)\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensor", "Subsubsection"],

Cell[BoxData[
    \(InitCTensor[coSys, \ metric]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\[ExponentialE]\^\(2\ \[Rho][t, x]\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ 2, 
              \[Rho][ t, x]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\[ExponentialE]\^\(2\ \[Rho][t, x]\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            dhPark`Tensor`Private`UnderBar[ "x"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Power[ E, 
          Times[ 2, 
            \[Rho][ t, x]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\("f"\_\("ab"\)\%\("  c"\)\),
          Subsuperscript[ "f", "ab", "  c"],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "\<\"0\"\>"}],
      SequenceForm[ 
        Subsuperscript[ "f", "ab", "  c"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((1, 0)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Derivative[ 1, 0][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "x"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((0, 1)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "x"]]], " = ", 
        Derivative[ 0, 1][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((0, 1)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Derivative[ 0, 1][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "x"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((1, 0)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "x"]]], " = ", 
        Derivative[ 1, 0][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "x"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "x"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((1, 0)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Derivative[ 1, 0][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "x"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "x"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "x"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((0, 1)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "x"]]], " = ", 
        Derivative[ 0, 1][ \[Rho]][ t, x]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "x"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "x"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(2\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "x"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ 2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{"-", 
            RowBox[{
              SuperscriptBox["\[Rho]", 
                TagBox[\((0, 2)\),
                  Derivative],
                MultilineFunction->None], "[", \(t, x\), "]"}]}], "+", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((2, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Plus[ 
          Times[ -1, 
            Derivative[ 0, 2][ \[Rho]][ t, x]], 
          Derivative[ 2, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 2)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((2, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            dhPark`Tensor`Private`UnderBar[ "x"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Plus[ 
          Derivative[ 0, 2][ \[Rho]][ t, x], 
          Times[ -1, 
            Derivative[ 2, 0][ \[Rho]][ t, x]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
        "2", " ", \(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 
        Times[ 2, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Laplacian", "Subsubsection"],

Cell[BoxData[
    \(Tdefine[S, "\<\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(lap\  = \ Metricg[ua, ub] CD[la, lb, S[]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[\(S\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ S, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[Del]", "b"], 
                S[ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[Del]", "a"], 
            CD[ lb, 
              S[ ]]],
          Editable->False], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(CDtoAffine[CDtoAffine[%]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"-", 
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(S\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ S, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              S[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(S\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ S, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "b"], 
                  S[ ]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lb, 
                S[ ]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(SumDum[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(S\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ S, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 2], 
                  S[ ]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", 2], 
              BD[ -2, 
                S[ ]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(S\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ S, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  S[ ]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", 1], 
              BD[ -1, 
                S[ ]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, ",
  Cell[BoxData[
      \(TraditionalForm\`R\  = \ \(2\ \[PartialD]\^2 \[Rho]\  = \ 
          2\ \(e\^\(\(-2\) \[Rho]\)\) \[Del]\&\[RightVector]\^2 \[Rho]\)\)]]
}], "Text"],

Cell[TextData[{
  "For a symmetric space ",
  Cell[BoxData[
      \(TraditionalForm\`R\  = \ \(-2\)\ k\)]],
  ", where ",
  Cell[BoxData[
      \(TraditionalForm\`k\)]],
  " is a constant"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Laplacian[\(-Log[1\  + \ \(k\/4\) \((\(-t\^2\)\  + \ x\^2)\)]\), 
      Simplify]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(16\ \[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\ k\)\/\((4 + k\ \
\((\(-t\^2\) + x\^2)\))\)\^2\)\)\)], "Output"]
}, Open  ]],

Cell["Therefore, the metric in symmetric space is given by", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Rho][t, 
        x]\  = \ \(-Log[
          1\  + \ \(k\/4\) \((\(-t\^2\)\  + \ x\^2)\)]\)\)]]], "Text",
  TextAlignment->Center]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{566, 679},
WindowMargins->{{11, Automatic}, {Automatic, 2}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 65, 1, 43, "Title"],
Cell[1785, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1862, 57, 53, 0, 54, "Subsubsection"],
Cell[1918, 59, 56, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1999, 64, 142, 2, 81, "Input"],
Cell[2144, 68, 264, 6, 93, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2457, 80, 59, 0, 32, "Subsubsection"],
Cell[2519, 82, 61, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2617, 88, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[2679, 92, 46, 1, 56, "Input"],
Cell[2728, 95, 184, 4, 30, "Print"],
Cell[2915, 101, 1795, 45, 33, "Print"],
Cell[4713, 148, 1767, 44, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6517, 197, 49, 1, 56, "Input"],
Cell[6569, 200, 350, 8, 31, "Print"],
Cell[6922, 210, 372, 9, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7331, 224, 46, 1, 56, "Input"],
Cell[7380, 227, 392, 8, 31, "Print"],
Cell[7775, 237, 2906, 70, 33, "Print"],
Cell[10684, 309, 2906, 70, 32, "Print"],
Cell[13593, 381, 2906, 70, 33, "Print"],
Cell[16502, 453, 2906, 70, 32, "Print"],
Cell[19411, 525, 2906, 70, 33, "Print"],
Cell[22320, 597, 2906, 70, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25263, 672, 47, 1, 56, "Input"],
Cell[25313, 675, 293, 6, 29, "Print"],
Cell[25609, 683, 4640, 114, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[30286, 802, 45, 1, 56, "Input"],
Cell[30334, 805, 184, 4, 29, "Print"],
Cell[30521, 811, 2212, 57, 33, "Print"],
Cell[32736, 870, 2176, 56, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34949, 931, 46, 1, 56, "Input"],
Cell[34998, 934, 1151, 32, 30, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[36198, 972, 34, 0, 32, "Subsubsection"],
Cell[36235, 974, 51, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[36311, 979, 74, 1, 56, "Input"],
Cell[36388, 982, 1590, 42, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[38015, 1029, 58, 1, 56, "Input"],
Cell[38076, 1032, 4844, 119, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[42957, 1156, 42, 1, 56, "Input"],
Cell[43002, 1159, 1623, 41, 55, "Output"]
}, Open  ]],
Cell[44640, 1203, 204, 5, 41, "Text"],
Cell[44847, 1210, 200, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[45072, 1222, 114, 2, 72, "Input"],
Cell[45189, 1226, 137, 2, 76, "Output"]
}, Open  ]],
Cell[45341, 1231, 68, 0, 41, "Text"],
Cell[45412, 1233, 182, 4, 42, "Text"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

