(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     47679,       1311]*)
(*NotebookOutlinePosition[     48447,       1338]*)
(*  CellTagsIndexPosition[     48403,       1334]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Impulsive Plane Gravitational Waves", "Title",
  FontSize->24],

Cell["\<\
See: Introducing Einstein's Relativity, Ray D'Inverno, Chap 20, pp \
280\
\>", "Text"],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Exact plane gravitational waves", "Subsection"],

Cell["\<\
We assume the line element (Rosen form) in double null coordinates:\
\>", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-2\) 
          d\[InvisibleSpace]u\ d\[InvisibleSpace]v\  + \ \(\(f\^2\)(
            u)\)\ d\[InvisibleSpace]y\^2\  + \ \(\(g\^2\)(
            u)\)\ d\[InvisibleSpace]z\^2\)]]], "Text",
  TextAlignment->Center],

Cell[BoxData[
    \(coSys = {u, v, y, z}; 
    metric = {{0, \(-1\), 0, 0}, {\(-1\), 0, 0, 0}, {0, 0, f[u]\^2, 0}, {0, 
          0, 0, g[u]\^2}};\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[coSys, metric]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{\(f[u]\^2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(y\^2\),
                    Superscript[ y, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ y, 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(g[u]\^2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(z\^2\),
                    Superscript[ z, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ z, 2]],
                Editable->False], ")"}]}], "-", 
          RowBox[{"2", " ", 
            RowBox[{"(", 
              
              InterpretationBox[\("d"\[InvisibleSpace]u\[InvisibleSpace]"d"\
\[InvisibleSpace]v\),
                SequenceForm[ "d", u, "d", v],
                Editable->False], ")"}]}]}]}],
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          Times[ 
            Power[ 
              f[ u], 2], 
            SequenceForm[ "d", 
              Superscript[ y, 2]]], 
          Times[ 
            Power[ 
              g[ u], 2], 
            SequenceForm[ "d", 
              Superscript[ z, 2]]], 
          Times[ -2, 
            SequenceForm[ "d", u, "d", v]]]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(InitCTensor[coSys, metric]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"u"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "u"]],
                      Editable->False], "\[InvisibleSpace]", \("y"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "u"]], 
                    dhPark`Tensor`Private`UnderBar[ "y"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "u"]], 
                  dhPark`Tensor`Private`UnderBar[ "y"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("y"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "y"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "y"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "y"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{
            SuperscriptBox["f", "\[Prime]",
              MultilineFunction->None], "[", "u", "]"}], \(f[u]\)]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "u"]], 
              dhPark`Tensor`Private`UnderBar[ "y"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "y"]]], " = ", 
        Times[ 
          Power[ 
            f[ u], -1], 
          Derivative[ 1][ f][ u]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"u"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "u"]],
                      Editable->False], "\[InvisibleSpace]", \("z"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "u"]], 
                    dhPark`Tensor`Private`UnderBar[ "z"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "u"]], 
                  dhPark`Tensor`Private`UnderBar[ "z"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("z"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "z"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "z"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "z"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{
            SuperscriptBox["g", "\[Prime]",
              MultilineFunction->None], "[", "u", "]"}], \(g[u]\)]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "u"]], 
              dhPark`Tensor`Private`UnderBar[ "z"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "z"]]], " = ", 
        Times[ 
          Power[ 
            g[ u], -1], 
          Derivative[ 1][ g][ u]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"y"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "y"]],
                      Editable->False], "\[InvisibleSpace]", \("y"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "y"]], 
                    dhPark`Tensor`Private`UnderBar[ "y"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "y"]], 
                  dhPark`Tensor`Private`UnderBar[ "y"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("v"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "v"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "y"]], 
                dhPark`Tensor`Private`UnderBar[ "y"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "v"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(f[u]\), " ", 
          RowBox[{
            SuperscriptBox["f", "\[Prime]",
              MultilineFunction->None], "[", "u", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "y"]], 
              dhPark`Tensor`Private`UnderBar[ "y"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "v"]]], " = ", 
        Times[ 
          f[ u], 
          Derivative[ 1][ f][ u]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"z"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "z"]],
                      Editable->False], "\[InvisibleSpace]", \("z"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "z"]], 
                    dhPark`Tensor`Private`UnderBar[ "z"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "z"]], 
                  dhPark`Tensor`Private`UnderBar[ "z"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("v"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "v"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "z"]], 
                dhPark`Tensor`Private`UnderBar[ "z"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "v"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(g[u]\), " ", 
          RowBox[{
            SuperscriptBox["g", "\[Prime]",
              MultilineFunction->None], "[", "u", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "z"]], 
              dhPark`Tensor`Private`UnderBar[ "z"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "v"]]], " = ", 
        Times[ 
          g[ u], 
          Derivative[ 1][ g][ u]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"u"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "u"]],
                          Editable->False], "\[InvisibleSpace]", \("y"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "u"]], 
                        dhPark`Tensor`Private`UnderBar[ "y"]],
                      Editable->False], "\[InvisibleSpace]", \("u"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "u"]], 
                      dhPark`Tensor`Private`UnderBar[ "y"]], 
                    dhPark`Tensor`Private`UnderBar[ "u"]],
                  Editable->False], "\[InvisibleSpace]", \("y"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "u"]], 
                    dhPark`Tensor`Private`UnderBar[ "y"]], 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "y"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "u"]], 
                  dhPark`Tensor`Private`UnderBar[ "y"]], 
                dhPark`Tensor`Private`UnderBar[ "u"]], 
              dhPark`Tensor`Private`UnderBar[ "y"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(f[u]\), " ", 
          RowBox[{
            SuperscriptBox["f", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "u", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "y"]], 
              dhPark`Tensor`Private`UnderBar[ "u"]], 
            dhPark`Tensor`Private`UnderBar[ "y"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          f[ u], 
          Derivative[ 2][ f][ u]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"u"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "u"]],
                          Editable->False], "\[InvisibleSpace]", \("z"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "u"]], 
                        dhPark`Tensor`Private`UnderBar[ "z"]],
                      Editable->False], "\[InvisibleSpace]", \("u"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "u"]], 
                      dhPark`Tensor`Private`UnderBar[ "z"]], 
                    dhPark`Tensor`Private`UnderBar[ "u"]],
                  Editable->False], "\[InvisibleSpace]", \("z"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "u"]], 
                    dhPark`Tensor`Private`UnderBar[ "z"]], 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "z"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "u"]], 
                  dhPark`Tensor`Private`UnderBar[ "z"]], 
                dhPark`Tensor`Private`UnderBar[ "u"]], 
              dhPark`Tensor`Private`UnderBar[ "z"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(g[u]\), " ", 
          RowBox[{
            SuperscriptBox["g", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "u", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "z"]], 
              dhPark`Tensor`Private`UnderBar[ "u"]], 
            dhPark`Tensor`Private`UnderBar[ "z"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          g[ u], 
          Derivative[ 2][ g][ u]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"u"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "u"]],
                  Editable->False], "\[InvisibleSpace]", \("u"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "u"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "u"]], 
              dhPark`Tensor`Private`UnderBar[ "u"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{
            RowBox[{\(g[u]\), " ", 
              RowBox[{
                SuperscriptBox["f", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "u", "]"}]}], "+", 
            RowBox[{\(f[u]\), " ", 
              RowBox[{
                SuperscriptBox["g", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "u", "]"}]}]}], \(f[u]\ g[
              u]\)]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "u"]], 
            dhPark`Tensor`Private`UnderBar[ "u"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ 
            f[ u], -1], 
          Power[ 
            g[ u], -1], 
          Plus[ 
            Times[ 
              g[ u], 
              Derivative[ 2][ f][ u]], 
            Times[ 
              f[ u], 
              Derivative[ 2][ g][ u]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
  "Therefore, with ",
  Cell[BoxData[
      \(TraditionalForm\`\(f'\)'\/f\  = \ h\)]],
  " the vacuum field equation becomes ",
  Cell[BoxData[
      \(TraditionalForm\`\(g'\)'\/g\  = \ \(-h\)\)]],
  ". Hence any choice of the arbitrary function ",
  Cell[BoxData[
      \(TraditionalForm\`h[u]\)]],
  " gives rise to a vacuum solution. Such exact solutions are called linearly \
polarized plane gravitational waves."
}], "Text"],

Cell["We carry out the coordinate transformation", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`U\  = \ u, \ \ \ \ \ V\  = \ 
      v\  + \ \(y\^2\) f\ f'\  + \ z\^2\ g\ g', \ \ \ \ \ Y\  = \ 
      f\ y, \ \ \ \ \ Z\  = \ g\ z\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(\(newCoSys = {u, 
          v + \ y\^2\/2\ f[u]\ \(f'\)[u]\  + \ z\^2\/2\ g[u]\ \(g'\)[u], 
          f[u] y, g[u] z};\)\)], "Input"],

Cell[BoxData[
    \(Tdefine[newg, "\<ba\>"]\)], "Input"],

Cell[BoxData[
    \(Ttransform[Metricg[la, lb], newg, coSys, newCoSys]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[
      newg[la, lb], {la, \(-1\), \(-4\), \(-1\)}, {lb, \(-1\), \(-4\), \
\(-1\)}]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{
            RowBox[{
              RowBox[{\(y\^2\), " ", \(f[u]\), " ", 
                RowBox[{
                  SuperscriptBox["f", "\[Prime]\[Prime]",
                    MultilineFunction->None], "[", "u", "]"}]}], "+", 
              RowBox[{\(z\^2\), " ", \(g[u]\), " ", 
                RowBox[{
                  SuperscriptBox["g", "\[Prime]\[Prime]",
                    MultilineFunction->None], "[", "u", "]"}]}]}], 
            ",", \(-1\), ",", "0", ",", "0"}], "}"}], 
        ",", \({\(-1\), 0, 0, 0}\), ",", \({0, 0, 1, 0}\), 
        ",", \({0, 0, 0, 1}\)}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[{U, V, Y, Z}, %]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[\(Y\^2\),
                Superscript[ Y, 2],
                Editable->False]}],
            SequenceForm[ "d", 
              Superscript[ Y, 2]],
            Editable->False], "+", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[\(Z\^2\),
                Superscript[ Z, 2],
                Editable->False]}],
            SequenceForm[ "d", 
              Superscript[ Z, 2]],
            Editable->False], "-", 
          RowBox[{"2", " ", 
            RowBox[{"(", 
              
              InterpretationBox[\("d"\[InvisibleSpace]U\[InvisibleSpace]"d"\
\[InvisibleSpace]V\),
                SequenceForm[ "d", U, "d", V],
                Editable->False], ")"}]}], "+", 
          RowBox[{
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(U\^2\),
                    Superscript[ U, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ U, 2]],
                Editable->False], ")"}], " ", 
            RowBox[{"(", 
              RowBox[{
                RowBox[{\(y\^2\), " ", \(f[u]\), " ", 
                  RowBox[{
                    SuperscriptBox["f", "\[Prime]\[Prime]",
                      MultilineFunction->None], "[", "u", "]"}]}], "+", 
                RowBox[{\(z\^2\), " ", \(g[u]\), " ", 
                  RowBox[{
                    SuperscriptBox["g", "\[Prime]\[Prime]",
                      MultilineFunction->None], "[", "u", "]"}]}]}], 
              ")"}]}]}]}],
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          SequenceForm[ "d", 
            Superscript[ Y, 2]], 
          SequenceForm[ "d", 
            Superscript[ Z, 2]], 
          Times[ -2, 
            SequenceForm[ "d", U, "d", V]], 
          Times[ 
            SequenceForm[ "d", 
              Superscript[ U, 2]], 
            Plus[ 
              Times[ 
                Power[ y, 2], 
                f[ u], 
                Derivative[ 2][ f][ u]], 
              Times[ 
                Power[ z, 2], 
                g[ u], 
                Derivative[ 2][ g][ u]]]]]],
      Editable->False]], "Output"]
}, Open  ]],

Cell["The line element is transformed into the Brinkmann form", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-\(h(
              U)\)\) \((Z\^2\  - \ Y\^2)\) d\[InvisibleSpace]U\^2\  - 
        2  d\[InvisibleSpace]U\ d\[InvisibleSpace]V\  + \ 
        d\[InvisibleSpace]Y\^2\  + \ d\[InvisibleSpace]Z\^2\)]]], "Text",
  TextAlignment->Center],

Cell[TextData[{
  "This shows the explicit dependence on the freely specifiable function ",
  Cell[BoxData[
      \(TraditionalForm\`h(U)\)]],
  "."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Impulsive plane gravitational waves", "Subsection"],

Cell["We consider a line element in the Rosen form defined by", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`f(u)\  = \ 
      1\  + \ u\ \(\[Theta](u)\), \ \ \ \ \ g(u)\  = \ 
      1\  - \ u\ \(\[Theta](u)\)\)]]], "Text",
  TextAlignment->Center],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Theta](u)\)]],
  " is the Heaviside step function."
}], "Text"],

Cell[BoxData[
    \(Unprotect[UnitStep]; Format[UnitStep[u_]]\  := \ \[Theta][u]; 
    Protect[UnitStep];\)], "Input"],

Cell[BoxData[
    \(Unprotect[DiracDelta]; Format[DiracDelta[u_]]\  := \ \[Delta][u]; 
    Protect[DiracDelta];\)], "Input"],

Cell[BoxData[
    \(f[u_]\  := \ 1\  + \ u\ UnitStep[u]; 
    g[u_]\  := \ 1\  - \ u\ UnitStep[u];\)], "Input"],

Cell[BoxData[
    \(\(CsimplifyRules\  = \ {u^_. \ DiracDelta[u]\  -> \ 0, \ 
          u^n_. \ \(DiracDelta'\)[u]\  -> \ \(-u^\((n - 1)\)\)\ DiracDelta[
                u]};\)\)], "Input"],

Cell[BoxData[
    \(InitCTensor[coSys, metric]\)], "Input"],

Cell["The Ricci tensor vanishes, but the Riemann tensor does not.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("R"\_"ab"\[InvisibleSpace]" = "\[InvisibleSpace]"0"\),
      
      SequenceForm[ 
        Subscript[ "R", "ab"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"u"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "u"]],
                          Editable->False], "\[InvisibleSpace]", \("y"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "u"]], 
                        dhPark`Tensor`Private`UnderBar[ "y"]],
                      Editable->False], "\[InvisibleSpace]", \("u"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "u"]], 
                      dhPark`Tensor`Private`UnderBar[ "y"]], 
                    dhPark`Tensor`Private`UnderBar[ "u"]],
                  Editable->False], "\[InvisibleSpace]", \("y"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "u"]], 
                    dhPark`Tensor`Private`UnderBar[ "y"]], 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "y"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "u"]], 
                  dhPark`Tensor`Private`UnderBar[ "y"]], 
                dhPark`Tensor`Private`UnderBar[ "u"]], 
              dhPark`Tensor`Private`UnderBar[ "y"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\[Delta][u]\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "y"]], 
              dhPark`Tensor`Private`UnderBar[ "u"]], 
            dhPark`Tensor`Private`UnderBar[ "y"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        DiracDelta[ u]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"u"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "u"]],
                          Editable->False], "\[InvisibleSpace]", \("z"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "u"]], 
                        dhPark`Tensor`Private`UnderBar[ "z"]],
                      Editable->False], "\[InvisibleSpace]", \("u"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "u"]], 
                      dhPark`Tensor`Private`UnderBar[ "z"]], 
                    dhPark`Tensor`Private`UnderBar[ "u"]],
                  Editable->False], "\[InvisibleSpace]", \("z"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "u"]], 
                    dhPark`Tensor`Private`UnderBar[ "z"]], 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "z"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "u"]], 
                  dhPark`Tensor`Private`UnderBar[ "z"]], 
                dhPark`Tensor`Private`UnderBar[ "u"]], 
              dhPark`Tensor`Private`UnderBar[ "z"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\[Delta][
                u]\/\((\(-1\) + u\ \[Theta][u])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "u"]], 
                dhPark`Tensor`Private`UnderBar[ "z"]], 
              dhPark`Tensor`Private`UnderBar[ "u"]], 
            dhPark`Tensor`Private`UnderBar[ "z"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, 
          DiracDelta[ u], 
          Power[ 
            Plus[ -1, 
              Times[ u, 
                UnitStep[ u]]], -2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
  "The solution has delta functions in the curvature and hence it is non-flat \
only when ",
  Cell[BoxData[
      \(TraditionalForm\`u\  = \ 0\)]],
  "."
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[TextData[{
  "In Brinkmann form, ",
  Cell[BoxData[
      \(TraditionalForm\`h(U)\  = \ \[Delta](U)\)]],
  ". Therefore, for ",
  Cell[BoxData[
      \(TraditionalForm\`U\  \[NotEqual] \ 0\)]],
  ", the line element reduces to the Minkowski space-time in double null \
coordinates. The hypersurface ",
  Cell[BoxData[
      \(TraditionalForm\`U\  = \ 0\)]],
  ", where the curvature is concentrated, separates two flat reions."
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[
      newg[la, lb], {la, \(-1\), \(-4\), \(-1\)}, {lb, \(-1\), \(-4\), \
\(-1\)}]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{
            RowBox[{
              RowBox[{\(z\^2\), " ", \((1 - u\ \[Theta][u])\), " ", 
                RowBox[{"(", 
                  RowBox[{\(\(-2\)\ \[Delta][u]\), "-", 
                    RowBox[{"u", " ", 
                      RowBox[{
                        SuperscriptBox["DiracDelta", "\[Prime]",
                          MultilineFunction->None], "[", "u", "]"}]}]}], 
                  ")"}]}], "+", 
              RowBox[{\(y\^2\), " ", \((1 + u\ \[Theta][u])\), " ", 
                RowBox[{"(", 
                  RowBox[{\(2\ \[Delta][u]\), "+", 
                    RowBox[{"u", " ", 
                      RowBox[{
                        SuperscriptBox["DiracDelta", "\[Prime]",
                          MultilineFunction->None], "[", "u", "]"}]}]}], 
                  ")"}]}]}], ",", \(-1\), ",", "0", ",", "0"}], "}"}], 
        ",", \({\(-1\), 0, 0, 0}\), ",", \({0, 0, 1, 0}\), 
        ",", \({0, 0, 0, 1}\)}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% // Csimplify\)], "Input"],

Cell[BoxData[
    \({{y\^2\ \[Delta][u] - z\^2\ \[Delta][u], \(-1\), 0, 0}, {\(-1\), 0, 0, 
        0}, {0, 0, 1, 0}, {0, 0, 0, 1}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[{U, V, Y, Z}, %]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{\((y\^2\ \[Delta][u] - z\^2\ \[Delta][u])\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(U\^2\),
                    Superscript[ U, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ U, 2]],
                Editable->False], ")"}]}], "+", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[\(Y\^2\),
                Superscript[ Y, 2],
                Editable->False]}],
            SequenceForm[ "d", 
              Superscript[ Y, 2]],
            Editable->False], "+", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[\(Z\^2\),
                Superscript[ Z, 2],
                Editable->False]}],
            SequenceForm[ "d", 
              Superscript[ Z, 2]],
            Editable->False], "-", 
          RowBox[{"2", " ", 
            RowBox[{"(", 
              
              InterpretationBox[\("d"\[InvisibleSpace]U\[InvisibleSpace]"d"\
\[InvisibleSpace]V\),
                SequenceForm[ "d", U, "d", V],
                Editable->False], ")"}]}]}]}],
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          Times[ 
            Plus[ 
              Times[ 
                Power[ y, 2], 
                DiracDelta[ u]], 
              Times[ -1, 
                Power[ z, 2], 
                DiracDelta[ u]]], 
            SequenceForm[ "d", 
              Superscript[ U, 2]]], 
          SequenceForm[ "d", 
            Superscript[ Y, 2]], 
          SequenceForm[ "d", 
            Superscript[ Z, 2]], 
          Times[ -2, 
            SequenceForm[ "d", U, "d", V]]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{717, 686},
WindowMargins->{{17, Automatic}, {Automatic, 4}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 68, 1, 43, "Title"],
Cell[1788, 52, 96, 3, 41, "Text"],
Cell[1887, 57, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1964, 62, 53, 0, 66, "Subsection"],
Cell[2020, 64, 91, 2, 41, "Text"],
Cell[2114, 68, 295, 5, 35, "Text"],
Cell[2412, 75, 159, 3, 104, "Input"],

Cell[CellGroupData[{
Cell[2596, 82, 59, 1, 56, "Input"],
Cell[2658, 85, 1788, 50, 54, "Output"]
}, Open  ]],
Cell[4461, 138, 59, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[4545, 143, 46, 1, 56, "Input"],
Cell[4594, 146, 392, 8, 31, "Print"],
Cell[4989, 156, 2938, 72, 47, "Print"],
Cell[7930, 230, 2938, 72, 47, "Print"],
Cell[10871, 304, 2915, 71, 34, "Print"],
Cell[13789, 377, 2915, 71, 32, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16741, 453, 47, 1, 56, "Input"],
Cell[16791, 456, 293, 6, 29, "Print"],
Cell[17087, 464, 4098, 99, 34, "Print"],
Cell[21188, 565, 4098, 99, 32, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25323, 669, 45, 1, 56, "Input"],
Cell[25371, 672, 184, 4, 29, "Print"],
Cell[25558, 678, 2377, 64, 47, "Print"]
}, Open  ]],
Cell[27950, 745, 445, 12, 97, "Text"],
Cell[28398, 759, 58, 0, 41, "Text"],
Cell[28459, 761, 245, 5, 35, "Text"],
Cell[28707, 768, 153, 3, 102, "Input"],
Cell[28863, 773, 56, 1, 56, "Input"],
Cell[28922, 776, 83, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[29030, 781, 121, 3, 56, "Input"],
Cell[29154, 786, 685, 16, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[29876, 807, 61, 1, 56, "Input"],
Cell[29940, 810, 2677, 73, 54, "Output"]
}, Open  ]],
Cell[32632, 886, 71, 0, 41, "Text"],
Cell[32706, 888, 316, 5, 35, "Text"],
Cell[33025, 895, 160, 5, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[33222, 905, 57, 0, 38, "Subsection"],
Cell[33282, 907, 71, 0, 41, "Text"],
Cell[33356, 909, 190, 4, 35, "Text"],
Cell[33549, 915, 134, 5, 41, "Text"],
Cell[33686, 922, 118, 2, 79, "Input"],
Cell[33807, 926, 124, 2, 79, "Input"],
Cell[33934, 930, 111, 2, 79, "Input"],
Cell[34048, 934, 189, 3, 79, "Input"],
Cell[34240, 939, 59, 1, 56, "Input"],
Cell[34302, 942, 75, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[34402, 946, 45, 1, 56, "Input"],
Cell[34450, 949, 184, 4, 29, "Print"],
Cell[34637, 955, 198, 5, 28, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34872, 965, 47, 1, 56, "Input"],
Cell[34922, 968, 293, 6, 29, "Print"],
Cell[35218, 976, 3907, 93, 34, "Print"],
Cell[39128, 1071, 4090, 99, 47, "Print"]
}, Open  ]],
Cell[43233, 1173, 229, 8, 41, "Text"],
Cell[43465, 1183, 491, 14, 91, "Text"],

Cell[CellGroupData[{
Cell[43981, 1201, 121, 3, 56, "Input"],
Cell[44105, 1206, 1049, 23, 100, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[45191, 1234, 47, 1, 56, "Input"],
Cell[45241, 1237, 145, 2, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[45423, 1244, 61, 1, 56, "Input"],
Cell[45487, 1247, 2164, 60, 54, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

