(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     74180,       2016]*)
(*NotebookOutlinePosition[     74943,       2043]*)
(*  CellTagsIndexPosition[     74899,       2039]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["InitCTensor and Show", "Subtitle",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate Basis", "Section"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {t, x};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(metric\  = \ {{\(-E\^\(2  \[Rho][t, x]\)\), 0}, {0, 
          E\^\(2  \[Rho][t, x]\)}}; \ MatrixForm[metric]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(-\[ExponentialE]\^\(2\ \[Rho][t, x]\)\), "0"},
            {"0", \(\[ExponentialE]\^\(2\ \[Rho][t, x]\)\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensor", "Subsubsection"],

Cell[BoxData[
    \(InitCTensor[coSys, \ metric]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Status[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"\<\"AutoCanFlag\"\>"},
                {"\<\"CoordinateBasisFlag\"\>"},
                {"\<\"EvaluateBDFlag\"\>"},
                {"\<\"MetricgFlag\"\>"},
                {"\<\"TorsionFreeFlag\"\>"},
                {"\<\"SyntaxCheckFlag\"\>"},
                {"\<\"InitCTensor\"\>"},
                {"\<\"Dimension\"\>"},
                {"\<\"Coordinates\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag",
             "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", 
            "InitCTensor", "Dimension", "Coordinates"}]], 
        "\[InvisibleSpace]", "\<\" == \"\>", "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"True"},
                {"True"},
                {"True"},
                {"False"},
                {"True"},
                {"False"},
                {"True"},
                {"2"},
                {\({t, x}\)}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {True, True, True, False, True, False, True, 2, {t, 
            x}}]]}],
      SequenceForm[ 
        MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag", 
          "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", "InitCTensor", 
          "Dimension", "Coordinates"}], " == ", 
        MatrixForm[ {True, True, True, False, True, False, True, 2, {t, x}}]],
      
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\[ExponentialE]\^\(2\ \[Rho][t, x]\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ 2, 
              \[Rho][ t, x]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\[ExponentialE]\^\(2\ \[Rho][t, x]\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            dhPark`Tensor`Private`UnderBar[ "x"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Power[ E, 
          Times[ 2, 
            \[Rho][ t, x]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\("f"\_\("ab"\)\%\("  c"\)\),
          Subsuperscript[ "f", "ab", "  c"],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "\<\"0\"\>"}],
      SequenceForm[ 
        Subsuperscript[ "f", "ab", "  c"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((1, 0)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Derivative[ 1, 0][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "x"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((0, 1)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "x"]]], " = ", 
        Derivative[ 0, 1][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((0, 1)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Derivative[ 0, 1][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "x"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((1, 0)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "x"]]], " = ", 
        Derivative[ 1, 0][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "x"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "x"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((1, 0)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Derivative[ 1, 0][ \[Rho]][ t, x]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "x"]],
                      Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "x"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "x"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          SuperscriptBox["\[Rho]", 
            TagBox[\((0, 1)\),
              Derivative],
            MultilineFunction->None], "[", \(t, x\), "]"}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "x"]]], " = ", 
        Derivative[ 0, 1][ \[Rho]][ t, x]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "x"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "x"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "x"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(2\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "x"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ 2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{"-", 
            RowBox[{
              SuperscriptBox["\[Rho]", 
                TagBox[\((0, 2)\),
                  Derivative],
                MultilineFunction->None], "[", \(t, x\), "]"}]}], "+", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((2, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Plus[ 
          Times[ -1, 
            Derivative[ 0, 2][ \[Rho]][ t, x]], 
          Derivative[ 2, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"x"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "x"]],
                  Editable->False], "\[InvisibleSpace]", \("x"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "x"]], 
                dhPark`Tensor`Private`UnderBar[ "x"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "x"]], 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 2)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((2, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "x"]], 
            dhPark`Tensor`Private`UnderBar[ "x"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Plus[ 
          Derivative[ 0, 2][ \[Rho]][ t, x], 
          Times[ -1, 
            Derivative[ 2, 0][ \[Rho]][ t, x]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
        "2", " ", \(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 
        Times[ 2, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Non-Coordinate Basis", "Section"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {t, x};\)\)], "Input"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-1\), 0}, {0, 1}};\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Orthonomal Basis", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(basisM\  = \ {{E\^\(-\[Rho][t, x]\), 0}, {0, E\^\(-\[Rho][t, x]\)}}; \ 
    MatrixForm[basisM]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), "0"},
            {"0", \(\[ExponentialE]\^\(-\[Rho][t, x]\)\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensor", "Subsubsection"],

Cell[BoxData[
    \(InitCTensor[coSys, \ metric, basisM]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Status[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"\<\"AutoCanFlag\"\>"},
                {"\<\"CoordinateBasisFlag\"\>"},
                {"\<\"EvaluateBDFlag\"\>"},
                {"\<\"MetricgFlag\"\>"},
                {"\<\"TorsionFreeFlag\"\>"},
                {"\<\"SyntaxCheckFlag\"\>"},
                {"\<\"InitCTensor\"\>"},
                {"\<\"Dimension\"\>"},
                {"\<\"Coordinates\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag",
             "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", 
            "InitCTensor", "Dimension", "Coordinates"}]], 
        "\[InvisibleSpace]", "\<\" == \"\>", "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"True"},
                {"False"},
                {"True"},
                {"False"},
                {"True"},
                {"False"},
                {"True"},
                {"2"},
                {\({t, x}\)}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {True, False, True, False, True, False, True, 2, {t, 
            x}}]]}],
      SequenceForm[ 
        MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag", 
          "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", "InitCTensor", 
          "Dimension", "Coordinates"}], " == ", 
        MatrixForm[ {True, False, True, False, True, False, True, 2, {t, 
          x}}]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-1\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 1]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", -1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "1"}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 2]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 1],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 1)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ f, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 0, 1][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((1, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ f, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 1, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 1)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 0, 1][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 1)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 0, 1][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((1, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 1, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((1, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 1, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&\
^\)\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`OverHat[ 1]],
                          Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`OverHat[ 1]], 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]], 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 1]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 2]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
        "2", " ", \(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 
        Times[ 2, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{566, 679},
WindowMargins->{{31, Automatic}, {Automatic, 5}},
Magnification->1,
StyleDefinitions -> "Demo.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 89, 2, 81, "Subtitle"],
Cell[1809, 53, 52, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[1886, 58, 35, 0, 77, "Section"],

Cell[CellGroupData[{
Cell[1946, 62, 53, 0, 57, "Subsubsection"],
Cell[2002, 64, 56, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[2083, 69, 142, 2, 79, "Input"],
Cell[2228, 73, 264, 6, 94, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2541, 85, 59, 0, 35, "Subsubsection"],
Cell[2603, 87, 61, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[2689, 92, 41, 1, 54, "Input"],
Cell[2733, 95, 1640, 41, 215, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[4422, 142, 37, 0, 35, "Subsubsection"],

Cell[CellGroupData[{
Cell[4484, 146, 46, 1, 54, "Input"],
Cell[4533, 149, 184, 4, 30, "Print"],
Cell[4720, 155, 1795, 45, 33, "Print"],
Cell[6518, 202, 1767, 44, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8322, 251, 49, 1, 54, "Input"],
Cell[8374, 254, 350, 8, 31, "Print"],
Cell[8727, 264, 372, 9, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9136, 278, 46, 1, 54, "Input"],
Cell[9185, 281, 392, 8, 31, "Print"],
Cell[9580, 291, 2906, 70, 33, "Print"],
Cell[12489, 363, 2906, 70, 32, "Print"],
Cell[15398, 435, 2906, 70, 33, "Print"],
Cell[18307, 507, 2906, 70, 32, "Print"],
Cell[21216, 579, 2906, 70, 33, "Print"],
Cell[24125, 651, 2906, 70, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27068, 726, 47, 1, 54, "Input"],
Cell[27118, 729, 293, 6, 29, "Print"],
Cell[27414, 737, 4640, 114, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[32091, 856, 45, 1, 54, "Input"],
Cell[32139, 859, 184, 4, 29, "Print"],
Cell[32326, 865, 2212, 57, 33, "Print"],
Cell[34541, 924, 2176, 56, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36754, 985, 46, 1, 54, "Input"],
Cell[36803, 988, 1151, 32, 30, "Print"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[38015, 1027, 39, 0, 49, "Section"],

Cell[CellGroupData[{
Cell[38079, 1031, 53, 0, 57, "Subsubsection"],
Cell[38135, 1033, 56, 1, 54, "Input"],
Cell[38194, 1036, 72, 1, 54, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[38303, 1042, 41, 0, 35, "Subsubsection"],

Cell[CellGroupData[{
Cell[38369, 1046, 127, 2, 79, "Input"],
Cell[38499, 1050, 259, 6, 94, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[38807, 1062, 59, 0, 35, "Subsubsection"],
Cell[38869, 1064, 69, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[38963, 1069, 41, 1, 54, "Input"],
Cell[39007, 1072, 1647, 41, 215, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[40703, 1119, 37, 0, 35, "Subsubsection"],

Cell[CellGroupData[{
Cell[40765, 1123, 46, 1, 54, "Input"],
Cell[40814, 1126, 184, 4, 30, "Print"],
Cell[41001, 1132, 1646, 41, 32, "Print"],
Cell[42650, 1175, 1642, 41, 32, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44329, 1221, 49, 1, 54, "Input"],
Cell[44381, 1224, 350, 8, 31, "Print"],
Cell[44734, 1234, 2975, 75, 37, "Print"],
Cell[47712, 1311, 2980, 75, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[50729, 1391, 46, 1, 54, "Input"],
Cell[50778, 1394, 3017, 75, 37, "Print"],
Cell[53798, 1471, 3017, 75, 37, "Print"],
Cell[56818, 1548, 3017, 75, 37, "Print"],
Cell[59838, 1625, 3017, 75, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[62892, 1705, 47, 1, 54, "Input"],
Cell[62942, 1708, 293, 6, 29, "Print"],
Cell[63238, 1716, 4565, 113, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[67840, 1834, 45, 1, 54, "Input"],
Cell[67888, 1837, 184, 4, 29, "Print"],
Cell[68075, 1843, 2415, 63, 33, "Print"],
Cell[70493, 1908, 2410, 63, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[72940, 1976, 46, 1, 54, "Input"],
Cell[72989, 1979, 1151, 32, 30, "Print"]
}, Open  ]]
}, Closed]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

