(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    261540,       6849]*)
(*NotebookOutlinePosition[    262308,       6876]*)
(*  CellTagsIndexPosition[    262264,       6872]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Kerr SpaceTime", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {t, r, \[Theta], \[Phi]};\)\)], "Input"],

Cell[BoxData[
    \(\[CapitalDelta]\  = \ 
      r\^2\  - \ 2  M\ r\  + \ a\^2; \[CapitalSigma]\  = \ 
      r\^2\  + \ a\^2\ Cos[\[Theta]]\^2;\)], "Input"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-\(\(\[CapitalDelta]\  - \ 
                    a\^2\ Sin[\[Theta]]\^2\)\/\[CapitalSigma]\)\), 0, 
            0, \(-\(\(a\ \(Sin[\[Theta]]\^2\) \((r\^2\  + \ 
                        a\^2\  - \ \
\[CapitalDelta])\)\)\/\[CapitalSigma]\)\)}, {0, \[CapitalSigma]\/\
\[CapitalDelta], 0, 0}, {0, 0, \[CapitalSigma], 
            0}, {\(-\(\(a\ \(Sin[\[Theta]]\^2\) \((r\^2\  + \ 
                        a\^2\  - \ \[CapitalDelta])\)\)\/\[CapitalSigma]\)\), 
            0, 0, \(\((r\^2\  + \ a\^2)\)\^2\  - \ \[CapitalDelta]\ a\^2\ \
Sin[\[Theta]]\^2\)\/\[CapitalSigma]\ Sin[\[Theta]]\^2}};\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensor", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[
      InitCTensor[coSys, \ metric, SimplifyMore -> True, 
        CalcRiemann -> False]]\)], "Input"],

Cell[BoxData[
    \({37.452999999999996`\ Second, Null}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\^2 - 2\ M\ r + r\^2 - 
                a\^2\ Sin[\[Theta]]\^2\)\/\(r\^2 + 
                a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Power[ a, 2], 
            Times[ -2, M, r], 
            Power[ r, 2], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ 
                Sin[ \[Theta]], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ a\ M\ r\ Sin[\[Theta]]\^2\)\/\(r\^2 + 
                a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -2, a, M, r, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(r\^2 + a\^2\ Cos[\[Theta]]\^2\)\/\(a\^2 - 
              2\ M\ r + r\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ -2, M, r], 
              Power[ r, 2]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ a, 2], 
              Power[ 
                Cos[ \[Theta]], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]"\+_\
\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\^2 + a\^2\ Cos[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ a, 2], 
            Power[ 
              Cos[ \[Theta]], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(Sin[\[Theta]]\^2\ \((\((a\^2 + r\^2)\)\^2 - 
                  a\^2\ \((a\^2 - 2\ M\ r + 
                        r\^2)\)\ Sin[\[Theta]]\^2)\)\)\/\(r\^2 + 
              a\^2\ Cos[\[Theta]]\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Plus[ 
            Power[ 
              Plus[ 
                Power[ a, 2], 
                Power[ r, 2]], 2], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Power[ 
                Sin[ \[Theta]], 2]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\("f"\_\("ab"\)\%\("  c"\)\),
          Subsuperscript[ "f", "ab", "  c"],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "\<\"0\"\>"}],
      SequenceForm[ 
        Subsuperscript[ "f", "ab", "  c"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(M\ \((\(-a\^2\) + 2\ M\ r - 
                  r\^2)\)\ \((a\^2 - 2\ r\^2 + 
                  a\^2\ Cos[
                      2\ \[Theta]])\)\)\/\(2\ \((r\^2 + a\^2\ \
Cos[\[Theta]]\^2)\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Rational[ 1, 2], M, 
          Plus[ 
            Times[ -1, 
              Power[ a, 2]], 
            Times[ 2, M, r], 
            Times[ -1, 
              Power[ r, 2]]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Plus[ 
            Power[ a, 2], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\^2\ M\ r\ Sin[
                  2\ \[Theta]]\)\/\((r\^2 + a\^2\ \
Cos[\[Theta]]\^2)\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ -1, 
          Power[ a, 2], M, r, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ M\ \((a\^2 + r\^2)\)\ \((a\^2 - 
                    2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\)\/\(\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ -2, M, 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Power[ a, 2], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ a\ M\ \((a\^2 - 2\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\)\/\(\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ -2, a, M, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Power[ a, 2], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(4\ a\^2\ M\ r\ Sin[
                  2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ -4, 
          Power[ a, 2], M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(8\ a\ M\ r\ Cot[\[Theta]]\)\/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ -8, a, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Cot[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(4\ a\ M\ \((a\^2 + 
                  r\ \((\(-2\)\ M + r)\))\)\ \((a\^2 - 2\ r\^2 + 
                  a\^2\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2\)\/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 4, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Plus[ 
            Power[ a, 2], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(8\ a\ M\ r\ \((a\^2 + r\^2)\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 8, a, M, r, 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(r\ \((a\^2 - M\ r)\) + 
              a\^2\ \((M - r)\)\ Cos[\[Theta]]\^2\)\/\(\((a\^2 + 
                  r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Times[ r, 
              Plus[ 
                Power[ a, 2], 
                Times[ -1, M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ M, 
                Times[ -1, r]], 
              Power[ 
                Cos[ \[Theta]], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(a\^2\ Cos[\[Theta]]\ Sin[\[Theta]]\)\/\(\((a\
\^2 + r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 
          Power[ a, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\^2\ Cos[\[Theta]]\ Sin[\[Theta]]\)\/\(r\
\^2 + a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -1, 
          Power[ a, 2], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\/\(r\^2 + a\^2\ Cos[\[Theta]]\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ r, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ a\ M\ \((a\^4 - 3\ a\^2\ r\^2 - 6\ r\^4 + 
                  a\^2\ \((a\^2 - r\^2)\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2\)\/\(\((a\^2 + 
                  r\ \((\(-2\)\ M + 
                        r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ 2, a, M, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Power[ a, 4], 
            Times[ -3, 
              Power[ a, 2], 
              Power[ r, 2]], 
            Times[ -6, 
              Power[ r, 4]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ -1, 
                  Power[ r, 2]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((a\^4\ M + 3\ a\^4\ r - 12\ a\^2\ M\ r\^2 + 
              8\ a\^2\ r\^3 - 16\ M\ r\^4 + 8\ r\^5 + 
              4\ a\^2\ r\ \((a\^2 + r\ \((\(-M\) + 2\ r)\))\)\ Cos[
                  2\ \[Theta]] - 
              a\^4\ \((M - r)\)\ Cos[4\ \[Theta]])\)/\((2\ \((a\^2 + 
                  r\ \((\(-2\)\ M + 
                        r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^2)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 4], M], 
            Times[ 3, 
              Power[ a, 4], r], 
            Times[ -12, 
              Power[ a, 2], M, 
              Power[ r, 2]], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ r, 3]], 
            Times[ -16, M, 
              Power[ r, 4]], 
            Times[ 8, 
              Power[ r, 5]], 
            Times[ 4, 
              Power[ a, 2], r, 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -1, M], 
                    Times[ 2, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ -1, 
              Power[ a, 4], 
              Plus[ M, 
                Times[ -1, r]], 
              Cos[ 
                Times[ 4, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ r\ \((a\^2 + 
                    r\ \((\(-2\)\ M + r)\))\)\)\/\(a\^2 + 2\ r\^2 + 
                a\^2\ Cos[2\ \[Theta]]\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -2, r, 
          Plus[ 
            Power[ a, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\^2\ Cos[\[Theta]]\ Sin[\[Theta]]\)\/\(r\
\^2 + a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ -1, 
          Power[ a, 2], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(8\ a\^3\ M\ r\ Cos[\[Theta]]\ \
Sin[\[Theta]]\^3\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^2\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ 8, 
          Power[ a, 3], M, r, 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((\((3\ a\^4 + 8\ a\^2\ M\ r + 
                  8\ a\^2\ r\^2 + 8\ r\^4 + 
                  4\ a\^2\ \((a\^2 + 2\ r\ \((\(-M\) + r)\))\)\ Cos[
                      2\ \[Theta]] + 
                  a\^4\ Cos[
                      4\ \[Theta]])\)\ Cot[\[Theta]])\)/\((2\ \((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^2)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 3, 
              Power[ a, 4]], 
            Times[ 8, 
              Power[ a, 2], M, r], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ r, 2]], 
            Times[ 8, 
              Power[ r, 4]], 
            Times[ 4, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ 2, r, 
                  Plus[ 
                    Times[ -1, M], r]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Cot[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((\((a\^2 + 
                    r\ \((\(-2\)\ M + r)\))\)\ \((a\^4\ M + 3\ a\^4\ r - 
                    4\ a\^2\ M\ r\^2 + 8\ a\^2\ r\^3 + 8\ r\^5 + 
                    4\ a\^2\ r\ \((a\^2 + r\ \((M + 2\ r)\))\)\ Cos[
                        2\ \[Theta]] - 
                    a\^4\ \((M - r)\)\ Cos[
                        4\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((8\ \((r\^2 + \
a\^2\ Cos[\[Theta]]\^2)\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Rational[ -1, 8], 
          Plus[ 
            Power[ a, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 4], M], 
            Times[ 3, 
              Power[ a, 4], r], 
            Times[ -4, 
              Power[ a, 2], M, 
              Power[ r, 2]], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ r, 3]], 
            Times[ 8, 
              Power[ r, 5]], 
            Times[ 4, 
              Power[ a, 2], r, 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ M, 
                    Times[ 2, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ -1, 
              Power[ a, 4], 
              Plus[ M, 
                Times[ -1, r]], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((\((3\ a\^6 + 10\ a\^4\ M\ r + 
                    11\ a\^4\ r\^2 + 16\ a\^2\ M\ r\^3 + 16\ a\^2\ r\^4 + 
                    8\ r\^6 + 
                    4\ a\^2\ \((a\^2 + 2\ r\^2)\)\ \((a\^2 + 
                          r\ \((\(-2\)\ M + r)\))\)\ Cos[2\ \[Theta]] + 
                    a\^4\ \((a\^2 - 2\ M\ r + r\^2)\)\ Cos[
                        4\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((16\ \((r\^2 + a\^2\ \
Cos[\[Theta]]\^2)\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 
          Rational[ -1, 16], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 6]], 
            Times[ 10, 
              Power[ a, 4], M, r], 
            Times[ 11, 
              Power[ a, 4], 
              Power[ r, 2]], 
            Times[ 16, 
              Power[ a, 2], M, 
              Power[ r, 3]], 
            Times[ 16, 
              Power[ a, 2], 
              Power[ r, 4]], 
            Times[ 8, 
              Power[ r, 6]], 
            Times[ 4, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ 2, 
                  Power[ r, 2]]], 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -2, M], r]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ a, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Calculation of Curvature Tensors", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[RicciR, Simplify]]\)], "Input"],

Cell[BoxData[
    \({74.39600000000002`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("R"\_"ab"\[InvisibleSpace]" = "\[InvisibleSpace]"0"\),
      
      SequenceForm[ 
        Subscript[ "R", "ab"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[
    \(Tcalc[ScalarR]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "0"}],
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 0],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[RiemannR, Simplify]]\)], "Input"],

Cell[BoxData[
    \({267.775`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ M\ r\ \((\(-3\)\ a\^2 + 2\ r\^2 - 
                    3\ a\^2\ Cos[2\ \[Theta]])\)\ \((\(-5\)\ a\^2 + 
                    4\ \((2\ M - r)\)\ r + 
                    a\^2\ Cos[2\ \[Theta]])\))\)/\((\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Times[ -3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ r, 2]], 
            Times[ -3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -5, 
              Power[ a, 2]], 
            Times[ 4, 
              Plus[ 
                Times[ 2, M], 
                Times[ -1, r]], r], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ M\ \((a\^2 - 6\ r\^2 + 
                  a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ a\ M\ r\ \((3\ a\^2 + 
                    r\ \((\(-4\)\ M + 3\ r)\))\)\ \((3\ a\^2 - 2\ r\^2 + 
                    3\ a\^2\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, a, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -4, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((a\ M\ \((a\^2 - 6\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((\(-5\)\ a\^2 - 4\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -5, 
              Power[ a, 2]], 
            Times[ -4, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ M\ \((a\^2 - 6\ r\^2 + 
                  a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ M\ r\ \((\(-3\)\ a\^2 + 2\ r\^2 - 
                  3\ a\^2\ Cos[
                      2\ \[Theta]])\)\ \((\(-2\)\ a\^2 + \((2\ M - r)\)\ r + 
                  a\^2\ Cos[
                      2\ \[Theta]])\))\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, M, r, 
          Plus[ 
            Times[ -3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ r, 2]], 
            Times[ -3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -2, 
              Power[ a, 2]], 
            Times[ 
              Plus[ 
                Times[ 2, M], 
                Times[ -1, r]], r], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ a\ M\ \((a\^2 - 6\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((\(-2\)\ a\^2 - r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -2, 
              Power[ a, 2]], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ a\ M\ r\ \((3\ a\^2 + 
                  r\ \((\(-2\)\ M + 3\ r)\))\)\ \((3\ a\^2 - 2\ r\^2 + 
                  3\ a\^2\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, a, M, r, 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ M\ r\ \((\(-a\^2\) + 2\ M\ r - 
                    r\^2)\)\ \((3\ a\^2 - 2\ r\^2 + 
                    3\ a\^2\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, M, r, 
          Plus[ 
            Times[ -1, 
              Power[ a, 2]], 
            Times[ 2, M, r], 
            Times[ -1, 
              Power[ r, 2]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ M\ \((a\^2 - 6\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(M\ r\ \((\(-3\)\ a\^2 + 2\ r\^2 - 
                  3\ a\^2\ Cos[2\ \[Theta]])\)\)\/\(2\ \((a\^2 + 
                  r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 2], M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Times[ -3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ r, 2]], 
            Times[ -3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ M\ r\ \((3\ a\^2 - 2\ r\^2 + 
                    3\ a\^2\ Cos[2\ \[Theta]])\)\ \((2\ a\^4 + r\^4 + 
                    a\^2\ r\ \((\(-2\)\ M + 3\ r)\) - 
                    a\^2\ \((a\^2 + r\ \((\(-2\)\ M + r)\))\)\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ 2, 
              Power[ a, 4]], 
            Power[ r, 4], 
            Times[ 
              Power[ a, 2], r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 3, r]]], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -2, M], r]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(12\ a\^2\ M\ \((a\^2 + 
                  r\^2)\)\ Cos[\[Theta]]\ \((a\^2 - 6\ r\^2 + 
                  a\^2\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^3\)\/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 12, 
          Power[ a, 2], M, 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Cos[ \[Theta]], 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((2\ M\ r\ \((3\ a\^2 - 2\ r\^2 + 
                  3\ a\^2\ Cos[2\ \[Theta]])\)\ \((5\ a\^4 + 4\ r\^4 + 
                  a\^2\ r\ \((\(-2\)\ M + 9\ r)\) - 
                  a\^2\ \((a\^2 + r\ \((\(-2\)\ M + r)\))\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 2, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ 5, 
              Power[ a, 4]], 
            Times[ 4, 
              Power[ r, 4]], 
            Times[ 
              Power[ a, 2], r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 9, r]]], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -2, M], r]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[WeylC, Simplify]]\)], "Input"],

Cell[BoxData[
    \({0.30000000000001137`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[WeylC]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"C"\_"[ab][cd]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "C", "[ab][cd]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ M\ r\ \((\(-3\)\ a\^2 + 2\ r\^2 - 
                    3\ a\^2\ Cos[2\ \[Theta]])\)\ \((\(-5\)\ a\^2 + 
                    4\ \((2\ M - r)\)\ r + 
                    a\^2\ Cos[2\ \[Theta]])\))\)/\((\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Times[ -3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ r, 2]], 
            Times[ -3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -5, 
              Power[ a, 2]], 
            Times[ 4, 
              Plus[ 
                Times[ 2, M], 
                Times[ -1, r]], r], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ M\ \((a\^2 - 6\ r\^2 + 
                  a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ a\ M\ r\ \((3\ a\^2 + 
                    r\ \((\(-4\)\ M + 3\ r)\))\)\ \((3\ a\^2 - 2\ r\^2 + 
                    3\ a\^2\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, a, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -4, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((a\ M\ \((a\^2 - 6\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((\(-5\)\ a\^2 - 4\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -5, 
              Power[ a, 2]], 
            Times[ -4, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ M\ \((a\^2 - 6\ r\^2 + 
                  a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ M\ r\ \((\(-3\)\ a\^2 + 2\ r\^2 - 
                  3\ a\^2\ Cos[
                      2\ \[Theta]])\)\ \((\(-2\)\ a\^2 + \((2\ M - r)\)\ r + 
                  a\^2\ Cos[
                      2\ \[Theta]])\))\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, M, r, 
          Plus[ 
            Times[ -3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ r, 2]], 
            Times[ -3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -2, 
              Power[ a, 2]], 
            Times[ 
              Plus[ 
                Times[ 2, M], 
                Times[ -1, r]], r], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ a\ M\ \((a\^2 - 6\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((\(-2\)\ a\^2 - r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ -2, 
              Power[ a, 2]], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ a\ M\ r\ \((3\ a\^2 + 
                  r\ \((\(-2\)\ M + 3\ r)\))\)\ \((3\ a\^2 - 2\ r\^2 + 
                  3\ a\^2\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, a, M, r, 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ M\ r\ \((\(-a\^2\) + 2\ M\ r - 
                    r\^2)\)\ \((3\ a\^2 - 2\ r\^2 + 
                    3\ a\^2\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, M, r, 
          Plus[ 
            Times[ -1, 
              Power[ a, 2]], 
            Times[ 2, M, r], 
            Times[ -1, 
              Power[ r, 2]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ M\ \((a\^2 - 6\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ M\ \((a\^2 - 6\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, M, 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(M\ r\ \((\(-3\)\ a\^2 + 2\ r\^2 - 
                  3\ a\^2\ Cos[2\ \[Theta]])\)\)\/\(2\ \((a\^2 + 
                  r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 2], M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Times[ -3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ r, 2]], 
            Times[ -3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ a\ M\ r\ \((3\ a\^2 + 
                    r\ \((\(-4\)\ M + 3\ r)\))\)\ \((3\ a\^2 - 2\ r\^2 + 
                    3\ a\^2\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, a, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -4, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((a\ M\ \((3\ a\^4 - 22\ a\^2\ r\^2 - 
                  12\ r\^4 + 
                  2\ a\^2\ \((a\^2 + 7\ r\^2)\)\ Cos[2\ \[Theta]] - 
                  a\^4\ Cos[4\ \[Theta]])\)\ Sin[
                2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ a, M, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 4]], 
            Times[ -22, 
              Power[ a, 2], 
              Power[ r, 2]], 
            Times[ -12, 
              Power[ r, 4]], 
            Times[ 2, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ 7, 
                  Power[ r, 2]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ -1, 
              Power[ a, 4], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ M\ r\ \((3\ a\^2 - 2\ r\^2 + 
                    3\ a\^2\ Cos[2\ \[Theta]])\)\ \((2\ a\^4 + r\^4 + 
                    a\^2\ r\ \((\(-2\)\ M + 3\ r)\) - 
                    a\^2\ \((a\^2 + r\ \((\(-2\)\ M + r)\))\)\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ 2, 
              Power[ a, 4]], 
            Power[ r, 4], 
            Times[ 
              Power[ a, 2], r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 3, r]]], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -2, M], r]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(12\ a\^2\ M\ \((a\^2 + 
                  r\^2)\)\ Cos[\[Theta]]\ \((a\^2 - 6\ r\^2 + 
                  a\^2\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^3\)\/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 12, 
          Power[ a, 2], M, 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Cos[ \[Theta]], 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((a\ M\ \((9\ a\^4 - 52\ a\^2\ r\^2 - 
                  48\ r\^4 + 
                  4\ a\^2\ \((2\ a\^2 + 5\ r\^2)\)\ Cos[2\ \[Theta]] - 
                  a\^4\ Cos[4\ \[Theta]])\)\ Sin[
                2\ \[Theta]])\)/\((2\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 2], a, M, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 9, 
              Power[ a, 4]], 
            Times[ -52, 
              Power[ a, 2], 
              Power[ r, 2]], 
            Times[ -48, 
              Power[ r, 4]], 
            Times[ 4, 
              Power[ a, 2], 
              Plus[ 
                Times[ 2, 
                  Power[ a, 2]], 
                Times[ 5, 
                  Power[ r, 2]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ -1, 
              Power[ a, 4], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ a\ M\ r\ \((3\ a\^2 + 
                  r\ \((\(-2\)\ M + 3\ r)\))\)\ \((3\ a\^2 - 2\ r\^2 + 
                  3\ a\^2\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, a, M, r, 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(12\ a\^2\ M\ \((a\^2 + 
                  r\^2)\)\ Cos[\[Theta]]\ \((a\^2 - 6\ r\^2 + 
                  a\^2\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^3\)\/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 12, 
          Power[ a, 2], M, 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Cos[ \[Theta]], 
          Plus[ 
            Power[ a, 2], 
            Times[ -6, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((2\ M\ r\ \((3\ a\^2 - 2\ r\^2 + 
                  3\ a\^2\ Cos[2\ \[Theta]])\)\ \((5\ a\^4 + 4\ r\^4 + 
                  a\^2\ r\ \((\(-2\)\ M + 9\ r)\) - 
                  a\^2\ \((a\^2 + r\ \((\(-2\)\ M + r)\))\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 2, M, r, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ -2, 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Plus[ 
            Times[ 5, 
              Power[ a, 4]], 
            Times[ 4, 
              Power[ r, 4]], 
            Times[ 
              Power[ a, 2], r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 9, r]]], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -2, M], r]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Laplacian", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Laplacian[f[r, \[Theta]], Simplify]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"(", 
        RowBox[{"2", " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{\(Cot[\[Theta]]\), " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((0, 1)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{
                SuperscriptBox["f", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(r, \[Theta]\), "]"}], "-", 
              
              RowBox[{"2", " ", "M", " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((1, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{"2", " ", "r", " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((1, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{\(a\^2\), " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((2, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "-", 
              RowBox[{"2", " ", "M", " ", "r", " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((2, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{\(r\^2\), " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((2, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}]}],
             ")"}]}], ")"}], 
      "/", \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature Invariant", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[\(riemannUp\  = \ Get["\<dhPark`RiemannUp4`\>"];\)]\)], "Input"],

Cell[BoxData[
    \({3.795000000000016`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Cset[RiemannR[ua, ub, uc, ud], riemannUp]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[
      SumDum[RiemannR[la, lb, lc, ld]\ RiemannR[ua, ub, uc, ud]]\  // \ 
        Simplify]\)], "Input"],

Cell[BoxData[
    \({144.56800000000004`\ Second, \(-\(\((96\ M\^2\ \((10\ a\^6 - 
                  180\ a\^4\ r\^2 + 240\ a\^2\ r\^4 - 32\ r\^6 + 
                  15\ a\^2\ \((a\^4 - 16\ a\^2\ r\^2 + 16\ r\^4)\)\ Cos[
                      2\ \[Theta]] + 
                  6\ a\^4\ \((a\^2 - 10\ r\^2)\)\ Cos[4\ \[Theta]] + 
                  a\^6\ Cos[
                      6\ \[Theta]])\))\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^6\)\)}\)], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{636, 689},
WindowMargins->{{16, Automatic}, {Automatic, 4}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 47, 1, 43, "Title"],
Cell[1767, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1844, 57, 53, 0, 54, "Subsubsection"],
Cell[1900, 59, 74, 1, 56, "Input"],
Cell[1977, 62, 156, 3, 58, "Input"],
Cell[2136, 67, 641, 10, 230, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2814, 82, 59, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[2898, 86, 129, 3, 79, "Input"],
Cell[3030, 91, 69, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[3148, 98, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[3210, 102, 46, 1, 56, "Input"],
Cell[3259, 105, 184, 4, 30, "Print"],
Cell[3446, 111, 2211, 59, 50, "Print"],
Cell[5660, 172, 2040, 52, 50, "Print"],
Cell[7703, 226, 2041, 54, 47, "Print"],
Cell[9747, 282, 1914, 49, 33, "Print"],
Cell[11664, 333, 2536, 69, 51, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14237, 407, 49, 1, 56, "Input"],
Cell[14289, 410, 350, 8, 31, "Print"],
Cell[14642, 420, 372, 9, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15051, 434, 46, 1, 56, "Input"],
Cell[15100, 437, 392, 8, 31, "Print"],
Cell[15495, 447, 3493, 91, 51, "Print"],
Cell[18991, 540, 3114, 77, 51, "Print"],
Cell[22108, 619, 3646, 95, 51, "Print"],
Cell[25757, 716, 3581, 92, 51, "Print"],
Cell[29341, 810, 3212, 80, 51, "Print"],
Cell[32556, 892, 3161, 77, 48, "Print"],
Cell[35720, 971, 3625, 94, 83, "Print"],
Cell[39348, 1067, 3288, 82, 51, "Print"],
Cell[42639, 1151, 3550, 92, 50, "Print"],
Cell[46192, 1245, 3286, 82, 50, "Print"],
Cell[49481, 1329, 3121, 77, 50, "Print"],
Cell[52605, 1408, 3025, 73, 44, "Print"],
Cell[55633, 1483, 3920, 103, 84, "Print"],
Cell[59556, 1588, 4479, 120, 113, "Print"],
Cell[64038, 1710, 3302, 82, 50, "Print"],
Cell[67343, 1794, 3170, 77, 50, "Print"],
Cell[70516, 1873, 3240, 80, 51, "Print"],
Cell[73759, 1955, 4065, 106, 113, "Print"],
Cell[77827, 2063, 4363, 116, 107, "Print"],
Cell[82193, 2181, 4572, 121, 132, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[86814, 2308, 57, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[86896, 2312, 64, 1, 56, "Input"],
Cell[86963, 2315, 68, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[87068, 2321, 45, 1, 56, "Input"],
Cell[87116, 2324, 184, 4, 29, "Print"],
Cell[87303, 2330, 198, 5, 28, "Print"]
}, Open  ]],
Cell[87516, 2338, 47, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[87588, 2343, 46, 1, 56, "Input"],
Cell[87637, 2346, 361, 9, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[88035, 2360, 66, 1, 56, "Input"],
Cell[88104, 2363, 58, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[88199, 2369, 47, 1, 56, "Input"],
Cell[88249, 2372, 293, 6, 29, "Print"],
Cell[88545, 2380, 5183, 134, 84, "Print"],
Cell[93731, 2516, 4619, 116, 51, "Print"],
Cell[98353, 2634, 5138, 132, 86, "Print"],
Cell[103494, 2768, 4961, 126, 83, "Print"],
Cell[108458, 2896, 4635, 116, 51, "Print"],
Cell[113096, 3014, 5014, 127, 82, "Print"],
Cell[118113, 3143, 4949, 125, 83, "Print"],
Cell[123065, 3270, 4976, 125, 83, "Print"],
Cell[128044, 3397, 4881, 123, 61, "Print"],
Cell[132928, 3522, 4633, 115, 51, "Print"],
Cell[137564, 3639, 4784, 119, 50, "Print"],
Cell[142351, 3760, 5575, 144, 109, "Print"],
Cell[147929, 3906, 4859, 122, 82, "Print"],
Cell[152791, 4030, 5445, 138, 84, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[158273, 4173, 63, 1, 56, "Input"],
Cell[158339, 4176, 70, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[158446, 4182, 44, 1, 56, "Input"],
Cell[158493, 4185, 192, 4, 29, "Print"],
Cell[158688, 4191, 5183, 134, 84, "Print"],
Cell[163874, 4327, 4619, 116, 51, "Print"],
Cell[168496, 4445, 5138, 132, 86, "Print"],
Cell[173637, 4579, 4961, 126, 83, "Print"],
Cell[178601, 4707, 4635, 116, 51, "Print"],
Cell[183239, 4825, 5014, 127, 82, "Print"],
Cell[188256, 4954, 4949, 125, 83, "Print"],
Cell[193208, 5081, 4976, 125, 83, "Print"],
Cell[198187, 5208, 4881, 123, 61, "Print"],
Cell[203071, 5333, 4633, 115, 51, "Print"],
Cell[207707, 5450, 4637, 115, 51, "Print"],
Cell[212347, 5567, 4784, 119, 50, "Print"],
Cell[217134, 5688, 5150, 132, 86, "Print"],
Cell[222287, 5822, 5071, 129, 83, "Print"],
Cell[227361, 5953, 5575, 144, 109, "Print"],
Cell[232939, 6099, 4859, 122, 82, "Print"],
Cell[237801, 6223, 5162, 131, 83, "Print"],
Cell[242966, 6356, 4965, 124, 83, "Print"],
Cell[247934, 6482, 4852, 121, 82, "Print"],
Cell[252789, 6605, 5445, 138, 84, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[258283, 6749, 34, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[258342, 6753, 68, 1, 56, "Input"],
Cell[258413, 6756, 2102, 54, 100, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[260564, 6816, 44, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[260633, 6820, 91, 1, 56, "Input"],
Cell[260727, 6823, 68, 1, 53, "Output"]
}, Open  ]],
Cell[260810, 6827, 74, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[260909, 6832, 131, 3, 102, "Input"],
Cell[261043, 6837, 469, 8, 152, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

