(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     15874,        474]*)
(*NotebookOutlinePosition[     16642,        501]*)
(*  CellTagsIndexPosition[     16598,        497]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Kerr SpaceTime in Newman-Penrose Formalism", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System", "Subsubsection"],

Cell[BoxData[
    \(\(coSys = {t, r, \[Theta], \[Phi]};\)\)], "Input"],

Cell[BoxData[
    \(\[CapitalDelta]\  = \ 
      r\^2\  - \ 2  M\ r\  + \ a\^2; \ \[CapitalSigma]\  = \ 
      r\^2\  + \ a\^2\ Cos[\[Theta]]\^2;\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Kerr Metric", "Subsubsection"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-\(\(\[CapitalDelta]\  - \ 
                    a\^2\ Sin[\[Theta]]\^2\)\/\[CapitalSigma]\)\), 0, 
            0, \(-\(\(a\ \(Sin[\[Theta]]\^2\) \((r\^2\  + \ 
                        a\^2\  - \ \
\[CapitalDelta])\)\)\/\[CapitalSigma]\)\)}, {0, \[CapitalSigma]\/\
\[CapitalDelta], 0, 0}, {0, 0, \[CapitalSigma], 
            0}, {\(-\(\(a\ \(Sin[\[Theta]]\^2\) \((r\^2\  + \ 
                        a\^2\  - \ \[CapitalDelta])\)\)\/\[CapitalSigma]\)\), 
            0, 0, \(\((r\^2\  + \ a\^2)\)\^2\  - \ \[CapitalDelta]\ a\^2\ \
Sin[\[Theta]]\^2\)\/\[CapitalSigma]\ Sin[\[Theta]]\^2}};\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Newman-Penrose Frame (See Chandrasekhar, pp 299)", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(nullVs\  = \ {\(1\/\[CapitalDelta]\) {r\^2\  + \ a\^2, \[CapitalDelta], 
            0, a}, \(1\/\(2\ \[CapitalSigma]\)\) {r\^2\  + \ 
              a\^2, \(-\[CapitalDelta]\), 0, 
            a}, \(1\/\(\(\@2\) \((r\  + \ 
                    I\ a\ Cos[\[Theta]])\)\)\) {I\ a\ Sin[\[Theta]], 0, 1, 
            I\/Sin[\[Theta]]}, \(1\/\(\(\@2\) \((r\  - \ 
                    I\ a\ Cos[\[Theta]])\)\)\) {\(-I\)\ a\ Sin[\[Theta]], 0, 
            1, \(-\(I\/Sin[\[Theta]]\)\)}}; MatrixForm[nullVs]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(\(a\^2 + r\^2\)\/\(a\^2 - 2\ M\ r + r\^2\)\), "1", 
              "0", \(a\/\(a\^2 - 2\ M\ r + r\^2\)\)},
            {\(\(a\^2 + 
                    r\^2\)\/\(2\ \((r\^2 + 
                        a\^2\ Cos[\[Theta]]\^2)\)\)\), \(\(\(-a\^2\) + 
                    2\ M\ r - 
                    r\^2\)\/\(2\ \((r\^2 + a\^2\ Cos[\[Theta]]\^2)\)\)\), 
              "0", \(a\/\(2\ \((r\^2 + a\^2\ Cos[\[Theta]]\^2)\)\)\)},
            {\(\(\[ImaginaryI]\ a\ Sin[\[Theta]]\)\/\(\@2\ \((r + \
\[ImaginaryI]\ a\ Cos[\[Theta]])\)\)\), 
              "0", \(1\/\(\@2\ \((r + \[ImaginaryI]\ a\ Cos[\[Theta]])\)\)\), \
\(\(\[ImaginaryI]\ Csc[\[Theta]]\)\/\(\@2\ \((r + \[ImaginaryI]\ a\ Cos[\
\[Theta]])\)\)\)},
            {\(-\(\(\[ImaginaryI]\ a\ Sin[\[Theta]]\)\/\(\@2\ \((r - \
\[ImaginaryI]\ a\ Cos[\[Theta]])\)\)\)\), 
              "0", \(1\/\(\@2\ \((r - \[ImaginaryI]\ a\ Cos[\[Theta]])\)\)\), \
\(-\(\(\[ImaginaryI]\ Csc[\[Theta]]\)\/\(\@2\ \((r - \[ImaginaryI]\ a\ Cos[\
\[Theta]])\)\)\)\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell["Check Newman-Penrose frame", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[
      Table[nullVs[\([i]\)]\  . \ metric\  . \ nullVs[\([j]\)], {i, 4}, {j, 
            4}]\  // \ Simplify]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"0", \(-1\), "0", "0"},
            {\(-1\), "0", "0", "0"},
            {"0", "0", "0", "1"},
            {"0", "0", "1", "0"}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialize CTensor using coSys and nullVs", "Subsubsection"],

Cell["SimplifyMore->True is CURIOUSLY required in VER 4!", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[InitCTensorNP[coSys, nullVs, SimplifyMore -> True]]\)], "Input"],

Cell[BoxData[
    \({40.058`\ Second, Null}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[SpinCoefficients]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("\[Epsilon] = \[Kappa] = \[Lambda] = \[Nu] = \[Sigma] \
= "\[InvisibleSpace]"0"\),
      SequenceForm[ 
      "\[Epsilon] = \[Kappa] = \[Lambda] = \[Nu] = \[Sigma] = ", "0"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[Alpha]"\[InvisibleSpace]" = "\[InvisibleSpace]\(r\ \
Cot[\[Theta]] - \[ImaginaryI]\ a\ \((Csc[\[Theta]] + Sin[\[Theta]])\)\)\/\(2\ \
\@2\ \((r - \[ImaginaryI]\ a\ Cos[\[Theta]])\)\^2\)\),
      SequenceForm[ "\[Alpha]", " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 2, 
            Rational[ -1, 2]], 
          Power[ 
            Plus[ r, 
              Times[ 
                Complex[ 0, -1], a, 
                Cos[ \[Theta]]]], -2], 
          Plus[ 
            Times[ r, 
              Cot[ \[Theta]]], 
            Times[ 
              Complex[ 0, -1], a, 
              Plus[ 
                Csc[ \[Theta]], 
                Sin[ \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[Beta]"\[InvisibleSpace]" = "\[InvisibleSpace]Cot[\
\[Theta]]\/\(\@2\ \((\(-2\)\ r - 2\ \[ImaginaryI]\ a\ Cos[\[Theta]])\)\)\),
      SequenceForm[ "\[Beta]", " = ", 
        Times[ 
          Power[ 2, 
            Rational[ -1, 2]], 
          Power[ 
            Plus[ 
              Times[ -2, r], 
              Times[ 
                Complex[ 0, -2], a, 
                Cos[ \[Theta]]]], -1], 
          Cot[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[Gamma]"\[InvisibleSpace]" = \
"\[InvisibleSpace]\(a\^2 - M\ r - \[ImaginaryI]\ a\ \((M - r)\)\ \
Cos[\[Theta]]\)\/\(2\ \((r - \[ImaginaryI]\ a\ Cos[\[Theta]])\)\^2\ \((r + \
\[ImaginaryI]\ a\ Cos[\[Theta]])\)\)\),
      SequenceForm[ "\[Gamma]", " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ r, 
              Times[ 
                Complex[ 0, -1], a, 
                Cos[ \[Theta]]]], -2], 
          Power[ 
            Plus[ r, 
              Times[ 
                Complex[ 0, 1], a, 
                Cos[ \[Theta]]]], -1], 
          Plus[ 
            Power[ a, 2], 
            Times[ -1, M, r], 
            Times[ 
              Complex[ 0, -1], a, 
              Plus[ M, 
                Times[ -1, r]], 
              Cos[ \[Theta]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[Mu]"\[InvisibleSpace]" = "\[InvisibleSpace]\(a\^2 \
+ r\ \((\(-2\)\ M + r)\)\)\/\(2\ \((r - \[ImaginaryI]\ a\ Cos[\[Theta]])\)\^2\
\ \((r + \[ImaginaryI]\ a\ Cos[\[Theta]])\)\)\),
      SequenceForm[ "\[Mu]", " = ", 
        Times[ 
          Rational[ 1, 2], 
          Plus[ 
            Power[ a, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ r, 
              Times[ 
                Complex[ 0, -1], a, 
                Cos[ \[Theta]]]], -2], 
          Power[ 
            Plus[ r, 
              Times[ 
                Complex[ 0, 1], a, 
                Cos[ \[Theta]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[Pi]"\[InvisibleSpace]" = "\[InvisibleSpace]\(\
\[ImaginaryI]\ a\ Sin[\[Theta]]\)\/\(\@2\ \((\[ImaginaryI]\ r + a\ Cos[\
\[Theta]])\)\^2\)\),
      SequenceForm[ "\[Pi]", " = ", 
        Times[ 
          Complex[ 0, 1], 
          Power[ 2, 
            Rational[ -1, 2]], a, 
          Power[ 
            Plus[ 
              Times[ 
                Complex[ 0, 1], r], 
              Times[ a, 
                Cos[ \[Theta]]]], -2], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[Rho]"\[InvisibleSpace]" = "\[InvisibleSpace]1\/\(r \
- \[ImaginaryI]\ a\ Cos[\[Theta]]\)\),
      SequenceForm[ "\[Rho]", " = ", 
        Power[ 
          Plus[ r, 
            Times[ 
              Complex[ 0, -1], a, 
              Cos[ \[Theta]]]], -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[Tau]"\[InvisibleSpace]" = "\[InvisibleSpace]\(\
\[ImaginaryI]\ a\ Sin[\[Theta]]\)\/\(\@2\ \((r\^2 + a\^2\ Cos[\[Theta]]\^2)\)\
\)\),
      SequenceForm[ "\[Tau]", " = ", 
        Times[ 
          Complex[ 0, 1], 
          Power[ 2, 
            Rational[ -1, 2]], a, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[PhiNP]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{
            
            InterpretationBox[\("\[CapitalPhi]"\_"00"\[InvisibleSpace]" = "\),
              
              SequenceForm[ 
                Subscript[ "\[CapitalPhi]", "00"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            
            InterpretationBox[\("\[CapitalPhi]"\_"01"\[InvisibleSpace]" = "\),
              
              SequenceForm[ 
                Subscript[ "\[CapitalPhi]", "01"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            
            InterpretationBox[\("\[CapitalPhi]"\_"02"\[InvisibleSpace]" = "\),
              
              SequenceForm[ 
                Subscript[ "\[CapitalPhi]", "02"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            
            InterpretationBox[\("\[CapitalPhi]"\_"11"\[InvisibleSpace]" = "\),
              
              SequenceForm[ 
                Subscript[ "\[CapitalPhi]", "11"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            
            InterpretationBox[\("\[CapitalPhi]"\_"12"\[InvisibleSpace]" = "\),
              
              SequenceForm[ 
                Subscript[ "\[CapitalPhi]", "12"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            
            InterpretationBox[\("\[CapitalPhi]"\_"22"\[InvisibleSpace]" = "\),
              
              SequenceForm[ 
                Subscript[ "\[CapitalPhi]", "22"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[\("\[CapitalLambda]"\[InvisibleSpace]" = "\),
              SequenceForm[ "\[CapitalLambda]", " = "],
              Editable->False]}],
          SequenceForm[ 
            SequenceForm[ 
              Subscript[ "\[CapitalPhi]", "00"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPhi]", "01"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPhi]", "02"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPhi]", "11"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPhi]", "12"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPhi]", "22"], " = "], 
            SequenceForm[ "\[CapitalLambda]", " = "]],
          Editable->False], "\[InvisibleSpace]", "\<\"0\"\>"}],
      SequenceForm[ 
        SequenceForm[ 
          SequenceForm[ 
            Subscript[ "\[CapitalPhi]", "00"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPhi]", "01"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPhi]", "02"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPhi]", "11"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPhi]", "12"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPhi]", "22"], " = "], 
          SequenceForm[ "\[CapitalLambda]", " = "]], "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[PsiNP]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{
            InterpretationBox[\("\[CapitalPsi]"\_"0"\[InvisibleSpace]" = "\),
              SequenceForm[ 
                Subscript[ "\[CapitalPsi]", "0"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[\("\[CapitalPsi]"\_"1"\[InvisibleSpace]" = "\),
              SequenceForm[ 
                Subscript[ "\[CapitalPsi]", "1"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[\("\[CapitalPsi]"\_"3"\[InvisibleSpace]" = "\),
              SequenceForm[ 
                Subscript[ "\[CapitalPsi]", "3"], " = "],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[\("\[CapitalPsi]"\_"4"\[InvisibleSpace]" = "\),
              SequenceForm[ 
                Subscript[ "\[CapitalPsi]", "4"], " = "],
              Editable->False]}],
          SequenceForm[ 
            SequenceForm[ 
              Subscript[ "\[CapitalPsi]", "0"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPsi]", "1"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPsi]", "3"], " = "], 
            SequenceForm[ 
              Subscript[ "\[CapitalPsi]", "4"], " = "]],
          Editable->False], "\[InvisibleSpace]", "\<\"0\"\>"}],
      SequenceForm[ 
        SequenceForm[ 
          SequenceForm[ 
            Subscript[ "\[CapitalPsi]", "0"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPsi]", "1"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPsi]", "3"], " = "], 
          SequenceForm[ 
            Subscript[ "\[CapitalPsi]", "4"], " = "]], "0"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("\[CapitalPsi]"\_"2"\[InvisibleSpace]" = "\
\[InvisibleSpace]M\/\((r - \[ImaginaryI]\ a\ Cos[\[Theta]])\)\^3\),
      SequenceForm[ 
        Subscript[ "\[CapitalPsi]", "2"], " = ", 
        Times[ M, 
          Power[ 
            Plus[ r, 
              Times[ 
                Complex[ 0, -1], a, 
                Cos[ \[Theta]]]], -3]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[PetrovType]\)], "Input"],

Cell[BoxData[
    \("Petrov type: D"\)], "Print"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{716, 680},
WindowMargins->{{20, Automatic}, {Automatic, 6}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 75, 1, 43, "Title"],
Cell[1795, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1872, 57, 42, 0, 54, "Subsubsection"],
Cell[1917, 59, 70, 1, 56, "Input"],
Cell[1990, 62, 158, 3, 58, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2185, 70, 36, 0, 32, "Subsubsection"],
Cell[2224, 72, 641, 10, 205, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2902, 87, 73, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[3000, 91, 531, 8, 193, "Input"],
Cell[3534, 101, 1154, 22, 193, "Output"]
}, Open  ]],
Cell[4703, 126, 42, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[4770, 130, 154, 3, 102, "Input"],
Cell[4927, 135, 283, 8, 125, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5259, 149, 66, 0, 32, "Subsubsection"],
Cell[5328, 151, 66, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[5419, 155, 91, 1, 56, "Input"],
Cell[5513, 158, 57, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5619, 165, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[5681, 169, 55, 1, 56, "Input"],
Cell[5739, 172, 246, 5, 28, "Print"],
Cell[5988, 179, 756, 22, 51, "Print"],
Cell[6747, 203, 503, 14, 50, "Print"],
Cell[7253, 219, 870, 26, 50, "Print"],
Cell[8126, 247, 739, 22, 50, "Print"],
Cell[8868, 271, 540, 16, 51, "Print"],
Cell[9411, 289, 333, 9, 47, "Print"],
Cell[9747, 300, 561, 17, 51, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10345, 322, 44, 1, 56, "Input"],
Cell[10392, 325, 3034, 74, 28, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13463, 404, 44, 1, 56, "Input"],
Cell[13510, 407, 1782, 41, 28, "Print"],
Cell[15295, 450, 413, 11, 46, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15745, 466, 49, 1, 56, "Input"],
Cell[15797, 469, 49, 1, 28, "Print"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

