(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    301152,       7976]*)
(*NotebookOutlinePosition[    301920,       8003]*)
(*  CellTagsIndexPosition[    301876,       7999]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Kerr-Newman SpaceTime", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Initialization of Component Tensors", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[InitCTensor["\<KerrNewman\>"]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Coordinate System = "\[InvisibleSpace]{t, 
          r, \[Theta], \[Phi]}\),
      SequenceForm[ "Coordinate System = ", {t, r, \[Theta], \[Phi]}],
      Editable->False]], "Print"],

Cell[BoxData[
    \({45.024`\ Second, Null}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\^2 + Q\^2 - 2\ M\ r + r\^2 - 
                a\^2\ Sin[\[Theta]]\^2\)\/\(r\^2 + 
                a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Power[ a, 2], 
            Power[ Q, 2], 
            Times[ -2, M, r], 
            Power[ r, 2], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ 
                Sin[ \[Theta]], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ \((\(-Q\^2\) + 
                    2\ M\ r)\)\ Sin[\[Theta]]\^2\)\/\(r\^2 + 
                a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, a, 
          Plus[ 
            Times[ -1, 
              Power[ Q, 2]], 
            Times[ 2, M, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(r\^2 + a\^2\ Cos[\[Theta]]\^2\)\/\(a\^2 + 
              Q\^2 - 2\ M\ r + r\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ -2, M, r], 
              Power[ r, 2]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ a, 2], 
              Power[ 
                Cos[ \[Theta]], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]"\+_\
\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\^2 + a\^2\ Cos[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ a, 2], 
            Power[ 
              Cos[ \[Theta]], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(Sin[\[Theta]]\^2\ \((\((a\^2 + r\^2)\)\^2 - 
                  a\^2\ \((a\^2 + Q\^2 - 2\ M\ r + 
                        r\^2)\)\ Sin[\[Theta]]\^2)\)\)\/\(r\^2 + 
              a\^2\ Cos[\[Theta]]\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Plus[ 
            Power[ 
              Plus[ 
                Power[ a, 2], 
                Power[ r, 2]], 2], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Power[ 
                Sin[ \[Theta]], 2]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\("f"\_\("ab"\)\%\("  c"\)\),
          Subsuperscript[ "f", "ab", "  c"],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "\<\"0\"\>"}],
      SequenceForm[ 
        Subsuperscript[ "f", "ab", "  c"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((\((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + r)\))\)\ \((a\^2\ M + 
                    2\ r\ \((Q\^2 - M\ r)\) + 
                    a\^2\ M\ Cos[
                        2\ \[Theta]])\))\)/\((2\ \((r\^2 + a\^2\ \
Cos[\[Theta]]\^2)\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Rational[ -1, 2], 
          Plus[ 
            Power[ a, 2], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 2, r, 
              Plus[ 
                Power[ Q, 2], 
                Times[ -1, M, r]]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(a\^2\ \((Q\^2 - 2\ M\ r)\)\ Sin[
                2\ \[Theta]]\)\/\(2\ \((r\^2 + a\^2\ Cos[\[Theta]]\^2)\)\^3\)\
\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ a, 2], 
          Plus[ 
            Power[ Q, 2], 
            Times[ -2, M, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ \((a\^2 + r\^2)\)\ \((a\^2\ M + 
                    2\ r\ \((Q\^2 - M\ r)\) + 
                    a\^2\ M\ Cos[2\ \[Theta]])\)\)\/\(\((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ -2, 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 2, r, 
              Plus[ 
                Power[ Q, 2], 
                Times[ -1, M, r]]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ \((a\^2\ M + 2\ r\ \((Q\^2 - M\ r)\) + 
                    a\^2\ M\ Cos[2\ \[Theta]])\)\)\/\(2\ \((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((r\^2 + a\^2\ Cos[\[Theta]]\^2)\)\^2\)\)\
\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ 
          Rational[ -1, 2], a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 2, r, 
              Plus[ 
                Power[ Q, 2], 
                Times[ -1, M, r]]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ a\^2\ \((Q\^2 - 2\ M\ r)\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^2\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ 2, 
          Power[ a, 2], 
          Plus[ 
            Power[ Q, 2], 
            Times[ -2, M, r]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(4\ a\ \((Q\^2 - 
                  2\ M\ r)\)\ Cot[\[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ \
Cos[2\ \[Theta]])\)\^2\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ 4, a, 
          Plus[ 
            Power[ Q, 2], 
            Times[ -2, M, r]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Cot[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((a\ \((a\^2 + Q\^2 + 
                  r\ \((\(-2\)\ M + r)\))\)\ \((a\^2\ M + 
                  2\ r\ \((Q\^2 - M\ r)\) + 
                  a\^2\ M\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((2\ \((r\^2 + \
a\^2\ Cos[\[Theta]]\^2)\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Rational[ 1, 2], a, 
          Plus[ 
            Power[ a, 2], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 2, r, 
              Plus[ 
                Power[ Q, 2], 
                Times[ -1, M, r]]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ \((Q\^2 - 2\ M\ r)\)\ \((a\^2 + 
                    r\^2)\)\ Sin[
                  2\ \[Theta]]\)\/\(2\ \((r\^2 + a\^2\ Cos[\[Theta]]\^2)\)\^3\
\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 
          Rational[ -1, 2], a, 
          Plus[ 
            Power[ Q, 2], 
            Times[ -2, M, r]], 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(r\ \((a\^2 + Q\^2 - M\ r)\) + 
              a\^2\ \((M - r)\)\ Cos[\[Theta]]\^2\)\/\(\((a\^2 + Q\^2 + 
                  r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Times[ r, 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ -1, M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ M, 
                Times[ -1, r]], 
              Power[ 
                Cos[ \[Theta]], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(a\^2\ Cos[\[Theta]]\ Sin[\[Theta]]\)\/\(\((a\
\^2 + Q\^2 + r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 
          Power[ a, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\^2\ Cos[\[Theta]]\ Sin[\[Theta]]\)\/\(r\
\^2 + a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -1, 
          Power[ a, 2], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\/\(r\^2 + a\^2\ Cos[\[Theta]]\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ r, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((2\ a\ \((a\^4\ M + 
                  2\ r\^3\ \((2\ Q\^2 - 3\ M\ r)\) + 
                  3\ a\^2\ r\ \((Q\^2 - M\ r)\) + 
                  a\^2\ \((a\^2\ M + r\ \((Q\^2 - M\ r)\))\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                  Q\^2 + r\ \((\(-2\)\ M + 
                        r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^2)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ 2, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 4], M], 
            Times[ 2, 
              Power[ r, 3], 
              Plus[ 
                Times[ 2, 
                  Power[ Q, 2]], 
                Times[ -3, M, r]]], 
            Times[ 3, 
              Power[ a, 2], r, 
              Plus[ 
                Power[ Q, 2], 
                Times[ -1, M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Times[ 
                  Power[ a, 2], M], 
                Times[ r, 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -1, M, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((a\^4\ M + 3\ a\^4\ r + 8\ a\^2\ Q\^2\ r - 
              12\ a\^2\ M\ r\^2 + 8\ a\^2\ r\^3 + 8\ Q\^2\ r\^3 - 
              16\ M\ r\^4 + 8\ r\^5 + 
              4\ a\^2\ r\ \((a\^2 + r\ \((\(-M\) + 2\ r)\))\)\ Cos[
                  2\ \[Theta]] - 
              a\^4\ \((M - r)\)\ Cos[4\ \[Theta]])\)/\((2\ \((a\^2 + Q\^2 + 
                  r\ \((\(-2\)\ M + 
                        r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^2)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 4], M], 
            Times[ 3, 
              Power[ a, 4], r], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ Q, 2], r], 
            Times[ -12, 
              Power[ a, 2], M, 
              Power[ r, 2]], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ r, 3]], 
            Times[ 8, 
              Power[ Q, 2], 
              Power[ r, 3]], 
            Times[ -16, M, 
              Power[ r, 4]], 
            Times[ 8, 
              Power[ r, 5]], 
            Times[ 4, 
              Power[ a, 2], r, 
              Plus[ 
                Power[ a, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -1, M], 
                    Times[ 2, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ -1, 
              Power[ a, 4], 
              Plus[ M, 
                Times[ -1, r]], 
              Cos[ 
                Times[ 4, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ r\ \((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + r)\))\)\)\/\(a\^2 + 2\ r\^2 + 
                a\^2\ Cos[2\ \[Theta]]\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -2, r, 
          Plus[ 
            Power[ a, 2], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\^2\ Cos[\[Theta]]\ Sin[\[Theta]]\)\/\(r\
\^2 + a\^2\ Cos[\[Theta]]\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ -1, 
          Power[ a, 2], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(4\ a\^3\ \((Q\^2 - 
                    2\ M\ r)\)\ Cos[\[Theta]]\ Sin[\[Theta]]\^3\)\/\((a\^2 + \
2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ -4, 
          Power[ a, 3], 
          Plus[ 
            Power[ Q, 2], 
            Times[ -2, M, r]], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((\((3\ a\^4 - 4\ a\^2\ Q\^2 + 
                  8\ a\^2\ M\ r + 8\ a\^2\ r\^2 + 8\ r\^4 + 
                  4\ a\^2\ \((a\^2 + Q\^2 + 2\ r\ \((\(-M\) + r)\))\)\ Cos[
                      2\ \[Theta]] + 
                  a\^4\ Cos[
                      4\ \[Theta]])\)\ Cot[\[Theta]])\)/\((2\ \((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^2)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 3, 
              Power[ a, 4]], 
            Times[ -4, 
              Power[ a, 2], 
              Power[ Q, 2]], 
            Times[ 8, 
              Power[ a, 2], M, r], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ r, 2]], 
            Times[ 8, 
              Power[ r, 4]], 
            Times[ 4, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ 2, r, 
                  Plus[ 
                    Times[ -1, M], r]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Cot[ \[Theta]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((\((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + r)\))\)\ \((a\^4\ M + 3\ a\^4\ r + 
                    4\ a\^2\ Q\^2\ r - 4\ a\^2\ M\ r\^2 + 8\ a\^2\ r\^3 + 
                    8\ r\^5 + 
                    4\ a\^2\ r\ \((a\^2 - Q\^2 + r\ \((M + 2\ r)\))\)\ Cos[
                        2\ \[Theta]] - 
                    a\^4\ \((M - r)\)\ Cos[
                        4\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -1, 
          Plus[ 
            Power[ a, 2], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 4], M], 
            Times[ 3, 
              Power[ a, 4], r], 
            Times[ 4, 
              Power[ a, 2], 
              Power[ Q, 2], r], 
            Times[ -4, 
              Power[ a, 2], M, 
              Power[ r, 2]], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ r, 3]], 
            Times[ 8, 
              Power[ r, 5]], 
            Times[ 4, 
              Power[ a, 2], r, 
              Plus[ 
                Power[ a, 2], 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ r, 
                  Plus[ M, 
                    Times[ 2, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ -1, 
              Power[ a, 4], 
              Plus[ M, 
                Times[ -1, r]], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((\((3\ a\^6 - 5\ a\^4\ Q\^2 + 
                    10\ a\^4\ M\ r + 11\ a\^4\ r\^2 - 8\ a\^2\ Q\^2\ r\^2 + 
                    16\ a\^2\ M\ r\^3 + 16\ a\^2\ r\^4 + 8\ r\^6 + 
                    4\ a\^2\ \((a\^2 + Q\^2 - 2\ M\ r + r\^2)\)\ \((a\^2 + 
                          2\ r\^2)\)\ Cos[2\ \[Theta]] + 
                    a\^4\ \((a\^2 + Q\^2 - 2\ M\ r + r\^2)\)\ Cos[
                        4\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((16\ \((r\^2 + a\^2\ \
Cos[\[Theta]]\^2)\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 
          Rational[ -1, 16], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Plus[ 
            Times[ 3, 
              Power[ a, 6]], 
            Times[ -5, 
              Power[ a, 4], 
              Power[ Q, 2]], 
            Times[ 10, 
              Power[ a, 4], M, r], 
            Times[ 11, 
              Power[ a, 4], 
              Power[ r, 2]], 
            Times[ -8, 
              Power[ a, 2], 
              Power[ Q, 2], 
              Power[ r, 2]], 
            Times[ 16, 
              Power[ a, 2], M, 
              Power[ r, 3]], 
            Times[ 16, 
              Power[ a, 2], 
              Power[ r, 4]], 
            Times[ 8, 
              Power[ r, 6]], 
            Times[ 4, 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Plus[ 
                Power[ a, 2], 
                Times[ 2, 
                  Power[ r, 2]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Calculation of Curvature Tensors", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[RicciR, Simplify]]\)], "Input"],

Cell[BoxData[
    \({128.32500000000002`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(4\ Q\^2\ \((3\ a\^2 + 
                    2\ \((Q\^2 - 2\ M\ r + r\^2)\) - 
                    a\^2\ Cos[
                        2\ \[Theta]])\)\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -4, 
          Power[ Q, 2], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ 2, 
              Plus[ 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]]], 
            Times[ -1, 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(8\ a\ Q\^2\ \((2\ a\^2 + Q\^2 + 
                  2\ r\ \((\(-M\) + 
                        r)\))\)\ Sin[\[Theta]]\^2\)\/\((a\^2 + 2\ r\^2 + a\^2\
\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 8, a, 
          Power[ Q, 2], 
          Plus[ 
            Times[ 2, 
              Power[ a, 2]], 
            Power[ Q, 2], 
            Times[ 2, r, 
              Plus[ 
                Times[ -1, M], r]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ Q\^2\)\/\(\((a\^2 + Q\^2 + 
                  r\ \((\(-2\)\ M + r)\))\)\ \((a\^2 + 2\ r\^2 + 
                  a\^2\ Cos[2\ \[Theta]])\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 2, 
          Power[ Q, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]"\+_\
\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ Q\^2\)\/\(a\^2 + 2\ r\^2 + 
                a\^2\ Cos[2\ \[Theta]]\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -2, 
          Power[ Q, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ Q\^2\ \((\(-3\)\ a\^4 - 2\ r\^4 - 
                  a\^2\ \((Q\^2 - 2\ M\ r + 5\ r\^2)\) + 
                  a\^2\ \((a\^2 + Q\^2 - 2\ M\ r + r\^2)\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 4, 
          Power[ Q, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ -3, 
              Power[ a, 4]], 
            Times[ -2, 
              Power[ r, 4]], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Times[ 5, 
                  Power[ r, 2]]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[
    \(Tcalc[ScalarR]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "0"}],
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 0],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[RiemannR, Simplify]]\)], "Input"],

Cell[BoxData[
    \({528.5699999999999`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ \((r\^2 + 
                    a\^2\ Cos[\[Theta]]\^2)\)\ \((\(-5\)\ a\^4\ Q\^2 - 
                    4\ a\^2\ Q\^4 + 27\ a\^4\ M\ r + 32\ a\^2\ M\ Q\^2\ r - 
                    48\ a\^2\ M\^2\ r\^2 + 24\ a\^2\ Q\^2\ r\^2 + 
                    24\ Q\^4\ r\^2 + 4\ a\^2\ M\ r\^3 - 64\ M\ Q\^2\ r\^3 + 
                    32\ M\^2\ r\^4 + 24\ Q\^2\ r\^4 - 16\ M\ r\^5 - 
                    4\ a\^2\ \((Q\^4 + M\ \((12\ M - 7\ r)\)\ r\^2 + 
                          2\ Q\^2\ r\ \((\(-4\)\ M + r)\) + 
                          a\^2\ \((Q\^2 - 6\ M\ r)\))\)\ Cos[2\ \[Theta]] + 
                    a\^4\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                        4\ \[Theta]])\))\)/\((\((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^4)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ a, 2], 
              Power[ 
                Cos[ \[Theta]], 2]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -4], 
          Plus[ 
            Times[ -5, 
              Power[ a, 4], 
              Power[ Q, 2]], 
            Times[ -4, 
              Power[ a, 2], 
              Power[ Q, 4]], 
            Times[ 27, 
              Power[ a, 4], M, r], 
            Times[ 32, 
              Power[ a, 2], M, 
              Power[ Q, 2], r], 
            Times[ -48, 
              Power[ a, 2], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 24, 
              Power[ a, 2], 
              Power[ Q, 2], 
              Power[ r, 2]], 
            Times[ 24, 
              Power[ Q, 4], 
              Power[ r, 2]], 
            Times[ 4, 
              Power[ a, 2], M, 
              Power[ r, 3]], 
            Times[ -64, M, 
              Power[ Q, 2], 
              Power[ r, 3]], 
            Times[ 32, 
              Power[ M, 2], 
              Power[ r, 4]], 
            Times[ 24, 
              Power[ Q, 2], 
              Power[ r, 4]], 
            Times[ -16, M, 
              Power[ r, 5]], 
            Times[ -4, 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 4], 
                Times[ M, 
                  Plus[ 
                    Times[ 12, M], 
                    Times[ -7, r]], 
                  Power[ r, 2]], 
                Times[ 2, 
                  Power[ Q, 2], r, 
                  Plus[ 
                    Times[ -4, M], r]], 
                Times[ 
                  Power[ a, 2], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -6, M, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 4, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ \((a\^2\ M + 4\ Q\^2\ r - 
                  6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ a\ \((a\^4\ \((2\ Q\^2 - 9\ M\ r)\) + 
                  2\ r\^2\ \((\(-3\)\ Q\^4 + 4\ Q\^2\ \((2\ M - r)\)\ r + 
                        M\ r\^2\ \((\(-4\)\ M + 3\ r)\))\) + 
                  a\^2\ \((Q\^4 + 3\ M\ \((4\ M - r)\)\ r\^2 - 
                        2\ Q\^2\ r\ \((4\ M + 3\ r)\))\) + 
                  a\^2\ \((Q\^4 + 3\ M\ \((4\ M - 3\ r)\)\ r\^2 + 
                        2\ Q\^2\ r\ \((\(-4\)\ M + r)\) + 
                        a\^2\ \((2\ Q\^2 - 9\ M\ r)\))\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                  Q\^2 + r\ \((\(-2\)\ M + 
                        r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Times[ 2, 
                  Power[ Q, 2]], 
                Times[ -9, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -3, 
                  Power[ Q, 4]], 
                Times[ 4, 
                  Power[ Q, 2], 
                  Plus[ 
                    Times[ 2, M], 
                    Times[ -1, r]], r], 
                Times[ M, 
                  Power[ r, 2], 
                  Plus[ 
                    Times[ -4, M], 
                    Times[ 3, r]]]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 4], 
                Times[ 3, M, 
                  Plus[ 
                    Times[ 4, M], 
                    Times[ -1, r]], 
                  Power[ r, 2]], 
                Times[ -2, 
                  Power[ Q, 2], r, 
                  Plus[ 
                    Times[ 4, M], 
                    Times[ 3, r]]]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 4], 
                Times[ 3, M, 
                  Plus[ 
                    Times[ 4, M], 
                    Times[ -3, r]], 
                  Power[ r, 2]], 
                Times[ 2, 
                  Power[ Q, 2], r, 
                  Plus[ 
                    Times[ -4, M], r]], 
                Times[ 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ 2, 
                      Power[ Q, 2]], 
                    Times[ -9, M, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((a\ \((\(-5\)\ a\^2 - 4\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((a\^2\ M + 4\ Q\^2\ r - 
                    6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, 
          Plus[ 
            Times[ -5, 
              Power[ a, 2]], 
            Times[ -4, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ \((a\^2\ M + 4\ Q\^2\ r - 
                  6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((\(-\(\(4\ a\^2\ \((Q\^2 - 2\ M\ r)\)\^2\ \
\((a\^2 + r\^2)\)\ Cos[\[Theta]]\^2\)\/\(a\^2 + 2\ r\^2 + 
                      a\^2\ Cos[2\ \[Theta]]\)\)\) + 
              r\ \((a\^2 + Q\^2 + r\ \((\(-2\)\ M + r)\))\)\ \((a\^2\ M + 
                    2\ r\ \((Q\^2 - M\ r)\) + a\^2\ M\ Cos[2\ \[Theta]])\) - 
              a\^2\ \((Q\^2 - 
                    2\ M\ r)\)\ \((\((a\^2 + 2\ r\^2)\)\ Cos[2\ \[Theta]] - 
                    a\^2\ \((\(-2\) + Cos[4\ \[Theta]])\))\) + 
              a\^4\ \((Q\^2 - 2\ M\ r)\)\ Cos[\[Theta]]\ Sin[\[Theta]]\ Sin[
                  
                  2\ \[Theta]] + \(a\^4\ \((Q\^2 - 2\ M\ r)\)\^2\ Sin[2\ \
\[Theta]]\^2\)\/\(a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]]\))\)/\((2\ \((r\^2 \
+ a\^2\ Cos[\[Theta]]\^2)\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -3], 
          Plus[ 
            Times[ -4, 
              Power[ a, 2], 
              Power[ 
                Plus[ 
                  Power[ Q, 2], 
                  Times[ -2, M, r]], 2], 
              Plus[ 
                Power[ a, 2], 
                Power[ r, 2]], 
              Power[ 
                Cos[ \[Theta]], 2], 
              Power[ 
                Plus[ 
                  Power[ a, 2], 
                  Times[ 2, 
                    Power[ r, 2]], 
                  Times[ 
                    Power[ a, 2], 
                    Cos[ 
                      Times[ 2, \[Theta]]]]], -1]], 
            Times[ r, 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -2, M], r]]], 
              Plus[ 
                Times[ 
                  Power[ a, 2], M], 
                Times[ 2, r, 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -1, M, r]]], 
                Times[ 
                  Power[ a, 2], M, 
                  Cos[ 
                    Times[ 2, \[Theta]]]]]], 
            Times[ -1, 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -2, M, r]], 
              Plus[ 
                Times[ 
                  Plus[ 
                    Power[ a, 2], 
                    Times[ 2, 
                      Power[ r, 2]]], 
                  Cos[ 
                    Times[ 2, \[Theta]]]], 
                Times[ -1, 
                  Power[ a, 2], 
                  Plus[ -2, 
                    Cos[ 
                      Times[ 4, \[Theta]]]]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -2, M, r]], 
              Cos[ \[Theta]], 
              Sin[ \[Theta]], 
              Sin[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Power[ 
                Plus[ 
                  Power[ Q, 2], 
                  Times[ -2, M, r]], 2], 
              Power[ 
                Plus[ 
                  Power[ a, 2], 
                  Times[ 2, 
                    Power[ r, 2]], 
                  Times[ 
                    Power[ a, 2], 
                    Cos[ 
                      Times[ 2, \[Theta]]]]], -1], 
              Power[ 
                Sin[ 
                  Times[ 2, \[Theta]]], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ a\ \((\(-2\)\ a\^2 - r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((a\^2\ M + 4\ Q\^2\ r - 
                    6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, a, 
          Plus[ 
            Times[ -2, 
              Power[ a, 2]], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ a\ \((a\^4\ \((4\ Q\^2 - 9\ M\ r)\) + 
                    a\^2\ \((Q\^4 - 5\ M\ Q\^2\ r + 
                          3\ M\ \((2\ M - r)\)\ r\^2)\) - 
                    2\ r\^2\ \((Q\^4 + M\ \((2\ M - 3\ r)\)\ r\^2 + 
                          Q\^2\ r\ \((\(-3\)\ M + 2\ r)\))\) + 
                    a\^2\ \((Q\^4 + 3\ M\ \((2\ M - 3\ r)\)\ r\^2 + 
                          Q\^2\ r\ \((\(-5\)\ M + 4\ r)\) + 
                          a\^2\ \((4\ Q\^2 - 9\ M\ r)\))\)\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Times[ 4, 
                  Power[ Q, 2]], 
                Times[ -9, M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 4], 
                Times[ -5, M, 
                  Power[ Q, 2], r], 
                Times[ 3, M, 
                  Plus[ 
                    Times[ 2, M], 
                    Times[ -1, r]], 
                  Power[ r, 2]]]], 
            Times[ -2, 
              Power[ r, 2], 
              Plus[ 
                Power[ Q, 4], 
                Times[ M, 
                  Plus[ 
                    Times[ 2, M], 
                    Times[ -3, r]], 
                  Power[ r, 2]], 
                Times[ 
                  Power[ Q, 2], r, 
                  Plus[ 
                    Times[ -3, M], 
                    Times[ 2, r]]]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 4], 
                Times[ 3, M, 
                  Plus[ 
                    Times[ 2, M], 
                    Times[ -3, r]], 
                  Power[ r, 2]], 
                Times[ 
                  Power[ Q, 2], r, 
                  Plus[ 
                    Times[ -5, M], 
                    Times[ 4, r]]], 
                Times[ 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ 4, 
                      Power[ Q, 2]], 
                    Times[ -9, M, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ \((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + r)\))\)\ \((a\^2\ \((Q\^2 - 
                          3\ M\ r)\) + 2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
                    a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, 
          Plus[ 
            Power[ a, 2], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ \((a\^2\ M + 4\ Q\^2\ r - 6\ M\ r\^2 + 
                    a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(a\^2\ \((Q\^2 - 3\ M\ r)\) + 
              2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
              a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[2\ \[Theta]]\)\/\(2\ \((a\^2 + 
                  Q\^2 + r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((Sin[\[Theta]]\^2\ \((4\ a\^2\ \((\(-Q\^2\) \
+ 2\ M\ r)\)\ \((a\^2 + r\^2)\)\ \((2\ r\^2\ \((4\ Q\^2 - 3\ M\ r)\) + 
                        a\^2\ \((\(-2\)\ Q\^2 + 9\ M\ r)\) - 
                        a\^2\ \((2\ Q\^2 - 9\ M\ r)\)\ Cos[
                            2\ \[Theta]])\)\ Sin[\[Theta]]\^2 + \((5\ a\^4\ Q\
\^2 - 18\ a\^4\ M\ r - 16\ a\^2\ Q\^2\ r\^2 + 4\ a\^2\ M\ r\^3 - 
                        8\ Q\^2\ r\^4 + 8\ M\ r\^5 + 
                        4\ a\^2\ \((r\^2\ \((4\ Q\^2 - 5\ M\ r)\) + 
                              a\^2\ \((Q\^2 - 3\ M\ r)\))\)\ Cos[
                            2\ \[Theta]] - 
                        a\^4\ \((Q\^2 - 6\ M\ r)\)\ Cos[
                            4\ \[Theta]])\)\ \((\((a\^2 + r\^2)\)\^2 - 
                        a\^2\ \((a\^2 + Q\^2 - 2\ M\ r + 
                              r\^2)\)\ Sin[\[Theta]]\^2)\))\))\)/\((\((a\^2 + 
                  Q\^2 + r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Plus[ 
            Times[ 4, 
              Power[ a, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ 2, M, r]], 
              Plus[ 
                Power[ a, 2], 
                Power[ r, 2]], 
              Plus[ 
                Times[ 2, 
                  Power[ r, 2], 
                  Plus[ 
                    Times[ 4, 
                      Power[ Q, 2]], 
                    Times[ -3, M, r]]], 
                Times[ 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ -2, 
                      Power[ Q, 2]], 
                    Times[ 9, M, r]]], 
                Times[ -1, 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ 2, 
                      Power[ Q, 2]], 
                    Times[ -9, M, r]], 
                  Cos[ 
                    Times[ 2, \[Theta]]]]], 
              Power[ 
                Sin[ \[Theta]], 2]], 
            Times[ 
              Plus[ 
                Times[ 5, 
                  Power[ a, 4], 
                  Power[ Q, 2]], 
                Times[ -18, 
                  Power[ a, 4], M, r], 
                Times[ -16, 
                  Power[ a, 2], 
                  Power[ Q, 2], 
                  Power[ r, 2]], 
                Times[ 4, 
                  Power[ a, 2], M, 
                  Power[ r, 3]], 
                Times[ -8, 
                  Power[ Q, 2], 
                  Power[ r, 4]], 
                Times[ 8, M, 
                  Power[ r, 5]], 
                Times[ 4, 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ 
                      Power[ r, 2], 
                      Plus[ 
                        Times[ 4, 
                          Power[ Q, 2]], 
                        Times[ -5, M, r]]], 
                    Times[ 
                      Power[ a, 2], 
                      Plus[ 
                        Power[ Q, 2], 
                        Times[ -3, M, r]]]], 
                  Cos[ 
                    Times[ 2, \[Theta]]]], 
                Times[ -1, 
                  Power[ a, 4], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -6, M, r]], 
                  Cos[ 
                    Times[ 4, \[Theta]]]]], 
              Plus[ 
                Power[ 
                  Plus[ 
                    Power[ a, 2], 
                    Power[ r, 2]], 2], 
                Times[ -1, 
                  Power[ a, 2], 
                  Plus[ 
                    Power[ a, 2], 
                    Power[ Q, 2], 
                    Times[ -2, M, r], 
                    Power[ r, 2]], 
                  Power[ 
                    Sin[ \[Theta]], 2]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((12\ a\^2\ \((a\^2 + 
                  r\^2)\)\ Cos[\[Theta]]\ \((a\^2\ M + 4\ Q\^2\ r - 
                  6\ M\ r\^2 + 
                  a\^2\ M\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^3)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 12, 
          Power[ a, 2], 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((Sin[\[Theta]]\^2\ \((\(-4\)\ a\^2\ \
\((\(-Q\^2\) + 2\ M\ r)\)\ \((a\^2 + 
                        r\^2)\)\ \((2\ r\^2\ \((2\ Q\^2 - 3\ M\ r)\) + 
                        a\^2\ \((\(-4\)\ Q\^2 + 9\ M\ r)\) - 
                        a\^2\ \((4\ Q\^2 - 9\ M\ r)\)\ Cos[
                            2\ \[Theta]])\)\ Sin[\[Theta]]\^2 + \((\(-13\)\ a\
\^4\ Q\^2 + 27\ a\^4\ M\ r + 4\ a\^2\ M\ r\^3 + 8\ Q\^2\ r\^4 - 16\ M\ r\^5 - 
                        4\ a\^2\ \((r\^2\ \((4\ Q\^2 - 7\ M\ r)\) + 
                              3\ a\^2\ \((Q\^2 - 2\ M\ r)\))\)\ Cos[
                            2\ \[Theta]] + 
                        a\^4\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                            4\ \[Theta]])\)\ \((\((a\^2 + r\^2)\)\^2 - 
                        a\^2\ \((a\^2 + Q\^2 - 2\ M\ r + 
                              r\^2)\)\ Sin[\[Theta]]\^2)\))\))\)/\((\((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Plus[ 
            Times[ -4, 
              Power[ a, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ 2, M, r]], 
              Plus[ 
                Power[ a, 2], 
                Power[ r, 2]], 
              Plus[ 
                Times[ 2, 
                  Power[ r, 2], 
                  Plus[ 
                    Times[ 2, 
                      Power[ Q, 2]], 
                    Times[ -3, M, r]]], 
                Times[ 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ -4, 
                      Power[ Q, 2]], 
                    Times[ 9, M, r]]], 
                Times[ -1, 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ 4, 
                      Power[ Q, 2]], 
                    Times[ -9, M, r]], 
                  Cos[ 
                    Times[ 2, \[Theta]]]]], 
              Power[ 
                Sin[ \[Theta]], 2]], 
            Times[ 
              Plus[ 
                Times[ -13, 
                  Power[ a, 4], 
                  Power[ Q, 2]], 
                Times[ 27, 
                  Power[ a, 4], M, r], 
                Times[ 4, 
                  Power[ a, 2], M, 
                  Power[ r, 3]], 
                Times[ 8, 
                  Power[ Q, 2], 
                  Power[ r, 4]], 
                Times[ -16, M, 
                  Power[ r, 5]], 
                Times[ -4, 
                  Power[ a, 2], 
                  Plus[ 
                    Times[ 
                      Power[ r, 2], 
                      Plus[ 
                        Times[ 4, 
                          Power[ Q, 2]], 
                        Times[ -7, M, r]]], 
                    Times[ 3, 
                      Power[ a, 2], 
                      Plus[ 
                        Power[ Q, 2], 
                        Times[ -2, M, r]]]], 
                  Cos[ 
                    Times[ 2, \[Theta]]]], 
                Times[ 
                  Power[ a, 4], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -3, M, r]], 
                  Cos[ 
                    Times[ 4, \[Theta]]]]], 
              Plus[ 
                Power[ 
                  Plus[ 
                    Power[ a, 2], 
                    Power[ r, 2]], 2], 
                Times[ -1, 
                  Power[ a, 2], 
                  Plus[ 
                    Power[ a, 2], 
                    Power[ Q, 2], 
                    Times[ -2, M, r], 
                    Power[ r, 2]], 
                  Power[ 
                    Sin[ \[Theta]], 2]]]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[WeylC, Simplify]]\)], "Input"],

Cell[BoxData[
    \({77.43099999999993`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[WeylC]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"C"\_"[ab][cd]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "C", "[ab][cd]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((\(-9\)\ a\^4\ Q\^2 - 8\ a\^2\ Q\^4 + 
                27\ a\^4\ M\ r + 40\ a\^2\ M\ Q\^2\ r - 
                48\ a\^2\ M\^2\ r\^2 + 12\ a\^2\ Q\^2\ r\^2 + 
                16\ Q\^4\ r\^2 + 4\ a\^2\ M\ r\^3 - 48\ M\ Q\^2\ r\^3 + 
                32\ M\^2\ r\^4 + 16\ Q\^2\ r\^4 - 16\ M\ r\^5 - 
                4\ a\^2\ \((2\ Q\^4 + M\ \((12\ M - 7\ r)\)\ r\^2 + 
                      Q\^2\ r\ \((\(-10\)\ M + 3\ r)\) + 
                      2\ a\^2\ \((Q\^2 - 3\ M\ r)\))\)\ Cos[2\ \[Theta]] + 
                a\^4\ \((Q\^2 - 3\ M\ r)\)\ Cos[4\ \[Theta]])\)/\((\((a\^2 + 
                    Q\^2 + r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ -9, 
              Power[ a, 4], 
              Power[ Q, 2]], 
            Times[ -8, 
              Power[ a, 2], 
              Power[ Q, 4]], 
            Times[ 27, 
              Power[ a, 4], M, r], 
            Times[ 40, 
              Power[ a, 2], M, 
              Power[ Q, 2], r], 
            Times[ -48, 
              Power[ a, 2], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 12, 
              Power[ a, 2], 
              Power[ Q, 2], 
              Power[ r, 2]], 
            Times[ 16, 
              Power[ Q, 4], 
              Power[ r, 2]], 
            Times[ 4, 
              Power[ a, 2], M, 
              Power[ r, 3]], 
            Times[ -48, M, 
              Power[ Q, 2], 
              Power[ r, 3]], 
            Times[ 32, 
              Power[ M, 2], 
              Power[ r, 4]], 
            Times[ 16, 
              Power[ Q, 2], 
              Power[ r, 4]], 
            Times[ -16, M, 
              Power[ r, 5]], 
            Times[ -4, 
              Power[ a, 2], 
              Plus[ 
                Times[ 2, 
                  Power[ Q, 4]], 
                Times[ M, 
                  Plus[ 
                    Times[ 12, M], 
                    Times[ -7, r]], 
                  Power[ r, 2]], 
                Times[ 
                  Power[ Q, 2], r, 
                  Plus[ 
                    Times[ -10, M], 
                    Times[ 3, r]]], 
                Times[ 2, 
                  Power[ a, 2], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -3, M, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 4, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ \((a\^2\ M + 4\ Q\^2\ r - 
                  6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ a\ \((3\ a\^2 + 2\ Q\^2 + 
                  r\ \((\(-4\)\ M + 3\ r)\))\)\ \((a\^2\ \((Q\^2 - 
                        3\ M\ r)\) + 2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
                  a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                  Q\^2 + r\ \((\(-2\)\ M + 
                        r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ Q, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -4, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((a\ \((\(-5\)\ a\^2 - 4\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((a\^2\ M + 4\ Q\^2\ r - 
                    6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, 
          Plus[ 
            Times[ -5, 
              Power[ a, 2]], 
            Times[ -4, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(6\ a\^2\ \((a\^2\ M + 4\ Q\^2\ r - 
                  6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)\
\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 6, 
          Power[ a, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((2\ \((\(-3\)\ a\^4\ Q\^2 - 2\ a\^2\ Q\^4 + 
                  9\ a\^4\ M\ r + 10\ a\^2\ M\ Q\^2\ r - 
                  12\ a\^2\ M\^2\ r\^2 + 6\ a\^2\ Q\^2\ r\^2 + 
                  4\ Q\^4\ r\^2 - 2\ a\^2\ M\ r\^3 - 12\ M\ Q\^2\ r\^3 + 
                  8\ M\^2\ r\^4 + 4\ Q\^2\ r\^4 - 4\ M\ r\^5 - 
                  2\ a\^2\ \((Q\^4 + M\ \((6\ M - 5\ r)\)\ r\^2 + 
                        Q\^2\ r\ \((\(-5\)\ M + 3\ r)\) + 
                        a\^2\ \((Q\^2 - 3\ M\ r)\))\)\ Cos[2\ \[Theta]] + 
                  a\^4\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                      4\ \[Theta]])\))\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 2, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ -3, 
              Power[ a, 4], 
              Power[ Q, 2]], 
            Times[ -2, 
              Power[ a, 2], 
              Power[ Q, 4]], 
            Times[ 9, 
              Power[ a, 4], M, r], 
            Times[ 10, 
              Power[ a, 2], M, 
              Power[ Q, 2], r], 
            Times[ -12, 
              Power[ a, 2], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 6, 
              Power[ a, 2], 
              Power[ Q, 2], 
              Power[ r, 2]], 
            Times[ 4, 
              Power[ Q, 4], 
              Power[ r, 2]], 
            Times[ -2, 
              Power[ a, 2], M, 
              Power[ r, 3]], 
            Times[ -12, M, 
              Power[ Q, 2], 
              Power[ r, 3]], 
            Times[ 8, 
              Power[ M, 2], 
              Power[ r, 4]], 
            Times[ 4, 
              Power[ Q, 2], 
              Power[ r, 4]], 
            Times[ -4, M, 
              Power[ r, 5]], 
            Times[ -2, 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 4], 
                Times[ M, 
                  Plus[ 
                    Times[ 6, M], 
                    Times[ -5, r]], 
                  Power[ r, 2]], 
                Times[ 
                  Power[ Q, 2], r, 
                  Plus[ 
                    Times[ -5, M], 
                    Times[ 3, r]]], 
                Times[ 
                  Power[ a, 2], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -3, M, r]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 4, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ a\ \((\(-2\)\ a\^2 - r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((a\^2\ M + 4\ Q\^2\ r - 
                    6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, a, 
          Plus[ 
            Times[ -2, 
              Power[ a, 2]], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ a\ \((3\ a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + 3\ r)\))\)\ \((a\^2\ \((Q\^2 - 
                          3\ M\ r)\) + 2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
                    a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, a, 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ \((a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + r)\))\)\ \((a\^2\ \((Q\^2 - 
                          3\ M\ r)\) + 2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
                    a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, 
          Plus[ 
            Power[ a, 2], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ \((a\^2\ M + 4\ Q\^2\ r - 6\ M\ r\^2 + 
                    a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(a\ \((a\^2\ M + 4\ Q\^2\ r - 6\ M\ r\^2 + 
                    a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]]\)\/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -2], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(a\^2\ \((Q\^2 - 3\ M\ r)\) + 
              2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
              a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[2\ \[Theta]]\)\/\(2\ \((a\^2 + 
                  Q\^2 + r\ \((\(-2\)\ M + r)\))\)\ \((r\^2 + 
                  a\^2\ Cos[\[Theta]]\^2)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ \[Theta]], 2]]], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((4\ a\ \((3\ a\^2 + 2\ Q\^2 + 
                  r\ \((\(-4\)\ M + 3\ r)\))\)\ \((a\^2\ \((Q\^2 - 
                        3\ M\ r)\) + 2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
                  a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                  Q\^2 + r\ \((\(-2\)\ M + 
                        r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 4, a, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Times[ 2, 
              Power[ Q, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -4, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ a\ \((\(-2\)\ a\^2 - r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((a\^2\ M + 4\ Q\^2\ r - 
                    6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, a, 
          Plus[ 
            Times[ -2, 
              Power[ a, 2]], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((2\ \((\(-3\)\ a\^6\ Q\^2 - a\^4\ Q\^4 + 
                    9\ a\^6\ M\ r + 5\ a\^4\ M\ Q\^2\ r - 
                    6\ a\^4\ M\^2\ r\^2 + 3\ a\^4\ Q\^2\ r\^2 + 
                    4\ a\^2\ Q\^4\ r\^2 + 7\ a\^4\ M\ r\^3 - 
                    12\ a\^2\ M\ Q\^2\ r\^3 + 8\ a\^2\ M\^2\ r\^4 + 
                    10\ a\^2\ Q\^2\ r\^4 - 6\ a\^2\ M\ r\^5 + 4\ Q\^2\ r\^6 - 
                    4\ M\ r\^7 - 
                    2\ a\^2\ \((a\^4\ \((Q\^2 - 3\ M\ r)\) + 
                          4\ a\^2\ r\^2\ \((Q\^2 - 2\ M\ r)\) + 
                          r\^2\ \((2\ Q\^4 + M\ \((4\ M - 5\ r)\)\ r\^2 + 
                                3\ Q\^2\ r\ \((\(-2\)\ M + r)\))\))\)\ Cos[
                        2\ \[Theta]] + 
                    a\^4\ \((Q\^2 - 3\ M\ r)\)\ \((a\^2 + Q\^2 - 2\ M\ r + 
                          r\^2)\)\ Cos[
                        4\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((\((a\^2 + 
                    Q\^2 + r\ \((\(-2\)\ M + 
                          r)\))\)\ \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \
\[Theta]])\)\^3)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Power[ Q, 2], 
              Times[ r, 
                Plus[ 
                  Times[ -2, M], r]]], -1], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ -3, 
              Power[ a, 6], 
              Power[ Q, 2]], 
            Times[ -1, 
              Power[ a, 4], 
              Power[ Q, 4]], 
            Times[ 9, 
              Power[ a, 6], M, r], 
            Times[ 5, 
              Power[ a, 4], M, 
              Power[ Q, 2], r], 
            Times[ -6, 
              Power[ a, 4], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 4], 
              Power[ Q, 2], 
              Power[ r, 2]], 
            Times[ 4, 
              Power[ a, 2], 
              Power[ Q, 4], 
              Power[ r, 2]], 
            Times[ 7, 
              Power[ a, 4], M, 
              Power[ r, 3]], 
            Times[ -12, 
              Power[ a, 2], M, 
              Power[ Q, 2], 
              Power[ r, 3]], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ M, 2], 
              Power[ r, 4]], 
            Times[ 10, 
              Power[ a, 2], 
              Power[ Q, 2], 
              Power[ r, 4]], 
            Times[ -6, 
              Power[ a, 2], M, 
              Power[ r, 5]], 
            Times[ 4, 
              Power[ Q, 2], 
              Power[ r, 6]], 
            Times[ -4, M, 
              Power[ r, 7]], 
            Times[ -2, 
              Power[ a, 2], 
              Plus[ 
                Times[ 
                  Power[ a, 4], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -3, M, r]]], 
                Times[ 4, 
                  Power[ a, 2], 
                  Power[ r, 2], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -2, M, r]]], 
                Times[ 
                  Power[ r, 2], 
                  Plus[ 
                    Times[ 2, 
                      Power[ Q, 4]], 
                    Times[ M, 
                      Plus[ 
                        Times[ 4, M], 
                        Times[ -5, r]], 
                      Power[ r, 2]], 
                    Times[ 3, 
                      Power[ Q, 2], r, 
                      Plus[ 
                        Times[ -2, M], r]]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((12\ a\^2\ \((a\^2 + 
                  r\^2)\)\ Cos[\[Theta]]\ \((a\^2\ M + 4\ Q\^2\ r - 
                  6\ M\ r\^2 + 
                  a\^2\ M\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^3)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 12, 
          Power[ a, 2], 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((a\ \((\(-5\)\ a\^2 - 4\ r\^2 + 
                    a\^2\ Cos[2\ \[Theta]])\)\ \((a\^2\ M + 4\ Q\^2\ r - 
                    6\ M\ r\^2 + a\^2\ M\ Cos[2\ \[Theta]])\)\ Sin[
                  2\ \[Theta]])\)/\((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\
\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, a, 
          Plus[ 
            Times[ -5, 
              Power[ a, 2]], 
            Times[ -4, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Sin[ 
            Times[ 2, \[Theta]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\((4\ a\ \((3\ a\^2 + Q\^2 + 
                    r\ \((\(-2\)\ M + 3\ r)\))\)\ \((a\^2\ \((Q\^2 - 
                          3\ M\ r)\) + 2\ r\^2\ \((\(-Q\^2\) + M\ r)\) + 
                    a\^2\ \((Q\^2 - 3\ M\ r)\)\ Cos[
                        2\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ \
r\^2 + a\^2\ Cos[2\ \[Theta]])\)\^3\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -4, a, 
          Plus[ 
            Times[ 3, 
              Power[ a, 2]], 
            Power[ Q, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], 
                Times[ 3, r]]]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]]], 
            Times[ 2, 
              Power[ r, 2], 
              Plus[ 
                Times[ -1, 
                  Power[ Q, 2]], 
                Times[ M, r]]], 
            Times[ 
              Power[ a, 2], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((12\ a\^2\ \((a\^2 + 
                  r\^2)\)\ Cos[\[Theta]]\ \((a\^2\ M + 4\ Q\^2\ r - 
                  6\ M\ r\^2 + 
                  a\^2\ M\ Cos[
                      2\ \[Theta]])\)\ Sin[\[Theta]]\^3)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 12, 
          Power[ a, 2], 
          Plus[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Cos[ \[Theta]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ 
              Power[ a, 2], M], 
            Times[ 4, 
              Power[ Q, 2], r], 
            Times[ -6, M, 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], M, 
              Cos[ 
                Times[ 2, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((\((\(-9\)\ a\^6\ Q\^2 - a\^4\ Q\^4 + 
                  27\ a\^6\ M\ r + 5\ a\^4\ M\ Q\^2\ r - 
                  6\ a\^4\ M\^2\ r\^2 + 3\ a\^4\ Q\^2\ r\^2 + 
                  4\ a\^2\ Q\^4\ r\^2 + 31\ a\^4\ M\ r\^3 - 
                  12\ a\^2\ M\ Q\^2\ r\^3 + 8\ a\^2\ M\^2\ r\^4 + 
                  28\ a\^2\ Q\^2\ r\^4 - 12\ a\^2\ M\ r\^5 + 16\ Q\^2\ r\^6 - 
                  16\ M\ r\^7 - 
                  4\ a\^2\ \((a\^2\ r\^2\ \((5\ Q\^2 - 13\ M\ r)\) + 
                        2\ a\^4\ \((Q\^2 - 3\ M\ r)\) + 
                        r\^2\ \((Q\^4 + M\ \((2\ M - 7\ r)\)\ r\^2 + 
                              3\ Q\^2\ r\ \((\(-M\) + r)\))\))\)\ Cos[
                      2\ \[Theta]] + 
                  a\^4\ \((Q\^2 - 3\ M\ r)\)\ \((a\^2 + Q\^2 - 2\ M\ r + 
                        r\^2)\)\ Cos[
                      4\ \[Theta]])\)\ Sin[\[Theta]]\^2)\)/\((a\^2 + 2\ r\^2 \
+ a\^2\ Cos[2\ \[Theta]])\)\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ 2, 
                Power[ r, 2]], 
              Times[ 
                Power[ a, 2], 
                Cos[ 
                  Times[ 2, \[Theta]]]]], -3], 
          Plus[ 
            Times[ -9, 
              Power[ a, 6], 
              Power[ Q, 2]], 
            Times[ -1, 
              Power[ a, 4], 
              Power[ Q, 4]], 
            Times[ 27, 
              Power[ a, 6], M, r], 
            Times[ 5, 
              Power[ a, 4], M, 
              Power[ Q, 2], r], 
            Times[ -6, 
              Power[ a, 4], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 3, 
              Power[ a, 4], 
              Power[ Q, 2], 
              Power[ r, 2]], 
            Times[ 4, 
              Power[ a, 2], 
              Power[ Q, 4], 
              Power[ r, 2]], 
            Times[ 31, 
              Power[ a, 4], M, 
              Power[ r, 3]], 
            Times[ -12, 
              Power[ a, 2], M, 
              Power[ Q, 2], 
              Power[ r, 3]], 
            Times[ 8, 
              Power[ a, 2], 
              Power[ M, 2], 
              Power[ r, 4]], 
            Times[ 28, 
              Power[ a, 2], 
              Power[ Q, 2], 
              Power[ r, 4]], 
            Times[ -12, 
              Power[ a, 2], M, 
              Power[ r, 5]], 
            Times[ 16, 
              Power[ Q, 2], 
              Power[ r, 6]], 
            Times[ -16, M, 
              Power[ r, 7]], 
            Times[ -4, 
              Power[ a, 2], 
              Plus[ 
                Times[ 
                  Power[ a, 2], 
                  Power[ r, 2], 
                  Plus[ 
                    Times[ 5, 
                      Power[ Q, 2]], 
                    Times[ -13, M, r]]], 
                Times[ 2, 
                  Power[ a, 4], 
                  Plus[ 
                    Power[ Q, 2], 
                    Times[ -3, M, r]]], 
                Times[ 
                  Power[ r, 2], 
                  Plus[ 
                    Power[ Q, 4], 
                    Times[ M, 
                      Plus[ 
                        Times[ 2, M], 
                        Times[ -7, r]], 
                      Power[ r, 2]], 
                    Times[ 3, 
                      Power[ Q, 2], r, 
                      Plus[ 
                        Times[ -1, M], r]]]]], 
              Cos[ 
                Times[ 2, \[Theta]]]], 
            Times[ 
              Power[ a, 4], 
              Plus[ 
                Power[ Q, 2], 
                Times[ -3, M, r]], 
              Plus[ 
                Power[ a, 2], 
                Power[ Q, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Cos[ 
                Times[ 4, \[Theta]]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Laplacian", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Laplacian[f[r, \[Theta]], Simplify]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"(", 
        RowBox[{"2", " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{\(Cot[\[Theta]]\), " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((0, 1)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{
                SuperscriptBox["f", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(r, \[Theta]\), "]"}], "-", 
              
              RowBox[{"2", " ", "M", " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((1, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{"2", " ", "r", " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((1, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{\(a\^2\), " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((2, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{\(Q\^2\), " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((2, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "-", 
              RowBox[{"2", " ", "M", " ", "r", " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((2, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}], 
              "+", 
              RowBox[{\(r\^2\), " ", 
                RowBox[{
                  SuperscriptBox["f", 
                    TagBox[\((2, 0)\),
                      Derivative],
                    MultilineFunction->None], "[", \(r, \[Theta]\), "]"}]}]}],
             ")"}]}], ")"}], 
      "/", \((a\^2 + 2\ r\^2 + a\^2\ Cos[2\ \[Theta]])\)}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{636, 689},
WindowMargins->{{16, Automatic}, {Automatic, 4}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 54, 1, 43, "Title"],
Cell[1774, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1851, 57, 60, 0, 54, "Subsubsection"],

Cell[CellGroupData[{
Cell[1936, 61, 70, 1, 56, "Input"],
Cell[2009, 64, 220, 4, 28, "Print"],
Cell[2232, 70, 57, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2338, 77, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[2400, 81, 46, 1, 56, "Input"],
Cell[2449, 84, 184, 4, 30, "Print"],
Cell[2636, 90, 2245, 60, 50, "Print"],
Cell[4884, 152, 2175, 57, 50, "Print"],
Cell[7062, 211, 2077, 55, 49, "Print"],
Cell[9142, 268, 1914, 49, 33, "Print"],
Cell[11059, 319, 2574, 70, 51, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13670, 394, 49, 1, 56, "Input"],
Cell[13722, 397, 350, 8, 31, "Print"],
Cell[14075, 407, 372, 9, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14484, 421, 46, 1, 56, "Input"],
Cell[14533, 424, 392, 8, 31, "Print"],
Cell[14928, 434, 3642, 95, 83, "Print"],
Cell[18573, 531, 3224, 81, 51, "Print"],
Cell[21800, 614, 3809, 100, 51, "Print"],
Cell[25612, 716, 3681, 95, 51, "Print"],
Cell[29296, 813, 3290, 83, 51, "Print"],
Cell[32589, 898, 3260, 81, 51, "Print"],
Cell[35852, 981, 3749, 98, 86, "Print"],
Cell[39604, 1081, 3365, 85, 51, "Print"],
Cell[42972, 1168, 3624, 94, 50, "Print"],
Cell[46599, 1264, 3341, 84, 50, "Print"],
Cell[49943, 1350, 3121, 77, 50, "Print"],
Cell[53067, 1429, 3025, 73, 44, "Print"],
Cell[56095, 1504, 4354, 116, 87, "Print"],
Cell[60452, 1622, 4732, 128, 113, "Print"],
Cell[65187, 1752, 3336, 83, 50, "Print"],
Cell[68526, 1837, 3170, 77, 50, "Print"],
Cell[71699, 1916, 3353, 84, 51, "Print"],
Cell[75055, 2002, 4202, 110, 88, "Print"],
Cell[79260, 2114, 4624, 124, 130, "Print"],
Cell[83887, 2240, 4846, 129, 132, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[88782, 2375, 57, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[88864, 2379, 64, 1, 56, "Input"],
Cell[88931, 2382, 69, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[89037, 2388, 45, 1, 56, "Input"],
Cell[89085, 2391, 184, 4, 29, "Print"],
Cell[89272, 2397, 2488, 68, 51, "Print"],
Cell[91763, 2467, 2383, 64, 51, "Print"],
Cell[94149, 2533, 2284, 61, 50, "Print"],
Cell[96436, 2596, 2113, 55, 50, "Print"],
Cell[98552, 2653, 2928, 81, 84, "Print"]
}, Open  ]],
Cell[101495, 2737, 47, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[101567, 2742, 46, 1, 56, "Input"],
Cell[101616, 2745, 361, 9, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[102014, 2759, 66, 1, 56, "Input"],
Cell[102083, 2762, 68, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[102188, 2768, 47, 1, 56, "Input"],
Cell[102238, 2771, 293, 6, 29, "Print"],
Cell[102534, 2779, 7162, 192, 198, "Print"],
Cell[109699, 2973, 4723, 119, 51, "Print"],
Cell[114425, 3094, 6808, 179, 155, "Print"],
Cell[121236, 3275, 5065, 129, 83, "Print"],
Cell[126304, 3406, 4739, 119, 51, "Print"],
Cell[131046, 3527, 7453, 197, 193, "Print"],
Cell[138502, 3726, 5053, 128, 83, "Print"],
Cell[143558, 3856, 6527, 168, 152, "Print"],
Cell[150088, 4026, 5249, 134, 84, "Print"],
Cell[155340, 4162, 4737, 118, 51, "Print"],
Cell[160080, 4282, 5153, 131, 79, "Print"],
Cell[165236, 4415, 8478, 223, 251, "Print"],
Cell[173717, 4640, 4983, 126, 84, "Print"],
Cell[178703, 4768, 8198, 212, 204, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[186938, 4985, 63, 1, 56, "Input"],
Cell[187004, 4988, 68, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[187109, 4994, 44, 1, 56, "Input"],
Cell[187156, 4997, 192, 4, 29, "Print"],
Cell[187351, 5003, 6954, 186, 175, "Print"],
Cell[194308, 5191, 4723, 119, 51, "Print"],
Cell[199034, 5312, 5580, 146, 108, "Print"],
Cell[204617, 5460, 5065, 129, 83, "Print"],
Cell[209685, 5591, 4739, 119, 51, "Print"],
Cell[214427, 5712, 6728, 177, 150, "Print"],
Cell[221158, 5891, 5053, 128, 83, "Print"],
Cell[226214, 6021, 5383, 137, 106, "Print"],
Cell[231600, 6160, 5249, 134, 84, "Print"],
Cell[236852, 6296, 4737, 118, 51, "Print"],
Cell[241592, 6416, 4741, 118, 51, "Print"],
Cell[246336, 6536, 5153, 131, 79, "Print"],
Cell[251492, 6669, 5592, 146, 108, "Print"],
Cell[257087, 6817, 5049, 128, 83, "Print"],
Cell[262139, 6947, 8113, 217, 223, "Print"],
Cell[270255, 7166, 4983, 126, 84, "Print"],
Cell[275241, 7294, 5070, 128, 83, "Print"],
Cell[280314, 7424, 5372, 136, 106, "Print"],
Cell[285689, 7562, 4976, 125, 84, "Print"],
Cell[290668, 7689, 7901, 209, 198, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[298618, 7904, 34, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[298677, 7908, 68, 1, 56, "Input"],
Cell[298748, 7911, 2376, 61, 123, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

