(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    107858,       2857]*)
(*NotebookOutlinePosition[    108626,       2884]*)
(*  CellTagsIndexPosition[    108582,       2880]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Killing Basis in 2-D Symmetric SpaceTime", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {\[Tau], \[Lambda]};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(coMetric\  = \ {{0, \(-\(1\/\(2 \((1\  - \ \(k\/4\) \[Tau]\ \
\[Lambda])\)\^2\)\)\)}, {\(-\(1\/\(2 \((1\  - \ \(k\/4\) \[Tau]\ \[Lambda])\)\
\^2\)\)\), 0}}; \ MatrixForm[coMetric]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              "0", \(-\(1\/\(2\ \((1 - \(k\ \[Lambda]\ \
\[Tau]\)\/4)\)\^2\)\)\)},
            {\(-\(1\/\(2\ \((1 - \(k\ \[Lambda]\ \[Tau]\)\/4)\)\^2\)\)\), 
              "0"}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Killing Basis", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`K\_a\%A\  = \ \(h\_a\^\[Mu]\) \[PartialD]\_\[Mu]\%A, \ \
\ \ \ \ K\_1\%A\  = \ \[PartialD]\_\[Tau]\%A\ \(-\ \(k\/4\)\) \
\(\[Lambda]\^2\) \[PartialD]\_\[Lambda]\%A, \ \ \ \ \ K\_2\%A\  = \ \(-\(k\/4\
\)\) \(\[Tau]\^2\) \[PartialD]\_\[Tau]\%A\ \(+\ \
\[PartialD]\_\[Lambda]\%A\)\)]]], "Text",
  TextAlignment->Center],

Cell[CellGroupData[{

Cell[BoxData[
    \(hMatrix\  = \ {{1, \ \(-\(k\/4\)\) \[Lambda]\^2}, {\(-\(k\/4\)\) \[Tau]\
\^2, 1}}; MatrixForm[hMatrix]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"1", \(-\(\(k\ \[Lambda]\^2\)\/4\)\)},
            {\(-\(\(k\ \[Tau]\^2\)\/4\)\), "1"}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell["Non-Coordinate Metric", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`g\_\(a\[InvisibleSpace]b\)\  = \ \(\(h\_a\^\[Mu]\) \(h\
\_b\^\[Nu]\) 
          g\_\[Mu]\[Nu]\  = \ \((h\ g\ \
h\^T)\)\_\(a\[InvisibleSpace]b\)\)\)]]], "Text",
  TextAlignment->Center],

Cell[CellGroupData[{

Cell[BoxData[
    \(nonCoMetric\  = \ 
      hMatrix\  . \ coMetric\  . \ Transpose[hMatrix]\  // \ Simplify; \ 
    MatrixForm[nonCoMetric]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(\(4\ k\ \[Lambda]\^2\)\/\((\(-4\) + k\ \[Lambda]\ \[Tau])\)\^2\
\), \(-\(\(16 + 
                      k\^2\ \[Lambda]\^2\ \[Tau]\^2\)\/\(2\ \((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\)\)\)},
            {\(-\(\(16 + 
                      k\^2\ \[Lambda]\^2\ \[Tau]\^2\)\/\(2\ \((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\)\)\), \(\(4\ k\ \[Tau]\^2\)\/\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensors", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[InitCTensor[coSys, \ nonCoMetric, hMatrix]]\)], "Input"],

Cell[BoxData[
    \({1.2519999999999998`\ Second, Null}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(4\ k\ \[Lambda]\^2\)\/\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 1]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 4, k, 
          Power[ \[Lambda], 2], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(16 + 
                k\^2\ \[Lambda]\^2\ \[Tau]\^2\)\/\(2\ \((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Rational[ -1, 2], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Plus[ 16, 
            Times[ 
              Power[ k, 2], 
              Power[ \[Lambda], 2], 
              Power[ \[Tau], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(4\ k\ \[Tau]\^2\)\/\((\(-4\) + k\ \[Lambda]\ \
\[Tau])\)\^2\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 2]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 4, k, 
          Power[ \[Tau], 2], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ k\ \[Tau]\)\/\(4 + 
                k\ \[Lambda]\ \[Tau]\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ f, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ -2, k, \[Tau], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ k\ \[Lambda]\)\/\(4 + 
              k\ \[Lambda]\ \[Tau]\)\)}],
      SequenceForm[ 
        Subsuperscript[ f, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ 2, k, \[Lambda], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ \((16\ k\ \[Lambda] + 
                  k\^3\ \[Lambda]\^3\ \[Tau]\^2)\)\)\/\(\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\ \((4 + k\ \[Lambda]\ \[Tau])\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 2, 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1], 
          Plus[ 
            Times[ 16, k, \[Lambda]], 
            Times[ 
              Power[ k, 3], 
              Power[ \[Lambda], 3], 
              Power[ \[Tau], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(16\ k\^2\ \[Lambda]\^3\)\/\(\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\ \((4 + k\ \[Lambda]\ \[Tau])\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ 16, 
          Power[ k, 2], 
          Power[ \[Lambda], 3], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ \((16\ k\ \[Tau] + 
                    k\^3\ \[Lambda]\^2\ \[Tau]\^3)\)\)\/\(\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\ \((4 + k\ \[Lambda]\ \[Tau])\)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ -2, 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1], 
          Plus[ 
            Times[ 16, k, \[Tau]], 
            Times[ 
              Power[ k, 3], 
              Power[ \[Lambda], 2], 
              Power[ \[Tau], 3]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(16\ k\^2\ \[Lambda]\^2\ \
\[Tau]\)\/\(\((\(-4\) + k\ \[Lambda]\ \[Tau])\)\^2\ \((4 + 
                    k\ \[Lambda]\ \[Tau])\)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ -16, 
          Power[ k, 2], 
          Power[ \[Lambda], 2], \[Tau], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(16\ k\^2\ \[Lambda]\ \
\[Tau]\^2\)\/\(\((\(-4\) + k\ \[Lambda]\ \[Tau])\)\^2\ \((4 + 
                    k\ \[Lambda]\ \[Tau])\)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ -16, 
          Power[ k, 2], \[Lambda], 
          Power[ \[Tau], 2], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ \((16\ k\ \[Lambda] + 
                    k\^3\ \[Lambda]\^3\ \[Tau]\^2)\)\)\/\(\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\ \((4 + k\ \[Lambda]\ \[Tau])\)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ -2, 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1], 
          Plus[ 
            Times[ 16, k, \[Lambda]], 
            Times[ 
              Power[ k, 3], 
              Power[ \[Lambda], 3], 
              Power[ \[Tau], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(16\ k\^2\ \[Tau]\^3\)\/\(\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\ \((4 + k\ \[Lambda]\ \[Tau])\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 16, 
          Power[ k, 2], 
          Power[ \[Tau], 3], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ \((16\ k\ \[Tau] + 
                  k\^3\ \[Lambda]\^2\ \[Tau]\^3)\)\)\/\(\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\ \((4 + k\ \[Lambda]\ \[Tau])\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ 2, 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Power[ 
            Plus[ 4, 
              Times[ k, \[Lambda], \[Tau]]], -1], 
          Plus[ 
            Times[ 16, k, \[Tau]], 
            Times[ 
              Power[ k, 3], 
              Power[ \[Lambda], 2], 
              Power[ \[Tau], 3]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&\
^\)\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`OverHat[ 1]],
                          Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`OverHat[ 1]], 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]], 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(16\ k + 8\ k\^2\ \[Lambda]\ \[Tau] + 
              k\^3\ \[Lambda]\^2\ \[Tau]\^2\)\/\(4\ \((\(-4\) + k\ \[Lambda]\ \
\[Tau])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 4], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Plus[ 
            Times[ 16, k], 
            Times[ 8, 
              Power[ k, 2], \[Lambda], \[Tau]], 
            Times[ 
              Power[ k, 3], 
              Power[ \[Lambda], 2], 
              Power[ \[Tau], 2]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(4\ k\^2\ \[Lambda]\^2\)\/\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 1]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -4, 
          Power[ k, 2], 
          Power[ \[Lambda], 2], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(16\ k + 
              k\^3\ \[Lambda]\^2\ \[Tau]\^2\)\/\(2\ \((\(-4\) + k\ \[Lambda]\ \
\[Tau])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2], 
          Plus[ 
            Times[ 16, k], 
            Times[ 
              Power[ k, 3], 
              Power[ \[Lambda], 2], 
              Power[ \[Tau], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(4\ k\^2\ \[Tau]\^2\)\/\((\(-4\) + k\ \
\[Lambda]\ \[Tau])\)\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 2]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -4, 
          Power[ k, 2], 
          Power[ \[Tau], 2], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-2\)\ k\)}],
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 
        Times[ -2, k]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Laplacian[f[\[Tau], \[Lambda]], Simplify]\)], "Input"],

Cell[BoxData[
    RowBox[{\(-\(1\/4\)\), " ", \(\((\(-4\) + k\ \[Lambda]\ \[Tau])\)\^2\), 
      " ", 
      RowBox[{
        SuperscriptBox["f", 
          TagBox[\((1, 1)\),
            Derivative],
          MultilineFunction->None], "[", \(\[Tau], \[Lambda]\), 
        "]"}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Killing Vectors", "Subsubsection"],

Cell[BoxData[
    \(\(\(Tdefine[K, "\<a\>"]\)\(\ \ \)\( (*\ 
      Killing\ vector\ *) \)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(killingEq\  = \ 
      CD[la, Metricg[lb, lc]\ K[uc]]\  + \ 
        CD[lb, Metricg[la, lc]\ K[uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(CDtoAffine[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(killingEq = SumDum[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              K[ 2]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "1"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], 1],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], 1], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], 1], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "2"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], 2],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], 2], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], 2], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              K[ 2]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "1"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], 1],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], 1], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], 1], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "2"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], 2],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], 2], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], 2], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              K[ 1]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "1"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], 1],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], 1], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], 1], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "2"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], 2],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], 2], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], 2], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              K[ 1]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "1"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], 1],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], 1], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], 1], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "2"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], 2],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], 2], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], 2], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k11 = 
      killingEq\ \((\(-4\)\  + \ k\ \[Lambda]\ \[Tau])\)\^2 /. \ {la -> \
\(-1\), lb -> \(-1\)} // Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"8", " ", "k", " ", \(\[Lambda]\^2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_1\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 1], 
              K[ 1]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\((16 + k\^2\ \[Lambda]\^2\ \[Tau]\^2)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_1\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 1], 
              K[ 2]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k12 = 
      killingEq\ 2 \((\(-4\)\  + \ k\ \[Lambda]\ \[Tau])\)\^2 /. \ {la -> \
\(-1\), lb -> \(-2\)}\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"8", " ", "k", " ", \(\[Lambda]\^2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_2\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 2], 
              K[ 1]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\((16 + k\^2\ \[Lambda]\^2\ \[Tau]\^2)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_2\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 2], 
              K[ 2]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\((16 + k\^2\ \[Lambda]\^2\ \[Tau]\^2)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_1\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 1], 
              K[ 1]],
            Editable->False], ")"}]}], "+", 
      RowBox[{"8", " ", "k", " ", \(\[Tau]\^2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_1\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 1], 
              K[ 2]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k22\  = \ 
      killingEq\ \((\(-4\)\  + \ k\ \[Lambda]\ \[Tau])\)\^2 /. \ {la -> \
\(-2\), lb -> \(-2\)}\  // Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\((16 + k\^2\ \[Lambda]\^2\ \[Tau]\^2)\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_2\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 2], 
              K[ 1]],
            Editable->False], ")"}]}], "+", 
      RowBox[{"8", " ", "k", " ", \(\[Tau]\^2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_2\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", 2], 
              K[ 2]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(hInv\  = \ Inverse[hMatrix]\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \({{1\/\(1 - 1\/16\ k\^2\ \[Lambda]\^2\ \[Tau]\^2\), \(4\ k\ \[Lambda]\^2\
\)\/\(16 - k\^2\ \[Lambda]\^2\ \[Tau]\^2\)}, {\(4\ k\ \[Tau]\^2\)\/\(16 - \
k\^2\ \[Lambda]\^2\ \[Tau]\^2\), 
        1\/\(1 - 1\/16\ k\^2\ \[Lambda]\^2\ \[Tau]\^2\)}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \({PD1, PD2}\  = \ 
      hInv\  . \ {K\_\[Tau]\%A, K\_\[Lambda]\%A}\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \({\(-\(\(4\ \((k\ \[Lambda]\^2\ K\_\[Lambda]\%A + 
                  4\ K\_\[Tau]\%A)\)\)\/\(\(-16\) + 
              k\^2\ \[Lambda]\^2\ \[Tau]\^2\)\)\), \(-\(\(4\ \((4\ K\_\
\[Lambda]\%A + k\ \[Tau]\^2\ K\_\[Tau]\%A)\)\)\/\(\(-16\) + 
              k\^2\ \[Lambda]\^2\ \[Tau]\^2\)\)\)}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Tau]\ PD1\  - \ \[Lambda]\ PD2\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \(\(4\ \((\(-\[Lambda]\)\ K\_\[Lambda]\%A + \[Tau]\ \
K\_\[Tau]\%A)\)\)\/\(4 + k\ \[Lambda]\ \[Tau]\)\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Cset[K[1], \ 
      a\  + \ c\ \[Tau]\/\(1\  + \ \(k\/4\) \[Lambda]\ \[Tau]\)]\)], "Input"],

Cell[BoxData[
    \(Cset[K[2], \ 
      b\  - \ c\ \[Lambda]\/\(1\  + \ \(k\/4\) \[Lambda]\ \[Tau]\)]\)], \
"Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(k11\ \  // \ Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k12\ \  // \ Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k22\ \  // \ Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell["Therefore, the three Killing vectors in Killing basis are", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`K\_1\  = \ K\_\[Tau]\%A, \ \ \ \ \ K\_2\  = \ 
      K\_\[Lambda]\%A, \ \ \ \ \ K\_3\  = \ \(1\/\(1\  + \ \(k\/4\) \[Tau]\ \
\[Lambda]\)\) \((\[Tau]\ K\_\[Tau]\%A\  - \ \[Lambda]\ \
K\_\[Lambda]\%A)\)\)]]], "Text",
  TextAlignment->Center]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{560, 667},
WindowMargins->{{20, Automatic}, {Automatic, 7}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 73, 1, 43, "Title"],
Cell[1793, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1870, 57, 53, 0, 54, "Subsubsection"],
Cell[1926, 59, 69, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2020, 64, 212, 3, 165, "Input"],
Cell[2235, 69, 333, 9, 125, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2617, 84, 38, 0, 32, "Subsubsection"],
Cell[2658, 86, 367, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[3050, 96, 135, 2, 95, "Input"],
Cell[3188, 100, 242, 6, 117, "Output"]
}, Open  ]],
Cell[3445, 109, 37, 0, 41, "Text"],
Cell[3485, 111, 235, 5, 36, "Text"],

Cell[CellGroupData[{
Cell[3745, 120, 153, 3, 102, "Input"],
Cell[3901, 125, 548, 12, 121, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[4498, 143, 60, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[4583, 147, 83, 1, 79, "Input"],
Cell[4669, 150, 69, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[4787, 157, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[4849, 161, 46, 1, 56, "Input"],
Cell[4898, 164, 184, 4, 30, "Print"],
Cell[5085, 170, 1852, 47, 50, "Print"],
Cell[6940, 219, 2030, 53, 50, "Print"],
Cell[8973, 274, 1846, 47, 50, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10856, 326, 49, 1, 56, "Input"],
Cell[10908, 329, 350, 8, 31, "Print"],
Cell[11261, 339, 2786, 69, 44, "Print"],
Cell[14050, 410, 2784, 69, 44, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16871, 484, 46, 1, 56, "Input"],
Cell[16920, 487, 3176, 79, 50, "Print"],
Cell[20099, 568, 3003, 74, 50, "Print"],
Cell[23105, 644, 3178, 79, 50, "Print"],
Cell[26286, 725, 3046, 75, 50, "Print"],
Cell[29335, 802, 3046, 75, 50, "Print"],
Cell[32384, 879, 3184, 79, 50, "Print"],
Cell[35571, 960, 2997, 74, 50, "Print"],
Cell[38571, 1036, 3170, 79, 50, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[41778, 1120, 47, 1, 56, "Input"],
Cell[41828, 1123, 293, 6, 29, "Print"],
Cell[42124, 1131, 4300, 106, 50, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[46461, 1242, 45, 1, 56, "Input"],
Cell[46509, 1245, 184, 4, 29, "Print"],
Cell[46696, 1251, 1883, 48, 50, "Print"],
Cell[48582, 1301, 2049, 54, 50, "Print"],
Cell[50634, 1357, 1877, 48, 50, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[52548, 1410, 46, 1, 56, "Input"],
Cell[52597, 1413, 392, 10, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[53026, 1428, 74, 1, 56, "Input"],
Cell[53103, 1431, 293, 8, 69, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[53445, 1445, 40, 0, 32, "Subsubsection"],
Cell[53488, 1447, 104, 2, 56, "Input"],

Cell[CellGroupData[{
Cell[53617, 1453, 132, 3, 79, "Input"],
Cell[53752, 1458, 3581, 89, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[57370, 1552, 46, 1, 56, "Input"],
Cell[57419, 1555, 9685, 239, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[67141, 1799, 54, 1, 56, "Input"],
Cell[67198, 1802, 31205, 777, 101, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[98440, 2584, 147, 3, 103, "Input"],
Cell[98590, 2589, 1699, 41, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[100326, 2635, 137, 3, 81, "Input"],
Cell[100466, 2640, 3362, 81, 78, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[103865, 2726, 153, 3, 103, "Input"],
Cell[104021, 2731, 1701, 41, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[105759, 2777, 76, 1, 56, "Input"],
Cell[105838, 2780, 274, 4, 142, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[106149, 2789, 115, 2, 58, "Input"],
Cell[106267, 2793, 320, 5, 74, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[106624, 2803, 81, 1, 56, "Input"],
Cell[106708, 2806, 133, 2, 74, "Output"]
}, Open  ]],
Cell[106856, 2811, 111, 2, 80, "Input"],
Cell[106970, 2815, 116, 3, 83, "Input"],

Cell[CellGroupData[{
Cell[107111, 2822, 54, 1, 56, "Input"],
Cell[107168, 2825, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[107240, 2831, 54, 1, 56, "Input"],
Cell[107297, 2834, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[107369, 2840, 54, 1, 56, "Input"],
Cell[107426, 2843, 35, 1, 53, "Output"]
}, Open  ]],
Cell[107476, 2847, 73, 0, 41, "Text"],
Cell[107552, 2849, 290, 5, 51, "Text"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

