(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     72155,       1931]*)
(*NotebookOutlinePosition[     72923,       1958]*)
(*  CellTagsIndexPosition[     72879,       1954]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Killing Vector in 2-D Symmetric SpaceTime", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {\[Tau], \[Lambda]};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(metric\  = \ {{0, \(-\(1\/\(2 \((1\  - \ \(k\/4\) \[Tau]\ \
\[Lambda])\)\^2\)\)\)}, {\(-\(1\/\(2 \((1\  - \ \(k\/4\) \[Tau]\ \[Lambda])\)\
\^2\)\)\), 0}}; \ MatrixForm[metric]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              "0", \(-\(1\/\(2\ \((1 - \(k\ \[Lambda]\ \
\[Tau]\)\/4)\)\^2\)\)\)},
            {\(-\(1\/\(2\ \((1 - \(k\ \[Lambda]\ \[Tau]\)\/4)\)\^2\)\)\), 
              "0"}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensors", "Subsubsection"],

Cell[BoxData[
    \(InitCTensor[coSys, \ metric]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Tau]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Lambda]"\+_\)}],
              
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(1\/\(2\ \((1 - \(k\ \[Lambda]\ \[Tau]\)\/4)\
\)\^2\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Rational[ -1, 2], 
          Power[ 
            Plus[ 1, 
              Times[ 
                Rational[ -1, 4], k, \[Lambda], \[Tau]]], -2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Tau]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ k\ \[Lambda]\)\/\(\(-4\) + 
                k\ \[Lambda]\ \[Tau]\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]], " = ", 
        Times[ -2, k, \[Lambda], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Lambda]"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Lambda]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Lambda]"\+_\)}],
              
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(2\ k\ \[Tau]\)\/\(\(-4\) + 
                k\ \[Lambda]\ \[Tau]\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]]], " = ", 
        Times[ -2, k, \[Tau], 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -1]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Tau]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Lambda]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Lambda]"\+_\)}],
              
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(64\ k\)\/\((\(-4\) + k\ \[Lambda]\ \[Tau])\)\
\^4\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 64, k, 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -4]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Tau]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Lambda]"\+_\)}],
              
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(8\ k\)\/\((\(-4\) + k\ \[Lambda]\ \
\[Tau])\)\^2\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Lambda]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 8, k, 
          Power[ 
            Plus[ -4, 
              Times[ k, \[Lambda], \[Tau]]], -2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-2\)\ k\)}],
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 
        Times[ -2, k]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Killing Vectors", "Subsubsection"],

Cell[BoxData[
    \(\(\(Tdefine[K, "\<a\>"]\)\(\ \)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(killingEq\  = \ 
      CD[la, Metricg[lb, lc]\ K[uc]]\  + \ 
        CD[lb, Metricg[la, lc]\ K[uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(CDtoAffine[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(killingEq = SumDum[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              K[ 2]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              K[ 2]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              K[ 1]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              K[ 1]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k11\  = \ killingEq\  /. \ {la -> \(-1\), lb -> \(-1\)}\)], "Input"],

Cell[BoxData[
    RowBox[{"-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["K", 
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], 
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False]],
              Subsuperscript[ K, 
                SequenceForm[ "", " "], 
                SequenceForm[ "", 2]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", 1], 
            K[ 2]],
          
          Editable->
            False], \(\((1 - \(k\ \[Lambda]\ \[Tau]\)\/4)\)\^2\)]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "=> ",
  Cell[BoxData[
      \(TraditionalForm\`K\^\[Lambda]\  = \ A[\[Lambda]]\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cset[K[2], \ AA[\[Lambda]]]; \ k11\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k22\  = \ killingEq\  /. \ {la -> \(-2\), lb -> \(-2\)}\)], "Input"],

Cell[BoxData[
    RowBox[{"-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["K", 
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], 
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False]],
              Subsuperscript[ K, 
                SequenceForm[ "", " "], 
                SequenceForm[ "", 1]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", 2], 
            K[ 1]],
          
          Editable->
            False], \(\((1 - \(k\ \[Lambda]\ \[Tau]\)\/4)\)\^2\)]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "=> ",
  Cell[BoxData[
      \(TraditionalForm\`K\^\[Tau]\  = \ B[\[Tau]]\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cset[K[1], \ BB[\[Tau]]]; \ k22\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k12\  = \ 
      killingEq\ \ \((\(-4\)\  + \ k\ \[Lambda]\ \[Tau])\)\^3\/16 /. \ {la -> \
\(-1\), lb -> \(-2\)} // Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{\(k\ \[Tau]\ AA[\[Lambda]]\), "+", \(k\ \[Lambda]\ BB[\[Tau]]\), 
      "-", 
      RowBox[{\(1\/2\), " ", \((\(-4\) + k\ \[Lambda]\ \[Tau])\), " ", 
        RowBox[{
          SuperscriptBox["AA", "\[Prime]",
            MultilineFunction->None], "[", "\[Lambda]", "]"}]}], "-", 
      RowBox[{\(1\/2\), " ", \((\(-4\) + k\ \[Lambda]\ \[Tau])\), " ", 
        RowBox[{
          SuperscriptBox["BB", "\[Prime]",
            MultilineFunction->None], "[", "\[Tau]", "]"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(BB[\[Tau]_]\  := \ \(-\(tt\/4\)\) k\ \[Tau]\^2\  - \ pp\ \[Tau]\  + \ 
        rr; \ AA[\[Lambda]_]\  := \ \(-\(rr\/4\)\) k\ \[Lambda]\^2\  + \ 
        pp\ \[Lambda]\  + \ tt; \ k12 // Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(K[1]\ \*"\"\<\!\(\[PartialD]\_\[Tau]\%A\)\>\""\  + \ 
        K[2]\ \*"\"\<\!\(\[PartialD]\_\[Lambda]\%A\)\>\""\ \  // \ 
      Collect[#, {pp, rr, tt}] &\)], "Input"],

Cell[BoxData[
    \(rr\ \(("\!\(\[PartialD]\_\[Tau]\%A\)" - 
            1\/4\ "\!\(\[PartialD]\_\[Lambda]\%A\)"\ k\ \[Lambda]\^2)\) + 
      pp\ \(("\!\(\[PartialD]\_\[Lambda]\%A\)"\ \[Lambda] - 
            "\!\(\[PartialD]\_\[Tau]\%A\)"\ \[Tau])\) + 
      tt\ \(("\!\(\[PartialD]\_\[Lambda]\%A\)" - 
            1\/4\ "\!\(\[PartialD]\_\[Tau]\%A\)"\ k\ \[Tau]\^2)\)\)], "Output"]
}, Open  ]],

Cell["Therefore, we have three Killing vectors", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`K\_1\  = \ \[PartialD]\_\[Tau]\ \(-\ \(k\/4\)\) \(\
\[Lambda]\^2\) \[PartialD]\_\[Lambda], \ \ \ \ \ K\_2\  = \ \(-\(k\/4\)\) \(\
\[Tau]\^2\) \[PartialD]\_\[Tau]\ \(+\ \[PartialD]\_\[Lambda]\), \ \ \ \ \ \
K\_3\  = \ \[Tau]\ \[PartialD]\_\[Tau]\ \(-\ \[Lambda]\)\ \[PartialD]\_\
\[Lambda]\)]]], "Text",
  TextAlignment->Center]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutation Relation of Killing Vectors", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(killingM\  = \ {{1, \ \(-\(k\/4\)\) \[Lambda]\^2}, {\(-\(k\/4\)\) \
\[Tau]\^2, 1}, {\[Tau], \(-\[Lambda]\)}}; \ MatrixForm[killingM]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"1", \(-\(\(k\ \[Lambda]\^2\)\/4\)\)},
            {\(-\(\(k\ \[Tau]\^2\)\/4\)\), "1"},
            {"\[Tau]", \(-\[Lambda]\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[killingM]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"[V\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_1\),
          Subscript[ 1],
          Editable->False], "\[InvisibleSpace]", "\<\", V\"\>", 
        "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_2\),
          Subscript[ 2],
          Editable->False], "\[InvisibleSpace]", "\<\"] = \"\>", 
        "\[InvisibleSpace]", \({\(-\(\(k\ \[Tau]\)\/2\)\), \(k\ \
\[Lambda]\)\/2}\)}],
      SequenceForm[ "[V", 
        Subscript[ 1], ", V", 
        Subscript[ 2], "] = ", {
        Times[ 
          Rational[ -1, 2], k, \[Tau]], 
        Times[ 
          Rational[ 1, 2], k, \[Lambda]]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"[V\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_1\),
          Subscript[ 1],
          Editable->False], "\[InvisibleSpace]", "\<\", V\"\>", 
        "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_3\),
          Subscript[ 3],
          Editable->False], "\[InvisibleSpace]", "\<\"] = \"\>", 
        "\[InvisibleSpace]", \({1, \(-\(\(k\ \[Lambda]\^2\)\/4\)\)}\)}],
      SequenceForm[ "[V", 
        Subscript[ 1], ", V", 
        Subscript[ 3], "] = ", {1, 
        Times[ 
          Rational[ -1, 4], k, 
          Power[ \[Lambda], 2]]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"[V\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_2\),
          Subscript[ 2],
          Editable->False], "\[InvisibleSpace]", "\<\", V\"\>", 
        "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_3\),
          Subscript[ 3],
          Editable->False], "\[InvisibleSpace]", "\<\"] = \"\>", 
        "\[InvisibleSpace]", \({\(k\ \[Tau]\^2\)\/4, \(-1\)}\)}],
      SequenceForm[ "[V", 
        Subscript[ 2], ", V", 
        Subscript[ 3], "] = ", {
        Times[ 
          Rational[ 1, 4], k, 
          Power[ \[Tau], 2]], -1}],
      Editable->False]], "Print"]
}, Open  ]],

Cell["Therefore", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\([K\_1, \ K\_2]\)\  = \ \(-\(k\/2\)\) 
        K\_3, \ \ \ \ \ \([K\_1, \ K\_3]\)\  = \ 
      K\_1, \ \ \ \ \ \([K\_2, \ K\_3]\)\  = \ \(-K\_2\)\)]]], "Text",
  TextAlignment->Center],

Cell["Explicitely,", "Text"],

Cell[BoxData[
    \(\({k1, k2, 
          k3}\  = \ {{1, \ \(-\(k\/4\)\) \[Lambda]\^2}, {\(-\(k\/4\)\) \[Tau]\
\^2, 1}, {\[Tau], \(-\[Lambda]\)}};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[k1, k2]\)], "Input"],

Cell[BoxData[
    \({\(-\(\(k\ \[Tau]\)\/2\)\), \(k\ \[Lambda]\)\/2}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[k1, k3]\)], "Input"],

Cell[BoxData[
    \({1, \(-\(\(k\ \[Lambda]\^2\)\/4\)\)}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[k2, k3]\)], "Input"],

Cell[BoxData[
    \({\(k\ \[Tau]\^2\)\/4, \(-1\)}\)], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{600, 678},
WindowMargins->{{26, Automatic}, {Automatic, 8}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 74, 1, 43, "Title"],
Cell[1794, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1871, 57, 53, 0, 54, "Subsubsection"],
Cell[1927, 59, 69, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2021, 64, 208, 3, 165, "Input"],
Cell[2232, 69, 333, 9, 125, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2614, 84, 60, 0, 32, "Subsubsection"],
Cell[2677, 86, 61, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2775, 92, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[2837, 96, 46, 1, 56, "Input"],
Cell[2886, 99, 184, 4, 30, "Print"],
Cell[3073, 105, 1981, 51, 56, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5091, 161, 46, 1, 56, "Input"],
Cell[5140, 164, 392, 8, 31, "Print"],
Cell[5535, 174, 2974, 70, 44, "Print"],
Cell[8512, 246, 3028, 71, 44, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[11577, 322, 47, 1, 56, "Input"],
Cell[11627, 325, 293, 6, 29, "Print"],
Cell[11923, 333, 4216, 99, 47, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16176, 437, 45, 1, 56, "Input"],
Cell[16224, 440, 184, 4, 29, "Print"],
Cell[16411, 446, 1917, 49, 47, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18365, 500, 46, 1, 56, "Input"],
Cell[18414, 503, 392, 10, 29, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[18855, 519, 40, 0, 32, "Subsubsection"],
Cell[18898, 521, 66, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[18989, 526, 132, 3, 79, "Input"],
Cell[19124, 531, 3581, 89, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22742, 625, 46, 1, 56, "Input"],
Cell[22791, 628, 9725, 239, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[32553, 872, 54, 1, 56, "Input"],
Cell[32610, 875, 31317, 777, 101, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[63964, 1657, 88, 1, 56, "Input"],
Cell[64055, 1660, 854, 22, 86, "Output"]
}, Open  ]],
Cell[64924, 1685, 113, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[65062, 1693, 67, 1, 56, "Input"],
Cell[65132, 1696, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[65204, 1702, 88, 1, 56, "Input"],
Cell[65295, 1705, 854, 22, 86, "Output"]
}, Open  ]],
Cell[66164, 1730, 107, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[66296, 1738, 64, 1, 56, "Input"],
Cell[66363, 1741, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[66435, 1747, 157, 3, 124, "Input"],
Cell[66595, 1752, 513, 10, 93, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[67145, 1767, 227, 3, 132, "Input"],
Cell[67375, 1772, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[67447, 1778, 187, 3, 81, "Input"],
Cell[67637, 1783, 383, 6, 110, "Output"]
}, Open  ]],
Cell[68035, 1792, 56, 0, 41, "Text"],
Cell[68094, 1794, 378, 6, 42, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[68509, 1805, 64, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[68598, 1809, 165, 2, 95, "Input"],
Cell[68766, 1813, 282, 7, 137, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[69085, 1825, 53, 1, 56, "Input"],
Cell[69141, 1828, 710, 19, 44, "Print"],
Cell[69854, 1849, 665, 17, 47, "Print"],
Cell[70522, 1868, 655, 17, 47, "Print"]
}, Open  ]],
Cell[71192, 1888, 25, 0, 41, "Text"],
Cell[71220, 1890, 236, 4, 42, "Text"],
Cell[71459, 1896, 28, 0, 41, "Text"],
Cell[71490, 1898, 161, 3, 72, "Input"],

Cell[CellGroupData[{
Cell[71676, 1905, 51, 1, 56, "Input"],
Cell[71730, 1908, 82, 1, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[71849, 1914, 51, 1, 56, "Input"],
Cell[71903, 1917, 70, 1, 72, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[72010, 1923, 51, 1, 56, "Input"],
Cell[72064, 1926, 63, 1, 72, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

