(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     39578,       1170]*)
(*NotebookOutlinePosition[     40347,       1197]*)
(*  CellTagsIndexPosition[     40303,       1193]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Newman-Penrose Formalism", "Title",
  TextAlignment->Center,
  TextJustification->0,
  FontSize->36],

Cell["See Chandrasekhar pp 40", "Text"],

Cell[CellGroupData[{

Cell["Orthonormal Basis", "Subsection"],

Cell[CellGroupData[{

Cell["Basis (Tetrad)", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`e\_a\%A\  = \ \(h\_a\^\[Mu]\) \[PartialD]\_\[Mu]\%A, \ \
\ \ \ \ \[PartialD]\_\[Mu]\%A\ \(\(=\)\(\ \)\(\(h\_\[Mu]\^a\) 
          e\_a\%A\)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Basis one-form", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`e\_A\%a\  = \ \(h\_\[Mu]\^a\) 
        d\[InvisibleSpace]x\_A\%\[Mu], \ \ \ \ \ \
d\[InvisibleSpace]x\_A\%\[Mu]\  = \ \(h\_a\^\[Mu]\) e\_a\%A\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Orthonomality", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\(h\_a\^\[Mu]\) 
        h\_\[Mu]\^b\  = \ \[Delta]\_a\^b, \ \ \ \ \ \(h\_\[Mu]\^a\) 
        h\_a\^\[Nu]\  = \ \[Delta]\_\[Mu]\^\[Nu], \ \ \ \ \ g\_\[Mu]\[Nu]\  = \
\ \(h\_\[Mu]\^a\) \(h\_\[Nu]\^b\) \[Eta]\_\(a\[InvisibleSpace]b\)\)]]], "Text",\

  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta]\_\(a\[InvisibleSpace]b\)\)]],
  " is a constant matrix."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Connection one-form", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Omega]\_\(A\[InvisibleSpace]a\)\^b\  = \ \
\(-e\_a\%B\) \[Del]\_A e\_B\%b\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Ricci-rotation coefficients", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Omega]\_\(a\[InvisibleSpace]b\)\^c\  = \ \
\(\(-e\_a\%A\) \[Omega]\_\(A\[InvisibleSpace]b\)\^c\  = \ \(\[CapitalGamma]\_\
\(a\[InvisibleSpace]b\)\^c\  = \ \((\(e\_a\%A\) \[Del]\_A e\_b\%B)\) 
            e\_B\%c\)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[CapitalGamma]\_\(a\[InvisibleSpace]b\[InvisibleSpace]\
c\)\  = \ \(\(1\/2\) \((f\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\)\  + \ 
              f\_\(c\[InvisibleSpace]a\[InvisibleSpace]b\)\  + \ 
              f\_\(c\[InvisibleSpace]b\[InvisibleSpace]a\))\)\  = \ \
\[CapitalGamma]\_\(a[b\[InvisibleSpace]c]\)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\(\([\)\(e\_a, \ 
      e\_b\)\(]\)\)\  = \ \(f\_\(a\[InvisibleSpace]b\)\^c\) 
        e\_c, \ \ \ \ \ f\_\(a\[InvisibleSpace]b\)\^c\  = \ \[CapitalGamma]\_\
\(a\[InvisibleSpace]b\)\^c\  - \ \
\[CapitalGamma]\_\(b\[InvisibleSpace]a\)\^c\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\  = \ \(\[PartialD]\_a \[CapitalGamma]\_\(b\
\[InvisibleSpace]c\[InvisibleSpace]d\)\  + \ \(\[CapitalGamma]\_\(a\
\[InvisibleSpace]e\[InvisibleSpace]d\)\) \
\[CapitalGamma]\_\(b\[InvisibleSpace]c\)\^e\  - \ \(\[CapitalGamma]\_\(a\
\[InvisibleSpace]b\)\^e\) \[CapitalGamma]\_\(e\[InvisibleSpace]c\
\[InvisibleSpace]d\)\  - \ \((a\  \[LeftRightArrow] \ 
              b)\)\  = \ \(R\_\(\(\([\)\(a\[InvisibleSpace]b\)\(]\)\)[c\
\[InvisibleSpace]d]\)\  = \ 
          R\_\(\(\([\)\(c\[InvisibleSpace]d\)\(]\)\)[a\[InvisibleSpace]b]\)\)\
\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_\(\(\([\)\(a\[InvisibleSpace]b\[InvisibleSpace]c\)\(\
]\)\) d\)\  = \ 0\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "Therefore, ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\)]],
  " has 20 independent components"
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Newman-Penrose Basis", "Subsection"],

Cell[CellGroupData[{

Cell["Orthogonality", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`l\ \[CenterDot]\ 
        m\  = \ \(l\ \[CenterDot]\ 
          m\&_\  = \ \(n\ \[CenterDot]\ m\  = \ \(n\ \[CenterDot]\ m\&_\  = \ 
            0\)\)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Null condition", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`l\ \[CenterDot]\ 
        l\  = \ \(n\ \[CenterDot]\ 
          n\  = \ \(m\ \[CenterDot]\ m\  = \ \(m\&_\ \[CenterDot]\ m\&_\  = \ 
            0\)\)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Normalization", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`l\ \[CenterDot]\ 
        n\  = \ \(-1\), \ \ \ \ \ m\ \[CenterDot]\ m\&_\  = \ 1\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell["Explicitely,", "Text"],

Cell[TextData[Cell[BoxData[
    FormBox[
      RowBox[{\(e\_1\  = \ l\), ",", "     ", \(e\_2\  = \ n\), ",", 
        "     ", \(e\_3\  = \ m\), ",", "     ", \(e\_4\  = \ m\&_\), ",", 
        "     ", 
        RowBox[{\(\[Eta]\_\(a\[InvisibleSpace]b\)\), " ", "=", " ", 
          RowBox[{"(", GridBox[{
                {"0", \(-1\), "0", "0"},
                {\(-1\), "0", "0", "0"},
                {"0", "0", "0", "1"},
                {"0", "0", "1", "0"}
                }], ")"}]}]}], TraditionalForm]]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`m\^A\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`m\&_\^A\)]],
  " are mutually complex conjugated"
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Newman-Penrose Basis from Metric", "Subsection"],

Cell[TextData[{
  "Let ",
  Cell[BoxData[
      \(TraditionalForm\`l'\^\[Mu]\  = \ 
        g\^\(1  \[Mu]\)\  + \ a\ g\^\(2  \[Mu]\)\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`n'\^\[Mu]\  = \ 
        g\^\(1  \[Mu]\)\  + \ b\ g\^\(2  \[Mu]\)\)]],
  " for non-vanishing ",
  Cell[BoxData[
      \(TraditionalForm\`g\^11\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`g\^22\)]],
  ". (If ",
  Cell[BoxData[
      \(TraditionalForm\`g\^11\  = \ 0\)]],
  ", then ",
  Cell[BoxData[
      \(TraditionalForm\`l'\^\[Mu]\  = \ g\^\(1  \[Mu]\)\)]],
  "). From the null condition, we have ",
  Cell[BoxData[
      \(TraditionalForm\`g\^11\  + \ 2  a\ g\^12\  + \ a\^2\ g\^22\  = \ 
        0\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`g\^11\  + \ 2  b\ g\^12\  + \ \(b\^2\) g\^22\  = \ 
        0\)]],
  ". From the normalization condition, we have ",
  Cell[BoxData[
      \(TraditionalForm\`g\^11\  + \ \((a\  + \ b)\) g\^12\  + \ 
          a\ b\ g\^22\  = \ \(-s\^2\)\)]],
  ". Then the Newman-Penrose null vectors are obtained by setting ",
  Cell[BoxData[
      \(TraditionalForm\`l\^\[Mu]\  = \ l'\^\[Mu]/s\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`n\^\[Mu]\  = \ n'\^\[Mu]/s\)]],
  "."
}], "Text",
  TextAlignment->Left,
  TextJustification->1]
}, Closed]],

Cell[CellGroupData[{

Cell["Basis Transformation", "Subsection"],

Cell[CellGroupData[{

Cell["Class I", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`l\  \[Rule] \ l, \ \ \ \ \ n\  \[Rule] \ 
      n\  + \ a\&_\ m\  + \ a\ m\&_\  + \ a\ a\&_\ l, \ \ \ \ \ m\  \[Rule] \ 
      m\  + \ a\ l, \ \ \ \ \ m\&_\  \[Rule] \ m\&_\  + \ a\&_\ l\)]]], "Text",\

  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Class II", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`l\  \[Rule] \ 
      l\  + \ b\&_\ m\  + \ b\ m\&_\  + \ b\ b\&_\ n, \ \ \ \ \ n\  \[Rule] \ 
      n, \ \ \ \ \ m\  \[Rule] \ m\  + \ b\ n, \ \ \ \ \ m\&_\  \[Rule] \ 
      m\&_\  + \ b\&_\ n\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Class III", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`l\  \[Rule] \ \(A\^\(-1\)\) l, \ \ \ \ \ n\  \[Rule] \ 
      A\ n, \ \ \ \ \ m\  \[Rule] \ \(e\^\(i\[InvisibleSpace]\[Theta]\)\) 
        m, \ \ \ \ \ m\&_\  \[Rule] \ \
\(e\^\(\(-i\)\[InvisibleSpace]\[Theta]\)\) m\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Formulas", "Subsection"],

Cell[CellGroupData[{

Cell["Directional derivatives", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`D\  = \ \(e\_1\%A\) \[Del]\_A, \ \ \ \ \ \
\[CapitalDelta]\  = \ \(e\_2\%A\) \[Del]\_A, \ \ \ \ \ \[Delta]\  = \ \
\(e\_3\%A\) \[Del]\_A, \ \ \ \ \ \[Delta]\&_\  = \ \(e\_4\%A\) \
\[Del]\_A\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Spin coefficients", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Epsilon]\  = \ \(1\/2\) \((\[CapitalGamma]\_112\  - \
\ \[CapitalGamma]\_134)\), \ \ \ \ \ \[Kappa]\  = \ \[CapitalGamma]\_113, \ \ \
\ \ \ \[Pi]\  = \ \(-\[CapitalGamma]\_124\), \ \ \ \ \ \[Gamma]\  = \ \
\(1\/2\) \((\[CapitalGamma]\_212\  - \ \[CapitalGamma]\_234)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\(\(\[Tau]\  = \ \[CapitalGamma]\_213\)\(,\)\(\ \ \ \ \
\ \)\(\[Nu]\  = \ \(-\[CapitalGamma]\_224\)\)\(,\)\(\ \ \ \ \ \)\(\[Beta]\  = \
\ \(1\/2\) \((\[CapitalGamma]\_312\  - \ \[CapitalGamma]\_334)\)\)\(,\)\(\ \ \
\ \ \ \)\(\[Sigma]\  = \ \[CapitalGamma]\_313\)\(,\)\(\ \ \ \ \ \)\)\)]]], \
"Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Lambda]\  = \ \(-\[CapitalGamma]\_424\), \ \ \ \ \ \
\[Rho]\  = \ \[CapitalGamma]\_413, \ \ \ \ \ \[Mu]\  = \ \
\(-\[CapitalGamma]\_324\), \ \ \ \ \ \[Alpha]\  = \ \(1\/2\) \((\
\[CapitalGamma]\_412\  - \ \[CapitalGamma]\_434)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`3\  \[LeftRightArrow] \ 4\)]],
  " corresponds to complex conjugation"
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Weyl tensor", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\  = \ 
      R\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\[InvisibleSpace]d\)\  - \ \
\(1\/2\) \((\(\[Eta]\_\(a\[InvisibleSpace]c\)\) 
                R\_\(b\[InvisibleSpace]d\)\  - \ \(\[Eta]\_\(a\
\[InvisibleSpace]d\)\) 
                R\_\(b\[InvisibleSpace]c\)\  - \ \(\[Eta]\_\(b\
\[InvisibleSpace]c\)\) 
                R\_\(a\[InvisibleSpace]d\)\  + \ \(\[Eta]\_\(b\
\[InvisibleSpace]d\)\) 
                R\_\(a\[InvisibleSpace]c\))\)\  + \ \(1\/6\) \
\(R(\(\[Eta]\_\(a\[InvisibleSpace]c\)\) \[Eta]\_\(b\[InvisibleSpace]d\)\  - \ \
\(\[Eta]\_\(a\[InvisibleSpace]d\)\) \[Eta]\_\(b\[InvisibleSpace]c\))\)\)]]], \
"Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "From the trace-free condition ",
  Cell[BoxData[
      \(TraditionalForm\`C\_abc\^b\  = \ 0\)]],
  " and the Bianchi identity ",
  Cell[BoxData[
      \(TraditionalForm\`C\_\(\((a\[InvisibleSpace]b\[InvisibleSpace]c)\) d\)\
\  = \ 0\)]],
  " we find"
}], "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1314\  = \ \(C\_2324\  = \ \(C\_1323\  = \ \
\(C\_1424\  = \ 
            0\)\)\), \ \ \ \ \ C\_1213\  = \ \(-C\_1334\), \ \ \ \ \ C\_1214\ \
 = \ C\_1434\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1223\  = \ 
      C\_2334, \ \ \ \ \ C\_1224\  = \ \(-C\_2434\), \ \ \ \ \ C\_1324\  = \ \
\(1\/2\) \((C\_1234\  - \ C\_1212)\), \ \ \ \ \ C\_1212\  = \ 
      C\_3434\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell["Explicitely,", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1212\  = \ 
      R\_1212\  - \ R\_12\  - \ R\/6, \ \ \ \ \ C\_1324\  = \ 
      R\_1324\  + \ R\/12, \ \ \ \ \ C\_1234\  = \ R\_1234\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_3434\  = \ 
      R\_3434\  + \ R\_34\  - \ R\/6, \ \ \ \ \ C\_1313\  = \ 
      R\_1313, \ \ \ \ \ C\_2323\  = \ R\_2323\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1314\  = \ \(0\  = \ 
        R\_1314\  - \ \(1\/2\) R\_11\), \ \ \ \ \ C\_2324\  = \ \(0\  = \ 
        R\_2324\  - \ \(1\/2\) R\_22\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1323\  = \ \(0\  = \ 
        R\_1323\  + \ \(1\/2\) R\_33\), \ \ \ \ \ C\_1213\  = \ 
      R\_1213\  - \ \(1\/2\) R\_13, \ \ \ \ \ C\_1334\  = \ 
      R\_1334\  - \ \(1\/2\) R\_13\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1223\  = \ 
      R\_1223\  + \ \(1\/2\) R\_23, \ \ \ \ \ C\_2334\  = \ 
      R\_2334\  - \ \(1\/2\) R\_23\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell["\<\
The ten-independent components of Weyl tensor are represented by the five \
complex scalars\
\>", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Psi]\_0\  = \ C\_1313, \ \ \ \ \ \[Psi]\_1\  = \ 
      C\_1213, \ \ \ \ \ \[Psi]\_2\  = \ \(-C\_1324\), \ \ \ \ \ \[Psi]\_3\  \
= \ \(-C\_1224\), \ \ \ \ \ \[Psi]\_4\  = \ C\_2424\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell["\<\
Using these scalars we can express the complete components of Weyl tensor:\
\>", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1212\  = \ \[Psi]\_2\  + \ \[Psi]\&_\_2, \ \ \ \ \ \
C\_1213\  = \ \[Psi]\_1, \ \ \ \ \ C\_1214\  = \ \[Psi]\&_\_1, \ \ \ \ \ \
C\_1223\  = \ \(-\[Psi]\&_\_3\), \ \ \ \ \ C\_1224\  = \ \(-\[Psi]\_3\)\)]]], \
"Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1234\  = \ \(-\[Psi]\_2\)\  + \ \[Psi]\&_\_2, \ \ \ \
\ \ C\_1313\  = \ \[Psi]\_0, \ \ \ \ \ C\_1314\  = \ 
      0, \ \ \ \ \ C\_1323\  = \ 
      0, \ \ \ \ \ C\_1324\  = \ \(-\[Psi]\_2\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_1334\  = \ \(-\[Psi]\_1\), \ \ \ \ \ C\_1414\  = \ \
\[Psi]\&_\_0, \ \ \ \ \ C\_1423\  = \ \(-\[Psi]\&_\_2\), \ \ \ \ \ C\_1424\  \
= \ 0, \ \ \ \ \ C\_1434\  = \ \[Psi]\&_\_1, \ \ \ \ \ C\_2323\  = \ \
\[Psi]\&_\_4\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`C\_2324\  = \ 
      0, \ \ \ \ \ C\_2334\  = \ \(-\[Psi]\&_\_3\), \ \ \ \ \ C\_2424\  = \ \
\[Psi]\_4, \ \ \ \ \ C\_2434\  = \ \[Psi]\_3, \ \ \ \ \ C\_3434\  = \ \
\[Psi]\_2\  + \ \[Psi]\&_\_2\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Ricci Tensors", "Subsubsection"],

Cell[TextData[{
  "The ten-components of Ricci tensors are defined in terms of \
three-dimensional hermitian matrix ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalPhi]\_\(A\[InvisibleSpace]B\)\  = \ \
\[CapitalPhi]\&_\_\(B\[InvisibleSpace]A\)\)]]
}], "Text",
  TextAlignment->Left,
  TextJustification->1],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[CapitalPhi]\_00\  = \ \(1\/2\) 
        R\_11, \ \ \ \ \ \[CapitalPhi]\_01\  = \ \(1\/2\) 
        R\_13, \ \ \ \ \ \[CapitalPhi]\_10\  = \ \(1\/2\) 
        R\_14, \ \ \ \ \ \[CapitalPhi]\_02\  = \ \(1\/2\) 
        R\_33, \ \ \ \ \ \[CapitalPhi]\_20\  = \ \(1\/2\) R\_44\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[CapitalPhi]\_11\  = \ \(1\/4\) \((R\_12\  + \ 
            R\_34)\), \ \ \ \ \ \[CapitalPhi]\_12\  = \ \(1\/2\) 
        R\_23, \ \ \ \ \ \[CapitalPhi]\_21\  = \ \(1\/2\) 
        R\_24, \ \ \ \ \ \[CapitalPhi]\_22\  = \ \(1\/2\) R\_22\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Procedures", "Subsection"],

Cell[CellGroupData[{

Cell["Step 1", "Subsubsection"],

Cell[TextData[{
  "Setup ",
  Cell[BoxData[
      \(TraditionalForm\`{l, n, m, m\&_}\)]],
  " from ",
  Cell[BoxData[
      \(TraditionalForm\`g\_\[Mu]\[Nu]\)]],
  " or by input"
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Step 2", "Subsubsection"],

Cell[TextData[{
  "Calculate ",
  Cell[BoxData[
      \(TraditionalForm\`f\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\)\  = \ \
\(f\_\(a\[InvisibleSpace]b\)\^d\) \[Eta]\_\(c\[InvisibleSpace]d\)\)]],
  " by ",
  Cell[BoxData[
      \(TraditionalForm\`\(\([\)\(e\_a, \ 
        e\_b\)\(]\)\)\  = \ \(f\_\(a\[InvisibleSpace]b\)\^c\) e\_c\)]]
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Step 3", "Subsubsection"],

Cell[TextData[{
  "Calculate ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalGamma]\_\(a\[InvisibleSpace]b\
\[InvisibleSpace]c\)\)]],
  " and assign ",
  Cell[BoxData[
      \(TraditionalForm\`\[Alpha], \[Beta],  ... \)]]
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Step 4", "Subsubsection"],

Cell[TextData[{
  "Calculate ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\)]],
  ": (See below)"
}], "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1212\  = \ 
      D\[InvisibleSpace]\[Gamma]\  - \ \[CapitalDelta]\[InvisibleSpace]\
\[Epsilon]\  + \ \[Pi]\ \[Tau]\ \  - \ \[Nu]\ \[Kappa]\  - \ \((\[Epsilon]\  \
+ \ \[Epsilon]\&_)\) \((\[Gamma]\  + \ \[Gamma]\&_)\)\  + \ \((\[Alpha]\  + \ \
\[Beta]\&_)\) \((\[Tau]\  + \ \[Pi]\&_)\)\  + \ c . c . \)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1213\  = \ 
      D\[InvisibleSpace]\[Tau]\  - \ \[CapitalDelta]\[InvisibleSpace]\[Kappa]\
\  + \ \((\[Epsilon]\  - \ \[Epsilon]\&_)\) \[Tau]\ \  - \ \[Kappa](
          3  \[Gamma]\  + \ \[Gamma]\&_)\  + \ \((\[Pi]\  + \ \[Tau]\&_)\) \
\[Sigma]\  + \ \((\[Pi]\&_\  + \ \[Tau])\) \[Rho]\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1214\  = \ R\&_\_1213\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1223\  = \ \(-D\)\[InvisibleSpace]\[Nu]\&_\  + \ \
\[CapitalDelta]\[InvisibleSpace]\[Pi]\&_\  + \ \((\[Epsilon]\  + \ 
              3  \[Epsilon]\&_)\) \[Nu]\&_\  + \ \((\[Gamma]\  - \ \
\[Gamma]\&_)\) \[Pi]\&_\  - \ \((\[Pi]\  + \ \[Tau]\&_)\) \[Lambda]\&_\  - \ \
\((\[Pi]\&_\  + \ \[Tau])\) \[Mu]\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1224\  = \ R\&_\_1223\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1234\  = \ R\_1324\  - \ R\&_\_1324\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1313\  = \ 
      D\[InvisibleSpace]\[Sigma]\  - \ \[Delta]\[InvisibleSpace]\[Kappa]\  - \
\ \[Kappa](\[Alpha]\&_\  + \ 
            3  \[Beta]\  - \ \[Pi]\&_\  + \ \[Tau])\  + \ \[Sigma](
          3  \[Epsilon]\  - \ \[Epsilon]\&_\  + \ \[Rho]\  + \ \
\[Rho]\&_)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1314\  = \ 
      D\[InvisibleSpace]\[Rho]\&_\  - \ \[Delta]\[InvisibleSpace]\[Kappa]\&_\ \
 - \ \(\[Kappa]\&_\)(
          3  \[Alpha]\&_\  + \ \[Beta]\  - \ \[Pi]\&_)\  - \ \[Kappa]\ \[Tau]\
\&_\  + \ \(\[Rho]\&_\)(\[Epsilon]\  + \ \[Epsilon]\&_\  + \ \[Rho]\&_)\  + \ \
\[Sigma]\ \[Sigma]\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1323\  = \ \(-D\)\[InvisibleSpace]\[Lambda]\&_\  + \
\ \[Delta]\[InvisibleSpace]\[Pi]\&_\  - \ \(\[Lambda]\&_\)(\[Epsilon]\  - \ 
            3  \[Epsilon]\&_\  + \ \[Rho]\&_)\  - \ \(\[Pi]\&_\)(\[Alpha]\&_\ \
 - \ \[Beta]\  + \ \[Pi]\&_)\  + \ \[Kappa]\ \[Nu]\&_\  - \ \[Sigma]\ \
\[Mu]\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1324\  = \ \(-D\)\[InvisibleSpace]\[Mu]\  + \ \
\[Delta]\[InvisibleSpace]\[Pi]\  + \ \[Mu](\[Epsilon]\  + \ \[Epsilon]\&_\  - \
\ \[Rho]\&_)\  + \ \[Pi](\[Alpha]\&_\  - \ \[Beta]\  - \ \[Pi]\&_)\  + \ \
\[Kappa]\ \[Nu]\  - \ \[Sigma]\ \[Lambda]\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1334\  = \ 
      D(\[Alpha]\&_\  - \ \[Beta])\  + \ \[Delta](\[Epsilon]\  - \ \[Epsilon]\
\&_)\  + \ \[Kappa](\[Mu]\  + \ \[Gamma]\  - \ \[Gamma]\&_)\  - \ \[Kappa]\&_\
\ \[Lambda]\&_\  - \ \[Pi]\ \[Sigma]\  + \ \[Pi]\&_\ \[Rho]\&_\  - \ \
\(\[Pi]\&_\)(\[Epsilon]\  - \ \[Epsilon]\&_)\n\t\ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \
\  - \ \[Sigma](\[Alpha]\  - \ \[Beta]\&_)\  - \ \[Beta]\ \[Rho]\&_\  + \ \(\
\[Alpha]\&_\)(
          2  \[Epsilon]\  - \ 2  \[Epsilon]\&_\  + \ \[Rho]\&_)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1414\  = \ R\&_\_1313\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1423\  = \ R\&_\_1324\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1424\  = \ R\&_\_1323\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_1434\  = \ \(-R\&_\_1334\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_2323\  = \ \(-\[CapitalDelta]\)\[InvisibleSpace]\
\[Lambda]\&_\  + \ \[Delta]\[InvisibleSpace]\[Nu]\&_\  - \ \(\[Nu]\&_\)(
          3  \[Alpha]\&_\  + \ \[Beta]\  + \ \[Pi]\&_\  - \ \[Tau])\  + \ \(\
\[Lambda]\&_\)(\[Mu]\  + \ \[Mu]\&_\  - \ \[Gamma]\  + \ 
            3  \[Gamma]\&_)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_2324\  = \ \(-\[CapitalDelta]\)\[InvisibleSpace]\
\[Mu]\  + \ \[Delta]\[InvisibleSpace]\[Nu]\  - \ \[Nu](\[Alpha]\&_\  + \ 
            3  \[Beta]\  - \ \[Tau])\  - \ \[Nu]\&_\ \[Pi]\  + \ \[Mu](\[Mu]\ \
 + \ \[Gamma]\  + \ \[Gamma]\&_)\  + \ \[Lambda]\ \[Lambda]\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_2334\  = \ \[CapitalDelta](\[Alpha]\&_\  - \ \
\[Beta])\  + \ \[Delta](\[Gamma]\  - \ \[Gamma]\&_)\  - \ \[Tau]\&_\ \
\[Lambda]\&_\  - \ \[Nu]\ \[Sigma]\  + \ \[Rho]\&_\ \[Nu]\&_\  + \ \[Mu]\ \
\[Tau]\  - \ \(\[Nu]\&_\)(\[Epsilon]\  - \ \[Epsilon]\&_)\n\t\ \ \ \ \ \ \ \ \
\ \ \ \ \ \ \  + \ \[Tau](\[Gamma]\  - \ \[Gamma]\&_)\  + \ \(\[Lambda]\&_\)(\
\[Alpha]\  - \ \[Beta]\&_)\  - \ \[Alpha]\&_\ \[Mu]\  - \ \[Beta](
          2  \[Gamma]\  - \ 2  \[Gamma]\&_\  - \ \[Mu])\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_2424\  = \ R\&_\_2323\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_2434\  = \ \(-R\&_\_2334\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_3434\  = \ \(-\(\[Delta](\[Alpha]\  - \ \[Beta]\&_)\
\)\)\  - \ \[Mu]\ \[Rho]\  + \ \[Lambda]\ \[Sigma]\  - \ \(1\/2\) \((\[Mu]\  \
- \ \[Mu]\&_)\) \((\[Epsilon]\  - \ \[Epsilon]\&_)\)\  - \ \(1\/2\) \((\[Rho]\
\  - \ \[Rho]\&_)\) \((\[Gamma]\  - \ \[Gamma]\&_)\)\n\t\t\ \ \ \ \ \ \ \ \ \ \
\ \ \  - \ \((\[Alpha]\  - \ \[Beta]\&_)\) \((\[Alpha]\&_\  - \ \[Beta])\)\  \
+ \ c . c\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Step 5", "Subsubsection"],

Cell[TextData[{
  "Calculate ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\)\)]],
  " and assign ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalPhi]\_\(A\[InvisibleSpace]B\)\)]]
}], "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_11\  = \ 2  R\_1314, \ \ \ \ \ R\_12\  = \ 
      R\_1212\  + \ R\_1323\  + \ R\&_\_1323, \ \ \ \ \ R\_13\  = \ 
      R\_1213\  + \ R\_1334, \ R\_14\  = \ R\&_\_13\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_22\  = \ 
      2  R\_2324, \ \ \ \ \ R\_23\  = \ \(-R\_1223\)\  + \ 
        R\_2334, \ \ \ \ \ R\_24\  = \ R\&_\_23\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_33\  = \ \(-2\) 
        R\_1323, \ \ \ \ \ R\_34\  = \ \(-R\_1324\)\  - \ R\&_\_1324\  - \ 
        R\_3434, \ \ \ \ \ R\_44\  = \ \(-2\) R\&_\_1323\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Step 6", "Subsubsection"],

Cell[TextData[{
  "Calculate ",
  Cell[BoxData[
      \(TraditionalForm\`C\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\)]],
  " and assign ",
  Cell[BoxData[
      \(TraditionalForm\`\[Psi]\_A\)]]
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Derivation of Step 4", "Subsection"],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[BoxData[
    \(\(Dimension = 4;\)\)], "Input"],

Cell[BoxData[
    \(Off[CoordinateBasisFlag]\)], "Input"],

Cell[CellGroupData[{

Cell["Metric", "Subsubsection"],

Cell[BoxData[
    \(\(metric = {{0, \(-1\), 0, 0}, {\(-1\), 0, 0, 0}, {0, 0, 0, 1}, {0, 0, 
            1, 0}};\)\)], "Input"],

Cell[BoxData[
    \(Cset[Metricg[la, lb], metric]\)], "Input"],

Cell[BoxData[
    \(Cset[Metricg[ua, ub], metric]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Ricci-rotation coefficients", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[CapitalGamma]\_112\  = \ \[Epsilon]\  + \ \[Epsilon]\
\&_, \ \ \ \ \ \[CapitalGamma]\_113\  = \ \[Kappa], \ \ \ \ \ \[CapitalGamma]\
\_114\  = \ \[Kappa]\&_, \ \ \ \ \ \[CapitalGamma]\_123\  = \ \(-\[Pi]\&_\), \
\ \ \ \ \ \[CapitalGamma]\_124\  = \ \(-\[Pi]\), \ \ \ \ \ \
\[CapitalGamma]\_134\  = \ \(-\[Epsilon]\)\  + \ \[Epsilon]\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[CapitalGamma]\_212\  = \ \[Gamma]\  + \ \[Gamma]\&_, \
\ \ \ \ \ \[CapitalGamma]\_213\  = \ \[Tau], \ \ \ \ \ \[CapitalGamma]\_214\  \
= \ \[Tau]\&_, \ \ \ \ \ \[CapitalGamma]\_223\  = \ \(-\[Nu]\&_\), \ \ \ \ \ \
\[CapitalGamma]\_224\  = \ \(-\[Nu]\), \ \ \ \ \ \[CapitalGamma]\_234\  = \ \
\(-\[Gamma]\)\  + \ \[Gamma]\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[CapitalGamma]\_312\  = \ \[Alpha]\&_\  + \ \[Beta], \
\ \ \ \ \ \[CapitalGamma]\_313\  = \ \[Sigma], \ \ \ \ \ \[CapitalGamma]\_314\
\  = \ \[Rho]\&_, \ \ \ \ \ \[CapitalGamma]\_323\  = \ \(-\[Lambda]\&_\), \ \ \
\ \ \ \[CapitalGamma]\_324\  = \ \(-\[Mu]\), \ \ \ \ \ \[CapitalGamma]\_334\  \
= \ \[Alpha]\&_\  - \ \[Beta]\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[CapitalGamma]\_412\  = \ \[Alpha]\  + \ \[Beta]\&_, \
\ \ \ \ \ \[CapitalGamma]\_413\  = \ \[Rho], \ \ \ \ \ \[CapitalGamma]\_414\  \
= \ \[Sigma]\&_, \ \ \ \ \ \[CapitalGamma]\_423\  = \ \(-\[Mu]\&_\), \ \ \ \ \
\ \[CapitalGamma]\_424\  = \ \(-\[Lambda]\), \ \ \ \ \ \[CapitalGamma]\_434\  \
= \ \(-\[Alpha]\)\  + \ \[Beta]\&_\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(rotCoeff[1, 1, 2] = \[Epsilon] + \[Epsilon]\&_; 
    rotCoeff[1, 1, 3] = \[Kappa]; rotCoeff[1, 1, 4] = \[Kappa]\&_; 
    rotCoeff[1, 2, 3] = \(-\[Pi]\&_\); rotCoeff[1, 2, 4] = \(-\[Pi]\); 
    rotCoeff[1, 3, 4] = \(-\[Epsilon]\) + \[Epsilon]\&_; 
    rotCoeff[2, 1, 2] = \[Gamma] + \[Gamma]\&_; rotCoeff[2, 1, 3] = \[Tau]; 
    rotCoeff[2, 1, 4] = \[Tau]\&_; rotCoeff[2, 2, 3] = \(-\[Nu]\&_\); 
    rotCoeff[2, 2, 4] = \(-\[Nu]\); 
    rotCoeff[2, 3, 4] = \(-\[Gamma]\) + \[Gamma]\&_; 
    rotCoeff[3, 1, 2] = \[Alpha]\&_ + \[Beta]; rotCoeff[3, 1, 3] = \[Sigma]; 
    rotCoeff[3, 1, 4] = \[Rho]\&_; rotCoeff[3, 2, 3] = \(-\[Lambda]\&_\); 
    rotCoeff[3, 2, 4] = \(-\[Mu]\); 
    rotCoeff[3, 3, 4] = \[Alpha]\&_ - \[Beta]; 
    rotCoeff[4, 1, 2] = \[Alpha] + \[Beta]\&_; rotCoeff[4, 1, 3] = \[Rho]; 
    rotCoeff[4, 1, 4] = \[Sigma]\&_; rotCoeff[4, 2, 3] = \(-\[Mu]\&_\); 
    rotCoeff[4, 2, 4] = \(-\[Lambda]\); 
    rotCoeff[4, 3, 4] = \(-\[Alpha]\) + \[Beta]\&_;\)], "Input"],

Cell[BoxData[
    \(\(spinCoeff = 
        Table[If[j == k, 0, 
            If[j\  > \ k, \ \(-rotCoeff[i, k, j]\), rotCoeff[i, j, k]]], {i, 
            4}, {j, 4}, {k, 4}];\)\)], "Input"],

Cell[BoxData[
    \(Cset[AffineG[la, lb, lc], spinCoeff]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Subsubsection"],

Cell[BoxData[
    \(Off[MetricgFlag]\)], "Input"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\  = \ \(D\_a\) \[CapitalGamma]\_\(b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\  - \ \(D\_b\) \[CapitalGamma]\_\(a\[InvisibleSpace]c\
\[InvisibleSpace]d\)\  + \ 
        curvMulti\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\[InvisibleSpace]d\)\
\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`curvMulti\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\
\[InvisibleSpace]d\)\  = \ \(g\^\(e\[InvisibleSpace]f\)\)(\(\[CapitalGamma]\_\
\(a\[InvisibleSpace]e\[InvisibleSpace]d\)\) \[CapitalGamma]\_\(b\
\[InvisibleSpace]c\[InvisibleSpace]f\)\  - \ \(\[CapitalGamma]\_\(b\
\[InvisibleSpace]e\[InvisibleSpace]d\)\) \
\[CapitalGamma]\_\(a\[InvisibleSpace]c\[InvisibleSpace]f\)\  - \ \(\
\[CapitalGamma]\_\(a\[InvisibleSpace]b\[InvisibleSpace]e\)\) \
\[CapitalGamma]\_\(f\[InvisibleSpace]c\[InvisibleSpace]d\)\  + \ \(\
\[CapitalGamma]\_\(b\[InvisibleSpace]a\[InvisibleSpace]e\)\) \
\[CapitalGamma]\_\(f\[InvisibleSpace]c\[InvisibleSpace]d\))\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(curvMulti[la_, lb_, lc_, ld_] := 
      Metricg[ue, uf] AffineG[la, le, ld] AffineG[lb, lc, lf] - 
        Metricg[ue, uf] AffineG[lb, le, ld] AffineG[la, lc, lf]\  - \ 
        Metricg[ue, uf] AffineG[la, lb, le] AffineG[lf, lc, ld]\  + \ 
        Metricg[ue, uf] AffineG[lb, la, le] AffineG[lf, lc, ld]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 2, 1, 
        2]\  = \ \(D \((\[Gamma] + \[Gamma]\&_)\)\  - \ \[CapitalDelta] \((\
\[Epsilon] + \[Epsilon]\&_)\)\  + \ 
            curvMulti[\(-1\), \(-2\), \(-1\), \(-2\)]\  // \ 
          SumDum\)\ \  // \ Simplify\)], "Input"],

Cell[BoxData[
    \(\(-\[Kappa]\)\ \[Nu] + \[Pi]\ \[Tau] + \[Pi]\ \((\[Beta] + \[Alpha]\&_)\
\) + \[Tau]\ \((\[Alpha] + \[Beta]\&_)\) + \[Pi]\&_\ \((\[Alpha] + \
\[Beta]\&_)\) + 
      D\ \((\[Gamma] + \[Gamma]\&_)\) - \[CapitalDelta]\ \((\[Epsilon] + \
\[Epsilon]\&_)\) - 
      2\ \((\[Gamma] + \[Gamma]\&_)\)\ \((\[Epsilon] + \[Epsilon]\&_)\) - \
\[Kappa]\&_\ \[Nu]\&_ + \[Pi]\&_\ \[Tau]\&_ + \((\[Beta] + \[Alpha]\&_)\)\ \
\[Tau]\&_\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 2, 1, 
        3] = \(D\ \[Tau]\  - \ \[CapitalDelta]\ \[Kappa]\  + \ 
            curvMulti[\(-1\), \(-2\), \(-1\), \(-3\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\(-3\)\ \[Gamma]\ \[Kappa] - \[CapitalDelta]\ \[Kappa] + \[Pi]\ \
\[Sigma] + 
      D\ \[Tau] + \[Epsilon]\ \[Tau] + \[Rho]\ \[Tau] + \[Rho]\ \[Pi]\&_ - \
\[Kappa]\ \[Gamma]\&_ - \[Tau]\ \[Epsilon]\&_ + \[Sigma]\ \[Tau]\&_\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 2, 2, 
        3] = \(\(-D\)\ \[Nu]\&_\  + \ \[CapitalDelta]\ \[Pi]\&_\  + 
            curvMulti[\(-1\), \(-2\), \(-2\), \(-3\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\[Gamma]\ \[Pi]\&_ + \[CapitalDelta]\ \[Pi]\&_ - \[Pi]\&_\ \[Gamma]\&_ \
- \[Tau]\ \[Mu]\&_ - \[Pi]\&_\ \[Mu]\&_ - 
      D\ \[Nu]\&_ + \[Epsilon]\ \[Nu]\&_ + 
      3\ \[Epsilon]\&_\ \[Nu]\&_ - \[Lambda]\&_\ \((\[Pi] + \[Tau]\&_)\)\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 3, 1, 
        3] = \(D\ \[Sigma] - \[Delta]\ \[Kappa]\  + \ 
            curvMulti[\(-1\), \(-3\), \(-1\), \(-3\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\(-3\)\ \[Beta]\ \[Kappa] - \[Delta]\ \[Kappa] + D\ \[Sigma] + 
      3\ \[Epsilon]\ \[Sigma] + \[Rho]\ \[Sigma] - \[Kappa]\ \[Tau] + \
\[Kappa]\ \[Pi]\&_ - \[Kappa]\ \[Alpha]\&_ - \[Sigma]\ \[Epsilon]\&_ + \
\[Sigma]\ \[Rho]\&_\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 3, 1, 
        4] = \(D\ \[Rho]\&_ - \[Delta]\ \[Kappa]\&_\  + 
            curvMulti[\(-1\), \(-3\), \(-1\), \(-4\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\(-\((\[Beta] + \[Delta] - \[Pi]\&_ + 
              3\ \[Alpha]\&_)\)\)\ \[Kappa]\&_ + \((D + \[Epsilon] + \
\[Epsilon]\&_)\)\ \[Rho]\&_ + \[Rho]\&_\^2 + \[Sigma]\ \[Sigma]\&_ - \[Kappa]\
\ \[Tau]\&_\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 3, 2, 
        3] = \(\(-D\)\ \[Lambda]\&_ + \[Delta]\ \[Pi]\&_\  + 
            curvMulti[\(-1\), \(-3\), \(-2\), \(-3\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\[Beta]\ \[Pi]\&_ + \[Delta]\ \[Pi]\&_ - \[Pi]\&_\^2 - \[Pi]\&_\ \
\[Alpha]\&_ - \[Sigma]\ \[Mu]\&_ + \[Kappa]\ \[Nu]\&_ - \[Lambda]\&_\ \((D + \
\[Epsilon] - 3\ \[Epsilon]\&_ + \[Rho]\&_)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 3, 2, 
        4] = \(\(-D\)\ \[Mu]\  + \ \[Delta]\ \[Pi]\  + \ 
            curvMulti[\(-1\), \(-3\), \(-2\), \(-4\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\(-\[Pi]\)\ \[Beta] + \[Pi]\ \[Delta] - 
      D\ \[Mu] + \[Epsilon]\ \[Mu] + \[Kappa]\ \[Nu] - \[Lambda]\ \[Sigma] - \
\[Pi]\ \[Pi]\&_ + \[Pi]\ \[Alpha]\&_ + \[Mu]\ \[Epsilon]\&_ - \[Mu]\ \
\[Rho]\&_\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[1, 3, 3, 
        4] = \(D \((\[Alpha]\&_ - \[Beta])\) + \[Delta] \((\[Epsilon] - \
\[Epsilon]\&_)\) + curvMulti[\(-1\), \(-3\), \(-3\), \(-4\)]\  // \ 
          SumDum\)\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \(\(-D\)\ \[Beta] + \[Delta]\ \[Epsilon] + \[Gamma]\ \[Kappa] + \[Kappa]\ \
\[Mu] - \[Pi]\ \[Sigma] - \[Alpha]\ \[Sigma] - \[Epsilon]\ \[Pi]\&_ + \
\[Sigma]\ \[Beta]\&_ - \[Kappa]\ \[Gamma]\&_ - \[Delta]\ \[Epsilon]\&_ + \
\[Pi]\&_\ \[Epsilon]\&_ - \[Kappa]\&_\ \[Lambda]\&_ - \[Beta]\ \[Rho]\&_ + \
\[Pi]\&_\ \[Rho]\&_ + \[Alpha]\&_\ \((D + 2\ \[Epsilon] - 
            2\ \[Epsilon]\&_ + \[Rho]\&_)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[2, 3, 2, 
        3] = \(\(-\[CapitalDelta]\)\ \[Lambda]\&_\  + \ \[Delta]\ \[Nu]\&_\  \
+ curvMulti[\(-2\), \(-3\), \(-2\), \(-3\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\[Lambda]\&_\ \((\(-\[Gamma]\) - \[CapitalDelta] + \[Mu] + 
            3\ \[Gamma]\&_ + \[Mu]\&_)\) + \((\(-\[Beta]\) + \[Delta] + \
\[Tau] - \[Pi]\&_ - 3\ \[Alpha]\&_)\)\ \[Nu]\&_\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[2, 3, 2, 
        4] = \(\(-\[CapitalDelta]\)\ \[Mu]\  + \ \[Delta]\ \[Nu]\  + \ 
            curvMulti[\(-2\), \(-3\), \(-2\), \(-4\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(\[Gamma]\ \[Mu] - \[CapitalDelta]\ \[Mu] + \[Mu]\^2 - 
      3\ \[Beta]\ \[Nu] + \[Delta]\ \[Nu] + \[Nu]\ \[Tau] - \[Nu]\ \
\[Alpha]\&_ + \[Mu]\ \[Gamma]\&_ + \[Lambda]\ \[Lambda]\&_ - \[Pi]\ \
\[Nu]\&_\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[2, 3, 3, 
        4] = \(\[CapitalDelta] \((\[Alpha]\&_ - \[Beta])\) + \[Delta] \((\
\[Gamma] - \[Gamma]\&_)\) + curvMulti[\(-2\), \(-3\), \(-3\), \(-4\)]\  // \ 
          SumDum\)\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \(\(-2\)\ \[Beta]\ \[Gamma] + \[Gamma]\ \[Delta] - \[Beta]\ \
\[CapitalDelta] + \[Beta]\ \[Mu] - \[Nu]\ \[Sigma] + \[Gamma]\ \[Tau] + \[Mu]\
\ \[Tau] + \((\[CapitalDelta] - \[Mu])\)\ \[Alpha]\&_ + \((2\ \[Beta] - \
\[Delta] - \[Tau])\)\ \[Gamma]\&_ + \[Alpha]\ \[Lambda]\&_ - \[Beta]\&_\ \
\[Lambda]\&_ - \[Epsilon]\ \[Nu]\&_ + \[Epsilon]\&_\ \[Nu]\&_ + \[Nu]\&_\ \
\[Rho]\&_ - \[Lambda]\&_\ \[Tau]\&_\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvLow[3, 4, 3, 
        4] = \(\(-\[Delta]\) \((\[Alpha] - \[Beta]\&_)\) - \(\[Delta]\&_\) \
\((\[Alpha]\&_ - \[Beta])\) + 
            curvMulti[\(-3\), \(-4\), \(-3\), \(-4\)]\  // \ SumDum\)\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    \(2\ \[Alpha]\ \[Beta] - \[Alpha]\ \[Delta] - \[Epsilon]\ \[Mu] - \
\[Gamma]\ \[Rho] - \[Mu]\ \[Rho] + \[Lambda]\ \[Sigma] + \((\(-2\)\ \[Beta] + \
\[Delta])\)\ \[Beta]\&_ + \[Rho]\ \[Gamma]\&_ + \[Beta]\ \[Delta]\&_ - \
\[Alpha]\&_\ \((2\ \[Alpha] - 
            2\ \[Beta]\&_ + \[Delta]\&_)\) + \[Mu]\ \[Epsilon]\&_ + \
\[Epsilon]\ \[Mu]\&_ - \[Epsilon]\&_\ \[Mu]\&_ + \[Gamma]\ \[Rho]\&_ - \
\[Gamma]\&_\ \[Rho]\&_ - \[Mu]\&_\ \[Rho]\&_ + \[Lambda]\&_\ \[Sigma]\&_\)], \
"Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{746, 674},
WindowMargins->{{39, Automatic}, {Automatic, 12}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 106, 3, 57, "Title"],
Cell[1826, 54, 39, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[1890, 58, 39, 0, 66, "Subsection"],

Cell[CellGroupData[{
Cell[1954, 62, 39, 0, 54, "Subsubsection"],
Cell[1996, 64, 255, 5, 38, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2288, 74, 39, 0, 32, "Subsubsection"],
Cell[2330, 76, 255, 5, 37, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2622, 86, 38, 0, 32, "Subsubsection"],
Cell[2663, 88, 347, 7, 36, "Text"],
Cell[3013, 97, 144, 5, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3194, 107, 44, 0, 32, "Subsubsection"],
Cell[3241, 109, 189, 4, 37, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3467, 118, 52, 0, 32, "Subsubsection"],
Cell[3522, 120, 331, 6, 37, "Text"],
Cell[3856, 128, 428, 7, 42, "Text"],
Cell[4287, 137, 350, 7, 35, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[4674, 149, 34, 0, 32, "Subsubsection"],
Cell[4711, 151, 714, 13, 37, "Text"],
Cell[5428, 166, 188, 4, 37, "Text"],
Cell[5619, 172, 190, 6, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[5858, 184, 42, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[5925, 188, 38, 0, 54, "Subsubsection"],
Cell[5966, 190, 266, 6, 35, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[6269, 201, 39, 0, 32, "Subsubsection"],
Cell[6311, 203, 266, 6, 35, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[6614, 214, 38, 0, 32, "Subsubsection"],
Cell[6655, 216, 195, 4, 35, "Text"],
Cell[6853, 222, 28, 0, 41, "Text"],
Cell[6884, 224, 572, 13, 93, "Text"],
Cell[7459, 239, 192, 8, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[7700, 253, 54, 0, 38, "Subsection"],
Cell[7757, 255, 1295, 42, 116, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[9089, 302, 42, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[9156, 306, 32, 0, 54, "Subsubsection"],
Cell[9191, 308, 302, 6, 35, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[9530, 319, 33, 0, 32, "Subsubsection"],
Cell[9566, 321, 307, 6, 35, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[9910, 332, 34, 0, 32, "Subsubsection"],
Cell[9947, 334, 332, 6, 35, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[10328, 346, 30, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[10383, 350, 48, 0, 54, "Subsubsection"],
Cell[10434, 352, 303, 6, 37, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[10774, 363, 42, 0, 32, "Subsubsection"],
Cell[10819, 365, 384, 6, 42, "Text"],
Cell[11206, 373, 399, 7, 42, "Text"],
Cell[11608, 382, 344, 6, 42, "Text"],
Cell[11955, 390, 151, 5, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[12143, 400, 36, 0, 32, "Subsubsection"],
Cell[12182, 402, 791, 16, 42, "Text"],
Cell[12976, 420, 281, 9, 41, "Text"],
Cell[13260, 431, 271, 6, 35, "Text"],
Cell[13534, 439, 284, 6, 42, "Text"],
Cell[13821, 447, 28, 0, 41, "Text"],
Cell[13852, 449, 250, 5, 42, "Text"],
Cell[14105, 456, 238, 5, 42, "Text"],
Cell[14346, 463, 252, 5, 42, "Text"],
Cell[14601, 470, 299, 6, 42, "Text"],
Cell[14903, 478, 224, 5, 42, "Text"],
Cell[15130, 485, 115, 3, 41, "Text"],
Cell[15248, 490, 296, 5, 35, "Text"],
Cell[15547, 497, 98, 2, 41, "Text"],
Cell[15648, 501, 318, 6, 35, "Text"],
Cell[15969, 509, 306, 6, 35, "Text"],
Cell[16278, 517, 332, 6, 35, "Text"],
Cell[16613, 525, 307, 6, 35, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[16957, 536, 38, 0, 32, "Subsubsection"],
Cell[16998, 538, 311, 8, 41, "Text"],
Cell[17312, 548, 388, 7, 42, "Text"],
Cell[17703, 557, 351, 6, 42, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[18103, 569, 32, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[18160, 573, 31, 0, 54, "Subsubsection"],
Cell[18194, 575, 190, 8, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[18421, 588, 31, 0, 32, "Subsubsection"],
Cell[18455, 590, 348, 9, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[18840, 604, 31, 0, 32, "Subsubsection"],
Cell[18874, 606, 238, 8, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[19149, 619, 31, 0, 32, "Subsubsection"],
Cell[19183, 621, 172, 6, 41, "Text"],
Cell[19358, 629, 417, 7, 35, "Text"],
Cell[19778, 638, 403, 7, 35, "Text"],
Cell[20184, 647, 138, 3, 35, "Text"],
Cell[20325, 652, 420, 7, 35, "Text"],
Cell[20748, 661, 138, 3, 35, "Text"],
Cell[20889, 666, 152, 3, 35, "Text"],
Cell[21044, 671, 381, 8, 35, "Text"],
Cell[21428, 681, 411, 8, 35, "Text"],
Cell[21842, 691, 407, 7, 35, "Text"],
Cell[22252, 700, 361, 6, 35, "Text"],
Cell[22616, 708, 594, 10, 59, "Text"],
Cell[23213, 720, 138, 3, 35, "Text"],
Cell[23354, 725, 138, 3, 35, "Text"],
Cell[23495, 730, 138, 3, 35, "Text"],
Cell[23636, 735, 143, 3, 35, "Text"],
Cell[23782, 740, 403, 7, 35, "Text"],
Cell[24188, 749, 383, 6, 35, "Text"],
Cell[24574, 757, 591, 9, 59, "Text"],
Cell[25168, 768, 138, 3, 35, "Text"],
Cell[25309, 773, 143, 3, 35, "Text"],
Cell[25455, 778, 496, 8, 66, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[25988, 791, 31, 0, 32, "Subsubsection"],
Cell[26022, 793, 218, 7, 41, "Text"],
Cell[26243, 802, 281, 5, 35, "Text"],
Cell[26527, 809, 234, 5, 35, "Text"],
Cell[26764, 816, 266, 5, 35, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[27067, 826, 31, 0, 32, "Subsubsection"],
Cell[27101, 828, 227, 8, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[27377, 842, 42, 0, 38, "Subsection"],
Cell[27422, 844, 52, 1, 56, "Input"],
Cell[27477, 847, 51, 1, 56, "Input"],
Cell[27531, 850, 57, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[27613, 855, 31, 0, 54, "Subsubsection"],
Cell[27647, 857, 126, 2, 56, "Input"],
Cell[27776, 861, 62, 1, 56, "Input"],
Cell[27841, 864, 62, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[27940, 870, 52, 0, 32, "Subsubsection"],
Cell[27995, 872, 451, 7, 35, "Text"],
Cell[28449, 881, 439, 7, 35, "Text"],
Cell[28891, 890, 438, 7, 35, "Text"],
Cell[29332, 899, 443, 7, 35, "Text"],
Cell[29778, 908, 998, 16, 332, "Input"],
Cell[30779, 926, 189, 4, 102, "Input"],
Cell[30971, 932, 69, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[31077, 938, 34, 0, 32, "Subsubsection"],
Cell[31114, 940, 49, 1, 56, "Input"],
Cell[31166, 943, 418, 8, 35, "Text"],
Cell[31587, 953, 746, 12, 38, "Text"],
Cell[32336, 967, 337, 5, 148, "Input"],

Cell[CellGroupData[{
Cell[32698, 976, 263, 5, 102, "Input"],
Cell[32964, 983, 450, 8, 76, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[33451, 996, 209, 4, 79, "Input"],
Cell[33663, 1002, 258, 5, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[33958, 1012, 214, 4, 79, "Input"],
Cell[34175, 1018, 268, 5, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34480, 1028, 200, 4, 79, "Input"],
Cell[34683, 1034, 262, 4, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34982, 1043, 202, 4, 79, "Input"],
Cell[35187, 1049, 234, 4, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[35458, 1058, 207, 4, 79, "Input"],
Cell[35668, 1064, 224, 3, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[35929, 1072, 203, 4, 79, "Input"],
Cell[36135, 1078, 234, 4, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36406, 1087, 228, 4, 102, "Input"],
Cell[36637, 1093, 434, 6, 76, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[37108, 1104, 214, 4, 79, "Input"],
Cell[37325, 1110, 215, 3, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[37577, 1118, 217, 4, 79, "Input"],
Cell[37797, 1124, 236, 4, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[38070, 1133, 238, 4, 102, "Input"],
Cell[38311, 1139, 433, 6, 76, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[38781, 1150, 253, 5, 102, "Input"],
Cell[39037, 1157, 501, 8, 76, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

