(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     36830,       1014]*)
(*NotebookOutlinePosition[     37598,       1041]*)
(*  CellTagsIndexPosition[     37554,       1037]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["2-D SpaceTime in Orthonomal Basis", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {t, x};\)\)], "Input"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-1\), 0}, {0, 1}};\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Orthonomal Basis", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(basisM\  = \ {{E\^\(-\[Rho][t, x]\), 0}, {0, E\^\(-\[Rho][t, x]\)}}; \ 
    MatrixForm[basisM]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), "0"},
            {"0", \(\[ExponentialE]\^\(-\[Rho][t, x]\)\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensors", "Subsubsection"],

Cell[BoxData[
    \(InitCTensor[coSys, \ metric, basisM]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-1\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 1]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", -1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "1"}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 2]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 1],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 1)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ f, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 0, 1][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((1, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ f, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 1, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 1)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 0, 1][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((0, 1)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 0, 1][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 1]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 1]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 2]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((1, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 1]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 2]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 1, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\
\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 2]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`OverHat[ 1]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(-\[Rho][t, x]\)\), " ", 
          RowBox[{
            SuperscriptBox["\[Rho]", 
              TagBox[\((1, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`OverHat[ 1]]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -1, 
              \[Rho][ t, x]]], 
          Derivative[ 1, 0][ \[Rho]][ t, x]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&\
^\)\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`OverHat[ 1]],
                          Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`OverHat[ 1]], 
                        dhPark`Tensor`Private`OverHat[ 2]],
                      Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`OverHat[ 1]], 
                      dhPark`Tensor`Private`OverHat[ 2]], 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`OverHat[ 1]], 
                    dhPark`Tensor`Private`OverHat[ 2]], 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]], 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(1\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 1]],
                  Editable->False], "\[InvisibleSpace]", \(1\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 1]], 
                dhPark`Tensor`Private`OverHat[ 1]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 1]], 
              dhPark`Tensor`Private`OverHat[ 1]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 1]], 
            dhPark`Tensor`Private`OverHat[ 1]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\(2\&^\)\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`OverHat[ 2]],
                  Editable->False], "\[InvisibleSpace]", \(2\&^\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`OverHat[ 2]], 
                dhPark`Tensor`Private`OverHat[ 2]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`OverHat[ 2]], 
              dhPark`Tensor`Private`OverHat[ 2]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`OverHat[ 2]], 
            dhPark`Tensor`Private`OverHat[ 2]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
        "2", " ", \(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
          RowBox[{"(", 
            RowBox[{
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((0, 2)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
              RowBox[{
                SuperscriptBox["\[Rho]", 
                  TagBox[\((2, 0)\),
                    Derivative],
                  MultilineFunction->None], "[", \(t, x\), "]"}]}], ")"}]}]}],
      
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 
        Times[ 2, 
          Power[ E, 
            Times[ -2, 
              \[Rho][ t, x]]], 
          Plus[ 
            Derivative[ 0, 2][ \[Rho]][ t, x], 
            Times[ -1, 
              Derivative[ 2, 0][ \[Rho]][ t, x]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Laplacian[f[t, x], Simplify]\)], "Input"],

Cell[BoxData[
    RowBox[{\(\[ExponentialE]\^\(\(-2\)\ \[Rho][t, x]\)\), " ", 
      RowBox[{"(", 
        RowBox[{
          RowBox[{
            SuperscriptBox["f", 
              TagBox[\((0, 2)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}], "-", 
          RowBox[{
            SuperscriptBox["f", 
              TagBox[\((2, 0)\),
                Derivative],
              MultilineFunction->None], "[", \(t, x\), "]"}]}], 
        ")"}]}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{564, 668},
WindowMargins->{{22, Automatic}, {Automatic, 6}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 66, 1, 43, "Title"],
Cell[1786, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1863, 57, 53, 0, 54, "Subsubsection"],
Cell[1919, 59, 56, 1, 56, "Input"],
Cell[1978, 62, 72, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2087, 68, 41, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[2153, 72, 127, 2, 81, "Input"],
Cell[2283, 76, 259, 6, 93, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2591, 88, 60, 0, 32, "Subsubsection"],
Cell[2654, 90, 69, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2760, 96, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[2822, 100, 46, 1, 56, "Input"],
Cell[2871, 103, 184, 4, 30, "Print"],
Cell[3058, 109, 1646, 41, 32, "Print"],
Cell[4707, 152, 1642, 41, 32, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6386, 198, 49, 1, 56, "Input"],
Cell[6438, 201, 350, 8, 31, "Print"],
Cell[6791, 211, 2975, 75, 37, "Print"],
Cell[9769, 288, 2980, 75, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[12786, 368, 46, 1, 56, "Input"],
Cell[12835, 371, 3017, 75, 37, "Print"],
Cell[15855, 448, 3017, 75, 37, "Print"],
Cell[18875, 525, 3017, 75, 37, "Print"],
Cell[21895, 602, 3017, 75, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[24949, 682, 47, 1, 56, "Input"],
Cell[24999, 685, 293, 6, 29, "Print"],
Cell[25295, 693, 4565, 113, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[29897, 811, 45, 1, 56, "Input"],
Cell[29945, 814, 184, 4, 29, "Print"],
Cell[30132, 820, 2415, 63, 33, "Print"],
Cell[32550, 885, 2410, 63, 33, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34997, 953, 46, 1, 56, "Input"],
Cell[35046, 956, 1151, 32, 30, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36234, 993, 61, 1, 56, "Input"],
Cell[36298, 996, 504, 14, 54, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

