(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     73773,       2017]*)
(*NotebookOutlinePosition[     74541,       2044]*)
(*  CellTagsIndexPosition[     74497,       2040]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Reissner-Nordstrom SpaceTime", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {t, r, \[Theta], \[Phi]};\)\)], "Input"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-A[r]\), 0, 0, 0}, {0, B[r], 0, 0}, {0, 0, r\^2, 
            0}, {0, 0, 0, r\^2\ Sin[\[Theta]]\^2}};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[coSys, metric]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{\(B[r]\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(r\^2\),
                    Superscript[ r, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ r, 2]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(A[r]\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(t\^2\),
                    Superscript[ t, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ t, 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(r\^2\), " ", 
            RowBox[{"(", 
              RowBox[{
                InterpretationBox[
                  RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                    InterpretationBox[\(\[Theta]\^2\),
                      Superscript[ \[Theta], 2],
                      Editable->False]}],
                  SequenceForm[ "d", 
                    Superscript[ \[Theta], 2]],
                  Editable->False], "+", 
                RowBox[{
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                        InterpretationBox[\(\[Phi]\^2\),
                          Superscript[ \[Phi], 2],
                          Editable->False]}],
                      SequenceForm[ "d", 
                        Superscript[ \[Phi], 2]],
                      Editable->False], ")"}], " ", \(Sin[\[Theta]]\^2\)}]}], 
              ")"}]}]}]}],
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          Times[ 
            B[ r], 
            SequenceForm[ "d", 
              Superscript[ r, 2]]], 
          Times[ -1, 
            A[ r], 
            SequenceForm[ "d", 
              Superscript[ t, 2]]], 
          Times[ 
            Power[ r, 2], 
            Plus[ 
              SequenceForm[ "d", 
                Superscript[ \[Theta], 2]], 
              Times[ 
                SequenceForm[ "d", 
                  Superscript[ \[Phi], 2]], 
                Power[ 
                  Sin[ \[Theta]], 2]]]]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensor", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[
      InitCTensor[coSys, \ metric, SimplifyMore \[Rule] True]]\)], "Input"],

Cell[BoxData[
    \({2.784`\ Second, Null}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Maxwell Field", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(maxwellMatrix\  = \ {{0, \(-\(\[CurlyPhi]'\)[r]\), 0, 
          0}, {\(\[CurlyPhi]'\)[r], 0, 0, 0}, {0, 0, 0, 
          M\ Sin[\[Theta]]}, {0, 0, \(-M\)\ Sin[\[Theta]], 0}}; \ 
    MatrixForm[maxwellMatrix]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"0", 
              RowBox[{"-", 
                RowBox[{
                  SuperscriptBox["\[CurlyPhi]", "\[Prime]",
                    MultilineFunction->None], "[", "r", "]"}]}], "0", "0"},
            {
              RowBox[{
                SuperscriptBox["\[CurlyPhi]", "\[Prime]",
                  MultilineFunction->None], "[", "r", "]"}], "0", "0", "0"},
            {"0", "0", "0", \(M\ Sin[\[Theta]]\)},
            {"0", "0", \(\(-M\)\ Sin[\[Theta]]\), "0"}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Cset[MaxwellF[la, lb], maxwellMatrix]\)], "Input"],

Cell[BoxData[
    \(\(maxwellUp\  = \ 
        Table[SumDum[
            Metricg[ua, uc]\ Metricg[ub, ud]\ MaxwellF[lc, ld]], {ua, 
            4}, {ub, 4}];\)\)], "Input"],

Cell[BoxData[
    \(Cset[MaxwellF[ua, ub], maxwellUp]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Energy-Momentum Tensor", "Subsubsection"],

Cell[BoxData[
    \(energyMomentumT[la_, lb_]\  := 
      SumDum[Metricg[uc, ud] MaxwellF[la, lc] MaxwellF[lb, ld]\  - \ \(1\/4\) 
            Metricg[la, lb] MaxwellF[uc, ud] MaxwellF[lc, ld]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(energyMomentumT[\(-1\), \(-1\)] // Expand\)], "Input"],

Cell[BoxData[
    RowBox[{\(\(M\^2\ A[r]\)\/\(2\ r\^4\)\), "+", 
      FractionBox[
        SuperscriptBox[
          RowBox[{
            SuperscriptBox["\[CurlyPhi]", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}], 
          "2"], \(2\ B[r]\)]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(energyMomentumT[\(-2\), \(-2\)] // Expand\)], "Input"],

Cell[BoxData[
    RowBox[{\(-\(\(M\^2\ B[r]\)\/\(2\ r\^4\)\)\), "-", 
      FractionBox[
        SuperscriptBox[
          RowBox[{
            SuperscriptBox["\[CurlyPhi]", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}], 
          "2"], \(2\ A[r]\)]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(energyMomentumT[\(-3\), \(-3\)] // Expand\)], "Input"],

Cell[BoxData[
    RowBox[{\(M\^2\/\(2\ r\^2\)\), "+", 
      FractionBox[
        RowBox[{\(r\^2\), " ", 
          SuperscriptBox[
            RowBox[{
              SuperscriptBox["\[CurlyPhi]", "\[Prime]",
                MultilineFunction->None], "[", "r", "]"}], "2"]}], \(2\ A[
            r]\ B[r]\)]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(energyMomentumT[\(-4\), \(-4\)]\  - \ 
        Sin[\[Theta]]\^2\ energyMomentumT[\(-3\), \(-3\)]\  // \ 
      Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Equations of Motion", "Subsubsection"],

Cell[TextData[{
  "Maxwell equation: ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\^a F\_\(a\[InvisibleSpace]b\)\  = \ 0\)]]
}], "Text"],

Cell[BoxData[
    \(maxwellEq[
        la_]\  := \(Metricg[ub, uc] CD[lb, MaxwellF[lc, la]]\  // \ 
          CDtoAffine\)\  // \ SumDum\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[maxwellEq[\(-i\)], {i, 2, 4}]\)], "Input"],

Cell[BoxData[
    \({0, 0, 0}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(maxwellEq[\(-1\)]\ B[r]\  // \ Expand\)], "Input"],

Cell[BoxData[
    RowBox[{
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{
            SuperscriptBox["\[CurlyPhi]", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}]}], "r"], "-", 
      FractionBox[
        RowBox[{
          RowBox[{
            SuperscriptBox["A", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}], " ", 
          RowBox[{
            SuperscriptBox["\[CurlyPhi]", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}]}], \(2\ A[r]\)], "-", 
      
      FractionBox[
        RowBox[{
          RowBox[{
            SuperscriptBox["B", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}], " ", 
          RowBox[{
            SuperscriptBox["\[CurlyPhi]", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}]}], \(2\ B[r]\)], "+", 
      
      RowBox[{
        SuperscriptBox["\[CurlyPhi]", "\[Prime]\[Prime]",
          MultilineFunction->None], "[", "r", "]"}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, the Maxwell equation gives ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyPhi]'\  = \ \(-\(Q\/r\^2\)\) \@\(A\ B\)\)]]
}], "Text"],

Cell[BoxData[
    \(\(maxwellSolRule\  = \ {\(\[CurlyPhi]'\)[
              r]\  -> \ \(-\(Q\/r\^2\)\) \@\(A[r]\ B[r]\)};\)\)], "Input"],

Cell[TextData[{
  "Because ",
  Cell[BoxData[
      \(TraditionalForm\`\(g\^\(a\[InvisibleSpace]b\)\) 
          T\_\(a\[InvisibleSpace]b\)\  = \ 0\)]],
  ", the Einstein equation becomes ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\)\  = \ \(-8\) \[Pi]\ G\ \
T\_\(a\[InvisibleSpace]b\)\)]]
}], "Text"],

Cell[BoxData[
    \(einsteinEq[la_, lb_]\  := \ 
      RicciR[la, lb]\  + \ 8  \[Pi]\ G\ energyMomentumT[la, lb]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(-\(\(r\ B[r]\)\/A[r]\)\) einsteinEq[\(-1\), \(-1\)]\  - \ 
        r\ *\ einsteinEq[\(-2\), \(-2\)]\  // \ Expand\)], "Input"],

Cell[BoxData[
    RowBox[{
      FractionBox[
        RowBox[{
          SuperscriptBox["A", "\[Prime]",
            MultilineFunction->None], "[", "r", "]"}], \(A[r]\)], "+", 
      FractionBox[
        RowBox[{
          SuperscriptBox["B", "\[Prime]",
            MultilineFunction->None], "[", "r", "]"}], \(B[r]\)]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "From this and the asymptotic flatness we have ",
  Cell[BoxData[
      \(TraditionalForm\`A\ B\  = \ 1\)]],
  "."
}], "Text"],

Cell[BoxData[
    \(B[r_]\  := \ 1\/A[r]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(einsteinEq[\(-3\), \(-3\)]\  /. \ maxwellSolRule\  // \ 
      Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{\(-1\), "+", \(\(4\ G\ M\^2\ \[Pi]\)\/r\^2\), 
      "+", \(\(4\ G\ \[Pi]\ Q\^2\)\/r\^2\), "+", \(A[r]\), "+", 
      RowBox[{"r", " ", 
        RowBox[{
          SuperscriptBox["A", "\[Prime]",
            MultilineFunction->None], "[", "r", "]"}]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "From this we have ",
  Cell[BoxData[
      \(TraditionalForm\`A\  = \ 
        1\  - \ \(2  G\ m\)\/r\  + \ 
          4  \[Pi]\ G\ \(Q\^2\  + \ M\^2\)\/r\^2\)]]
}], "Text"],

Cell[BoxData[
    \(A[r_]\  := \ 
      1\  - \ \(2  G\ m\)\/r\  + \ 
        4  \[Pi]\ G\ \(Q\^2\  + \ M\^2\)\/r\^2\)], "Input"],

Cell["The remaining equations are automatically satisfied", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(einsteinEq[\(-1\), \(-1\)]\  /. \ maxwellSolRule\  // \ 
      Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(einsteinEq[\(-4\), \(-4\)]\  - \ 
        Sin[\[Theta]]\^2\ einsteinEq[\(-3\), \(-3\)]\  // \ 
      Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Solution", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[coSys, metric]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          FractionBox[
            InterpretationBox[
              RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                InterpretationBox[\(r\^2\),
                  Superscript[ r, 2],
                  Editable->False]}],
              SequenceForm[ "d", 
                Superscript[ r, 2]],
              
              Editable->
                False], \(1 + \(4\ G\ \[Pi]\ \((M\^2 + Q\^2)\)\)\/r\^2 - \(2\ \
G\ m\)\/r\)], "+", 
          
          RowBox[{\((\(-1\) - \(4\ G\ \[Pi]\ \((M\^2 + Q\^2)\)\)\/r\^2 + \(2\ \
G\ m\)\/r)\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(t\^2\),
                    Superscript[ t, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ t, 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(r\^2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Theta]\^2\),
                    Superscript[ \[Theta], 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ \[Theta], 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(r\^2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Phi]\^2\),
                    Superscript[ \[Phi], 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ \[Phi], 2]],
                Editable->False], ")"}], " ", \(Sin[\[Theta]]\^2\)}]}]}],
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          Times[ 
            Power[ 
              Plus[ 1, 
                Times[ 4, G, Pi, 
                  Plus[ 
                    Power[ M, 2], 
                    Power[ Q, 2]], 
                  Power[ r, -2]], 
                Times[ -2, G, m, 
                  Power[ r, -1]]], -1], 
            SequenceForm[ "d", 
              Superscript[ r, 2]]], 
          Times[ 
            Plus[ -1, 
              Times[ -4, G, Pi, 
                Plus[ 
                  Power[ M, 2], 
                  Power[ Q, 2]], 
                Power[ r, -2]], 
              Times[ 2, G, m, 
                Power[ r, -1]]], 
            SequenceForm[ "d", 
              Superscript[ t, 2]]], 
          Times[ 
            Power[ r, 2], 
            SequenceForm[ "d", 
              Superscript[ \[Theta], 2]]], 
          Times[ 
            Power[ r, 2], 
            SequenceForm[ "d", 
              Superscript[ \[Phi], 2]], 
            Power[ 
              Sin[ \[Theta]], 2]]]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(maxwellSolRule\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          SuperscriptBox["\[CurlyPhi]", "\[Prime]",
            MultilineFunction->None], "[", "r", "]"}], 
        "\[Rule]", \(-\(Q\/r\^2\)\)}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Calculate Component Tensors using the Solution", "Subsubsection"],

Cell[BoxData[
    \(InitCTensor[coSys, metric, SimplifyMore -> True]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                    m\ r)\)\ \((r\^2 + 
                    G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                          2\ m\ r)\))\)\)\/r\^5\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -1, G, 
          Power[ r, -5], 
          Plus[ 
            Times[ 4, 
              Power[ M, 2], Pi], 
            Times[ 4, Pi, 
              Power[ Q, 2]], 
            Times[ -1, m, r]], 
          Plus[ 
            Power[ r, 2], 
            Times[ G, 
              Plus[ 
                Times[ 4, 
                  Power[ M, 2], Pi], 
                Times[ 4, Pi, 
                  Power[ Q, 2]], 
                Times[ -2, m, r]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ \((\(-4\)\ M\^2\ \[Pi] - 4\ \[Pi]\ Q\^2 + 
                  m\ r)\)\)\/\(r\ \((r\^2 + 
                  G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 2\ m\ r)\))\)\)\)}],
      
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ G, 
          Power[ r, -1], 
          Plus[ 
            Times[ -4, 
              Power[ M, 2], Pi], 
            Times[ -4, Pi, 
              Power[ Q, 2]], 
            Times[ m, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ G, 
                Plus[ 
                  Times[ 4, 
                    Power[ M, 2], Pi], 
                  Times[ 4, Pi, 
                    Power[ Q, 2]], 
                  Times[ -2, m, r]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                  m\ r)\)\)\/\(r\ \((r\^2 + 
                  G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 2\ m\ r)\))\)\)\)}],
      
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ G, 
          Power[ r, -1], 
          Plus[ 
            Times[ 4, 
              Power[ M, 2], Pi], 
            Times[ 4, Pi, 
              Power[ Q, 2]], 
            Times[ -1, m, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ G, 
                Plus[ 
                  Times[ 4, 
                    Power[ M, 2], Pi], 
                  Times[ 4, Pi, 
                    Power[ Q, 2]], 
                  Times[ -2, m, r]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(1\/r\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Power[ r, -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(1\/r\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Power[ r, -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(4\ G\ M\^2\ \[Pi] + 4\ G\ \[Pi]\ Q\^2 - 
                2\ G\ m\ r + r\^2\)\/r\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -1, 
          Power[ r, -1], 
          Plus[ 
            Times[ 4, G, 
              Power[ M, 2], Pi], 
            Times[ 4, G, Pi, 
              Power[ Q, 2]], 
            Times[ -2, G, m, r], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(Cot[\[Theta]]\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Cot[ \[Theta]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(\((r\^2 + 
                    G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                          2\ m\ r)\))\)\ Sin[\[Theta]]\^2\)\/r\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ -1, 
          Power[ r, -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ G, 
              Plus[ 
                Times[ 4, 
                  Power[ M, 2], Pi], 
                Times[ 4, Pi, 
                  Power[ Q, 2]], 
                Times[ -2, m, r]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-Cos[\[Theta]]\)\ Sin[\[Theta]]\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ -1, 
          Cos[ \[Theta]], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Calculation of Weyl Tensor", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[WeylC, Simplify]]\)], "Input"],

Cell[BoxData[
    \({2.043000000000003`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[WeylC]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"C"\_"[ab][cd]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "C", "[ab][cd]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ G\ \((\(-4\)\ M\^2\ \[Pi] - 
                  4\ \[Pi]\ Q\^2 + m\ r)\)\)\/r\^4\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 2, G, 
          Power[ r, -4], 
          Plus[ 
            Times[ -4, 
              Power[ M, 2], Pi], 
            Times[ -4, Pi, 
              Power[ Q, 2]], 
            Times[ m, r]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                  m\ r)\)\ \((r\^2 + 
                  G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                        2\ m\ r)\))\)\)\/r\^4\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, 
          Power[ r, -4], 
          Plus[ 
            Times[ 4, 
              Power[ M, 2], Pi], 
            Times[ 4, Pi, 
              Power[ Q, 2]], 
            Times[ -1, m, r]], 
          Plus[ 
            Power[ r, 2], 
            Times[ G, 
              Plus[ 
                Times[ 4, 
                  Power[ M, 2], Pi], 
                Times[ 4, Pi, 
                  Power[ Q, 2]], 
                Times[ -2, m, r]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(1\/r\^4\)\((G\ \((4\ M\^2\ \[Pi] + 
                4\ \[Pi]\ Q\^2 - m\ r)\)\ \((r\^2 + 
                G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                      2\ m\ r)\))\)\ Sin[\[Theta]]\^2)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, 
          Power[ r, -4], 
          Plus[ 
            Times[ 4, 
              Power[ M, 2], Pi], 
            Times[ 4, Pi, 
              Power[ Q, 2]], 
            Times[ -1, m, r]], 
          Plus[ 
            Power[ r, 2], 
            Times[ G, 
              Plus[ 
                Times[ 4, 
                  Power[ M, 2], Pi], 
                Times[ 4, Pi, 
                  Power[ Q, 2]], 
                Times[ -2, m, r]]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ \((\(-4\)\ M\^2\ \[Pi] - 4\ \[Pi]\ Q\^2 + 
                  m\ r)\)\)\/\(r\^2 + 
              G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 2\ m\ r)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, 
          Plus[ 
            Times[ -4, 
              Power[ M, 2], Pi], 
            Times[ -4, Pi, 
              Power[ Q, 2]], 
            Times[ m, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ G, 
                Plus[ 
                  Times[ 4, 
                    Power[ M, 2], Pi], 
                  Times[ 4, Pi, 
                    Power[ Q, 2]], 
                  Times[ -2, m, r]]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(\(G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
                    m\ r)\)\ Sin[\[Theta]]\^2\)\/\(r\^2 + 
                G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 2\ m\ r)\)\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, G, 
          Plus[ 
            Times[ 4, 
              Power[ M, 2], Pi], 
            Times[ 4, Pi, 
              Power[ Q, 2]], 
            Times[ -1, m, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ G, 
                Plus[ 
                  Times[ 4, 
                    Power[ M, 2], Pi], 
                  Times[ 4, Pi, 
                    Power[ Q, 2]], 
                  Times[ -2, m, r]]]], -1], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(2\ G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 
              m\ r)\)\ Sin[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 2, G, 
          Plus[ 
            Times[ 4, 
              Power[ M, 2], Pi], 
            Times[ 4, Pi, 
              Power[ Q, 2]], 
            Times[ -1, m, r]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Laplace Equation", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Laplacian[f[r], Simplify]\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/r\^2\), 
      RowBox[{"(", 
        RowBox[{
          RowBox[{\((\(-2\)\ G\ m + 2\ r)\), " ", 
            RowBox[{
              SuperscriptBox["f", "\[Prime]",
                MultilineFunction->None], "[", "r", "]"}]}], "+", 
          
          RowBox[{\((r\^2 + 
                G\ \((4\ M\^2\ \[Pi] + 4\ \[Pi]\ Q\^2 - 2\ m\ r)\))\), " ", 
            RowBox[{
              SuperscriptBox["f", "\[Prime]\[Prime]",
                MultilineFunction->None], "[", "r", "]"}]}]}], 
        ")"}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature Invariant", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[\(riemannUp\  = \ Get["\<dhPark`RiemannUp4`\>"];\)]\)], "Input"],

Cell[BoxData[
    \({3.3049999999999997`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Cset[RiemannR[ua, ub, uc, ud], riemannUp]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[
      SumDum[RiemannR[la, lb, lc, ld]\ RiemannR[ua, ub, uc, ud]]\  // \ 
        Simplify]\)], "Input"],

Cell[BoxData[
    \({0.9710000000000001`\ Second, \(\(1\/r\^8\)\((16\ G\^2\ \((56\ M\^4\ \
\[Pi]\^2 + 56\ \[Pi]\^2\ Q\^4 - 24\ m\ \[Pi]\ Q\^2\ r + 3\ m\^2\ r\^2 + 
              8\ M\^2\ \[Pi]\ \((14\ \[Pi]\ Q\^2 - 
                    3\ m\ r)\))\))\)\)}\)], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{600, 667},
WindowMargins->{{18, Automatic}, {Automatic, 4}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 61, 1, 43, "Title"],
Cell[1781, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1858, 57, 53, 0, 54, "Subsubsection"],
Cell[1914, 59, 74, 1, 56, "Input"],
Cell[1991, 62, 155, 2, 80, "Input"],

Cell[CellGroupData[{
Cell[2171, 68, 59, 1, 56, "Input"],
Cell[2233, 71, 2682, 71, 77, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[4964, 148, 59, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[5048, 152, 103, 2, 79, "Input"],
Cell[5154, 156, 56, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5259, 163, 38, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[5322, 167, 241, 4, 125, "Input"],
Cell[5566, 173, 628, 15, 127, "Output"]
}, Open  ]],
Cell[6209, 191, 70, 1, 56, "Input"],
Cell[6282, 194, 172, 4, 125, "Input"],
Cell[6457, 200, 66, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[6560, 206, 47, 0, 32, "Subsubsection"],
Cell[6610, 208, 206, 3, 195, "Input"],

Cell[CellGroupData[{
Cell[6841, 215, 74, 1, 56, "Input"],
Cell[6918, 218, 281, 7, 75, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7236, 230, 74, 1, 56, "Input"],
Cell[7313, 233, 286, 7, 75, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7636, 245, 74, 1, 56, "Input"],
Cell[7713, 248, 321, 8, 75, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8071, 261, 152, 3, 80, "Input"],
Cell[8226, 266, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[8310, 273, 44, 0, 32, "Subsubsection"],
Cell[8357, 275, 141, 4, 41, "Text"],
Cell[8501, 281, 149, 3, 102, "Input"],

Cell[CellGroupData[{
Cell[8675, 288, 68, 1, 56, "Input"],
Cell[8746, 291, 43, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8826, 297, 70, 1, 56, "Input"],
Cell[8899, 300, 997, 27, 72, "Output"]
}, Open  ]],
Cell[9911, 330, 163, 4, 44, "Text"],
Cell[10077, 336, 136, 2, 71, "Input"],
Cell[10216, 340, 326, 9, 41, "Text"],
Cell[10545, 351, 125, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[10695, 357, 148, 2, 99, "Input"],
Cell[10846, 361, 333, 9, 72, "Output"]
}, Open  ]],
Cell[11194, 373, 144, 5, 41, "Text"],
Cell[11341, 380, 53, 1, 75, "Input"],

Cell[CellGroupData[{
Cell[11419, 385, 104, 2, 79, "Input"],
Cell[11526, 389, 291, 6, 72, "Output"]
}, Open  ]],
Cell[11832, 398, 192, 6, 49, "Text"],
Cell[12027, 406, 129, 3, 75, "Input"],
Cell[12159, 411, 67, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[12251, 415, 104, 2, 79, "Input"],
Cell[12358, 419, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[12430, 425, 142, 3, 80, "Input"],
Cell[12575, 430, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[12659, 437, 33, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[12717, 441, 59, 1, 56, "Input"],
Cell[12779, 444, 3165, 88, 165, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15981, 537, 47, 1, 56, "Input"],
Cell[16031, 540, 228, 6, 68, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[16308, 552, 71, 0, 32, "Subsubsection"],
Cell[16382, 554, 81, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[16488, 559, 46, 1, 56, "Input"],
Cell[16537, 562, 392, 8, 31, "Print"],
Cell[16932, 572, 3361, 84, 75, "Print"],
Cell[20296, 658, 3373, 85, 50, "Print"],
Cell[23672, 745, 3370, 85, 50, "Print"],
Cell[27045, 832, 2808, 66, 44, "Print"],
Cell[29856, 900, 2790, 66, 44, "Print"],
Cell[32649, 968, 3105, 75, 47, "Print"],
Cell[35757, 1045, 2821, 66, 39, "Print"],
Cell[38581, 1113, 3246, 80, 47, "Print"],
Cell[41830, 1195, 2887, 68, 37, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[44766, 1269, 51, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[44842, 1273, 63, 1, 56, "Input"],
Cell[44908, 1276, 68, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[45013, 1282, 44, 1, 56, "Input"],
Cell[45060, 1285, 192, 4, 29, "Print"],
Cell[45255, 1291, 4165, 101, 47, "Print"],
Cell[49423, 1394, 4593, 112, 76, "Print"],
Cell[54019, 1508, 4643, 114, 68, "Print"],
Cell[58665, 1624, 4574, 111, 50, "Print"],
Cell[63242, 1737, 4633, 113, 50, "Print"],
Cell[67878, 1852, 4348, 103, 35, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[72275, 1961, 41, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[72341, 1965, 58, 1, 56, "Input"],
Cell[72402, 1968, 545, 14, 93, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[72996, 1988, 44, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[73065, 1992, 91, 1, 56, "Input"],
Cell[73159, 1995, 69, 1, 53, "Output"]
}, Open  ]],
Cell[73243, 1999, 74, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[73342, 2004, 131, 3, 102, "Input"],
Cell[73476, 2009, 269, 4, 118, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

