(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     90758,       2385]*)
(*NotebookOutlinePosition[     91526,       2412]*)
(*  CellTagsIndexPosition[     91482,       2408]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Robertson-Walker SpaceTime", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {\[Tau], r, \[Theta], \[Phi]};\)\)], "Input"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-1\), 0, 0, 0}, {0, 
            a[\[Tau]]\^2\/\(1\  - \ \[Kappa]\ r\^2\), 0, 0}, {0, 
            0, \(a[\[Tau]]\^2\) r\^2, 0}, {0, 0, 
            0, \(a[\[Tau]]\^2\) r\^2\ Sin[\[Theta]]\^2}};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[coSys, metric]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{"-", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Tau]\^2\),
                    Superscript[ \[Tau], 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ \[Tau], 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(a[\[Tau]]\^2\), " ", 
            RowBox[{"(", 
              RowBox[{
                FractionBox[
                  InterpretationBox[
                    RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                      InterpretationBox[\(r\^2\),
                        Superscript[ r, 2],
                        Editable->False]}],
                    SequenceForm[ "d", 
                      Superscript[ r, 2]],
                    Editable->False], \(1 - r\^2\ \[Kappa]\)], "+", 
                RowBox[{\(r\^2\), " ", 
                  RowBox[{"(", 
                    RowBox[{
                      InterpretationBox[
                        RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                          InterpretationBox[\(\[Theta]\^2\),
                            Superscript[ \[Theta], 2],
                            Editable->False]}],
                        SequenceForm[ "d", 
                          Superscript[ \[Theta], 2]],
                        Editable->False], "+", 
                      RowBox[{
                        RowBox[{"(", 
                          InterpretationBox[
                            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                              InterpretationBox[\(\[Phi]\^2\),
                                Superscript[ \[Phi], 2],
                                Editable->False]}],
                            SequenceForm[ "d", 
                              Superscript[ \[Phi], 2]],
                            Editable->False], ")"}], 
                        " ", \(Sin[\[Theta]]\^2\)}]}], ")"}]}]}], ")"}]}]}]}],
      
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          Times[ -1, 
            SequenceForm[ "d", 
              Superscript[ \[Tau], 2]]], 
          Times[ 
            Power[ 
              a[ \[Tau]], 2], 
            Plus[ 
              Times[ 
                Power[ 
                  Plus[ 1, 
                    Times[ -1, 
                      Power[ r, 2], \[Kappa]]], -1], 
                SequenceForm[ "d", 
                  Superscript[ r, 2]]], 
              Times[ 
                Power[ r, 2], 
                Plus[ 
                  SequenceForm[ "d", 
                    Superscript[ \[Theta], 2]], 
                  Times[ 
                    SequenceForm[ "d", 
                      Superscript[ \[Phi], 2]], 
                    Power[ 
                      Sin[ \[Theta]], 2]]]]]]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensors", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[InitCTensor[coSys, \ metric, SimplifyMore -> True]]\)], "Input"],

Cell[BoxData[
    \({2.564`\ Second, Null}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Tau]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{
            SuperscriptBox["a", "\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", 
            "]"}], \(a[\[Tau]]\)]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Power[ 
            a[ \[Tau]], -1], 
          Derivative[ 1][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Tau]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{
            SuperscriptBox["a", "\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", 
            "]"}], \(a[\[Tau]]\)]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ 
          Power[ 
            a[ \[Tau]], -1], 
          Derivative[ 1][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Tau]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{
            SuperscriptBox["a", "\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", 
            "]"}], \(a[\[Tau]]\)]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Times[ 
          Power[ 
            a[ \[Tau]], -1], 
          Derivative[ 1][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{\(a[\[Tau]]\), " ", 
            RowBox[{
              SuperscriptBox["a", "\[Prime]",
                MultilineFunction->None], "[", "\[Tau]", "]"}]}], \(1 - 
            r\^2\ \[Kappa]\)]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]], " = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ r, 2], \[Kappa]]], -1], 
          a[ \[Tau]], 
          Derivative[ 1][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(r\ \[Kappa]\)\/\(1 - r\^2\ \[Kappa]\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ r, \[Kappa], 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ r, 2], \[Kappa]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(1\/r\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Power[ r, -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(1\/r\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Power[ r, -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(r\^2\), " ", \(a[\[Tau]]\), " ", 
          RowBox[{
            SuperscriptBox["a", "\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]], " = ", 
        Times[ 
          Power[ r, 2], 
          a[ \[Tau]], 
          Derivative[ 1][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\ \((\(-1\) + r\^2\ \[Kappa])\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ r, 
          Plus[ -1, 
            Times[ 
              Power[ r, 2], \[Kappa]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(Cot[\[Theta]]\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Cot[ \[Theta]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(r\^2\), " ", \(a[\[Tau]]\), " ", \(Sin[\[Theta]]\^2\), " ", 
          
          RowBox[{
            SuperscriptBox["a", "\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]]], " = ", 
        Times[ 
          Power[ r, 2], 
          a[ \[Tau]], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Derivative[ 1][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\ \((\(-1\) + 
              r\^2\ \[Kappa])\)\ Sin[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ r, 
          Plus[ -1, 
            Times[ 
              Power[ r, 2], \[Kappa]]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-Cos[\[Theta]]\)\ Sin[\[Theta]]\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ -1, 
          Cos[ \[Theta]], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Tau]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{\(a[\[Tau]]\), " ", 
            RowBox[{
              SuperscriptBox["a", "\[Prime]\[Prime]",
                MultilineFunction->None], "[", "\[Tau]", "]"}]}], \(1 - 
            r\^2\ \[Kappa]\)]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ r, 2], \[Kappa]]], -1], 
          a[ \[Tau]], 
          Derivative[ 2][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Tau]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(r\^2\), " ", \(a[\[Tau]]\), " ", 
          RowBox[{
            SuperscriptBox["a", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ r, 2], 
          a[ \[Tau]], 
          Derivative[ 2][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Tau]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(r\^2\), " ", \(a[\[Tau]]\), " ", \(Sin[\[Theta]]\^2\), " ", 
          
          RowBox[{
            SuperscriptBox["a", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", "]"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ r, 2], 
          a[ \[Tau]], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Derivative[ 2][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{\(r\^2\), " ", \(a[\[Tau]]\^2\), " ", 
            RowBox[{"(", 
              RowBox[{"\[Kappa]", "+", 
                SuperscriptBox[
                  RowBox[{
                    SuperscriptBox["a", "\[Prime]",
                      MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
              ")"}]}], \(\(-1\) + r\^2\ \[Kappa]\)]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Plus[ -1, 
              Times[ 
                Power[ r, 2], \[Kappa]]], -1], 
          Power[ 
            a[ \[Tau]], 2], 
          Plus[ \[Kappa], 
            Power[ 
              Derivative[ 1][ a][ \[Tau]], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          
          RowBox[{\(r\^2\), " ", \(a[\[Tau]]\^2\), " ", \(Sin[\[Theta]]\^2\), 
            " ", 
            RowBox[{"(", 
              RowBox[{"\[Kappa]", "+", 
                SuperscriptBox[
                  RowBox[{
                    SuperscriptBox["a", "\[Prime]",
                      MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
              ")"}]}], \(\(-1\) + r\^2\ \[Kappa]\)]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Plus[ -1, 
              Times[ 
                Power[ r, 2], \[Kappa]]], -1], 
          Power[ 
            a[ \[Tau]], 2], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Plus[ \[Kappa], 
            Power[ 
              Derivative[ 1][ a][ \[Tau]], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-r\^4\), " ", \(a[\[Tau]]\^2\), " ", \(Sin[\[Theta]]\^2\), 
          " ", 
          RowBox[{"(", 
            RowBox[{"\[Kappa]", "+", 
              SuperscriptBox[
                RowBox[{
                  SuperscriptBox["a", "\[Prime]",
                    MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
            ")"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -1, 
          Power[ r, 4], 
          Power[ 
            a[ \[Tau]], 2], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Plus[ \[Kappa], 
            Power[ 
              Derivative[ 1][ a][ \[Tau]], 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Tau]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Tau]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{"3", " ", 
            RowBox[{
              SuperscriptBox["a", "\[Prime]\[Prime]",
                MultilineFunction->None], "[", "\[Tau]", 
              "]"}]}], \(a[\[Tau]]\)]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Tau]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 3, 
          Power[ 
            a[ \[Tau]], -1], 
          Derivative[ 2][ a][ \[Tau]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        FractionBox[
          RowBox[{\(2\ \[Kappa]\), "+", 
            RowBox[{"2", " ", 
              SuperscriptBox[
                RowBox[{
                  SuperscriptBox["a", "\[Prime]",
                    MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
            "+", 
            RowBox[{\(a[\[Tau]]\), " ", 
              RowBox[{
                SuperscriptBox["a", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "\[Tau]", 
                "]"}]}]}], \(\(-1\) + r\^2\ \[Kappa]\)]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ 
            Plus[ -1, 
              Times[ 
                Power[ r, 2], \[Kappa]]], -1], 
          Plus[ 
            Times[ 2, \[Kappa]], 
            Times[ 2, 
              Power[ 
                Derivative[ 1][ a][ \[Tau]], 2]], 
            Times[ 
              a[ \[Tau]], 
              Derivative[ 2][ a][ \[Tau]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]"\+_\
\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-r\^2\), " ", 
          RowBox[{"(", 
            RowBox[{\(2\ \[Kappa]\), "+", 
              RowBox[{"2", " ", 
                SuperscriptBox[
                  RowBox[{
                    SuperscriptBox["a", "\[Prime]",
                      MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
              "+", 
              RowBox[{\(a[\[Tau]]\), " ", 
                RowBox[{
                  SuperscriptBox["a", "\[Prime]\[Prime]",
                    MultilineFunction->None], "[", "\[Tau]", "]"}]}]}], 
            ")"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ r, 2], 
          Plus[ 
            Times[ 2, \[Kappa]], 
            Times[ 2, 
              Power[ 
                Derivative[ 1][ a][ \[Tau]], 2]], 
            Times[ 
              a[ \[Tau]], 
              Derivative[ 2][ a][ \[Tau]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{\(-r\^2\), " ", \(Sin[\[Theta]]\^2\), " ", 
          RowBox[{"(", 
            RowBox[{\(2\ \[Kappa]\), "+", 
              RowBox[{"2", " ", 
                SuperscriptBox[
                  RowBox[{
                    SuperscriptBox["a", "\[Prime]",
                      MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
              "+", 
              RowBox[{\(a[\[Tau]]\), " ", 
                RowBox[{
                  SuperscriptBox["a", "\[Prime]\[Prime]",
                    MultilineFunction->None], "[", "\[Tau]", "]"}]}]}], 
            ")"}]}]}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ -1, 
          Power[ r, 2], 
          Power[ 
            Sin[ \[Theta]], 2], 
          Plus[ 
            Times[ 2, \[Kappa]], 
            Times[ 2, 
              Power[ 
                Derivative[ 1][ a][ \[Tau]], 2]], 
            Times[ 
              a[ \[Tau]], 
              Derivative[ 2][ a][ \[Tau]]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{"-", 
          FractionBox[
            RowBox[{"6", " ", 
              RowBox[{"(", 
                RowBox[{"\[Kappa]", "+", 
                  SuperscriptBox[
                    RowBox[{
                      SuperscriptBox["a", "\[Prime]",
                        MultilineFunction->None], "[", "\[Tau]", "]"}], "2"], 
                  "+", 
                  RowBox[{\(a[\[Tau]]\), " ", 
                    RowBox[{
                      SuperscriptBox["a", "\[Prime]\[Prime]",
                        MultilineFunction->None], "[", "\[Tau]", "]"}]}]}], 
                ")"}]}], \(a[\[Tau]]\^2\)]}]}],
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 
        Times[ -6, 
          Power[ 
            a[ \[Tau]], -2], 
          Plus[ \[Kappa], 
            Power[ 
              Derivative[ 1][ a][ \[Tau]], 2], 
            Times[ 
              a[ \[Tau]], 
              Derivative[ 2][ a][ \[Tau]]]]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Einstein Equation", "Subsubsection"],

Cell[TextData[{
  "Einstein tensor: ",
  Cell[BoxData[
      \(TraditionalForm\`G\_\(a\[InvisibleSpace]b\)\  = \ 
        R\_\(a\[InvisibleSpace]b\)\  - \ \(1\/2\) \
\(g\_\(a\[InvisibleSpace]b\)\) R\)]]
}], "Text"],

Cell[BoxData[
    \(einstein[la_, lb_]\  := \ 
      RicciR[la, lb]\  - \ \(1\/2\) Metricg[la, lb]\ ScalarR[] // 
        Csimplify\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(einstein[\(-1\), \(-1\)]\)], "Input"],

Cell[BoxData[
    RowBox[{"-", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            RowBox[{"\[Kappa]", "+", 
              SuperscriptBox[
                RowBox[{
                  SuperscriptBox["a", "\[Prime]",
                    MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
            ")"}]}], \(a[\[Tau]]\^2\)]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(einstein[\(-2\), \(-2\)]\)], "Input"],

Cell[BoxData[
    FractionBox[
      RowBox[{\(-\[Kappa]\), "-", 
        SuperscriptBox[
          RowBox[{
            SuperscriptBox["a", "\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", "]"}], "2"], "-", 
        RowBox[{"2", " ", \(a[\[Tau]]\), " ", 
          RowBox[{
            SuperscriptBox["a", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", "]"}]}]}], \(\(-1\) + 
        r\^2\ \[Kappa]\)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(einstein[\(-3\), \(-3\)]\  - \ \(r\^2\) \((1\  - \ \[Kappa]\ r\^2)\) 
          einstein[\(-2\), \(-2\)]\  // Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(einstein[\(-4\), \(-4\)]\  - \ \(Sin[\[Theta]]\^2\) 
          einstein[\(-3\), \(-3\)] // Simplify\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Energy-momentum tensor: ",
  Cell[BoxData[
      \(TraditionalForm\`T\_\(a\[InvisibleSpace]b\)\  = \ \((\[Rho]\  + \ 
                p)\) \(\[Tau]\_a\) \[Tau]\_b\  + \ 
          p\ g\_\(a\[InvisibleSpace]b\)\)]]
}], "Text"],

Cell[BoxData[
    \(Tdefine[EnergyMomentumT, \ "\<T\>", "\<ba\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(energyMomentumMatrix\  = \ {{\[Rho], 0, 0, 
          0}, {0, \(a[\[Tau]]\^2\/\(1\  - \ \[Kappa]\ r\^2\)\) p, 0, 0}, {0, 
          0, \(r\^2\) \(a[\[Tau]]\^2\) p, 0}, {0, 0, 
          0, \(r\^2\) \(Sin[\[Theta]]\^2\) \(a[\[Tau]]\^2\) p}}; \ 
    MatrixForm[energyMomentumMatrix]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"\[Rho]", "0", "0", "0"},
            {"0", \(\(p\ a[\[Tau]]\^2\)\/\(1 - r\^2\ \[Kappa]\)\), "0", "0"},
            {"0", "0", \(p\ r\^2\ a[\[Tau]]\^2\), "0"},
            {"0", "0", "0", \(p\ r\^2\ a[\[Tau]]\^2\ Sin[\[Theta]]\^2\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Cset[EnergyMomentumT[la, lb], energyMomentumMatrix]\)], "Input"],

Cell[TextData[{
  "Equation of motion: ",
  Cell[BoxData[
      \(TraditionalForm\`G\_\(a\[InvisibleSpace]b\)\  - \ \[CapitalLambda]\ g\
\_\(a\[InvisibleSpace]b\)\  = \ 8  \[Pi]\ G\ T\_\(a\[InvisibleSpace]b\)\)]]
}], "Text"],

Cell[BoxData[
    \(eq[la_, lb_]\  := \ 
      einstein[la, lb]\  - \ \[CapitalLambda]\ Metricg[la, lb]\  + \ 
          8  \[Pi]\ G\ EnergyMomentumT[la, lb]\  // CsimplifyMore\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq1\  = \ eq[\(-1\), \(-1\)]\)], "Input"],

Cell[BoxData[
    RowBox[{"\[CapitalLambda]", "+", \(8\ G\ \[Pi]\ \[Rho]\), "-", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            RowBox[{"\[Kappa]", "+", 
              SuperscriptBox[
                RowBox[{
                  SuperscriptBox["a", "\[Prime]",
                    MultilineFunction->None], "[", "\[Tau]", "]"}], "2"]}], 
            ")"}]}], \(a[\[Tau]]\^2\)]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq2\  = \ 
      eq[\(-2\), \(-2\)]\ \(3\/2\) \((1\  - \ \[Kappa]\ r\^2)\)\/a[\[Tau]]\^2\
\  + \ \(1\/2\) eq[\(-1\), \(-1\)]\  // \ Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{\(-\[CapitalLambda]\), "+", \(4\ G\ \[Pi]\ \((3\ p + \[Rho])\)\), 
      "+", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{
            SuperscriptBox["a", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "\[Tau]", 
            "]"}]}], \(a[\[Tau]]\)]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Einstein's Static Universe", "Subsubsection"],

Cell[BoxData[
    \(\(staticRule\  = \ {a[\[Tau]]\  -> \ a, \ \(a'\)[\[Tau]]\  -> \ 
            0, \ \(\(a'\)'\)[\[Tau]]\  -> \ 0};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({eq1, eq2}\  /. \ staticRule\)], "Input"],

Cell[BoxData[
    \({\(-\(\(3\ \[Kappa]\)\/a\^2\)\) + \[CapitalLambda] + 
        8\ G\ \[Pi]\ \[Rho], \(-\[CapitalLambda]\) + 
        4\ G\ \[Pi]\ \((3\ p + \[Rho])\)}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "From these we find ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalLambda]\  > \ 0\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Kappa]\  = \ 1\)]]
}], "Text"],

Cell[TextData[{
  "For a dust case (",
  Cell[BoxData[
      \(TraditionalForm\`p\  = \ 0\)]],
  "), ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalLambda]\  = \ 4  \[Pi]\ G\ \[Rho]\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`a\  = \ \@\(1\/\(4  \[Pi]\ G\ \[Rho]\)\)\)]]
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["de Sitter Spacetime", "Subsubsection"],

Cell[TextData[{
  "In this case ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalLambda]\  > \ 0\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`\[Rho]\  = \ \(p\  = \ 0\)\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`\[Kappa]\  = \ 0\)]]
}], "Text"],

Cell[BoxData[
    \(\(deSitterRule\  = \ {\[Rho]\  -> \ 0, \ 
          p\  -> \ 0, \ \[Kappa]\  -> \ 0};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({eq1, \ eq2}\  /. \ deSitterRule\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"\[CapitalLambda]", "-", 
          FractionBox[
            RowBox[{"3", " ", 
              SuperscriptBox[
                RowBox[{
                  SuperscriptBox["a", "\[Prime]",
                    MultilineFunction->None], "[", "\[Tau]", "]"}], 
                "2"]}], \(a[\[Tau]]\^2\)]}], ",", 
        RowBox[{\(-\[CapitalLambda]\), "+", 
          FractionBox[
            RowBox[{"3", " ", 
              RowBox[{
                SuperscriptBox["a", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "\[Tau]", 
                "]"}]}], \(a[\[Tau]]\)]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, ",
  Cell[BoxData[
      \(TraditionalForm\`a[\[Tau]]\  = \ 
        C\ e\^\(\(\@\(\[CapitalLambda]\/3\)\) t\)\)]]
}], "Text"]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{605, 660},
WindowMargins->{{14, Automatic}, {Automatic, 6}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 43, "Title"],
Cell[1779, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1856, 57, 53, 0, 54, "Subsubsection"],
Cell[1912, 59, 79, 1, 56, "Input"],
Cell[1994, 62, 248, 4, 103, "Input"],

Cell[CellGroupData[{
Cell[2267, 70, 59, 1, 56, "Input"],
Cell[2329, 73, 3194, 80, 99, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5572, 159, 60, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[5657, 163, 91, 1, 79, "Input"],
Cell[5751, 166, 56, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5856, 173, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[5918, 177, 46, 1, 56, "Input"],
Cell[5967, 180, 392, 8, 31, "Print"],
Cell[6362, 190, 2980, 73, 47, "Print"],
Cell[9345, 265, 3064, 74, 47, "Print"],
Cell[12412, 341, 3046, 74, 47, "Print"],
Cell[15461, 417, 3136, 77, 44, "Print"],
Cell[18600, 496, 2879, 69, 41, "Print"],
Cell[21482, 567, 2808, 66, 44, "Print"],
Cell[24293, 635, 2790, 66, 44, "Print"],
Cell[27086, 703, 3072, 73, 33, "Print"],
Cell[30161, 778, 2907, 69, 33, "Print"],
Cell[33071, 849, 2821, 66, 39, "Print"],
Cell[35895, 917, 3139, 76, 35, "Print"],
Cell[39037, 995, 2969, 72, 35, "Print"],
Cell[42009, 1069, 2887, 68, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44933, 1142, 47, 1, 56, "Input"],
Cell[44983, 1145, 293, 6, 29, "Print"],
Cell[45279, 1153, 4380, 106, 44, "Print"],
Cell[49662, 1261, 4311, 101, 33, "Print"],
Cell[53976, 1364, 4380, 104, 35, "Print"],
Cell[58359, 1470, 4621, 112, 47, "Print"],
Cell[62983, 1584, 4703, 116, 47, "Print"],
Cell[67689, 1702, 4622, 111, 35, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[72348, 1818, 45, 1, 56, "Input"],
Cell[72396, 1821, 184, 4, 29, "Print"],
Cell[72583, 1827, 2050, 53, 47, "Print"],
Cell[74636, 1882, 2554, 67, 47, "Print"],
Cell[77193, 1951, 2587, 67, 33, "Print"],
Cell[79783, 2020, 2647, 69, 35, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[82467, 2094, 46, 1, 56, "Input"],
Cell[82516, 2097, 1238, 33, 50, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[83803, 2136, 42, 0, 32, "Subsubsection"],
Cell[83848, 2138, 214, 6, 44, "Text"],
Cell[84065, 2146, 144, 3, 119, "Input"],

Cell[CellGroupData[{
Cell[84234, 2153, 57, 1, 56, "Input"],
Cell[84294, 2156, 374, 10, 75, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[84705, 2171, 57, 1, 56, "Input"],
Cell[84765, 2174, 460, 11, 72, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[85262, 2190, 151, 2, 80, "Input"],
Cell[85416, 2194, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[85488, 2200, 132, 2, 81, "Input"],
Cell[85623, 2204, 35, 1, 53, "Output"]
}, Open  ]],
Cell[85673, 2208, 244, 6, 41, "Text"],
Cell[85920, 2216, 78, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[86023, 2221, 313, 5, 174, "Input"],
Cell[86339, 2228, 388, 8, 151, "Output"]
}, Open  ]],
Cell[86742, 2239, 84, 1, 79, "Input"],
Cell[86829, 2242, 224, 5, 41, "Text"],
Cell[87056, 2249, 189, 3, 102, "Input"],

Cell[CellGroupData[{
Cell[87270, 2256, 61, 1, 56, "Input"],
Cell[87334, 2259, 424, 10, 75, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[87795, 2274, 173, 3, 102, "Input"],
Cell[87971, 2279, 327, 8, 72, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[88347, 2293, 51, 0, 32, "Subsubsection"],
Cell[88401, 2295, 147, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[88573, 2301, 61, 1, 56, "Input"],
Cell[88637, 2304, 183, 3, 69, "Output"]
}, Open  ]],
Cell[88835, 2310, 196, 7, 41, "Text"],
Cell[89034, 2319, 301, 10, 65, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[89372, 2334, 44, 0, 32, "Subsubsection"],
Cell[89419, 2336, 268, 10, 41, "Text"],
Cell[89690, 2348, 120, 2, 56, "Input"],

Cell[CellGroupData[{
Cell[89835, 2354, 65, 1, 56, "Input"],
Cell[89903, 2357, 668, 17, 75, "Output"]
}, Open  ]],
Cell[90586, 2377, 156, 5, 61, "Text"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

