(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     98493,       2587]*)
(*NotebookOutlinePosition[     99261,       2614]*)
(*  CellTagsIndexPosition[     99217,       2610]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Schwarzschild SpaceTime", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Coordinate System and Metric", "Subsubsection"],

Cell[BoxData[
    \(\(coSys\  = \ {t, r, \[Theta], \[Phi]};\)\)], "Input"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-\((1\  - \ \(2  G\ M\)\/r)\)\), 0, 0, 0}, {0, 
            1\/\(1\  - \ \(2  G\ M\)\/r\), 0, 0}, {0, 0, r\^2, 0}, {0, 0, 0, 
            r\^2\ Sin[\[Theta]]\^2}};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[coSys, metric]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          FractionBox[
            RowBox[{"r", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                    InterpretationBox[\(r\^2\),
                      Superscript[ r, 2],
                      Editable->False]}],
                  SequenceForm[ "d", 
                    Superscript[ r, 2]],
                  Editable->False], ")"}]}], \(\(-2\)\ G\ M + r\)], "+", 
          RowBox[{\((\(-1\) + \(2\ G\ M\)\/r)\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(t\^2\),
                    Superscript[ t, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ t, 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(r\^2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Theta]\^2\),
                    Superscript[ \[Theta], 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ \[Theta], 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(r\^2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Phi]\^2\),
                    Superscript[ \[Phi], 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ \[Phi], 2]],
                Editable->False], ")"}], " ", \(Sin[\[Theta]]\^2\)}]}]}],
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          Times[ r, 
            Power[ 
              Plus[ 
                Times[ -2, G, M], r], -1], 
            SequenceForm[ "d", 
              Superscript[ r, 2]]], 
          Times[ 
            Plus[ -1, 
              Times[ 2, G, M, 
                Power[ r, -1]]], 
            SequenceForm[ "d", 
              Superscript[ t, 2]]], 
          Times[ 
            Power[ r, 2], 
            SequenceForm[ "d", 
              Superscript[ \[Theta], 2]]], 
          Times[ 
            Power[ r, 2], 
            SequenceForm[ "d", 
              Superscript[ \[Phi], 2]], 
            Power[ 
              Sin[ \[Theta]], 2]]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization of Component Tensors", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[InitCTensor[coSys, \ metric, SimplifyMore -> True]]\)], "Input"],

Cell[BoxData[
    \({2.564`\ Second, Null}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Show Results", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Metricg]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"g"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "g", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-1\) + \(2\ G\ M\)\/r\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "t"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Plus[ -1, 
          Times[ 2, G, M, 
            Power[ r, -1]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(1\/\(1 - \(2\ G\ M\)\/r\)\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Power[ 
          Plus[ 1, 
            Times[ -2, G, M, 
              Power[ r, -1]]], -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]"\+_\
\),
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\^2\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Power[ r, 2]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\+_\)\
,
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(r\^2\ Sin[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]], " = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[Structuref]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("f"\_\("[ab]"\)\%\("    c"\)\),
          Subsuperscript[ "f", "[ab]", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "f", "[ab]", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\("f"\_\("ab"\)\%\("  c"\)\),
          Subsuperscript[ "f", "ab", "  c"],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "\<\"0\"\>"}],
      SequenceForm[ 
        Subsuperscript[ "f", "ab", "  c"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[AffineG]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Symmetry: \"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("\[CapitalGamma]"\_\("(ab)"\)\%\("    c"\)\),
          Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"],
          Editable->False]}],
      SequenceForm[ "Symmetry: ", 
        Subsuperscript[ "\[CapitalGamma]", "(ab)", "    c"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\ \((\(-2\)\ G\ M + r)\)\)\/r\^3\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ G, M, 
          Power[ r, -3], 
          Plus[ 
            Times[ -2, G, M], r]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "t"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "t"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\)\/\(\(-2\)\ G\ M\ r + r\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "t"]]], " = ", 
        Times[ G, M, 
          Power[ 
            Plus[ 
              Times[ -2, G, M, r], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\)\/\(2\ G\ M\ r - r\^2\)\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ G, M, 
          Power[ 
            Plus[ 
              Times[ 2, G, M, r], 
              Times[ -1, 
                Power[ r, 2]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(1\/r\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Power[ r, -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+_\),
                      
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(1\/r\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Power[ r, -1]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(2\ G\ M - r\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Plus[ 
          Times[ 2, G, M], 
          Times[ -1, r]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Theta]\
"\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(Cot[\[Theta]]\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]]], " = ", 
        Cot[ \[Theta]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "r"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\((2\ G\ M - r)\)\ Sin[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "r"]]], " = ", 
        Times[ 
          Plus[ 
            Times[ 2, G, M], 
            Times[ -1, r]], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\+_\),
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-Cos[\[Theta]]\)\ Sin[\[Theta]]\)}],
      SequenceForm[ 
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]]], " = ", 
        Times[ -1, 
          Cos[ \[Theta]], 
          Sin[ \[Theta]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ G\ M\)\/r\^3\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 2, G, M, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\ \((2\ G\ M - r)\)\)\/r\^2\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Plus[ 
            Times[ 2, G, M], 
            Times[ -1, r]], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\ \((2\ G\ M - 
                  r)\)\ Sin[\[Theta]]\^2\)\/r\^2\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Plus[ 
            Times[ 2, G, M], 
            Times[ -1, r]], 
          Power[ r, -2], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\)\/\(\(-2\)\ G\ M + r\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Power[ 
            Plus[ 
              Times[ -2, G, M], r], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\ Sin[\[Theta]]\^2\)\/\(\(-2\)\ G\ M + 
              r\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Power[ 
            Plus[ 
              Times[ -2, G, M], r], -1], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-2\)\ G\ M\ r\ Sin[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, G, M, r, 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RicciR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"(ab)"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "(ab)"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("R"\_"ab"\[InvisibleSpace]" = "\[InvisibleSpace]"0"\),
      
      SequenceForm[ 
        Subscript[ "R", "ab"], " = ", "0"],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[ScalarR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", "0"}],
      SequenceForm[ 
        Subsuperscript[ R, "", ""], " = ", 0],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Geodesic Equations", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Geodesic[1, Simplify]\)], "Input"],

Cell[BoxData[
    \(\(2\ G\ M\ \("r"\& . \)\ \("t"\& . \)\)\/\(\(-2\)\ G\ M\ r + r\^2\) + 
      "t"\&".."\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Geodesic[2, Simplify]\)], "Input"],

Cell[BoxData[
    \(\(G\ M\ \("r"\& . \)\^2\)\/\(2\ G\ M\ r - r\^2\) + \(G\ M\ \((\(-2\)\ G\
\ M + r)\)\ \("t"\& . \)\^2\)\/r\^3 + \((2\ G\ M - 
            r)\)\ \("\[Theta]"\& . \)\^2 + 
      "r"\&".." + \((2\ G\ M - 
            r)\)\ \("\[Phi]"\& . \)\^2\ Sin[\[Theta]]\^2\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Geodesic[3, Simplify]\)], "Input"],

Cell[BoxData[
    \(\(2\ \("r"\& . \)\ \("\[Theta]"\& . \)\)\/r + "\[Theta]"\&".." - 
      Cos[\[Theta]]\ \("\[Phi]"\& . \)\^2\ Sin[\[Theta]]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Geodesic[4, Simplify]\)], "Input"],

Cell[BoxData[
    \(\(2\ \("r"\& . \)\ \("\[Phi]"\& . \)\)\/r + 
      2\ Cot[\[Theta]]\ \("\[Theta]"\& . \)\ \("\[Phi]"\& . \) + 
      "\[Phi]"\&".."\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Calculation of Weyl Tensor", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[Tcalc[WeylC, Simplify]]\)], "Input"],

Cell[BoxData[
    \({0.5310000000000006`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[WeylC]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"C"\_"[ab][cd]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "C", "[ab][cd]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "r"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(2\ G\ M\)\/r\^3\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "r"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 2, G, M, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\ \((2\ G\ M - r)\)\)\/r\^2\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Plus[ 
            Times[ 2, G, M], 
            Times[ -1, r]], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"t"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "t"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "t"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("t"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "t"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "t"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "t"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "t"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "t"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\ \((2\ G\ M - 
                  r)\)\ Sin[\[Theta]]\^2\)\/r\^2\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "t"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "t"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Plus[ 
            Times[ 2, G, M], 
            Times[ -1, r]], 
          Power[ r, -2], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\)\/\(\(-2\)\ G\ M + r\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Power[ 
            Plus[ 
              Times[ -2, G, M], r], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "r"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "r"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], "\[InvisibleSpace]", \("r"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "r"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "r"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "r"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "r"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "r"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(G\ M\ Sin[\[Theta]]\^2\)\/\(\(-2\)\ G\ M + 
              r\)\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "r"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "r"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ G, M, 
          Power[ 
            Plus[ 
              Times[ -2, G, M], r], -1], 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["C", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Theta]"\+_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                          Editable->False], 
                        "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                        dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
                      Editable->False], 
                    "\[InvisibleSpace]", \("\[Theta]"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                      dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]],
                  Editable->False], "\[InvisibleSpace]", \("\[Phi]"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                    dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ C, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                  dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(\(-2\)\ G\ M\ r\ Sin[\[Theta]]\^2\)}],
      SequenceForm[ 
        Subsuperscript[ C, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
                dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
              dhPark`Tensor`Private`UnderBar[ "\[Theta]"]], 
            dhPark`Tensor`Private`UnderBar[ "\[Phi]"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ -2, G, M, r, 
          Power[ 
            Sin[ \[Theta]], 2]]],
      Editable->False]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Laplace Equation", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Laplacian[f[r], Simplify]\)], "Input"],

Cell[BoxData[
    FractionBox[
      RowBox[{
        RowBox[{\((\(-2\)\ G\ M + 2\ r)\), " ", 
          RowBox[{
            SuperscriptBox["f", "\[Prime]",
              MultilineFunction->None], "[", "r", "]"}]}], "+", 
        RowBox[{"r", " ", \((\(-2\)\ G\ M + r)\), " ", 
          RowBox[{
            SuperscriptBox["f", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "r", 
            "]"}]}]}], \(r\^2\)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature Invariant", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[\(riemannUp\  = \ Get["\<dhPark`RiemannUp4`\>"];\)]\)], "Input"],

Cell[BoxData[
    \({3.2349999999999994`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Cset[RiemannR[ua, ub, uc, ud], riemannUp]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[
      SumDum[RiemannR[la, lb, lc, ld]\ RiemannR[ua, ub, uc, ud]]\  // \ 
        Simplify]\)], "Input"],

Cell[BoxData[
    \({0.7910000000000004`\ Second, \(48\ G\^2\ M\^2\)\/r\^6}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Kruskal Coordinates", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-\((1\  - \ \(2  \
M\)\/r)\)\) 
          d\[InvisibleSpace]t\^2\  + \ \(d\[InvisibleSpace]r\^2\)\/\(1\  - \ \
\(2  M\)\/r\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(\(Dimension = 2;\)\)], "Input"],

Cell[BoxData[
    \(\(coSys = {t, r};\)\)], "Input"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`{u, 
        v}\  = \ {\(\@\(r\/\(2  M\) - 1\)\) \(e\^\(r\/\(4  M\)\)\) \(sinh(
            t\/\(4  M\))\), \ \(\@\(r\/\(2  m\) - 
                1\)\) \(e\^\(r\/\(4  M\)\)\) \(cosh(
            t\/\(4  M\))\)}\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(\(newCoSys = {\(\@\(r\/\(2  M\)\  - \ 1\)\) \(E\^\(r\/\(4  M\)\)\) 
            Sinh[t\/\(4  M\)], \(\@\(r\/\(2  M\)\  - \ 
                  1\)\) \(E\^\(r\/\(4  M\)\)\) 
            Cosh[t\/\(4  M\)]};\)\)], "Input"],

Cell[BoxData[
    \(Tdefine[newg, "\<ba\>"]\)], "Input"],

Cell[BoxData[
    \(Cset[
      Metricg[la, 
        lb], {{\(-\((1\  - \ \(2  M\)\/r)\)\), 0}, {0, 
          1\/\(1\  - \ \(2  M\)\/r\)}}]\)], "Input"],

Cell[BoxData[
    \(Ttransform[Metricg[la, lb], newg, coSys, newCoSys, Simplify]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[newg[\(-i\), \(-j\)], {i, 2}, {j, 2}]\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \({{\(-\(\(32\ \[ExponentialE]\^\(-\(r\/\(2\ M\)\)\)\ M\^3\)\/r\)\), 
        0}, {0, \(32\ \[ExponentialE]\^\(-\(r\/\(2\ M\)\)\)\ M\^3\)\/r}}\)], \
"Output"]
}, Open  ]],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(\(32 
               M\^3\)\/r\) \(\(e\^\(-\(r\/\(2  
                      M\)\)\)\)(\(-d\)\[InvisibleSpace]t'\^2\  + \ 
            d\[InvisibleSpace]r'\^2)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Double Null Coordinates", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-\((1\  - \ \(2  \
M\)\/r)\)\) 
          d\[InvisibleSpace]t\^2\  + \ \(d\[InvisibleSpace]r\^2\)\/\(1\  - \ \
\(2  M\)\/r\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(\(Dimension = 2;\)\)], "Input"],

Cell[BoxData[
    \(\(coSys = {t, r};\)\)], "Input"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`{u, 
        v}\  = \ {t\  + \ r\  + \ 2  M\ \(ln(r\  - \ 2  M)\), \ 
        t\  - \ r\  - \ 2  M\ \(ln(r\  - \ 2  M)\)}\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(\(newCoSys = {t\  + \ r\  + \ 2  M\ Log[r\  - \ 2  M], \ 
          t\  - \ r\  - \ 2  M\ Log[r\  - \ 2  M]};\)\)], "Input"],

Cell[BoxData[
    \(Tdefine[newg, "\<ba\>"]\)], "Input"],

Cell[BoxData[
    \(Cset[
      Metricg[la, 
        lb], {{\(-\((1\  - \ \(2  M\)\/r)\)\), 0}, {0, 
          1\/\(1\  - \ \(2  M\)\/r\)}}]\)], "Input"],

Cell[BoxData[
    \(Ttransform[Metricg[la, lb], newg, coSys, newCoSys]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[newg[\(-i\), \(-j\)], {i, 2}, {j, 2}]\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \({{0, \(-\(1\/2\)\) + M\/r}, {\(-\(1\/2\)\) + M\/r, 0}}\)], "Output"]
}, Open  ]],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-\((1\  - \ \(2  \
M\)\/r)\)\) d\[InvisibleSpace]u\ d\[InvisibleSpace]v\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Eddington-Finkelstein Coordinates", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-\((1\  - \ \(2  \
M\)\/r)\)\) 
          d\[InvisibleSpace]t\^2\  + \ \(d\[InvisibleSpace]r\^2\)\/\(1\  - \ \
\(2  M\)\/r\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(\(Dimension = 2;\)\)], "Input"],

Cell[BoxData[
    \(\(coSys = {t, r};\)\)], "Input"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`{t', r'}\  = \ {t\  + \ 2  M\ \(ln(r\  - \ 2  M)\), \ 
        r}\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(\(newCoSys = {t\  + \ 2  M\ Log[r\  - \ 2  M], \ r};\)\)], "Input"],

Cell[BoxData[
    \(Tdefine[newg, "\<ba\>"]\)], "Input"],

Cell[BoxData[
    \(Cset[
      Metricg[la, 
        lb], {{\(-\((1\  - \ \(2  M\)\/r)\)\), 0}, {0, 
          1\/\(1\  - \ \(2  M\)\/r\)}}]\)], "Input"],

Cell[BoxData[
    \(Ttransform[Metricg[la, lb], newg, coSys, newCoSys]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[newg[\(-i\), \(-j\)], {i, 2}, {j, 2}]\  // \ Simplify\)], "Input"],

Cell[BoxData[
    \({{\(-1\) + \(2\ M\)\/r, \(2\ M\)\/r}, {\(2\ M\)\/r, 
        1 + \(2\ M\)\/r}}\)], "Output"]
}, Open  ]],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-\((1\  - \ \(2  \
M\)\/r)\)\) d\[InvisibleSpace]t'\^2\  + \ \(\(4  M\)\/r\) 
          d\[InvisibleSpace]t'\ d\[InvisibleSpace]r\  + \ \((1\  + \ \(2  M\)\
\/r)\)\ d\[InvisibleSpace]r\^2\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{567, 660},
WindowMargins->{{14, Automatic}, {Automatic, 6}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 56, 1, 43, "Title"],
Cell[1776, 52, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1853, 57, 53, 0, 54, "Subsubsection"],
Cell[1909, 59, 74, 1, 56, "Input"],
Cell[1986, 62, 217, 3, 154, "Input"],

Cell[CellGroupData[{
Cell[2228, 69, 59, 1, 56, "Input"],
Cell[2290, 72, 2778, 74, 96, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5117, 152, 60, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[5202, 156, 91, 1, 79, "Input"],
Cell[5296, 159, 56, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5401, 166, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[5463, 170, 46, 1, 56, "Input"],
Cell[5512, 173, 184, 4, 30, "Print"],
Cell[5699, 179, 1760, 44, 44, "Print"],
Cell[7462, 225, 1789, 45, 55, "Print"],
Cell[9254, 272, 1770, 44, 33, "Print"],
Cell[11027, 318, 1840, 47, 35, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[12904, 370, 49, 1, 56, "Input"],
Cell[12956, 373, 350, 8, 31, "Print"],
Cell[13309, 383, 372, 9, 29, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13718, 397, 46, 1, 56, "Input"],
Cell[13767, 400, 392, 8, 31, "Print"],
Cell[14162, 410, 2835, 68, 44, "Print"],
Cell[17000, 480, 2863, 69, 43, "Print"],
Cell[19866, 551, 2886, 70, 43, "Print"],
Cell[22755, 623, 2808, 66, 44, "Print"],
Cell[25566, 691, 2790, 66, 44, "Print"],
Cell[28359, 759, 2854, 68, 33, "Print"],
Cell[31216, 829, 2821, 66, 39, "Print"],
Cell[34040, 897, 2930, 71, 35, "Print"],
Cell[36973, 970, 2887, 68, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[39897, 1043, 47, 1, 56, "Input"],
Cell[39947, 1046, 293, 6, 29, "Print"],
Cell[40243, 1054, 3940, 94, 44, "Print"],
Cell[44186, 1150, 4101, 97, 44, "Print"],
Cell[48290, 1249, 4169, 100, 47, "Print"],
Cell[52462, 1351, 4075, 96, 43, "Print"],
Cell[56540, 1449, 4139, 99, 47, "Print"],
Cell[60682, 1550, 4142, 96, 35, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[64861, 1651, 45, 1, 56, "Input"],
Cell[64909, 1654, 184, 4, 29, "Print"],
Cell[65096, 1660, 198, 5, 28, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[65331, 1670, 46, 1, 56, "Input"],
Cell[65380, 1673, 361, 9, 29, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[65790, 1688, 43, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[65858, 1692, 54, 1, 56, "Input"],
Cell[65915, 1695, 120, 2, 72, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[66072, 1702, 54, 1, 56, "Input"],
Cell[66129, 1705, 291, 5, 100, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[66457, 1715, 54, 1, 56, "Input"],
Cell[66514, 1718, 156, 2, 72, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[66707, 1725, 54, 1, 56, "Input"],
Cell[66764, 1728, 165, 3, 73, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[66978, 1737, 51, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[67054, 1741, 63, 1, 56, "Input"],
Cell[67120, 1744, 69, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[67226, 1750, 44, 1, 56, "Input"],
Cell[67273, 1753, 192, 4, 29, "Print"],
Cell[67468, 1759, 3940, 94, 44, "Print"],
Cell[71411, 1855, 4101, 97, 44, "Print"],
Cell[75515, 1954, 4169, 100, 47, "Print"],
Cell[79687, 2056, 4075, 96, 43, "Print"],
Cell[83765, 2154, 4139, 99, 47, "Print"],
Cell[87907, 2255, 4142, 96, 35, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[92098, 2357, 41, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[92164, 2361, 58, 1, 56, "Input"],
Cell[92225, 2364, 445, 11, 69, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[92719, 2381, 44, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[92788, 2385, 91, 1, 79, "Input"],
Cell[92882, 2388, 69, 1, 53, "Output"]
}, Open  ]],
Cell[92966, 2392, 74, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[93065, 2397, 131, 3, 102, "Input"],
Cell[93199, 2402, 89, 1, 72, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[93337, 2409, 44, 0, 32, "Subsubsection"],
Cell[93384, 2411, 269, 6, 56, "Text"],
Cell[93656, 2419, 51, 1, 56, "Input"],
Cell[93710, 2422, 52, 1, 56, "Input"],
Cell[93765, 2425, 325, 7, 47, "Text"],
Cell[94093, 2434, 238, 4, 145, "Input"],
Cell[94334, 2440, 56, 1, 56, "Input"],
Cell[94393, 2443, 153, 4, 111, "Input"],
Cell[94549, 2449, 93, 1, 79, "Input"],

Cell[CellGroupData[{
Cell[94667, 2454, 92, 1, 79, "Input"],
Cell[94762, 2457, 176, 3, 78, "Output"]
}, Open  ]],
Cell[94953, 2463, 305, 6, 47, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[95295, 2474, 48, 0, 32, "Subsubsection"],
Cell[95346, 2476, 269, 6, 56, "Text"],
Cell[95618, 2484, 51, 1, 56, "Input"],
Cell[95672, 2487, 52, 1, 56, "Input"],
Cell[95727, 2490, 235, 5, 35, "Text"],
Cell[95965, 2497, 144, 2, 79, "Input"],
Cell[96112, 2501, 56, 1, 56, "Input"],
Cell[96171, 2504, 153, 4, 111, "Input"],
Cell[96327, 2510, 83, 1, 79, "Input"],

Cell[CellGroupData[{
Cell[96435, 2515, 92, 1, 79, "Input"],
Cell[96530, 2518, 88, 1, 69, "Output"]
}, Open  ]],
Cell[96633, 2522, 216, 4, 42, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[96886, 2531, 58, 0, 32, "Subsubsection"],
Cell[96947, 2533, 269, 6, 56, "Text"],
Cell[97219, 2541, 51, 1, 56, "Input"],
Cell[97273, 2544, 52, 1, 56, "Input"],
Cell[97328, 2547, 179, 4, 35, "Text"],
Cell[97510, 2553, 87, 1, 56, "Input"],
Cell[97600, 2556, 56, 1, 56, "Input"],
Cell[97659, 2559, 153, 4, 111, "Input"],
Cell[97815, 2565, 83, 1, 79, "Input"],

Cell[CellGroupData[{
Cell[97923, 2570, 92, 1, 79, "Input"],
Cell[98018, 2573, 112, 2, 69, "Output"]
}, Open  ]],
Cell[98145, 2578, 332, 6, 42, "Text"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

