(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     34751,        882]*)
(*NotebookOutlinePosition[     35514,        909]*)
(*  CellTagsIndexPosition[     35470,        905]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Ttransform", "Subtitle",
  TextAlignment->Center,
  TextJustification->0],

Cell["See MathTensor Book, pp 158", "Text"],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Familiar Calculation", "Section"],

Cell[BoxData[
    \(\(Dimension = 3;\)\)], "Input"],

Cell[BoxData[
    \(Tdefine[{oldg, newg}, "\<ba\>"]\)], "Input"],

Cell[BoxData[
    \(Cset[newg[la, lb], {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}]\)], "Input"],

Cell[BoxData[
    \(coSys = {r, \[Theta], \[Phi]}; \ 
    newCoSys\  = \ {r\ Sin[\[Theta]]\ Cos[\[Phi]], 
        r\ Sin[\[Theta]]\ Sin[\[Phi]], r\ Cos[\[Theta]]};\)], "Input"],

Cell[BoxData[
    \(Ttransform[oldg, newg[la, lb], coSys, newCoSys, \ Simplify]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[oldg[\(-i\), \(-j\)], {i, 3}, {j, 3}]\)], "Input"],

Cell[BoxData[
    \({{1, 0, 0}, {0, r\^2, 0}, {0, 0, r\^2\ Sin[\[Theta]]\^2}}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Simple Shear", "Section"],

Cell[BoxData[
    \(coSys = {x, y, z}; \ newCoSys\  = \ {x\  + \ K\ y, y, z};\)], "Input"],

Cell[BoxData[
    \(\(Cset[newg[la, lb], {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}];\)\)], "Input"],

Cell[BoxData[
    \(Ttransform[oldg, newg[la, lb], coSys, newCoSys]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[oldg[\(-i\), \(-j\)], {i, 3}, {j, 3}]\)], "Input"],

Cell[BoxData[
    \({{1, K, 0}, {K, 1 + K\^2, 0}, {0, 0, 1}}\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(Cset[newg[ua, ub], {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}];\)\)], "Input"],

Cell[BoxData[
    \(Ttransform[oldg, newg[ua, ub], coSys, newCoSys]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[oldg[i, j], {i, 3}, {j, 3}]\)], "Input"],

Cell[BoxData[
    \({{1 + K\^2, \(-K\), 0}, {\(-K\), 1, 0}, {0, 0, 1}}\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Special Relativity", "Section"],

Cell[BoxData[
    \(\(Dimension = 4;\)\)], "Input"],

Cell[BoxData[
    \(\(\[Gamma]\  = \ \((1\  - \ v\^2\/c\^2)\)\^\(\(-1\)/2\);\)\)], "Input"],

Cell[BoxData[
    \(coSys = {x, y, z, t}; \ 
    newCoSys\  = \ {\[Gamma] \((x\  - \ v\ t)\), y, 
        z, \[Gamma] \((t\  - \ v\/c\^2\ x)\)};\)], "Input"],

Cell[BoxData[
    \(Tdefine[newF, "\<-ba\>"]\)], "Input"],

Cell[BoxData[
    \(Ttransform[MaxwellF[ua, ub], newF, coSys, newCoSys]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[newF[i, j], {i, 1, 4}, {j, 1, 4}]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{"0", ",", 
            RowBox[{
              FractionBox[
                InterpretationBox[
                  SubsuperscriptBox["F", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                          SequenceForm[ "", 1],
                          Editable->False], "\[InvisibleSpace]", "2"}],
                      SequenceForm[ 
                        SequenceForm[ "", 1], 2],
                      Editable->False]],
                  Subsuperscript[ F, 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", 1], 2]],
                  Editable->False], \(\@\(1 - v\^2\/c\^2\)\)], "+", 
              FractionBox[
                RowBox[{"v", " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\)\
,
                            SequenceForm[ "", 2],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 2], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 2], 4]],
                    Editable->False]}], \(c\^2\ \@\(1 - v\^2\/c\^2\)\)]}], 
            ",", 
            RowBox[{
              FractionBox[
                InterpretationBox[
                  SubsuperscriptBox["F", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                          SequenceForm[ "", 1],
                          Editable->False], "\[InvisibleSpace]", "3"}],
                      SequenceForm[ 
                        SequenceForm[ "", 1], 3],
                      Editable->False]],
                  Subsuperscript[ F, 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", 1], 3]],
                  Editable->False], \(\@\(1 - v\^2\/c\^2\)\)], "+", 
              FractionBox[
                RowBox[{"v", " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]3\)\
,
                            SequenceForm[ "", 3],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 3], 4]],
                    Editable->False]}], \(c\^2\ \@\(1 - v\^2\/c\^2\)\)]}], 
            ",", 
            RowBox[{
              FractionBox[
                RowBox[{\(c\^2\), " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 4]],
                    Editable->False]}], \(c\^2 - v\^2\)], "-", 
              FractionBox[
                RowBox[{\(v\^2\), " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 4]],
                    Editable->False]}], \(c\^2 - v\^2\)]}]}], "}"}], ",", 
        RowBox[{"{", 
          RowBox[{
            RowBox[{
              RowBox[{"-", 
                FractionBox[
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "2"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 2],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 2]],
                    Editable->False], \(\@\(1 - v\^2\/c\^2\)\)]}], "-", 
              FractionBox[
                RowBox[{"v", " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\)\
,
                            SequenceForm[ "", 2],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 2], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 2], 4]],
                    Editable->False]}], \(c\^2\ \@\(1 - v\^2\/c\^2\)\)]}], 
            ",", "0", ",", 
            InterpretationBox[
              SubsuperscriptBox["F", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "3"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], 3],
                  Editable->False]],
              Subsuperscript[ F, 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", 2], 3]],
              Editable->False], ",", 
            RowBox[{
              FractionBox[
                RowBox[{"v", " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "2"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 2],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 2]],
                    Editable->False]}], \(\@\(1 - v\^2\/c\^2\)\)], "+", 
              FractionBox[
                InterpretationBox[
                  SubsuperscriptBox["F", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                          SequenceForm[ "", 2],
                          Editable->False], "\[InvisibleSpace]", "4"}],
                      SequenceForm[ 
                        SequenceForm[ "", 2], 4],
                      Editable->False]],
                  Subsuperscript[ F, 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", 2], 4]],
                  Editable->False], \(\@\(1 - v\^2\/c\^2\)\)]}]}], "}"}], ",",
         
        RowBox[{"{", 
          RowBox[{
            RowBox[{
              RowBox[{"-", 
                FractionBox[
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "3"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 3],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 3]],
                    Editable->False], \(\@\(1 - v\^2\/c\^2\)\)]}], "-", 
              FractionBox[
                RowBox[{"v", " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]3\)\
,
                            SequenceForm[ "", 3],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 3], 4]],
                    Editable->False]}], \(c\^2\ \@\(1 - v\^2\/c\^2\)\)]}], 
            ",", 
            RowBox[{"-", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                        SequenceForm[ "", 2],
                        Editable->False], "\[InvisibleSpace]", "3"}],
                    SequenceForm[ 
                      SequenceForm[ "", 2], 3],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", 2], 3]],
                Editable->False]}], ",", "0", ",", 
            RowBox[{
              FractionBox[
                RowBox[{"v", " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "3"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 3],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 3]],
                    Editable->False]}], \(\@\(1 - v\^2\/c\^2\)\)], "+", 
              FractionBox[
                InterpretationBox[
                  SubsuperscriptBox["F", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]3\),
                          SequenceForm[ "", 3],
                          Editable->False], "\[InvisibleSpace]", "4"}],
                      SequenceForm[ 
                        SequenceForm[ "", 3], 4],
                      Editable->False]],
                  Subsuperscript[ F, 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", 3], 4]],
                  Editable->False], \(\@\(1 - v\^2\/c\^2\)\)]}]}], "}"}], ",",
         
        RowBox[{"{", 
          RowBox[{
            RowBox[{
              FractionBox[
                RowBox[{\(c\^2\), " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 4]],
                    Editable->False]}], \(\(-c\^2\) + v\^2\)], "-", 
              FractionBox[
                RowBox[{\(v\^2\), " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False], "\[InvisibleSpace]", "4"}],
                        SequenceForm[ 
                          SequenceForm[ "", 1], 4],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 1], 4]],
                    Editable->False]}], \(\(-c\^2\) + v\^2\)]}], ",", 
            RowBox[{
              RowBox[{"-", 
                FractionBox[
                  RowBox[{"v", " ", 
                    InterpretationBox[
                      SubsuperscriptBox["F", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
1\),
                              SequenceForm[ "", 1],
                              Editable->False], "\[InvisibleSpace]", "2"}],
                          SequenceForm[ 
                            SequenceForm[ "", 1], 2],
                          Editable->False]],
                      Subsuperscript[ F, 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], 
                        SequenceForm[ 
                          SequenceForm[ "", 1], 2]],
                      Editable->False]}], \(\@\(1 - v\^2\/c\^2\)\)]}], "-", 
              FractionBox[
                InterpretationBox[
                  SubsuperscriptBox["F", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                          SequenceForm[ "", 2],
                          Editable->False], "\[InvisibleSpace]", "4"}],
                      SequenceForm[ 
                        SequenceForm[ "", 2], 4],
                      Editable->False]],
                  Subsuperscript[ F, 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", 2], 4]],
                  Editable->False], \(\@\(1 - v\^2\/c\^2\)\)]}], ",", 
            RowBox[{
              RowBox[{"-", 
                FractionBox[
                  RowBox[{"v", " ", 
                    InterpretationBox[
                      SubsuperscriptBox["F", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
1\),
                              SequenceForm[ "", 1],
                              Editable->False], "\[InvisibleSpace]", "3"}],
                          SequenceForm[ 
                            SequenceForm[ "", 1], 3],
                          Editable->False]],
                      Subsuperscript[ F, 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], 
                        SequenceForm[ 
                          SequenceForm[ "", 1], 3]],
                      Editable->False]}], \(\@\(1 - v\^2\/c\^2\)\)]}], "-", 
              FractionBox[
                InterpretationBox[
                  SubsuperscriptBox["F", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]3\),
                          SequenceForm[ "", 3],
                          Editable->False], "\[InvisibleSpace]", "4"}],
                      SequenceForm[ 
                        SequenceForm[ "", 3], 4],
                      Editable->False]],
                  Subsuperscript[ F, 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", 3], 4]],
                  Editable->False], \(\@\(1 - v\^2\/c\^2\)\)]}], ",", "0"}], 
          "}"}]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Kruskal Coordinates", "Section"],

Cell[BoxData[
    \(\(Dimension = 2;\)\)], "Input"],

Cell[BoxData[
    \(coSys = {r, t}; \ 
    newCoSys\  = \ {\(\@\(r\/\(2  M\) - 1\)\) \(E\^\(r\/\(4  M\)\)\) 
          Cosh[t\/\(4  M\)], \(\@\(r\/\(2  M\) - 1\)\) \(E\^\(r\/\(4  M\)\)\) 
          Sinh[t\/\(4  M\)]};\)], "Input"],

Cell[BoxData[
    \(Tdefine[newg, "\<ba\>"]\)], "Input"],

Cell[BoxData[
    \(Cset[
      Metricg[la, 
        lb], {{1\/\(1\  - \ \(2  M\)\/r\), 
          0}, {0, \(-\((1\  - \ \(2  M\)\/r)\)\)}}]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Timing[
      Ttransform[Metricg[la, lb], newg, coSys, newCoSys, 
        Simplify]]\)], "Input"],

Cell[BoxData[
    \({3.385`\ Second, Null}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[newg[\(-i\), \(-j\)], {i, 2}, {j, 2}]\)], "Input"],

Cell[BoxData[
    \({{\(32\ \[ExponentialE]\^\(-\(r\/\(2\ M\)\)\)\ M\^3\)\/r, 
        0}, {0, \(-\(\(32\ \[ExponentialE]\^\(-\(r\/\(2\ M\)\)\)\ M\^3\)\/r\)\
\)}}\)], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{586, 685},
WindowMargins->{{26, Automatic}, {Automatic, 8}},
Magnification->1,
StyleDefinitions -> "Demo.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 79, 2, 81, "Subtitle"],
Cell[1799, 53, 43, 0, 45, "Text"],
Cell[1845, 55, 52, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[1922, 60, 39, 0, 77, "Section"],
Cell[1964, 62, 51, 1, 54, "Input"],
Cell[2018, 65, 64, 1, 54, "Input"],
Cell[2085, 68, 86, 1, 77, "Input"],
Cell[2174, 71, 176, 3, 100, "Input"],
Cell[2353, 76, 92, 1, 77, "Input"],

Cell[CellGroupData[{
Cell[2470, 81, 76, 1, 54, "Input"],
Cell[2549, 84, 91, 1, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2689, 91, 31, 0, 49, "Section"],
Cell[2723, 93, 90, 1, 77, "Input"],
Cell[2816, 96, 91, 1, 77, "Input"],
Cell[2910, 99, 80, 1, 77, "Input"],

Cell[CellGroupData[{
Cell[3015, 104, 76, 1, 54, "Input"],
Cell[3094, 107, 74, 1, 55, "Output"]
}, Open  ]],
Cell[3183, 111, 91, 1, 77, "Input"],
Cell[3277, 114, 80, 1, 77, "Input"],

Cell[CellGroupData[{
Cell[3382, 119, 66, 1, 54, "Input"],
Cell[3451, 122, 84, 1, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[3584, 129, 37, 0, 49, "Section"],
Cell[3624, 131, 51, 1, 54, "Input"],
Cell[3678, 134, 91, 1, 75, "Input"],
Cell[3772, 137, 157, 3, 90, "Input"],
Cell[3932, 142, 57, 1, 54, "Input"],
Cell[3992, 145, 84, 1, 77, "Input"],

Cell[CellGroupData[{
Cell[4101, 150, 72, 1, 54, "Input"],
Cell[4176, 153, 29465, 680, 739, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[33690, 839, 38, 0, 49, "Section"],
Cell[33731, 841, 51, 1, 54, "Input"],
Cell[33785, 844, 230, 4, 169, "Input"],
Cell[34018, 850, 56, 1, 54, "Input"],
Cell[34077, 853, 153, 4, 109, "Input"],

Cell[CellGroupData[{
Cell[34255, 861, 117, 3, 77, "Input"],
Cell[34375, 866, 56, 1, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34468, 872, 76, 1, 54, "Input"],
Cell[34547, 875, 176, 3, 79, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

