(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    137014,       3607]*)
(*NotebookOutlinePosition[    140496,       3696]*)
(*  CellTagsIndexPosition[    140452,       3692]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Derivative Operators", "Title",
  TextAlignment->Left,
  TextJustification->0,
  FontSize->24],

Cell["Ref: General Relativity - R. M. Wald", "Text"],

Cell[BoxData[
    \(<< \ dhPark`IndexNotation`\)], "Input"],

Cell[BoxData[
    \(<< \ dhPark`IndexNotationPreDefine`\)], "Input"],

Cell[BoxData[
    \(DefineTensor[\[Delta], "\<ba\>"]\)], "Input"],

Cell[BoxData[
    \(DefineTensor[f, "\<\>"]\)], "Input"],

Cell[CellGroupData[{

Cell["Ordinary Derivative", "Section"],

Cell[BoxData[
    \(DefineOperator[OD, \ "\<\[PartialD]\>", CD]\)], "Input"],

Cell[CellGroupData[{

Cell["Linearity", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, \ A_\  + \ B_]\  := \ OD[la, \ A]\  + OD[la, \ B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[la, RicciR[lb, lc] + MaxwellF[lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["F", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ F, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "a"], 
          MaxwellF[ lb, lc]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["R", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ R, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "a"], 
          RicciR[ lb, lc]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Leibnitz rule", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, \ A_\ B_]\  := \ OD[la, A]\ B\  + \ A\ OD[la, B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[la, \ RicciR[lb, lc] MaxwellF[uc, ud]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "d"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              RicciR[ lb, lc]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "d"],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "d"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              MaxwellF[ uc, ud]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutativity with contraction", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, \[Delta][lb_, uc_]] := 
      0\  /; \ \((DnIndexQ[lb]\  && \ 
              UpIndexQ[uc])\)\  || \ \((UpIndexQ[lb]\  && \ 
              DnIndexQ[uc])\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[ua, \[Delta][lb, uc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutativity with indices", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, OD[lb_, expr_]]\  := 
      OD[lb, OD[la, expr]] /; \ \(! OrderedQ[{la, lb}]\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[lb, OD[la, RicciR[ua, ub]]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["R", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False]],
              Subsuperscript[ R, 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "b"], 
            RicciR[ ua, ub]],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\[PartialD]", "a"], 
        OD[ lb, 
          RicciR[ ua, ub]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Covariant Derivatives", "Section"],

Cell[BoxData[
    \(DefineOperator[{CD, \ 
        TildeCD}, \ {"\<\[Del]\>", \*"\"\<\!\(\[Del]\&~\)\>\""}, 
      CD]\)], "Input"],

Cell[CellGroupData[{

Cell["Linearity", "Subsubsection"],

Cell[BoxData[
    \(CD[la_, \ A_\  + \ B_]\  := \ CD[la, \ A]\  + \ CD[la, \ B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, RicciR[lb, lc] + MaxwellF[lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["F", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ F, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          MaxwellF[ lb, lc]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["R", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ R, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          RicciR[ lb, lc]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCD[la_, \ A_\  + \ B_]\  := \ 
      TildeCD[la, \ A]\  + \ TildeCD[la, \ B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[la, RicciR[lb, lc] + MaxwellF[lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["F", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ F, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\!\(\[Del]\&~\)", "a"], 
          MaxwellF[ lb, lc]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["R", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ R, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\!\(\[Del]\&~\)", "a"], 
          RicciR[ lb, lc]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Leibnitz rule", "Subsubsection"],

Cell[BoxData[
    \(CD[la_, \ A_\ B_]\  := \ CD[la, A]\ B\  + \ A\ CD[la, B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, \ RicciR[lb, lc] MaxwellF[uc, ud]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              RicciR[ lb, lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "d"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "d"],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "d"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              MaxwellF[ uc, ud]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCD[la_, \ A_\ B_]\  := \ 
      TildeCD[la, A]\ B\  + \ A\ TildeCD[la, B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[la, \ RicciR[lb, lc] MaxwellF[uc, ud]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "d"],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "d"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              MaxwellF[ uc, ud]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "d"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              RicciR[ lb, lc]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutativity with contraction", "Subsubsection"],

Cell[BoxData[
    \(CD[la_, \[Delta][lb_, uc_]] := 
      0\  /; \ \((DnIndexQ[lb]\  && \ 
              UpIndexQ[uc])\)\  || \ \((UpIndexQ[lb]\  && \ 
              DnIndexQ[uc])\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, \[Delta][lb, uc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCD[la_, \[Delta][lb_, uc_]] := 
      0\  /; \ \((DnIndexQ[lb]\  && \ 
              UpIndexQ[uc])\)\  || \ \((UpIndexQ[lb]\  && \ 
              DnIndexQ[uc])\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[ua, \[Delta][lb, uc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Consistency with directional derivative", "Subsubsection"],

Cell[BoxData[
    \(CD[la_, f_[]]\  := \ OD[la, f[]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, f[]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\[PartialD]", "a"], 
        f[ ]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCD[la_, f_[]]\  := \ OD[la, f[]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[la, f[]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\[PartialD]", "a"], 
        f[ ]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Torsionic and Torsion-free", "Subsubsection"],

Cell[TextData[{
  "Torsionic derivative operator: ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\_a\(\[Del]\_b 
            f\)\  = \ \[Del]\_b\(\[Del]\_a 
              f\)\  - \ \(t\_\(a\[InvisibleSpace]b\)\^p\) \[Del]\_p f\)]]
}], "Text"],

Cell[BoxData[
    \(CD[la_, CD[lb_, f_[]]]\  := \ 
      Module[{pair = DummyPair[NormalIndexQ]}, 
          CD[lb, CD[la, f[]]] - \ 
            Torsiont[la, lb, pair[\([2]\)]] 
              CD[pair[\([1]\)], f[]]] /; \ \(! OrderedQ[{la, 
              lb}]\)\)], "Input"],

Cell[TextData[{
  "Because ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\_a f \[Equal] \[PartialD]\_a f\)]]
}], "Text"],

Cell[BoxData[
    \(CD[la_, OD[lb_, f_[]]]\  := \ 
      Module[{pair = DummyPair[NormalIndexQ]}, 
          CD[lb, OD[la, f[]]] - \ 
            Torsiont[la, lb, pair[\([2]\)]] 
              CD[pair[\([1]\)], f[]]] /; \ \(! OrderedQ[{la, 
              lb}]\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[lb, CD[la, f[]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              f[ ]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          OD[ lb, 
            f[ ]]],
        Editable->False], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Torsion-free derivative operator: ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Del]\&~\)\_a\(\(\[Del]\&~\)\_b 
            f\)\  = \ \(\[Del]\&~\)\_b\(\(\[Del]\&~\)\_a f\)\)]]
}], "Text"],

Cell[BoxData[
    \(TildeCD[la_, OD[lb_, f_[]]]\  := \ 
      TildeCD[lb, OD[la, f[]]] /; \ \(! OrderedQ[{la, lb}]\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[lb, TildeCD[la, f[]]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "b"], 
            f[ ]],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\!\(\[Del]\&~\)", "a"], 
        OD[ lb, 
          f[ ]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Relation between Derivative Operators", "Section"],

Cell[BoxData[
    \(DefineTensor[{\[Mu], \[Nu], \[Xi], \[Eta]}, "\<a\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f[]\ \[Mu][lb]\  + \ \[Nu][lb]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "b"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      InterpretationBox[
        SubsuperscriptBox["\[Nu]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
            SequenceForm[ "", "b"],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False]],
        Subsuperscript[ \[Nu], 
          SequenceForm[ "", "b"], 
          SequenceForm[ "", " "]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Linearity and locality of ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Del]\_a\)\&~\ \(-\ \[Del]\_a\)\)]]
}], "Subsubsection"],

Cell[TextData[{
  "We show that ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\ \(-\ \(\[Del]\&~\)\_a\))\) \((f\ \[Mu]\
\_b\  + \ \[Nu]\_b)\)\  = \ \(f(\[Del]\_a\ \(-\ \(\[Del]\&~\)\_a\))\) \
\[Mu]\_b\  + \ \((\[Del]\_a\ \(-\ \(\[Del]\&~\)\_a\))\) \[Nu]\_b\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, \ %]\  - \ TildeCD[la, \ %]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Nu]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Nu], 
              SequenceForm[ "", "b"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          \[Nu][ lb]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              \[Mu][ lb]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              \[Mu][ lb]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Nu]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Nu], 
              SequenceForm[ "", "b"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\!\(\[Del]\&~\)", "a"], 
          \[Nu][ lb]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, there exists a tensor field ",
  Cell[BoxData[
      \(TraditionalForm\`\[Gamma]\_\(a\[InvisibleSpace]b\)\^c\)]],
  " such that ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Del]\&~\)\_a \[Mu]\_b\  = \ \[Del]\_a \[Mu]\_b\  \
- \ \(\[Gamma]\_\(a\[InvisibleSpace]b\)\^p\) \[Mu]\_p\)]]
}], "Text"],

Cell[BoxData[
    \(DefineTensor[\[Gamma], "\<abc\>"]\)], "Input"],

Cell[BoxData[
    \(TildeCD[la_, \[Mu][lb_]] := 
      Module[{pair = DummyPair[NormalIndexQ]}, 
        CD[la, \[Mu][lb]] - \[Gamma][la, lb, pair[\([2]\)]] \[Mu][
              pair[\([1]\)]]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[la, \[Mu][lb]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Mu]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Mu], 
              SequenceForm[ "", "b"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          \[Mu][ lb]],
        Editable->False], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
              SequenceForm[ "", "q"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "q"], 
            SequenceForm[ "", " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Because ",
  Cell[BoxData[
      \(TraditionalForm\`\((\(\[Del]\_a\)\&~\ \(-\ \[Del]\_a\))\) \
\((\(\[Mu]\_b\) \[Xi]\^b)\)\  = \ 0\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Mu][lb] \[Xi][ub]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[Mu]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
            SequenceForm[ "", "b"],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False]],
        Subsuperscript[ \[Mu], 
          SequenceForm[ "", "b"], 
          SequenceForm[ "", " "]],
        Editable->False], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Xi]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
            SequenceForm[ "", "b"],
            Editable->False]],
        Subsuperscript[ \[Xi], 
          SequenceForm[ "", " "], 
          SequenceForm[ "", "b"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[la, %]\  - \ CD[la, %] // Expand\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"-", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Xi]", 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False]],
                  Subsuperscript[ \[Xi], 
                    SequenceForm[ "", " "], 
                    SequenceForm[ "", "b"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[Del]", "a"], 
                \[Xi][ ub]],
              Editable->False], ")"}]}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "b"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "b"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              \[Xi][ ub]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "b"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "r"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "r"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
              SequenceForm[ "", "r"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "r"], 
            SequenceForm[ "", " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "b"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "we have ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Del]\&~\)\_a \[Xi]\^c\  = \ \[Del]\_a \[Xi]\^c\  \
+ \ \(\[Gamma]\_\(a\[InvisibleSpace]p\)\^c\) \[Xi]\^p\)]]
}], "Text"],

Cell[BoxData[
    \(TildeCD[la_, \[Xi][uc_]] := 
      Module[{pair = DummyPair[NormalIndexQ]}, 
        CD[la, \[Xi][uc]] + \[Gamma][la, pair[\([1]\)], uc] \[Xi][
              pair[\([2]\)]]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCD[la, \[Xi][uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                SequenceForm[ "", "c"],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "c"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          \[Xi][ uc]],
        Editable->False], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"s\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "s"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "s"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "s"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"s"\),
              SequenceForm[ "", "s"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "s"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Notice that ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Del]\_a\)\&~\(\[Del]\_b 
              f\)\  - \ \[Del]\_a\(\[Del]\_b 
              f\) = \(-\[Gamma]\_\(a\[InvisibleSpace]b\)\^c\) \[Del]\_c 
            f\)]],
  ". Using the definition for the torsion we find"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((TildeCD[la, TildeCD[lb, f[]]] - 
          CD[la, CD[lb, f[]]])\) - \((TildeCD[lb, TildeCD[la, f[]]] - \ 
          CD[lb, CD[la, f[]]])\)\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"t"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "t"], 
              f[ ]],
            Editable->False], ")"}]}], " ", 
      InterpretationBox[
        SubsuperscriptBox["t", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "a"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "a"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\"t\"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "t"],
            Editable->False]],
        Subsuperscript[ t, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "b"], "a"], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], "t"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, we obtain ",
  Cell[BoxData[
      \(TraditionalForm\`\[Gamma]\_\(a\[InvisibleSpace]b\)\^c\  - \ \
\[Gamma]\_\(b\[InvisibleSpace]a\)\^c\  = \ \(-t\_\(a\[InvisibleSpace]b\)\^c\)\
\)]]
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Section"],

Cell[CellGroupData[{

Cell[TextData[{
  "Linearity and locality of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \(+\
\ t\_\(a\[InvisibleSpace]b\)\^c\)\)\)\) \[Del]\_c\)]]
}], "Subsubsection"],

Cell["We show that", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \((f\ \[Mu]\_c\  + \ \
\[Nu]\_c)\)\  = \ \(f(\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \(+\ 
                          t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \
\[Mu]\_b\[IndentingNewLine]\ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \
\ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \
\ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \  + \ \((\[Del]\_a\(\[Del]\_b\ \(-\ \
\[Del]\_b\(\[Del]\_a\ \(+\ 
                          t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \
\[Nu]\_b\)]]], "Text",
  TextAlignment->Center],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, CD[lb, f[]\ \[Mu][lc]\  + \ \[Nu][lc]]]\  - \ 
        CD[lb, CD[la, f[] \[Mu][lc]\  + \ \[Nu][lc]]]\  + \ 
        Torsiont[la, lb, ud] CD[ld, f[] \[Mu][lc]\  + \ \[Nu][lc]] // 
      Expand\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CD[ lb, 
            \[Nu][ lc]]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "b"], 
          CD[ la, 
            \[Nu][ lc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CD[ lb, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CD[ la, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "d"], 
              \[Nu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "d"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "d"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"u"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "u"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"u\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "u"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "u"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Can[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CD[ lb, 
            \[Nu][ lc]]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "b"], 
          CD[ la, 
            \[Nu][ lc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CD[ lb, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CD[ la, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Nu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, there is a tensor ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\)\^d\)]],
  " such that ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \[Mu]\_c\  = \ \(-R\_\
\(a\[InvisibleSpace]b\[InvisibleSpace]c\)\^d\) \[Mu]\_d\)]],
  "."
}], "Text"],

Cell[BoxData[
    \(DefineTensor[
      CurvatureR, \*"\"\<\!\(R\&~\)\>\"", "\<-bacd-abdc\>"]\)], "Input"],

Cell[BoxData[
    \(CD[la_, CD[lb_, \[Mu][lc_]]]\  := \ 
      Module[{pair = DummyPair[NormalIndexQ]}, 
          CD[lb, CD[la, \[Mu][lc]]]\  - \ 
            Torsiont[la, lb, pair[\([2]\)]] CD[pair[\([1]\)], \[Mu][lc]] - 
            CurvatureR[la, lb, lc, pair[\([2]\)]] \[Mu][
                pair[\([1]\)]]]\  /; \ \(! OrderedQ[{la, lb}]\)\  && \ 
          DnIndexQ[lc]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[lb, CD[la, \[Mu][lc]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Mu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CD[ lb, 
            \[Mu][ lc]]],
        Editable->False], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"y"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "y"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"y\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "y"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "y"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"a\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "a"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "a"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"y\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "y"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "a"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "y"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"y"\),
              SequenceForm[ "", "y"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "y"], 
            SequenceForm[ "", " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Can[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Mu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CD[ lb, 
            \[Mu][ lc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "c"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "c"], "p"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "For scalar, ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \((\(\[Mu]\_c\) \[Xi]\
\^c)\) = 0\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, CD[lb, \[Mu][lc] \[Xi][uc]]]\  - \ 
      CD[lb, CD[la, \[Mu][lc] \[Xi][uc]]]\  + \ 
      Torsiont[la, lb, ud] CD[ld, \[Mu][lc] \[Xi][uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "c"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Xi][ uc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CD[ lb, 
                \[Xi][ uc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "c"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Xi][ uc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CD[ la, 
                \[Xi][ uc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CD[ lb, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "c"]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          RowBox[{
            InterpretationBox[
              RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["\[Mu]", 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False]],
                      Subsuperscript[ \[Mu], 
                        SequenceForm[ "", "c"], 
                        SequenceForm[ "", " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[Del]", "b"], 
                    \[Mu][ lc]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[Del]", "a"], 
                CD[ lb, 
                  \[Mu][ lc]]],
              Editable->False], "-", 
            RowBox[{
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[Del]"\_"z"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["\[Mu]", 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False]],
                      Subsuperscript[ \[Mu], 
                        SequenceForm[ "", "c"], 
                        SequenceForm[ "", " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[Del]", "z"], 
                    \[Mu][ lc]],
                  Editable->False], ")"}], " ", 
              InterpretationBox[
                SubsuperscriptBox["t", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"a\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "a"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"z\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "z"],
                    Editable->False]],
                Subsuperscript[ t, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "a"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "z"]],
                Editable->False]}], "-", 
            RowBox[{
              InterpretationBox[
                SubsuperscriptBox[\(R\&~\), 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                SequenceForm[ "", "b"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"a\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "b"], "a"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "a"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"z\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], " "], "z"],
                    Editable->False]],
                Subsuperscript[ 
                  OverTilde[ R], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "a"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "], "z"]],
                Editable->False], " ", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"z"\),
                    SequenceForm[ "", "z"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "z"], 
                  SequenceForm[ "", " "]],
                Editable->False]}]}], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "c"]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False], " ", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["\[Xi]", 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False]],
                      Subsuperscript[ \[Xi], 
                        SequenceForm[ "", " "], 
                        SequenceForm[ "", "c"]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[Del]", "d"], 
                    \[Xi][ uc]],
                  Editable->False], ")"}], " ", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}], "+", 
            RowBox[{
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["\[Mu]", 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False]],
                      Subsuperscript[ \[Mu], 
                        SequenceForm[ "", "c"], 
                        SequenceForm[ "", " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[Del]", "d"], 
                    \[Mu][ lc]],
                  Editable->False], ")"}], " ", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}]}], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Can[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Xi][ "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CD[ lb, 
                \[Xi][ "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Xi][ "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CD[ la, 
                \[Xi][ "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                    SequenceForm[ "", "q"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Xi][ "u$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
              SequenceForm[ "", "q"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "q"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
              SequenceForm[ "", "q"],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \[Xi]\^c\  = \ \
\(R\_\(a\[InvisibleSpace]b\[InvisibleSpace]p\)\^c\) \[Xi]\^p\)]]
}], "Text"],

Cell[BoxData[
    \(CD[la_, CD[lb_, \[Xi][uc_]]]\  := \ 
      Module[{pair = DummyPair[NormalIndexQ]}, 
          CD[lb, CD[la, \[Xi][uc]]]\  - \ 
            Torsiont[la, lb, pair[\([2]\)]] CD[pair[\([1]\)], \[Xi][uc]] + 
            CurvatureR[la, lb, pair[\([1]\)], uc] \[Xi][
                pair[\([2]\)]]]\  /; \ \(! OrderedQ[{la, lb}]\)\  && \ 
          UpIndexQ[uc]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[lb, CD[la, \[Xi][uc]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Xi][ uc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CD[ lb, 
            \[Xi][ uc]]],
        Editable->False], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"$20"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "$20"], 
              \[Xi][ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"   \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "a"], "   "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"$20\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "$20"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "a"], "   "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "$20"]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"a\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"$20\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "a"], "$20"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "a"], "$20"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"   \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "   "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "   "], "c"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "a"], "$20"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "   "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"   "\),
              SequenceForm[ "", "   "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$20"\),
              SequenceForm[ "", "$20"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", "   "], 
            SequenceForm[ "", "$20"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Can[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Xi][ uc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CD[ lb, 
            \[Xi][ uc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Xi][ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{639, 694},
WindowMargins->{{7, Automatic}, {Automatic, 0}},
Visible->True,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"], 
    "Vdefine"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
            OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
            NotationBoxTag, TagStyle -> "NotationTemplateStyle"], " ", 
          "\[DoubleLongLeftRightArrow]", " ", 
          TagBox[ 
            RowBox[ {"Vector", "[", "\[Placeholder]", "]"}], NotationBoxTag, 
            TagStyle -> "NotationTemplateStyle"]}], "]"}], 
    "ivec"->OverscriptBox[ "\[DotlessI]", "^"], 
    "jvec"->OverscriptBox[ "\[DotlessJ]", "^"], 
    "vect"->OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
    "xvec"->OverscriptBox[ "x", "^"], "yvec"->OverscriptBox[ "y", "^"], 
    "zvec"->OverscriptBox[ "z", "^"], "rvec"->OverscriptBox[ "r", "^"], 
    "tvec"->OverscriptBox[ "\[Theta]", "^"], 
    "pvec"->OverscriptBox[ "\[Phi]", "^"]},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 100, 3, 43, "Title"],
Cell[1820, 54, 52, 0, 41, "Text"],
Cell[1875, 56, 59, 1, 56, "Input"],
Cell[1937, 59, 68, 1, 56, "Input"],
Cell[2008, 62, 65, 1, 56, "Input"],
Cell[2076, 65, 56, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2157, 70, 38, 0, 78, "Section"],
Cell[2198, 72, 76, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2299, 77, 34, 0, 54, "Subsubsection"],
Cell[2336, 79, 90, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2451, 84, 74, 1, 56, "Input"],
Cell[2528, 87, 2479, 61, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[5056, 154, 38, 0, 32, "Subsubsection"],
Cell[5097, 156, 89, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[5211, 161, 74, 1, 56, "Input"],
Cell[5288, 164, 4741, 113, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10078, 283, 55, 0, 32, "Subsubsection"],
Cell[10136, 285, 194, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[10355, 293, 57, 1, 56, "Input"],
Cell[10415, 296, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10499, 303, 51, 0, 32, "Subsubsection"],
Cell[10553, 305, 119, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[10697, 311, 63, 1, 56, "Input"],
Cell[10763, 314, 1517, 37, 55, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[12341, 358, 40, 0, 50, "Section"],
Cell[12384, 360, 131, 3, 56, "Input"],

Cell[CellGroupData[{
Cell[12540, 367, 34, 0, 54, "Subsubsection"],
Cell[12577, 369, 92, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[12694, 374, 74, 1, 56, "Input"],
Cell[12771, 377, 2459, 61, 54, "Output"]
}, Open  ]],
Cell[15245, 441, 114, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[15384, 447, 79, 1, 56, "Input"],
Cell[15466, 450, 2495, 61, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[18010, 517, 38, 0, 32, "Subsubsection"],
Cell[18051, 519, 89, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[18165, 524, 74, 1, 56, "Input"],
Cell[18242, 527, 4721, 113, 55, "Output"]
}, Open  ]],
Cell[22978, 643, 111, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[23114, 649, 79, 1, 56, "Input"],
Cell[23196, 652, 4757, 113, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[28002, 771, 55, 0, 32, "Subsubsection"],
Cell[28060, 773, 194, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[28279, 781, 57, 1, 56, "Input"],
Cell[28339, 784, 35, 1, 53, "Output"]
}, Open  ]],
Cell[28389, 788, 199, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[28613, 796, 62, 1, 56, "Input"],
Cell[28678, 799, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[28762, 806, 64, 0, 32, "Subsubsection"],
Cell[28829, 808, 65, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[28919, 813, 44, 1, 56, "Input"],
Cell[28966, 816, 345, 9, 54, "Output"]
}, Open  ]],
Cell[29326, 828, 70, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[29421, 833, 49, 1, 56, "Input"],
Cell[29473, 836, 345, 9, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[29867, 851, 51, 0, 32, "Subsubsection"],
Cell[29921, 853, 245, 6, 41, "Text"],
Cell[30169, 861, 274, 6, 148, "Input"],
Cell[30446, 869, 124, 4, 41, "Text"],
Cell[30573, 875, 274, 6, 148, "Input"],

Cell[CellGroupData[{
Cell[30872, 885, 52, 1, 56, "Input"],
Cell[30927, 888, 2626, 66, 54, "Output"]
}, Open  ]],
Cell[33568, 957, 209, 5, 41, "Text"],
Cell[33780, 964, 129, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[33934, 970, 62, 1, 56, "Input"],
Cell[33999, 973, 595, 16, 54, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[34655, 996, 56, 0, 50, "Section"],
Cell[34714, 998, 85, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[34824, 1003, 63, 1, 56, "Input"],
Cell[34890, 1006, 1137, 29, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36064, 1040, 148, 4, 56, "Subsubsection"],
Cell[36215, 1046, 282, 6, 41, "Text"],

Cell[CellGroupData[{
Cell[36522, 1056, 67, 1, 56, "Input"],
Cell[36592, 1059, 3399, 83, 54, "Output"]
}, Open  ]],
Cell[40006, 1145, 326, 8, 41, "Text"],
Cell[40335, 1155, 66, 1, 56, "Input"],
Cell[40404, 1158, 206, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[40635, 1166, 55, 1, 56, "Input"],
Cell[40693, 1169, 2827, 70, 56, "Output"]
}, Open  ]],
Cell[43535, 1242, 165, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[43725, 1251, 52, 1, 56, "Input"],
Cell[43780, 1254, 955, 25, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44772, 1284, 73, 1, 56, "Input"],
Cell[44848, 1287, 5301, 129, 55, "Output"]
}, Open  ]],
Cell[50164, 1419, 195, 5, 41, "Text"],
Cell[50362, 1426, 206, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[50593, 1434, 55, 1, 56, "Input"],
Cell[50651, 1437, 2827, 70, 54, "Output"]
}, Open  ]],
Cell[53493, 1510, 304, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[53822, 1522, 174, 3, 125, "Input"],
Cell[53999, 1527, 1964, 50, 54, "Output"]
}, Open  ]],
Cell[55978, 1580, 224, 6, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[56251, 1592, 28, 0, 50, "Section"],

Cell[CellGroupData[{
Cell[56304, 1596, 217, 5, 54, "Subsubsection"],
Cell[56524, 1603, 28, 0, 41, "Text"],
Cell[56555, 1605, 722, 11, 63, "Text"],

Cell[CellGroupData[{
Cell[57302, 1620, 231, 4, 125, "Input"],
Cell[57536, 1626, 14478, 352, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[72051, 1983, 39, 1, 56, "Input"],
Cell[72093, 1986, 9504, 230, 77, "Output"]
}, Open  ]],
Cell[81612, 2219, 414, 10, 41, "Text"],
Cell[82029, 2231, 106, 2, 56, "Input"],
Cell[82138, 2235, 388, 7, 217, "Input"],

Cell[CellGroupData[{
Cell[82551, 2246, 58, 1, 56, "Input"],
Cell[82612, 2249, 6375, 158, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[89024, 2412, 39, 1, 56, "Input"],
Cell[89066, 2415, 6375, 158, 56, "Output"]
}, Open  ]],
Cell[95456, 2576, 234, 6, 41, "Text"],

Cell[CellGroupData[{
Cell[95715, 2586, 179, 3, 102, "Input"],
Cell[95897, 2591, 17743, 419, 100, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[113677, 3015, 39, 1, 56, "Input"],
Cell[113719, 3018, 9613, 235, 56, "Output"]
}, Open  ]],
Cell[123347, 3256, 277, 6, 41, "Text"],
Cell[123627, 3264, 388, 7, 217, "Input"],

Cell[CellGroupData[{
Cell[124040, 3275, 58, 1, 56, "Input"],
Cell[124101, 3278, 6419, 158, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[130557, 3441, 39, 1, 56, "Input"],
Cell[130599, 3444, 6375, 158, 56, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

