(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    463380,      11254]*)
(*NotebookOutlinePosition[    466862,      11343]*)
(*  CellTagsIndexPosition[    466818,      11339]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Kaluza-Klein Formalism", "Title",
  TextAlignment->Left,
  TextJustification->0,
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`KKTheory`\)], "Input"],

Cell[CellGroupData[{

Cell["Defined Tensors", "Section"],

Cell[CellGroupData[{

Cell["KKClass Objects", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \({RiemannR[lA, lB, lC, lD], RicciR[lA, lB], ScalarR[], 
      AffineG[lA, lB, uC], Metricg[lA, lB], Structuref[lA, lB, uC], 
      GaugeF[l\[Mu], l\[Nu], ua], GaugeA[l\[Mu], ua]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"A"\)\
,
                          SequenceForm[ "", "A"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"B\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "A"], "B"],
                      Editable->False], "\[InvisibleSpace]", "\<\"C\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "A"], "B"], "C"],
                  Editable->False], "\[InvisibleSpace]", "\<\"D\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "A"], "B"], "C"], "D"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "A"], "B"], "C"], "D"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"A"\),
                  SequenceForm[ "", "A"],
                  Editable->False], "\[InvisibleSpace]", "\<\"B\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "A"], "B"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "A"], "B"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"A"\),
                      SequenceForm[ "", "A"],
                      Editable->False], "\[InvisibleSpace]", "\<\"B\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "A"], "B"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "A"], "B"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"C\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "C"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "A"], "B"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "C"]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"A"\),
                  SequenceForm[ "", "A"],
                  Editable->False], "\[InvisibleSpace]", "\<\"B\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "A"], "B"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "A"], "B"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"A"\),
                      SequenceForm[ "", "A"],
                      Editable->False], "\[InvisibleSpace]", "\<\"B\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "A"], "B"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "A"], "B"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"C\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "C"],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "A"], "B"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "C"]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "a"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "a"]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["A", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\),
                  SequenceForm[ "", "\[Mu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "a"],
              Editable->False]],
          Subsuperscript[ A, 
            SequenceForm[ 
              SequenceForm[ "", "\[Mu]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "a"]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Base Objects", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \({BaseRiemannR[l\[Mu], l\[Nu], l\[Rho], l\[Sigma]], 
      BaseRicciR[l\[Mu], l\[Nu]], BaseScalarR[], 
      BaseAffineG[l\[Mu], l\[Nu], u\[Rho]], BaseMetricg[l\[Mu], l\[Nu]], 
      BaseStructuref[l\[Mu], l\[Nu], u\[Rho]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Mu]"\),
                          SequenceForm[ "", "\[Mu]"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Rho]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Rho]"], 
                "\[Sigma]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Rho]"], 
              "\[Sigma]"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\),
                  SequenceForm[ "", "\[Mu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[\(\(R\&~\)\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ 
            OverTilde[ R], "", ""],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Rho]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Rho]"]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\),
                  SequenceForm[ "", "\[Mu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(f\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Rho]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ f], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Rho]"]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Fibre Objects", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \({FibreRiemannR[la, lb, lc, ld], FibreRicciR[la, lb], FibreScalarR[], 
      FibreAffineG[la, lb, uc], FibreMetricg[la, lb], 
      FibreStructuref[la, lb, uc]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "c"], "d"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "c"], "d"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(R\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ R], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[\(\(R\&^\)\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ 
            OverHat[ R], "", ""],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(f\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ f], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Ordinary Derivative", "Section"],

Cell[TextData[{
  "Notice that ",
  Cell[BoxData[
      \(TraditionalForm\`D\_A\  = \ \(\((D\_\[Mu], \ 
            D\_a)\)\  = \ \((\[PartialD]\_\[Mu]\ \(-\ 
                  g\)\ \[Kappa]\ \(A\_\[Mu]\^a\) \[PartialD]\_a, \ \
\[PartialD]\_a)\)\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(BD[lA, lB, RicciR[lC, lD]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("D"\_"A"\), "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{\("D"\_"B"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["R", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"C"\),
                      SequenceForm[ "", "C"],
                      Editable->False], "\[InvisibleSpace]", "\<\"D\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "C"], "D"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ R, 
                SequenceForm[ 
                  SequenceForm[ "", "C"], "D"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "B"], 
            RicciR[ lC, lD]],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "D", "A"], 
        BD[ lB, 
          RicciR[ lC, lD]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Block-Diagonal Basis", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[
      TableKK[Metricg[lA, lB], {lA, lB}] /. BlockDiagonalRule]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Alpha]"\),
                        SequenceForm[ "", "\[Alpha]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Beta]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Alpha]"], "\[Beta]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Alpha]"], "\[Beta]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False], "0"},
            {"0", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableKK[Structuref[lA, lB, uC], {lA, lB, uC}] /. 
      BlockDiagonalRule\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{"0", ",", 
                RowBox[{\(-g\), " ", "\[Kappa]", " ", 
                  InterpretationBox[
                    SubsuperscriptBox["F", 
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"\[Alpha]"\),
                                SequenceForm[ "", "\[Alpha]"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"\[Beta]\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "\[Alpha]"], "\[Beta]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "\[Alpha]"], "\[Beta]"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"a\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], "a"],
                        Editable->False]],
                    Subsuperscript[ F, 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "\[Alpha]"], "\[Beta]"], " "], 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], "a"]],
                    Editable->False]}]}], "}"}], ",", \({0, 0}\)}], "}"}], 
        ",", \({{0, 0}, {0, 0}}\)}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Connection", "Section"],

Cell[BoxData[
    \(affineUpG[lA_, lB_, uC_] := 
      SumDumKK[AffineG[lA, lB, uC]\  // \ AffineToMetric]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[l\[Mu], l\[Nu], u\[Rho]] /. BlockDiagonalRule // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Rho]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Rho]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              BaseMetricg[ l\[Mu], l\[Nu]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Rho]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Rho]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              BaseMetricg[ "l@1", l\[Nu]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Rho]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Rho]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Nu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Nu]"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[lA_, lB_, uC_], BaseAffineG[lA, lB, uC], 
      DnIndexQ[lA] && GreekIndexQ[lA] && DnIndexQ[lB] && GreekIndexQ[lB] && 
        UpIndexQ[uC] && GreekIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l\[Mu], l\[Nu], u\[Rho]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\),
                  SequenceForm[ "", "\[Mu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"],
              Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
          Editable->False], 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], "\[Rho]"],
          Editable->False]],
      Subsuperscript[ 
        OverTilde[ \[CapitalGamma]], 
        SequenceForm[ 
          SequenceForm[ 
            SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
        SequenceForm[ 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], "\[Rho]"]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[l\[Mu], l\[Nu], ua] /. BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ l\[Mu], l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "a"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "a"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[
      AffineG[lA_, lB_, 
        uC_], \(-FibreMetricg[uC, 
            ua]\) \((BD[la, BaseMetricg[lA, lB]] + 
              g\ \[Kappa]\ GaugeF[lA, lB, la])\)/2, 
      DnIndexQ[lA] && GreekIndexQ[lA] && DnIndexQ[lB] && GreekIndexQ[lB] && 
        UpIndexQ[uC] && NormalIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l\[Mu], l\[Nu], ua] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ l\[Mu], l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "a"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "a"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[l\[Mu], la, u\[Nu]]\  /. \ BlockDiagonalRule // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Nu]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Nu]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[lA_, lB_, uC_], 
      BaseMetricg[uC, 
          u\[Rho]] \((BD[lB, BaseMetricg[lA, l\[Rho]]] + 
              g\ \[Kappa]\ GaugeF[lA, l\[Rho], lB])\)/2, 
      DnIndexQ[lA] && GreekIndexQ[lA] && DnIndexQ[lB] && NormalIndexQ[lB] && 
        UpIndexQ[uC] && GreekIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l\[Mu], la, u\[Nu]] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Nu]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Nu]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[l\[Mu], la, ub]\  /. \ BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Phi]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Phi], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "\[Mu]"], 
            FibreMetricg[ la, "l$1"]],
          Editable->False], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Phi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "b"], "p"],
            Editable->False]],
        Subsuperscript[ \[Phi], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "b"], "p"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[lA_, lB_, uC_], 
      FibreMetricg[uC, uc] BD[lA, FibreMetricg[lB, lc]]/2, 
      DnIndexQ[lA] && GreekIndexQ[lA] && DnIndexQ[lB] && NormalIndexQ[lB] && 
        UpIndexQ[uC] && NormalIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l\[Mu], la, ub] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Phi]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Phi], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "\[Mu]"], 
            FibreMetricg[ la, "l$1"]],
          Editable->False], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Phi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "b"], "p"],
            Editable->False]],
        Subsuperscript[ \[Phi], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "b"], "p"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[la, l\[Mu], u\[Nu]]\  /. \ BlockDiagonalRule // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Nu]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Nu]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[lA_, lB_, uC_], 
      BaseMetricg[uC, 
          u\[Rho]] \((BD[lA, BaseMetricg[lB, l\[Rho]]] + 
              g\ \[Kappa]\ GaugeF[lB, l\[Rho], lA])\)/2, 
      DnIndexQ[lA] && NormalIndexQ[lA] && DnIndexQ[lB] && GreekIndexQ[lB] && 
        UpIndexQ[uC] && GreekIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[la, l\[Mu], u\[Nu]] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Nu]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Nu]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Nu]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[la, l\[Mu], ub]\  /. \ BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Phi]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Phi], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "\[Mu]"], 
            FibreMetricg[ la, "l$1"]],
          Editable->False], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Phi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "b"], "p"],
            Editable->False]],
        Subsuperscript[ \[Phi], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "b"], "p"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[lA_, lB_, uC_], 
      FibreMetricg[uC, uc] BD[lB, FibreMetricg[lA, lc]]/2, 
      DnIndexQ[lA] && NormalIndexQ[lA] && DnIndexQ[lB] && GreekIndexQ[lB] && 
        UpIndexQ[uC] && NormalIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[la, l\[Mu], ub] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Phi]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Phi], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "\[Mu]"], 
            FibreMetricg[ la, "l$1"]],
          Editable->False], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Phi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "b"], "p"],
            Editable->False]],
        Subsuperscript[ \[Phi], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "b"], "p"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[la, lb, u\[Mu]]\  /. \ BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(-\(1\/2\)\), " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Gamma]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                SequenceForm[ "", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Mu]"],
            Editable->False]],
        Subsuperscript[ \[Gamma], 
          SequenceForm[ 
            SequenceForm[ "", "\[Chi]"], " "], 
          SequenceForm[ 
            SequenceForm[ "", " "], "\[Mu]"]],
        Editable->False], " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[\("D"\^"\[Chi]"\),
              Superscript[ "D", "\[Chi]"],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Phi]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Phi], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Superscript[ "D", "\[Chi]"], 
            FibreMetricg[ la, lb]],
          Editable->False], ")"}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[
      AffineG[lA_, lB_, uC_], \(-BaseMetricg[uC, u\[Rho]]\) 
        BD[l\[Rho], FibreMetricg[lA, lB]]/2, 
      DnIndexQ[lA] && NormalIndexQ[lA] && DnIndexQ[lB] && NormalIndexQ[lB] && 
        UpIndexQ[uC] && GreekIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[la, lb, u\[Mu]] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(-\(1\/2\)\), " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Gamma]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                SequenceForm[ "", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Mu]"],
            Editable->False]],
        Subsuperscript[ \[Gamma], 
          SequenceForm[ 
            SequenceForm[ "", "\[Chi]"], " "], 
          SequenceForm[ 
            SequenceForm[ "", " "], "\[Mu]"]],
        Editable->False], " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[\("D"\^"\[Chi]"\),
              Superscript[ "D", "\[Chi]"],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Phi]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Phi], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Superscript[ "D", "\[Chi]"], 
            FibreMetricg[ la, lb]],
          Editable->False], ")"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[la, lb, uc]\  /. \ BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              FibreMetricg[ la, lb]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              FibreMetricg[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[lA_, lB_, uC_], FibreAffineG[lA, lB, uC], 
      DnIndexQ[lA] && NormalIndexQ[lA] && DnIndexQ[lB] && NormalIndexQ[lB] && 
        UpIndexQ[uC] && NormalIndexQ[uC]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[la, lb, uc]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], " "],
          Editable->False], 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], "c"],
          Editable->False]],
      Subsuperscript[ 
        OverHat[ \[CapitalGamma]], 
        SequenceForm[ 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"], " "], 
        SequenceForm[ 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], "c"]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Section"],

Cell[BoxData[
    \(riemann[lA_, lB_, lC_, uD_] := 
      SumDumKK[RiemannR[lA, lB, lC, uD] // RiemannToAffine]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(riemann[l\[Mu], l\[Nu], l\[Rho], u\[Sigma]]\ \  /. \ BlockDiagonalRule // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"-", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Rho]"],
                    Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Rho]"], "\[Chi]"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Rho]"], "\[Chi]"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False]}], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "\[Sigma]"]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Sigma]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], "\[Rho]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Chi]"]],
          Editable->False]}], "+", 
      InterpretationBox[
        RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]\
"\),
                        SequenceForm[ "", "\[Nu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Nu]"], "\[Rho]"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "\[Sigma]"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Sigma]"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "\[Mu]"], 
          BaseAffineG[ l\[Nu], l\[Rho], u\[Sigma]]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{\("D"\_"\[Nu]"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Rho]"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Rho]"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "\[Sigma]"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Rho]"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Sigma]"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "\[Nu]"], 
          BaseAffineG[ l\[Mu], l\[Rho], u\[Sigma]]],
        Editable->False], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Sigma]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", l\[Nu]]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Rho]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Rho]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ l\[Mu], l\[Rho]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Sigma]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]\
"\),
                        SequenceForm[ "", "\[Nu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Nu]"], "\[Rho]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], "\[Rho]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ l\[Nu], l\[Rho]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Sigma]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]\
"\),
                        SequenceForm[ "", "\[Nu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Nu]"], "\[Rho]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], "\[Rho]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ l\[Nu], l\[Rho]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Sigma]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Rho]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Rho]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", l\[Rho]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Sigma]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Rho]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Rho]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Rho]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Sigma]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Rho]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Rho]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Rho]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Sigma]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Rho]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Rho]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ l\[Mu], l\[Rho]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Sigma]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], "\[Rho]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Sigma]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], "\[Rho]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], "\[Rho]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Sigma]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Sigma]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], "p"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Rho]"\
\),
                      SequenceForm[ "", "\[Rho]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Rho]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Rho]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Rho]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ricciDn[lA_, lB_] := 
      SumDumKK[RicciR[lA, lB] // RiemannToAffine]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricciDn[l\[Mu], l\[Nu]] /. BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Nu]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Nu]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], "\[Chi]"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Chi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Psi]"]],
          Editable->False]}], "-", 
      InterpretationBox[
        RowBox[{\("D"\_"\[Chi]"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "\[Chi]"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Chi]"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "\[Chi]"], 
          BaseAffineG[ l\[Mu], l\[Nu], "u@1"]],
        Editable->False], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              FibreMetricg[ "u$1", "u$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ l\[Mu], l\[Nu]]],
            Editable->False], ")"}]}], "+", 
      InterpretationBox[
        RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Nu]"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Nu]"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "\[Chi]"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Nu]"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Chi]"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "\[Mu]"], 
          BaseAffineG[ "l@1", l\[Nu], "u@1"]],
        Editable->False], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              FibreMetricg[ "u$1", "u$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Nu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Nu]"], 
              FibreMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Mu]"\),
                            SequenceForm[ "", "\[Mu]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "q"], 
                  BaseMetricg[ l\[Mu], l\[Nu]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BD[ "l$2", 
                BaseMetricg[ l\[Mu], l\[Nu]]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Mu]"\),
                            SequenceForm[ "", "\[Mu]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Mu]"], "\[Nu]"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              GaugeF[ l\[Mu], l\[Nu], "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@2", l\[Nu]]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@2", l\[Mu]]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ "l@1", l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", "l@2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ l\[Mu], l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Chi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Chi]"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"\[Nu]"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Phi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], "q"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Phi], 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "\[Nu]"], 
                  FibreMetricg[ "l$1", "l$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              BD[ l\[Nu], 
                FibreMetricg[ "l$1", "l$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ l\[Mu], l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                      SequenceForm[ "", "q"],
                      Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "q"], "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "q"], "r"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "q"], "r"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                  SequenceForm[ "", "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "r"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "r"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              FibreMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Nu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                        SequenceForm[ "", "r"],
                        Editable->False], "\[InvisibleSpace]", "\<\"s\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "r"], "s"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "r"], "s"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Nu]"], 
              FibreMetricg[ "l$3", "l$4"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"s\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "s"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "s"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                  SequenceForm[ "", "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "r"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "r"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]\
"\),
                        SequenceForm[ "", "\[Mu]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ l\[Mu], l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@2", l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@2", l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]"\),
                  SequenceForm[ "", "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Chi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Nu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Nu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", l\[Nu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              FibreMetricg[ "u$1", "u$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", "l@2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                  SequenceForm[ "", "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "r"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "r"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "r"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "r"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@2", l\[Mu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], "p"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]"\),
                  SequenceForm[ "", "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Chi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Nu]"\)\
,
                      SequenceForm[ "", "\[Nu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Nu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Nu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Nu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], "q"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricciDn[l\[Mu], la] /. BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Psi]"\),
                            SequenceForm[ "", "\[Psi]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Psi]"], "\[Mu]"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "a"], 
                  BaseMetricg[ "l@2", l\[Mu]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              BD[ la, 
                BaseMetricg[ "l@2", l\[Mu]]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Psi]"\),
                            SequenceForm[ "", "\[Psi]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                        Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Psi]"], "\[Mu]"], "a"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Mu]"], "a"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              GaugeF[ "l@2", l\[Mu], la]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Psi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Phi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Phi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Phi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Phi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Phi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Phi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@2", "l@3"]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Chi]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Chi]"], 
              BaseMetricg[ "u@1", "u@2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@2", l\[Mu]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Chi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]"\),
                  SequenceForm[ "", "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Phi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Psi]"], "\[Phi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Psi]"], "\[Phi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]\
"\),
                        SequenceForm[ "", "\[Phi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Phi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Phi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@3", l\[Mu]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@1", "l@2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              BaseMetricg[ "u@1", "u@2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Chi]"\),
                            SequenceForm[ "", "\[Chi]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "a"], 
                  BaseMetricg[ "l@1", "l@2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              BD[ la, 
                BaseMetricg[ "l@1", "l@2"]]],
            Editable->False], ")"}]}], "+", 
      InterpretationBox[
        RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverHat[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "a"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "\[Mu]"], 
          FibreAffineG[ "l$1", la, "u$1"]],
        Editable->False], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              FibreMetricg[ "u$1", "u$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              FibreMetricg[ la, "l$2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Chi]"\),
                            SequenceForm[ "", "\[Chi]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                        Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Chi]"], "\[Psi]"], "a"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"], "a"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              GaugeF[ "l@1", "l@2", la]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Phi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "q"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Phi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "\[Mu]"], 
                  FibreMetricg[ la, "l$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BD[ l\[Mu], 
                FibreMetricg[ la, "l$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]"\),
                  SequenceForm[ "", "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Chi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ "l@1", l\[Mu]]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Psi]"\),
                Superscript[ "D", "\[Psi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Psi]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@2", l\[Mu]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", "l@2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              FibreMetricg[ la, "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"q"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]\
"\),
                        SequenceForm[ "", "\[Psi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Mu]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "q"], 
              BaseMetricg[ "l@2", l\[Mu]]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                      SequenceForm[ "", "q"],
                      Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "q"], "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "q"], "r"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "q"], "r"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                  SequenceForm[ "", "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "r"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "r"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              FibreMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                      SequenceForm[ "", "r"],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "r"], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "r"], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "r"], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                  SequenceForm[ "", "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "r"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "r"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              BaseMetricg[ "u@1", "u@2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Chi]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Chi]"], 
              BaseMetricg[ "u@1", "u@2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]"\
\),
                      SequenceForm[ "", "\[Psi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Mu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Mu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Psi]"], "\[Mu]"], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Psi]"], "\[Mu]"], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Mu]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\),
                  SequenceForm[ "", "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Phi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Chi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]"\
\),
                      SequenceForm[ "", "\[Psi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Phi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Psi]"], "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Psi]"], "\[Phi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Psi]"], "\[Phi]"], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Psi]"\),
                Superscript[ "D", "\[Psi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Psi]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Psi]"\),
                Superscript[ "D", "\[Psi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Psi]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Chi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], "q"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Psi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Chi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\),
                  SequenceForm[ "", "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Phi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Phi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "\[Phi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Phi]"], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricciDn[la, lb] /. BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Chi]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Chi]"], 
              BaseMetricg[ "u@1", "u@2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"\[Psi]"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "\[Psi]"], 
              FibreMetricg[ la, lb]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverHat[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "p"], 
          FibreAffineG[ la, lb, "u$1"]],
        Editable->False], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"\[Psi]"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Phi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Phi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "\[Psi]"], 
                  FibreMetricg[ la, lb]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              BD[ "l@2", 
                FibreMetricg[ la, lb]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Chi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\),
                  SequenceForm[ "", "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Phi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Phi]"\),
                Superscript[ "D", "\[Phi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Phi]"], 
              FibreMetricg[ la, lb]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Chi]"\),
                            SequenceForm[ "", "\[Chi]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "b"], 
                  BaseMetricg[ "l@1", "l@2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BD[ lb, 
                BaseMetricg[ "l@1", "l@2"]]],
            Editable->False], ")"}]}], "+", 
      InterpretationBox[
        RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverHat[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "b"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "a"], 
          FibreAffineG[ "l$1", lb, "u$1"]],
        Editable->False], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Chi]"\),
                            SequenceForm[ "", "\[Chi]"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "\[Chi]"], "\[Psi]"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              GaugeF[ "l@1", "l@2", lb]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "u@1", "u@2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              BaseMetricg[ "l@1", "l@2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\),
                  SequenceForm[ "", "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Xi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], "\[Xi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Phi]"], "\[Xi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]\
"\),
                        SequenceForm[ "", "\[Phi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Phi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Phi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@3", "l@2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Xi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Xi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Xi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              BaseMetricg[ "l@1", "l@4"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "b"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "b"], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l@1", "l@2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Psi]"\),
                Superscript[ "D", "\[Psi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Psi]"], 
              FibreMetricg[ la, lb]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Psi]"\),
                Superscript[ "D", "\[Psi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Psi]"], 
              FibreMetricg[ la, "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Chi]"\),
                Superscript[ "D", "\[Chi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Chi]"], 
              FibreMetricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"\[Psi]"\),
                Superscript[ "D", "\[Psi]"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Phi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Phi], 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "\[Psi]"], 
              FibreMetricg[ lb, "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Phi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ \[Phi], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "u@1", "u@2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], "b"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&^\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverHat[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], "p"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Psi]"], "p"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Psi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Psi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Xi]"\),
                  SequenceForm[ "", "\[Xi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Xi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Phi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Phi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Xi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Phi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]\
"\),
                        SequenceForm[ "", "\[Phi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Phi]"], "\[Psi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Phi]"], "\[Psi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              BaseMetricg[ "l@3", "l@2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Xi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Xi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Xi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "b"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Xi]"], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "g", " ", "\[Kappa]", " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Psi]"\),
                  SequenceForm[ "", "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Psi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Chi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Chi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Psi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Chi]"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\),
                  SequenceForm[ "", "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Xi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "\[Xi]"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Phi]"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "\[Xi]"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]\
"\),
                        SequenceForm[ "", "\[Chi]"],
                        Editable->False], 
                      "\[InvisibleSpace]", "\<\"\[Xi]\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "\[Chi]"], "\[Xi]"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Xi]"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              BaseMetricg[ "l@1", "l@4"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\),
                      SequenceForm[ "", "\[Phi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Phi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Phi]"], "\[Psi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], "\[Psi]"], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", \(g\^2\), " ", \(\[Kappa]\^2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\),
                  SequenceForm[ "", "\[Chi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Psi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Chi]"], "\[Psi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\),
                  SequenceForm[ "", "\[Phi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Xi]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], "\[Xi]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "\[Phi]"], "\[Xi]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Chi]"\
\),
                      SequenceForm[ "", "\[Chi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Xi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Chi]"], "\[Xi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Chi]"], "\[Xi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "b"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Chi]"], "\[Xi]"], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Phi]"\
\),
                      SequenceForm[ "", "\[Phi]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Psi]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Phi]"], "\[Psi]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Phi]"], "\[Psi]"], " "],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Phi]"], "\[Psi]"], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{645, 694},
WindowMargins->{{9, Automatic}, {0, Automatic}},
Visible->True,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"], 
    "Vdefine"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
            OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
            NotationBoxTag, TagStyle -> "NotationTemplateStyle"], " ", 
          "\[DoubleLongLeftRightArrow]", " ", 
          TagBox[ 
            RowBox[ {"Vector", "[", "\[Placeholder]", "]"}], NotationBoxTag, 
            TagStyle -> "NotationTemplateStyle"]}], "]"}], 
    "ivec"->OverscriptBox[ "\[DotlessI]", "^"], 
    "jvec"->OverscriptBox[ "\[DotlessJ]", "^"], 
    "vect"->OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
    "xvec"->OverscriptBox[ "x", "^"], "yvec"->OverscriptBox[ "y", "^"], 
    "zvec"->OverscriptBox[ "z", "^"], "rvec"->OverscriptBox[ "r", "^"], 
    "tvec"->OverscriptBox[ "\[Theta]", "^"], 
    "pvec"->OverscriptBox[ "\[Phi]", "^"]},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 102, 3, 43, "Title"],
Cell[1822, 54, 54, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1901, 59, 34, 0, 78, "Section"],

Cell[CellGroupData[{
Cell[1960, 63, 40, 0, 54, "Subsubsection"],

Cell[CellGroupData[{
Cell[2025, 67, 210, 3, 125, "Input"],
Cell[2238, 72, 10368, 254, 57, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[12655, 332, 37, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[12717, 336, 255, 4, 102, "Input"],
Cell[12975, 342, 8297, 204, 56, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[21321, 552, 38, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[21384, 556, 192, 3, 102, "Input"],
Cell[21579, 561, 7884, 194, 56, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[29524, 762, 38, 0, 50, "Section"],
Cell[29565, 764, 263, 7, 41, "Text"],

Cell[CellGroupData[{
Cell[29853, 775, 59, 1, 56, "Input"],
Cell[29915, 778, 1477, 37, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[31441, 821, 39, 0, 50, "Section"],

Cell[CellGroupData[{
Cell[31505, 825, 107, 2, 79, "Input"],
Cell[31615, 829, 2505, 57, 95, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34157, 891, 106, 2, 79, "Input"],
Cell[34266, 895, 2484, 54, 56, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[36799, 955, 29, 0, 50, "Section"],
Cell[36831, 957, 119, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[36975, 963, 102, 2, 79, "Input"],
Cell[37080, 967, 7684, 181, 69, "Output"]
}, Open  ]],
Cell[44779, 1151, 210, 3, 125, "Input"],

Cell[CellGroupData[{
Cell[45014, 1158, 65, 1, 56, "Input"],
Cell[45082, 1161, 1546, 39, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[46665, 1205, 90, 1, 79, "Input"],
Cell[46758, 1208, 5188, 126, 69, "Output"]
}, Open  ]],
Cell[51961, 1337, 327, 7, 171, "Input"],

Cell[CellGroupData[{
Cell[52313, 1348, 67, 1, 56, "Input"],
Cell[52383, 1351, 5188, 126, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[57608, 1482, 101, 2, 79, "Input"],
Cell[57712, 1486, 5262, 126, 69, "Output"]
}, Open  ]],
Cell[62989, 1615, 325, 6, 171, "Input"],

Cell[CellGroupData[{
Cell[63339, 1625, 67, 1, 56, "Input"],
Cell[63409, 1628, 5262, 126, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[68708, 1759, 90, 1, 79, "Input"],
Cell[68801, 1762, 2291, 56, 69, "Output"]
}, Open  ]],
Cell[71107, 1821, 247, 4, 148, "Input"],

Cell[CellGroupData[{
Cell[71379, 1829, 63, 1, 56, "Input"],
Cell[71445, 1832, 2291, 56, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[73773, 1893, 101, 2, 79, "Input"],
Cell[73877, 1897, 5262, 126, 69, "Output"]
}, Open  ]],
Cell[79154, 2026, 325, 6, 171, "Input"],

Cell[CellGroupData[{
Cell[79504, 2036, 67, 1, 56, "Input"],
Cell[79574, 2039, 5262, 126, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[84873, 2170, 90, 1, 79, "Input"],
Cell[84966, 2173, 2291, 56, 69, "Output"]
}, Open  ]],
Cell[87272, 2232, 247, 4, 125, "Input"],

Cell[CellGroupData[{
Cell[87544, 2240, 63, 1, 56, "Input"],
Cell[87610, 2243, 2291, 56, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[89938, 2304, 90, 1, 79, "Input"],
Cell[90031, 2307, 2439, 59, 69, "Output"]
}, Open  ]],
Cell[92485, 2369, 270, 5, 125, "Input"],

Cell[CellGroupData[{
Cell[92780, 2378, 63, 1, 56, "Input"],
Cell[92846, 2381, 2439, 59, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[95322, 2445, 86, 1, 79, "Input"],
Cell[95411, 2448, 7166, 169, 69, "Output"]
}, Open  ]],
Cell[102592, 2620, 214, 3, 148, "Input"],

Cell[CellGroupData[{
Cell[102831, 2627, 52, 1, 56, "Input"],
Cell[102886, 2630, 1493, 39, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[104428, 2675, 28, 0, 50, "Section"],
Cell[104459, 2677, 124, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[104608, 2683, 117, 2, 79, "Input"],
Cell[104728, 2687, 61365, 1482, 259, "Output"]
}, Open  ]],
Cell[166108, 4172, 104, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[166237, 4178, 84, 1, 56, "Input"],
Cell[166324, 4181, 116751, 2795, 520, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[283112, 6981, 80, 1, 56, "Input"],
Cell[283195, 6984, 102763, 2430, 479, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[385995, 9419, 76, 1, 56, "Input"],
Cell[386074, 9422, 77278, 1828, 397, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

