(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    297683,       7168]*)
(*NotebookOutlinePosition[    298451,       7195]*)
(*  CellTagsIndexPosition[    298407,       7191]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Linearized Field Equation and Linealized Gravitational Waves", "Title",
  FontSize->24],

Cell["\<\
See: Introducing Einstein's Relativity, Ray D'Inverno, Chap 20, pp \
271\
\>", "Text"],

Cell[BoxData[
    \(<< \ dhPark`Linealization`\)], "Input"],

Cell[CellGroupData[{

Cell["Linealization", "Section"],

Cell[BoxData[
    \(\(DefineTensor[{\[Eta], h}, "\<ba\>"];\)\)], "Input"],

Cell[TextData[{
  "Let ",
  Cell[BoxData[
      \(TraditionalForm\`g\_\(a\[InvisibleSpace]b\)\  = \ \[Eta]\_\(a\
\[InvisibleSpace]b\)\  + \ \[Epsilon]\ h\_\(a\[InvisibleSpace]b\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(metricRule\  = 
      ObjectRule[
        Metricg[la_, 
          lb_], \[Eta][la, lb]\  + \ \[Epsilon]\ h[la, lb]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "lb_"}],
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", 
              Pattern[ la, 
                Blank[ ]]], 
            Pattern[ lb, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            RowBox[{"\[Epsilon]", " ", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}], "+", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}], "]"}], "/;", "True"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(etaRule\  = {ObjectRule[OD[la_, \[Eta][lb_, lc_]], 0], \ 
        ObjectRule[\[Eta][la_, ua_], 4, PairQ[{la, ua}]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^la_\),
                Superscript[ "\[PartialD]", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Eta]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]lb_\),
                        SequenceForm[ "", 
                          Pattern[ lb, 
                            Blank[ ]]],
                        Editable->False], "\[InvisibleSpace]", "lc_"}],
                    SequenceForm[ 
                      SequenceForm[ "", 
                        Pattern[ lb, 
                          Blank[ ]]], 
                      Pattern[ lc, 
                        Blank[ ]]],
                    Editable->False]],
                Subsuperscript[ \[Eta], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      Pattern[ lb, 
                        Blank[ ]]], 
                    Pattern[ lc, 
                      Blank[ ]]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", 
                Pattern[ la, 
                  Blank[ ]]], 
              \[Eta][ 
                Pattern[ lb, 
                  Blank[ ]], 
                Pattern[ lc, 
                  Blank[ ]]]],
            Editable->False], "\[RuleDelayed]", \(DumFresh[0] /; True\)}], 
        ",", 
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Eta]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "ua_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ ua, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ \[Eta], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ ua, 
                  Blank[ ]]]],
            Editable->False], 
          "\[RuleDelayed]", \(DumFresh[4] /; PairQ[{la, ua}]\)}]}], 
      "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "We shall neglect terms of second order or higher in ",
  Cell[BoxData[
      \(TraditionalForm\`\[Epsilon]\)]],
  " and raise and lower indices with ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta]\_\(a\[InvisibleSpace]b\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell["Connection", "Subsection"],

Cell[BoxData[
    \(\(\(SetAttributes[\[Epsilon], Constant]\)\(\ \)\( (*\ 
      for\ OD\ *) \)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG = AffineG[la, lb, uc]\  // \ AffineToMetric\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                SequenceForm[ "", "c"],
                Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "c"], "p"]],
        Editable->False], " ", 
      RowBox[{"(", 
        RowBox[{
          RowBox[{"-", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["g", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ g, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  Metricg[ la, lb]],
                Editable->False], ")"}]}], "+", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              Metricg[ lb, "l$1"]],
            Editable->False], "+", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              Metricg[ la, "l$1"]],
            Editable->False]}], ")"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[affineUpG, Join[{metricRule}, etaRule]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              h[ la, lb]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{"\[Epsilon]", " ", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"]],
                Editable->False]}], "+", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                      SequenceForm[ "", "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "c"], "p"]],
              Editable->False]}], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              h[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{"\[Epsilon]", " ", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"]],
                Editable->False]}], "+", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                      SequenceForm[ "", "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "c"], "p"]],
              Editable->False]}], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              h[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{"\[Epsilon]", " ", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"]],
                Editable->False]}], "+", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                      SequenceForm[ "", "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "c"], "p"]],
              Editable->False]}], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Normal[Series[%, {\[Epsilon], 0, 1}]]\)], "Input"],

Cell[BoxData[
    RowBox[{"\[Epsilon]", " ", 
      RowBox[{"(", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ la, lb]],
                Editable->False], ")"}], " ", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                      SequenceForm[ "", "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "c"], "p"]],
              Editable->False]}], "+", 
          RowBox[{\(1\/2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  h[ lb, "l$1"]],
                Editable->False], ")"}], " ", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                      SequenceForm[ "", "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "c"], "p"]],
              Editable->False]}], "+", 
          RowBox[{\(1\/2\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "b"], 
                  h[ la, "l$1"]],
                Editable->False], ")"}], " ", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                      SequenceForm[ "", "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "p"],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "c"], "p"]],
              Editable->False]}]}], ")"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affine = Absorb[%, \[Eta], {OD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              h[ lb, uc]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              h[ la, uc]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"c"\),
                Superscript[ "\[PartialD]", "c"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "c"], 
              h[ la, lb]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineRule = ObjectRule[AffineG[la_, lb_, uc_], affine]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[CapitalGamma]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lb_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "uc_"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]], 
              Pattern[ uc, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]], 
            Pattern[ uc, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                              SequenceForm[ "", "b"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "b"], " "],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"c\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], "c"],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], "c"]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "a"], 
                    h[ lb, uc]],
                  Editable->False], ")"}]}], "+", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                              SequenceForm[ "", "a"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "a"], " "],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"c\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], "c"],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], "c"]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "b"], 
                    h[ la, uc]],
                  Editable->False], ")"}]}], "-", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"c"\),
                      Superscript[ "\[PartialD]", "c"],
                      Editable->False], "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                              SequenceForm[ "", "a"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"b\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "b"],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "c"], 
                    h[ la, lb]],
                  Editable->False], ")"}]}]}], "]"}], "/;", 
        "True"}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RiemannR[la, lb, lc, ld]\  // \ RiemannToAffine\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                SequenceForm[ "", "d"],
                Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "d"], "q"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", "d"], "q"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False], " ", 
      RowBox[{"(", 
        RowBox[{
          RowBox[{
            RowBox[{"-", 
              InterpretationBox[
                SubsuperscriptBox["\[CapitalGamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "r"],
                    Editable->False]],
                Subsuperscript[ \[CapitalGamma], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "r"]],
                Editable->False]}], " ", 
            InterpretationBox[
              SubsuperscriptBox["\[CapitalGamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"r\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "r"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "r"], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "q"],
                  Editable->False]],
              Subsuperscript[ \[CapitalGamma], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "r"], " "], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "q"]],
              Editable->False]}], "+", 
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox["\[CapitalGamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"r\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "r"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "r"], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "q"],
                  Editable->False]],
              Subsuperscript[ \[CapitalGamma], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "r"], " "], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "q"]],
              Editable->False], " ", 
            InterpretationBox[
              SubsuperscriptBox["\[CapitalGamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "r"],
                  Editable->False]],
              Subsuperscript[ \[CapitalGamma], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], " "], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "r"]],
              Editable->False]}], "+", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[CapitalGamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "q"],
                    Editable->False]],
                Subsuperscript[ \[CapitalGamma], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              AffineG[ lb, lc, "u$2"]],
            Editable->False], "-", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[CapitalGamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "q"],
                    Editable->False]],
                Subsuperscript[ \[CapitalGamma], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              AffineG[ la, lc, "u$2"]],
            Editable->False]}], ")"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Normal[
      Series[ApplyRules[%, 
          Join[{metricRule}, {affineRule}, etaRule]], {\[Epsilon], 0, 
          1}]]\)], "Input"],

Cell[BoxData[
    RowBox[{"\[Epsilon]", " ", 
      RowBox[{"(", 
        RowBox[{
          RowBox[{
            RowBox[{"(", 
              RowBox[{
                RowBox[{\(-\(1\/2\)\), " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          RowBox[{
                            InterpretationBox[\("\[PartialD]"\^"q"\),
                              Superscript[ "\[PartialD]", "q"],
                              Editable->False], "\[InvisibleSpace]", 
                            InterpretationBox[
                              SubsuperscriptBox["h", 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                      SequenceForm[ "", "b"],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\"c\"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", "b"], "c"],
                                  Editable->False], 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                      SequenceForm[ "", " "],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\" \"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", " "], " "],
                                  Editable->False]],
                              Subsuperscript[ h, 
                                SequenceForm[ 
                                  SequenceForm[ "", "b"], "c"], 
                                SequenceForm[ 
                                  SequenceForm[ "", " "], " "]],
                              Editable->False]}],
                          SequenceForm[ 
                            Superscript[ "\[PartialD]", "q"], 
                            h[ lb, lc]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "a"], 
                        OD[ "u$2", 
                          h[ lb, lc]]],
                      Editable->False], ")"}]}], "+", 
                RowBox[{\(1\/2\), " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          
                          RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]",
                             
                            InterpretationBox[
                              SubsuperscriptBox["h", 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"c"\),
                                      SequenceForm[ "", "c"],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\" \"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", "c"], " "],
                                  Editable->False], 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                      SequenceForm[ "", " "],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\"q\"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", " "], "q"],
                                  Editable->False]],
                              Subsuperscript[ h, 
                                SequenceForm[ 
                                  SequenceForm[ "", "c"], " "], 
                                SequenceForm[ 
                                  SequenceForm[ "", " "], "q"]],
                              Editable->False]}],
                          SequenceForm[ 
                            Subscript[ "\[PartialD]", "b"], 
                            h[ lc, "u$2"]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "a"], 
                        OD[ lb, 
                          h[ lc, "u$2"]]],
                      Editable->False], ")"}]}], "+", 
                RowBox[{\(1\/2\), " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          
                          RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]",
                             
                            InterpretationBox[
                              SubsuperscriptBox["h", 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                      SequenceForm[ "", "b"],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\" \"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", "b"], " "],
                                  Editable->False], 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                      SequenceForm[ "", " "],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\"q\"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", " "], "q"],
                                  Editable->False]],
                              Subsuperscript[ h, 
                                SequenceForm[ 
                                  SequenceForm[ "", "b"], " "], 
                                SequenceForm[ 
                                  SequenceForm[ "", " "], "q"]],
                              Editable->False]}],
                          SequenceForm[ 
                            Subscript[ "\[PartialD]", "c"], 
                            h[ lb, "u$2"]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "a"], 
                        OD[ lc, 
                          h[ lb, "u$2"]]],
                      Editable->False], ")"}]}]}], ")"}], " ", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                      SequenceForm[ "", "d"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "d"], "q"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}], "-", 
          RowBox[{
            RowBox[{"(", 
              RowBox[{
                RowBox[{\(1\/2\), " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          
                          RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]",
                             
                            InterpretationBox[
                              SubsuperscriptBox["h", 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"c"\),
                                      SequenceForm[ "", "c"],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\" \"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", "c"], " "],
                                  Editable->False], 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                      SequenceForm[ "", " "],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\"q\"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", " "], "q"],
                                  Editable->False]],
                              Subsuperscript[ h, 
                                SequenceForm[ 
                                  SequenceForm[ "", "c"], " "], 
                                SequenceForm[ 
                                  SequenceForm[ "", " "], "q"]],
                              Editable->False]}],
                          SequenceForm[ 
                            Subscript[ "\[PartialD]", "b"], 
                            h[ lc, "u$2"]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "a"], 
                        OD[ lb, 
                          h[ lc, "u$2"]]],
                      Editable->False], ")"}]}], "-", 
                RowBox[{\(1\/2\), " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          RowBox[{
                            InterpretationBox[\("\[PartialD]"\^"q"\),
                              Superscript[ "\[PartialD]", "q"],
                              Editable->False], "\[InvisibleSpace]", 
                            InterpretationBox[
                              SubsuperscriptBox["h", 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                      SequenceForm[ "", "a"],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\"c\"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", "a"], "c"],
                                  Editable->False], 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                      SequenceForm[ "", " "],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\" \"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", " "], " "],
                                  Editable->False]],
                              Subsuperscript[ h, 
                                SequenceForm[ 
                                  SequenceForm[ "", "a"], "c"], 
                                SequenceForm[ 
                                  SequenceForm[ "", " "], " "]],
                              Editable->False]}],
                          SequenceForm[ 
                            Superscript[ "\[PartialD]", "q"], 
                            h[ la, lc]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "b"], 
                        OD[ "u$2", 
                          h[ la, lc]]],
                      Editable->False], ")"}]}], "+", 
                RowBox[{\(1\/2\), " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          
                          RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]",
                             
                            InterpretationBox[
                              SubsuperscriptBox["h", 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                      SequenceForm[ "", "a"],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\" \"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", "a"], " "],
                                  Editable->False], 
                                InterpretationBox[
                                  RowBox[{
                                    
                                    InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                      SequenceForm[ "", " "],
                                      Editable->False], 
                                    "\[InvisibleSpace]", "\<\"q\"\>"}],
                                  SequenceForm[ 
                                    SequenceForm[ "", " "], "q"],
                                  Editable->False]],
                              Subsuperscript[ h, 
                                SequenceForm[ 
                                  SequenceForm[ "", "a"], " "], 
                                SequenceForm[ 
                                  SequenceForm[ "", " "], "q"]],
                              Editable->False]}],
                          SequenceForm[ 
                            Subscript[ "\[PartialD]", "c"], 
                            h[ la, "u$2"]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "b"], 
                        OD[ lc, 
                          h[ la, "u$2"]]],
                      Editable->False], ")"}]}]}], ")"}], " ", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                      SequenceForm[ "", "d"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "d"], "q"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}]}], ")"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(riemann = Absorb[%, \[Eta], {OD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ld]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ lc, 
                h[ lb, ld]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ ld, 
                h[ lb, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              OD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              OD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Eta][ub, ud] riemann\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"(", 
        RowBox[{
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                SequenceForm[ "", "b"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"d\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "d"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "b"], "d"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      h[ lb, ld]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  OD[ lc, 
                    h[ lb, ld]]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                SequenceForm[ "", "b"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "c"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "b"], "c"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "d"], 
                      h[ lb, lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  OD[ ld, 
                    h[ lb, lc]]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"d\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "d"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "d"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      h[ la, ld]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "b"], 
                  OD[ lc, 
                    h[ la, ld]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "c"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "c"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "d"], 
                      h[ la, lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "b"], 
                  OD[ ld, 
                    h[ la, lc]]],
                Editable->False], ")"}]}]}], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Eta]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "b"], "d"],
            Editable->False]],
        Subsuperscript[ \[Eta], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "b"], "d"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricci = Absorb[%, \[Eta], {OD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "b"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ub]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ lc, 
                h[ lb, ub]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"b"\),
                    Superscript[ "\[PartialD]", "b"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "b"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ ub, 
                h[ lb, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              OD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              OD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Eta][ua, uc] ricci\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"(", 
        RowBox[{
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                SequenceForm[ "", "b"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "b"], " "],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"b\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], "b"],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "b"], " "], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], "b"]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      h[ lb, ub]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  OD[ lc, 
                    h[ lb, ub]]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\("\[PartialD]"\^"b"\),
                        Superscript[ "\[PartialD]", "b"],
                        Editable->False], "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                SequenceForm[ "", "b"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "c"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "b"], "c"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Superscript[ "\[PartialD]", "b"], 
                      h[ lb, lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  OD[ ub, 
                    h[ lb, lc]]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"d"\),
                    Superscript[ "\[PartialD]", "d"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"d\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "d"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "d"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      h[ la, ld]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "d"], 
                  OD[ lc, 
                    h[ la, ld]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"d"\),
                    Superscript[ "\[PartialD]", "d"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "c"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "c"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "d"], 
                      h[ la, lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "d"], 
                  OD[ ld, 
                    h[ la, lc]]],
                Editable->False], ")"}]}]}], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Eta]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"],
            Editable->False]],
        Subsuperscript[ \[Eta], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", "a"], "c"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scalar = Dum[Absorb[%, \[Eta], {OD}]] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{"\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  h[ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "u$2", 
                h[ "l$2", "l$1"]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricci\  - \ \(1\/2\) Metricg[la, lc] scalar\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{"\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"p"\),
                      Superscript[ "\[PartialD]", "p"],
                      Editable->False], "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"q"\),
                                  SequenceForm[ "", "q"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "q"], " "],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"q\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], "q"],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "q"], " "], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], "q"]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "p"], 
                        h[ "l$2", "u$2"]],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "p"], 
                    OD[ "l$1", 
                      h[ "l$2", "u$2"]]],
                  Editable->False], ")"}]}], "-", 
            RowBox[{"\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"p"\),
                      Superscript[ "\[PartialD]", "p"],
                      Editable->False], "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\("\[PartialD]"\^"q"\),
                          Superscript[ "\[PartialD]", "q"],
                          Editable->False], "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"q"\),
                                  SequenceForm[ "", "q"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"p\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "q"], "p"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "q"], "p"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Superscript[ "\[PartialD]", "q"], 
                        h[ "l$2", "l$1"]],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "p"], 
                    OD[ "u$2", 
                      h[ "l$2", "l$1"]]],
                  Editable->False], ")"}]}]}], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "b"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ub]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ lc, 
                h[ lb, ub]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"b"\),
                    Superscript[ "\[PartialD]", "b"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "b"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ ub, 
                h[ lb, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              OD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              OD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(einstein = 
      Normal[Series[
            ApplyRules[%, Join[{metricRule}, etaRule]], {\[Epsilon], 0, 1}]] // 
        Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "l$1", 
                h[ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ lc, 
                h[ la, "l$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ lc, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ "u$1", 
                h[ lc, "l$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ "l$1", "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ lc, 
                h[ "l$1", "u$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  h[ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "u$2", 
                h[ "l$2", "l$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Bianchi identities", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvRule = ObjectRule[RiemannR[la_, lb_, lc_, ld_], riemann]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["R", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                        SequenceForm[ "", 
                          Pattern[ la, 
                            Blank[ ]]],
                        Editable->False], "\[InvisibleSpace]", "lb_"}],
                    SequenceForm[ 
                      SequenceForm[ "", 
                        Pattern[ la, 
                          Blank[ ]]], 
                      Pattern[ lb, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lc_"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]], 
                    Pattern[ lb, 
                      Blank[ ]]], 
                  Pattern[ lc, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "ld_"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]], 
                Pattern[ lc, 
                  Blank[ ]]], 
              Pattern[ ld, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]], 
              Pattern[ lc, 
                Blank[ ]]], 
            Pattern[ ld, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                  SequenceForm[ "", "b"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"d\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "b"], "d"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "d"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "c"], 
                        h[ lb, ld]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "a"], 
                    OD[ lc, 
                      h[ lb, ld]]],
                  Editable->False], ")"}]}], "-", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                  SequenceForm[ "", "b"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"c\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "b"], "c"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "c"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "d"], 
                        h[ lb, lc]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "a"], 
                    OD[ ld, 
                      h[ lb, lc]]],
                  Editable->False], ")"}]}], "-", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                  SequenceForm[ "", "a"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"d\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "a"], "d"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "d"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "c"], 
                        h[ la, ld]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "b"], 
                    OD[ lc, 
                      h[ la, ld]]],
                  Editable->False], ")"}]}], "+", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                  SequenceForm[ "", "a"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"c\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "a"], "c"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "c"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "d"], 
                        h[ la, lc]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "b"], 
                    OD[ ld, 
                      h[ la, lc]]],
                  Editable->False], ")"}]}]}], "]"}], "/;", 
        "True"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(bianchi1 = 
      Antisymmetrize[{la, lb, lc}, OD[la, RiemannR[lb, lc, ld, le]]] // 
        Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/3\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                SequenceForm[ "", "b"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "b"], "c"], "d"],
                        Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"], "d"], "e"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], "d"], "e"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              RiemannR[ lb, lc, ld, le]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/3\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "c"], "d"],
                        Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"], "d"], "e"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], "d"], "e"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              RiemannR[ la, lc, ld, le]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/3\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"b\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "b"], "d"],
                        Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"], "d"], "e"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], "d"], "e"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "c"], 
              RiemannR[ la, lb, ld, le]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell["The Bianchi identy is satisfied", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[bianchi1, Join[{curvRule}, Join[{metricRule}, etaRule]]] // 
      Can\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Gauge transformations", "Subsection"],

Cell[BoxData[
    \(DefineTensor[\[Xi], "\<a\>"]\)], "Input"],

Cell[TextData[{
  "Under the coordinate transformation ",
  Cell[BoxData[
      \(TraditionalForm\`x'\^a\  = \ x\^a\  + \ \[Epsilon]\ \[Xi]\^a\)]],
  ", the metric transforms ",
  Cell[BoxData[
      \(TraditionalForm\`\(h\_\(a\[InvisibleSpace]b\)\%'\)\  = \ 
        h\_\(a\[InvisibleSpace]b\)\  - \ 
          2 \[PartialD]\_\((a\)\[Xi]\_\(\(b\)\()\)\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(gaugeTr = 
      ObjectRule[h[la_, lb_], 
        h[la, lb]\  - \ OD[la, \[Xi][lb]]\  - \ OD[lb, \[Xi][la]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["h", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "lb_"}],
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ h, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", 
              Pattern[ la, 
                Blank[ ]]], 
            Pattern[ lb, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox["h", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ h, 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False], "-", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Xi]", 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False]],
                  Subsuperscript[ \[Xi], 
                    SequenceForm[ "", "b"], 
                    SequenceForm[ "", " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[Xi][ lb]],
              Editable->False], "-", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Xi]", 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False]],
                  Subsuperscript[ \[Xi], 
                    SequenceForm[ "", "a"], 
                    SequenceForm[ "", " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[Xi][ la]],
              Editable->False]}], "]"}], "/;", "True"}]}]], "Output"]
}, Open  ]],

Cell["\<\
The Riemann and the Ricci tensors are invariant under the gauge \
transformation\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(riemann\  - \ ApplyRules[riemann, {gaugeTr}] // Can\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ricci\  - \ ApplyRules[ricci, {gaugeTr}]\  // \ Dum\) // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "c"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "p"], 
                      \[Xi][ lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  OD[ "l$1", 
                    \[Xi][ lc]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ la, 
                OD[ "l$1", 
                  \[Xi][ lc]]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "p"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      \[Xi][ "l$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  OD[ lc, 
                    \[Xi][ "l$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ la, 
                OD[ lc, 
                  \[Xi][ "l$1"]]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "c"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "p"], 
                      \[Xi][ lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  OD[ "l$1", 
                    \[Xi][ lc]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ "u$1", 
                OD[ "l$1", 
                  \[Xi][ lc]]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "p"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      \[Xi][ "l$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  OD[ lc, 
                    \[Xi][ "l$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ "u$1", 
                OD[ lc, 
                  \[Xi][ "l$1"]]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", " "], 
                          SequenceForm[ "", "p"]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "p"], 
                      \[Xi][ "u$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  OD[ "l$1", 
                    \[Xi][ "u$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ lc, 
                OD[ "l$1", 
                  \[Xi][ "u$1"]]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\("\[PartialD]"\^"p"\),
                        Superscript[ "\[PartialD]", "p"],
                        Editable->False], "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "p"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Superscript[ "\[PartialD]", "p"], 
                      \[Xi][ "l$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  OD[ "u$1", 
                    \[Xi][ "l$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ lc, 
                OD[ "u$1", 
                  \[Xi][ "l$1"]]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dnup[%, {OD}]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Let ",
  Cell[BoxData[
      \(TraditionalForm\`\[Psi]\_\(a\[InvisibleSpace]b\)\  = \ 
        h\_\(a\[InvisibleSpace]b\)\  - \ \(1\/2\) \(\[Eta]\_\(a\
\[InvisibleSpace]b\)\) h\)]]
}], "Text"],

Cell[BoxData[
    \(DefineTensor[\[Psi], "\<ba\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(psiRule\  = \ {ObjectRule[
          h[la_, lb_], \[Psi][la, lb]\  + \ \(1\/2\) \[Eta][la, lb] 
              h[lc, uc]], \ 
        ObjectRule[\[Psi][la_, ua_], \(-h[la, ua]\), 
          PairQ[{la, ua}]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["h", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lb_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ h, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]]],
            Editable->False], "\[RuleDelayed]", 
          RowBox[{
            RowBox[{"DumFresh", "[", 
              RowBox[{
                RowBox[{\(1\/2\), " ", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$\
24"\),
                            SequenceForm[ "", "$24"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"   \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "$24"], "   "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"  \
 "\),
                            SequenceForm[ "", "   "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"$24\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "   "], "$24"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "$24"], "   "], 
                      SequenceForm[ 
                        SequenceForm[ "", "   "], "$24"]],
                    Editable->False], " ", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Eta]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Eta], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}], "+", 
                InterpretationBox[
                  SubsuperscriptBox["\[Psi]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Psi], 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}], "]"}], "/;", "True"}]}], ",", 
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Psi]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "ua_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ ua, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ \[Psi], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ ua, 
                  Blank[ ]]]],
            Editable->False], "\[RuleDelayed]", 
          RowBox[{
            RowBox[{"DumFresh", "[", 
              RowBox[{"-", 
                InterpretationBox[
                  SubsuperscriptBox["h", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$\
25"\),
                          SequenceForm[ "", "$25"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"   \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "$25"], "   "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"   \
"\),
                          SequenceForm[ "", "   "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"$25\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "   "], "$25"],
                      Editable->False]],
                  Subsuperscript[ h, 
                    SequenceForm[ 
                      SequenceForm[ "", "$25"], "   "], 
                    SequenceForm[ 
                      SequenceForm[ "", "   "], "$25"]],
                  Editable->False]}], "]"}], "/;", \(PairQ[{la, ua}]\)}]}]}], 
      "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Under the gauge transformation ",
  Cell[BoxData[
      \(TraditionalForm\`\[Psi]\_\(a\[InvisibleSpace]b\)\)]],
  " transfoms as"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[h[la, lb]\  - \ \(1\/2\) \[Eta][la, lb] h[lc, uc], 
        Join[{gaugeTr}, psiRule, etaRule]] // \ Dnup[#, {OD}] &\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              \[Xi][ lb]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", "a"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "b"], 
          \[Xi][ la]],
        Editable->False], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["h", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$28"\),
                  SequenceForm[ "", "$28"],
                  Editable->False], "\[InvisibleSpace]", "\<\"   \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "$28"], "   "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"   "\),
                  SequenceForm[ "", "   "],
                  Editable->False], "\[InvisibleSpace]", "\<\"$28\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "   "], "$28"],
              Editable->False]],
          Subsuperscript[ h, 
            SequenceForm[ 
              SequenceForm[ "", "$28"], "   "], 
            SequenceForm[ 
              SequenceForm[ "", "   "], "$28"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["h", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$30"\),
                  SequenceForm[ "", "$30"],
                  Editable->False], "\[InvisibleSpace]", "\<\"   \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "$30"], "   "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"   "\),
                  SequenceForm[ "", "   "],
                  Editable->False], "\[InvisibleSpace]", "\<\"$30\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "   "], "$30"],
              Editable->False]],
          Subsuperscript[ h, 
            SequenceForm[ 
              SequenceForm[ "", "$30"], "   "], 
            SequenceForm[ 
              SequenceForm[ "", "   "], "$30"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["h", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$32"\),
                  SequenceForm[ "", "$32"],
                  Editable->False], "\[InvisibleSpace]", "\<\"   \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "$32"], "   "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"   "\),
                  SequenceForm[ "", "   "],
                  Editable->False], "\[InvisibleSpace]", "\<\"$32\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "   "], "$32"],
              Editable->False]],
          Subsuperscript[ h, 
            SequenceForm[ 
              SequenceForm[ "", "$32"], "   "], 
            SequenceForm[ 
              SequenceForm[ "", "   "], "$32"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"$26"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"   "\),
                    SequenceForm[ "", "   "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$26"\),
                    SequenceForm[ "", "$26"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", "   "], 
                  SequenceForm[ "", "$26"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "$26"], 
              \[Xi][ "u$26"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      InterpretationBox[
        SubsuperscriptBox["\[Psi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ \[Psi], 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              \[Xi][ lb]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", "a"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "b"], 
          \[Xi][ la]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              \[Xi][ "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      InterpretationBox[
        SubsuperscriptBox["\[Psi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ \[Psi], 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "from which we find ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\_a \[Psi]'\^\(a\[InvisibleSpace]b\)\  = \
\ \[PartialD]\_a \[Psi]\^\(a\[InvisibleSpace]b\)\  - \ \[PartialD]\^2 \
\[Xi]\^b\)]]
}], "Text"],

Cell["The Einstein tensor becomes", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[einstein, Join[psiRule, etaRule]] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  \[Psi][ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "l$1", 
                \[Psi][ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  \[Psi][ la, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ lc, 
                \[Psi][ la, "l$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  \[Psi][ lc, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ "u$1", 
                \[Psi][ lc, "l$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ "l$1", "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ lc, 
                h[ "l$1", "u$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "c"], 
              OD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  \[Psi][ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              OD[ "u$2", 
                \[Psi][ "l$2", "l$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"r"\),
                            SequenceForm[ "", "r"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "r"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"r\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "r"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "r"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "r"]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  h[ "l$3", "u$3"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              OD[ "u$2", 
                h[ "l$3", "u$3"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ "u$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq = \(Absorb[%, \[Eta], {OD}] // Can\) // Dnup[#, {OD}] &\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  \[Psi][ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              OD[ "u$1", 
                \[Psi][ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  \[Psi][ lc, "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              OD[ "l$1", 
                \[Psi][ lc, "u$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  \[Psi][ la, "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "c"], 
              OD[ "l$1", 
                \[Psi][ la, "u$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"q"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "q"], 
                  \[Psi][ "u$2", "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              OD[ "l$2", 
                \[Psi][ "u$2", "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, if we choose the Einstein gauge ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\_a \[Psi]\^\(a\[InvisibleSpace]b\)\  = \ \
\(\[PartialD]\_a h\^\(a\[InvisibleSpace]b\)\  - \ \(1\/2\) \[PartialD]\^b 
                h = \ 0\)\)]],
  ", we have the wave equation"
}], "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`G\_\(a\[InvisibleSpace]b\)\  = \ \(\(\[Epsilon]\/2\) \
\[PartialD]\^2 \[Psi]\_\(a\[InvisibleSpace]b\)\  = \ \(-\[Kappa]\)\ T\_\(a\
\[InvisibleSpace]b\)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell[BoxData[
    \(einsteinGauge = 
      ObjectRule[OD[ua_, \[Psi][lb_, lc_]], 0, 
        PairQ[{ua, lb}] || PairQ[{ua, lc}]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{
          InterpretationBox[\("\[PartialD]"\^ua_\),
            Superscript[ "\[PartialD]", 
              Pattern[ ua, 
                Blank[ ]]],
            Editable->False], "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Psi]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]lb_\),
                    SequenceForm[ "", 
                      Pattern[ lb, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lc_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ lb, 
                      Blank[ ]]], 
                  Pattern[ lc, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ \[Psi], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ lb, 
                    Blank[ ]]], 
                Pattern[ lc, 
                  Blank[ ]]]],
            Editable->False]}],
        SequenceForm[ 
          Superscript[ "\[PartialD]", 
            Pattern[ ua, 
              Blank[ ]]], 
          \[Psi][ 
            Pattern[ lb, 
              Blank[ ]], 
            Pattern[ lc, 
              Blank[ ]]]],
        Editable->False], 
      "\[RuleDelayed]", \(DumFresh[0] /; 
        PairQ[{ua, lb}] || PairQ[{ua, lc}]\)}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[eq, einsteinGauge]\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Psi]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Psi], 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[Psi][ la, lc]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "p"], 
            OD[ "u$1", 
              \[Psi][ la, lc]]],
          Editable->False], ")"}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "The gauge is not completely fixed because we can always carry out \
additional transformation with ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\^2 \[Xi]\^a\  = \ 0\)]],
  ", which leaves the Einstein gauge unaltered."
}], "Text"],

Cell[TextData[{
  "In vacuum, the Einstein equation becomes ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\^2 h\_\(a\[InvisibleSpace]b\)\  = \ 
        0\)]],
  " in the Einstein gauge ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\_a h\_b\^a\  - \ \(1\/2\) \[PartialD]\_b 
              h\  = \ 0\)]]
}], "Text"],

Cell[TextData[{
  "The vacuum equation implies ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\^2 R\_\(a\[InvisibleSpace]b\
\[InvisibleSpace]c\[InvisibleSpace]d\)\  = \ 0\)]],
  " which obeys the wave equation. This implies that, in the linealized \
theory, gravitational effects propagate with the light velocity."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Linearized plane gravitational waves", "Section"],

Cell[TextData[{
  "We choose coordinates ",
  Cell[BoxData[
      \(TraditionalForm\`\((t, x, y, z)\)\)]],
  " and assume ",
  Cell[BoxData[
      \(TraditionalForm\`h\_\(a\[InvisibleSpace]b\)\  = \ \(h\_\(a\
\[InvisibleSpace]b\)\)(t, x)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(assumeRule = {ObjectRule[OD[ld_, OD[lc_, h[la_, lb_]]], 
          0, \((lc\  === \ \(-3\)\  || \ lc\  === \ \(-4\))\)], 
        ObjectRule[OD[lc_, OD[ld_, h[la_, lb_]]], 
          0, \((lc\  === \ \(-3\)\  || \ lc\  === \ \(-4\))\)]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^ld_\),
                Superscript[ "\[PartialD]", 
                  Pattern[ ld, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^lc_\),
                    Superscript[ "\[PartialD]", 
                      Pattern[ lc, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\
\),
                            SequenceForm[ "", 
                              Pattern[ la, 
                                Blank[ ]]],
                            Editable->False], "\[InvisibleSpace]", "lb_"}],
                        SequenceForm[ 
                          SequenceForm[ "", 
                            Pattern[ la, 
                              Blank[ ]]], 
                          Pattern[ lb, 
                            Blank[ ]]],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 
                          Pattern[ la, 
                            Blank[ ]]], 
                        Pattern[ lb, 
                          Blank[ ]]]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", 
                    Pattern[ lc, 
                      Blank[ ]]], 
                  h[ 
                    Pattern[ la, 
                      Blank[ ]], 
                    Pattern[ lb, 
                      Blank[ ]]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", 
                Pattern[ ld, 
                  Blank[ ]]], 
              OD[ 
                Pattern[ lc, 
                  Blank[ ]], 
                h[ 
                  Pattern[ la, 
                    Blank[ ]], 
                  Pattern[ lb, 
                    Blank[ ]]]]],
            Editable->False], 
          "\[RuleDelayed]", \(DumFresh[0] /; 
            lc === \(-3\) || lc === \(-4\)\)}], ",", 
        RowBox[{
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^lc_\),
                Superscript[ "\[PartialD]", 
                  Pattern[ lc, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^ld_\),
                    Superscript[ "\[PartialD]", 
                      Pattern[ ld, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\
\),
                            SequenceForm[ "", 
                              Pattern[ la, 
                                Blank[ ]]],
                            Editable->False], "\[InvisibleSpace]", "lb_"}],
                        SequenceForm[ 
                          SequenceForm[ "", 
                            Pattern[ la, 
                              Blank[ ]]], 
                          Pattern[ lb, 
                            Blank[ ]]],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 
                          Pattern[ la, 
                            Blank[ ]]], 
                        Pattern[ lb, 
                          Blank[ ]]]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", 
                    Pattern[ ld, 
                      Blank[ ]]], 
                  h[ 
                    Pattern[ la, 
                      Blank[ ]], 
                    Pattern[ lb, 
                      Blank[ ]]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", 
                Pattern[ lc, 
                  Blank[ ]]], 
              OD[ 
                Pattern[ ld, 
                  Blank[ ]], 
                h[ 
                  Pattern[ la, 
                    Blank[ ]], 
                  Pattern[ lb, 
                    Blank[ ]]]]],
            Editable->False], 
          "\[RuleDelayed]", \(DumFresh[0] /; 
            lc === \(-3\) || lc === \(-4\)\)}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(riemannComp\  = \ 
        Table[riemann, {la, \(-1\), \(-4\), \(-1\)}, {lb, \(-1\), \(-4\), \
\(-1\)}, {lc, \(-1\), \(-4\), \(-1\)}, {ld, \(-1\), \(-4\), \(-1\)}];\)\)], \
"Input"],

Cell["The 20 component Riemann tensors fall into the three groups", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(group1 = 
      ApplyRules[{riemannComp[\([1, 2, 3, 4]\)], 
          riemannComp[\([1, 3, 3, 4]\)], riemannComp[\([1, 4, 3, 4]\)], 
          riemannComp[\([2, 3, 3, 4]\)], riemannComp[\([2, 4, 3, 4]\)], 
          riemannComp[\([3, 4, 3, 4]\)]}, assumeRule]\)], "Input"],

Cell[BoxData[
    \({0, 0, 0, 0, 0, 0}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(group2 = 
      ApplyRules[{riemannComp[\([1, 2, 1, 2]\)], 
          riemannComp[\([1, 2, 1, 3]\)], riemannComp[\([1, 2, 1, 4]\)], 
          riemannComp[\([1, 2, 2, 3]\)], riemannComp[\([1, 2, 2, 4]\)]}, 
        assumeRule]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "1"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 1],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 1], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -1]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 2], 
                  OD[ -2, 
                    h[ -1, -1]]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "1"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 1],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 1], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -2, -1]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -2, 
                    h[ -2, -1]]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "2"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 2],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 2], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -2]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -2, 
                    h[ -1, -2]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "2"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 2],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 2], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 1], 
                      h[ -2, -2]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -1, 
                    h[ -2, -2]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -2, 
                    h[ -1, -3]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 1], 
                      h[ -2, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -1, 
                    h[ -2, -3]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -2, 
                    h[ -1, -4]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 1], 
                      h[ -2, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -1, 
                    h[ -2, -4]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 2], 
                  OD[ -2, 
                    h[ -1, -3]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -2, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -2, 
                    h[ -2, -3]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 2], 
                  OD[ -2, 
                    h[ -1, -4]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -2, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  OD[ -2, 
                    h[ -2, -4]]],
                Editable->False], ")"}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(group3 = 
      ApplyRules[{riemannComp[\([1, 3, 1, 3]\)], 
          riemannComp[\([1, 3, 1, 4]\)], riemannComp[\([1, 3, 2, 3]\)], 
          riemannComp[\([1, 3, 2, 4]\)], riemannComp[\([1, 4, 1, 4]\)], 
          riemannComp[\([1, 4, 2, 4]\)], riemannComp[\([2, 3, 2, 3]\)], 
          riemannComp[\([2, 3, 2, 4]\)], riemannComp[\([2, 4, 2, 4]\)]}, 
        assumeRule]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "3"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 3],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 3], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 1], 
                    h[ -3, -3]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                OD[ -1, 
                  h[ -3, -3]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 1], 
                    h[ -3, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                OD[ -1, 
                  h[ -3, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "3"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 3],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 3], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -3]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                OD[ -2, 
                  h[ -3, -3]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                OD[ -2, 
                  h[ -3, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
4\),
                              SequenceForm[ "", 4],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 4], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 4], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 1], 
                    h[ -4, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                OD[ -1, 
                  h[ -4, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
4\),
                              SequenceForm[ "", 4],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 4], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 4], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -4, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                OD[ -2, 
                  h[ -4, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "3"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 3],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 3], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -3]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 2], 
                OD[ -2, 
                  h[ -3, -3]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 2], 
                OD[ -2, 
                  h[ -3, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
4\),
                              SequenceForm[ "", 4],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 4], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 4], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -4, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 2], 
                OD[ -2, 
                  h[ -4, -4]]],
              Editable->False], ")"}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "The vacuum equation ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\)\  = \ 0\)]],
  " and group 1 result in group2 vanishing. Thus, only the group3 are \
non-zero and these only involve the components ",
  Cell[BoxData[
      \(TraditionalForm\`h\_33\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`h\_34\)]],
  ", and ",
  Cell[BoxData[
      \(TraditionalForm\`h\_44\)]],
  "."
}], "Text"],

Cell[TextData[{
  "With the ansatz ",
  Cell[BoxData[
      \(TraditionalForm\`h\_\(a\[InvisibleSpace]b\)\  = \ \(h\_\(a\
\[InvisibleSpace]b\)\)(t - x)\)]],
  " and the asymptotic flatness the Einstein gauge condition gives"
}], "Text"],

Cell[TextData[Cell[BoxData[
    FormBox[
      RowBox[{\(h\_\(a\[InvisibleSpace]b\)\), " ", "=", " ", 
        RowBox[{"(", GridBox[{
              {"0", "0", "0", "0"},
              {"0", "0", "0", "0"},
              {"0", "0", \(h\_33\), \(h\_34\)},
              {"0", "0", \(h\_34\), \(-h\_33\)}
              }], ")"}]}], TraditionalForm]]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "That is, ",
  Cell[BoxData[
      \(TraditionalForm\`h\_\(a\[InvisibleSpace]b\)\)]],
  " only depends on two functions ",
  Cell[BoxData[
      \(TraditionalForm\`h\_33\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`h\_34\)]]
}], "Text"]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{695, 680},
WindowMargins->{{19, Automatic}, {Automatic, 5}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 93, 1, 43, "Title"],
Cell[1813, 52, 96, 3, 41, "Text"],
Cell[1912, 57, 59, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1996, 62, 32, 0, 78, "Section"],
Cell[2031, 64, 73, 1, 56, "Input"],
Cell[2107, 67, 194, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[2326, 76, 148, 4, 79, "Input"],
Cell[2477, 82, 3535, 87, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6049, 174, 148, 2, 79, "Input"],
Cell[6200, 178, 3576, 92, 78, "Output"]
}, Open  ]],
Cell[9791, 273, 259, 7, 41, "Text"],

Cell[CellGroupData[{
Cell[10075, 284, 32, 0, 66, "Subsection"],
Cell[10110, 286, 110, 2, 56, "Input"],

Cell[CellGroupData[{
Cell[10245, 292, 86, 1, 56, "Input"],
Cell[10334, 295, 5337, 125, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15708, 425, 83, 1, 56, "Input"],
Cell[15794, 428, 11111, 250, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[26942, 683, 70, 1, 56, "Input"],
Cell[27015, 686, 8236, 189, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[35288, 880, 65, 1, 56, "Input"],
Cell[35356, 883, 4403, 100, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[39796, 988, 88, 1, 56, "Input"],
Cell[39887, 991, 7494, 178, 94, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[47430, 1175, 31, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[47486, 1179, 80, 1, 56, "Input"],
Cell[47569, 1182, 12954, 306, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[60560, 1493, 154, 4, 125, "Input"],
Cell[60717, 1499, 16863, 341, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[77617, 1845, 66, 1, 56, "Input"],
Cell[77686, 1848, 7645, 181, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[85368, 2034, 55, 1, 56, "Input"],
Cell[85426, 2037, 9306, 207, 94, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[94769, 2249, 64, 1, 56, "Input"],
Cell[94836, 2252, 8014, 190, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[102887, 2447, 53, 1, 56, "Input"],
Cell[102943, 2450, 9711, 216, 94, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[112691, 2671, 77, 1, 56, "Input"],
Cell[112771, 2674, 4208, 100, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[117016, 2779, 76, 1, 72, "Input"],
Cell[117095, 2782, 13811, 316, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[130943, 3103, 158, 4, 125, "Input"],
Cell[131104, 3109, 14199, 337, 110, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[145352, 3452, 40, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[145417, 3456, 93, 1, 79, "Input"],
Cell[145513, 3459, 11827, 267, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[157377, 3731, 129, 3, 102, "Input"],
Cell[157509, 3736, 9420, 211, 69, "Output"]
}, Open  ]],
Cell[166944, 3950, 47, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[167016, 3954, 114, 2, 102, "Input"],
Cell[167133, 3958, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[167217, 3965, 43, 0, 38, "Subsection"],
Cell[167263, 3967, 61, 1, 56, "Input"],
Cell[167327, 3970, 370, 9, 43, "Text"],

Cell[CellGroupData[{
Cell[167722, 3983, 141, 3, 79, "Input"],
Cell[167866, 3988, 4016, 98, 55, "Output"]
}, Open  ]],
Cell[171897, 4089, 104, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[172026, 4096, 84, 1, 56, "Input"],
Cell[172113, 4099, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[172185, 4105, 102, 2, 56, "Input"],
Cell[172290, 4109, 10528, 250, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[182855, 4364, 46, 1, 56, "Input"],
Cell[182904, 4367, 35, 1, 53, "Output"]
}, Open  ]],
Cell[182954, 4371, 211, 6, 44, "Text"],
Cell[183168, 4379, 63, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[183256, 4384, 239, 5, 149, "Input"],
Cell[183498, 4391, 8263, 200, 95, "Output"]
}, Open  ]],
Cell[191776, 4594, 160, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[191961, 4603, 159, 2, 99, "Input"],
Cell[192123, 4607, 10320, 254, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[202480, 4866, 41, 1, 56, "Input"],
Cell[202524, 4869, 4318, 107, 54, "Output"]
}, Open  ]],
Cell[206857, 4979, 232, 6, 41, "Text"],
Cell[207092, 4987, 43, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[207160, 4991, 84, 1, 56, "Input"],
Cell[207247, 4994, 24195, 574, 151, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[231479, 5573, 91, 1, 56, "Input"],
Cell[231573, 5576, 8867, 208, 69, "Output"]
}, Open  ]],
Cell[240455, 5787, 310, 7, 44, "Text"],
Cell[240768, 5796, 267, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[241060, 5805, 141, 3, 79, "Input"],
Cell[241204, 5810, 1952, 54, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[243193, 5869, 62, 1, 56, "Input"],
Cell[243258, 5872, 1983, 48, 69, "Output"]
}, Open  ]],
Cell[245256, 5923, 259, 6, 66, "Text"],
Cell[245518, 5931, 331, 9, 44, "Text"],
Cell[245852, 5942, 335, 7, 66, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[246236, 5955, 55, 0, 50, "Section"],
Cell[246294, 5957, 253, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[246572, 5969, 269, 4, 148, "Input"],
Cell[246844, 5975, 6226, 160, 78, "Output"]
}, Open  ]],
Cell[253085, 6138, 203, 4, 79, "Input"],
Cell[253291, 6144, 75, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[253391, 6148, 292, 5, 148, "Input"],
Cell[253686, 6155, 52, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[253775, 6161, 259, 5, 148, "Input"],
Cell[254037, 6168, 24367, 535, 151, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[278441, 6708, 405, 7, 194, "Input"],
Cell[278849, 6717, 17448, 398, 151, "Output"]
}, Open  ]],
Cell[296312, 7118, 438, 15, 66, "Text"],
Cell[296753, 7135, 236, 6, 41, "Text"],
Cell[296992, 7143, 406, 10, 95, "Text"],
Cell[297401, 7155, 266, 10, 41, "Text"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

