(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    450902,      11111]*)
(*NotebookOutlinePosition[    454386,      11200]*)
(*  CellTagsIndexPosition[    454342,      11196]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Stationary SpaceTime", "Title",
  TextAlignment->Left,
  TextJustification->0,
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`ThreeOne`\)], "Input"],

Cell[CellGroupData[{

Cell["Defined Tensors", "Section"],

Cell[CellGroupData[{

Cell["GRClass", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \({RiemannR[l\[Mu], l\[Nu], l\[Rho], l\[Sigma]], RicciR[l\[Mu], l\[Nu]], 
      ScalarR[], AffineG[l\[Mu], l\[Nu], u\[Rho]], Metricg[l\[Mu], l\[Nu]], 
      Structuref[l\[Mu], l\[Nu], u\[Rho]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\
\[Mu]"\),
                          SequenceForm[ "", "\[Mu]"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Rho]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
              
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Rho]"], 
                "\[Sigma]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], "\[Rho]"], 
              "\[Sigma]"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\),
                  SequenceForm[ "", "\[Mu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Rho]"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Rho]"]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\),
                  SequenceForm[ "", "\[Mu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "\[Mu]"], "\[Nu]"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Mu]"\)\
,
                      SequenceForm[ "", "\[Mu]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Nu]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Mu]"], "\[Nu]"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Rho]\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "\[Rho]"],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "\[Mu]"], "\[Nu]"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "\[Rho]"]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["BaseClass", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \({BaseRiemannR[li, lj, lk, ll], BaseRicciR[li, lj], BaseScalarR[], 
      BaseAffineG[li, lj, uk], BaseMetricg[li, lj], GaugeF[li, lj], 
      GaugeA[li]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"j\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "j"],
                      Editable->False], "\[InvisibleSpace]", "\<\"k\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"], "k"],
                  Editable->False], "\[InvisibleSpace]", "\<\"l\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"], "k"], "l"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], "k"], "l"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(R\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ R], 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[\(\(R\&~\)\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ 
            OverTilde[ R], "", ""],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"k\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "k"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "k"]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], ",", 
        InterpretationBox[
          SubsuperscriptBox["\[Omega]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
              SequenceForm[ "", "i"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Omega], 
            SequenceForm[ "", "i"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Basis Derivative", "Section"],

Cell[TextData[{
  "Notice that ",
  Cell[BoxData[
      \(TraditionalForm\`D\_\[Mu]\  = \ \(\((D\_t, \ 
            D\_i)\)\  = \ \((\[PartialD]\_t\(\(,\)\(\ \)\(\[PartialD]\_i\ \(-\
\ \[Omega]\_i\) \[PartialD]\_t\)\))\)\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(BD[l\[Mu], l\[Nu], RicciR[l\[Rho], l\[Sigma]]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("D"\_"\[Mu]"\), "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{\("D"\_"\[Nu]"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["R", 
                InterpretationBox[
                  RowBox[{
                    
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"\[Rho]"\
\),
                      SequenceForm[ "", "\[Rho]"],
                      Editable->False], 
                    "\[InvisibleSpace]", "\<\"\[Sigma]\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "\[Rho]"], "\[Sigma]"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ R, 
                SequenceForm[ 
                  SequenceForm[ "", "\[Rho]"], "\[Sigma]"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "\[Nu]"], 
            RicciR[ l\[Rho], l\[Sigma]]],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "D", "\[Mu]"], 
        BD[ l\[Nu], 
          RicciR[ l\[Rho], l\[Sigma]]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Block-Diagonal Basis", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[
      TableKK[Metricg[l\[Mu], l\[Nu]], {l\[Mu], l\[Nu]}] /. 
        BlockDiagonalRule]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              RowBox[{"-", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}], "0"},
            {"0", 
              FractionBox[
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False], 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]]}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[
      TableKK[Metricg[u\[Mu], u\[Nu]], {u\[Mu], u\[Nu]}] /. 
        BlockDiagonalRule]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              RowBox[{"-", 
                FractionBox["1", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]]}], "0"},
            {"0", 
              RowBox[{
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"]],
                  Editable->False], " ", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}]}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[
      TableKK[Metricg[l\[Mu], u\[Nu]], {l\[Mu], u\[Nu]}] /. 
        BlockDiagonalRule]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"1", "0"},
            {"0", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "b"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "b"]],
                Editable->False]}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableKK[
        Structuref[l\[Mu], l\[Nu], u\[Rho]], {l\[Mu], l\[Nu], u\[Rho]}] /. 
      BlockDiagonalRule\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\({{0, 0}, {0, 0}}\), ",", 
        RowBox[{"{", 
          RowBox[{\({0, 0}\), ",", 
            RowBox[{"{", 
              RowBox[{
                RowBox[{"-", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Pi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ Pi, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}], ",", "0"}], "}"}]}], "}"}]}], 
      "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableKK[
        Structuref[l\[Mu], l\[Nu], l\[Rho]], {l\[Mu], l\[Nu], l\[Rho]}] /. 
      BlockDiagonalRule\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\({{0, 0}, {0, 0}}\), ",", 
        RowBox[{"{", 
          RowBox[{\({0, 0}\), ",", 
            RowBox[{"{", 
              RowBox[{
                RowBox[{
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False], " ", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Pi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ Pi, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}], ",", "0"}], "}"}]}], "}"}]}], 
      "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Stationary SpaceTime", "Section"],

Cell[BoxData[
    \(\(StationaryRule = {BD[l0, expr_] \[RuleDelayed] \ 0, 
          BD[u0, expr_] \[RuleDelayed] \ 0};\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Connection", "Section"],

Cell[BoxData[
    \(affineUpG[la_, lb_, uc_] := 
      SumDumKK[AffineG[la, lb, uc]\  // \ AffineToMetric]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(affineUpG[li, lj, uk] //. BlockDiagonalRule // 
        Absorb[#, BaseMetricg] &\) // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"k"\),
                  SequenceForm[ "", "k"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "k"], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "k"], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                        SequenceForm[ "", "j"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "j"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "j"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "i"], 
              BaseMetricg[ lj, "l$1"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"k"\),
                  SequenceForm[ "", "k"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "k"], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "k"], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "j"], 
              BaseMetricg[ li, "l$1"]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"k"\),
                Superscript[ "D", "k"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "k"], 
              BaseMetricg[ li, lj]],
            Editable->False], ")"}]}], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                    SequenceForm[ "", "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"k\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "k"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "j"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "k"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"k\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "k"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "k"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"k"\),
                  Superscript[ "D", "k"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "k"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[la_, lb_, uc_], 
      BaseAffineG[la, lb, 
          uc]\  + \ \(\((BaseMetricg[la, lb] BD[uc, f[]] - 
                BaseMetricg[lb, uc] BD[la, f[]] - 
                BaseMetricg[la, uc] BD[lb, f[]])\)/2\)/f[], 
      DnIndexQ[la] && NormalIndexQ[la] && DnIndexQ[lb] && NormalIndexQ[lb] && 
        UpIndexQ[uc] && NormalIndexQ[uc]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[li, lj, uk] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\"k\"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "k"],
            Editable->False]],
        Subsuperscript[ 
          OverTilde[ \[CapitalGamma]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], "k"]],
        Editable->False], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                    SequenceForm[ "", "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"k\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "k"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "j"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "k"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"k\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "k"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "k"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"k"\),
                  Superscript[ "D", "k"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "k"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[li, lj, u0] //. BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        FractionBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox["\[Gamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Gamma], 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False], " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  f[ ]],
                Editable->False], ")"}]}], 
          RowBox[{"2", " ", 
            SuperscriptBox[
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False], "3"]}]]}], "+", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Gamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Gamma], 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "0"], 
            BaseMetricg[ li, lj]],
          Editable->False], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[
      AffineG[la_, lb_, 
        uc_], \(-BaseMetricg[la, lb]\) \(BD[l0, f[]]/2\)/
            f[]\^3 + \(BD[l0, BaseMetricg[la, lb]]/2\)/f[]\^2 - 
        GaugeF[la, lb]/2, 
      DnIndexQ[la] && NormalIndexQ[la] && DnIndexQ[lb] && NormalIndexQ[lb] && 
        UpIndexQ[uc] && OneIndexQ[uc]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[li, lj, u0]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        FractionBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox["\[Gamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Gamma], 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False], " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  f[ ]],
                Editable->False], ")"}]}], 
          RowBox[{"2", " ", 
            SuperscriptBox[
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False], "3"]}]]}], "+", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["\[Gamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Gamma], 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "0"], 
            BaseMetricg[ li, lj]],
          Editable->False], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(affineUpG[li, l0, uj] //. BlockDiagonalRule // 
        Absorb[#, BaseMetricg] &\) // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                  SequenceForm[ "", "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "j"], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "j"], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ li, "l$1"]],
            Editable->False], ")"}]}], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "j"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "j"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{\(1\/2\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "j"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "j"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[la_, lb_, uc_], 
      BaseMetricg[uc, ud] BD[l0, BaseMetricg[la, ld]]/2 - 
        BaseMetricg[la, uc] \(BD[l0, f[]]/2\)/f[] - \ 
        f[]\^2\ GaugeF[la, uc]/2, 
      DnIndexQ[la] && NormalIndexQ[la] && DnIndexQ[lb] && OneIndexQ[lb] && 
        UpIndexQ[uc] && NormalIndexQ[uc]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[li, l0, uj] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                  SequenceForm[ "", "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "j"], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "j"], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ li, "l$1"]],
            Editable->False], ")"}]}], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "j"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "j"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{\(1\/2\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "j"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "j"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[li, l0, u0] //. BlockDiagonalRule\)], "Input"],

Cell[BoxData[
    FractionBox[
      InterpretationBox[
        RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "i"], 
          f[ ]],
        Editable->False], 
      RowBox[{"2", " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}]]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[la_, lb_, uc_], \(BD[la, f[]]/2\)/f[], 
      DnIndexQ[la] && NormalIndexQ[la] && DnIndexQ[lb] && OneIndexQ[lb] && 
        UpIndexQ[uc] && OneIndexQ[uc]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[li, l0, u0]\)], "Input"],

Cell[BoxData[
    FractionBox[
      InterpretationBox[
        RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "i"], 
          f[ ]],
        Editable->False], 
      RowBox[{"2", " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}]]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(affineUpG[l0, li, uj] //. BlockDiagonalRule // 
        Absorb[#, BaseMetricg] &\) // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                  SequenceForm[ "", "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "j"], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "j"], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ li, "l$1"]],
            Editable->False], ")"}]}], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "j"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "j"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{\(1\/2\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "j"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "j"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[la_, lb_, uc_], 
      BaseMetricg[uc, ud] BD[l0, BaseMetricg[lb, ld]]/2 - 
        BaseMetricg[lb, uc] \(BD[l0, f[]]/2\)/f[] - f[]\^2\ GaugeF[lb, uc]/2, 
      DnIndexQ[la] && OneIndexQ[la] && DnIndexQ[lb] && NormalIndexQ[lb] && 
        UpIndexQ[uc] && NormalIndexQ[uc]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l0, li, uj] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                  SequenceForm[ "", "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "j"], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "j"], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ li, "l$1"]],
            Editable->False], ")"}]}], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "j"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "j"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{\(1\/2\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "j"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "j"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[l0, li, u0] //. BlockDiagonalRule\)], "Input"],

Cell[BoxData[
    FractionBox[
      InterpretationBox[
        RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "i"], 
          f[ ]],
        Editable->False], 
      RowBox[{"2", " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}]]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[la_, lb_, uc_], \(BD[lb, f[]]/2\)/f[], 
      DnIndexQ[la] && OneIndexQ[la] && DnIndexQ[lb] && NormalIndexQ[lb] && 
        UpIndexQ[uc] && OneIndexQ[uc]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l0, li, u0]\)], "Input"],

Cell[BoxData[
    FractionBox[
      InterpretationBox[
        RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "i"], 
          f[ ]],
        Editable->False], 
      RowBox[{"2", " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}]]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(affineUpG[l0, l0, ui] //. BlockDiagonalRule // 
        Absorb[#, BaseMetricg] &\) // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[\("D"\^"i"\),
              Superscript[ "D", "i"],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False]}],
          SequenceForm[ 
            Superscript[ "D", "i"], 
            f[ ]],
          Editable->False], ")"}], " ", 
      InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
        Subsuperscript[ f, "", ""],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[la_, lb_, uc_], BD[uc, f[]]\ *\ f[]/2, 
      DnIndexQ[la] && OneIndexQ[la] && DnIndexQ[lb] && OneIndexQ[lb] && 
        UpIndexQ[uc] && NormalIndexQ[uc]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l0, l0, ui] // Can\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[\("D"\^"i"\),
              Superscript[ "D", "i"],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False]}],
          SequenceForm[ 
            Superscript[ "D", "i"], 
            f[ ]],
          Editable->False], ")"}], " ", 
      InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
        Subsuperscript[ f, "", ""],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineUpG[l0, l0, u0] /. BlockDiagonalRule // Can\)], "Input"],

Cell[BoxData[
    FractionBox[
      InterpretationBox[
        RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "0"], 
          f[ ]],
        Editable->False], 
      RowBox[{"2", " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}]]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ObjectSet[AffineG[la_, lb_, uc_], \(BD[l0, f[]]/2\)/f[], 
      DnIndexQ[la] && DnIndexQ[lb] && UpIndexQ[uc] && 
        AllQ[{la, lb, uc}, OneIndexQ]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineG[l0, l0, u0]\)], "Input"],

Cell[BoxData[
    FractionBox[
      InterpretationBox[
        RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "0"], 
          f[ ]],
        Editable->False], 
      RowBox[{"2", " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}]]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableKK[AffineG[l\[Mu], l\[Nu], ui], {l\[Mu], l\[Nu]}] // 
      Map[Can, #, {2}] &\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{
            RowBox[{\(1\/2\), " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("D"\^"i"\),
                      Superscript[ "D", "i"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "D", "i"], 
                    f[ ]],
                  Editable->False], ")"}], " ", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}], ",", 
            RowBox[{
              RowBox[{\(1\/2\), " ", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"]],
                  Editable->False], " ", 
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Gamma]", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"p\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "p"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ \[Gamma], 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "p"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "D", "0"], 
                      BaseMetricg[ la, "l$1"]],
                    Editable->False], ")"}]}], "-", 
              FractionBox[
                RowBox[{
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"i\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "i"],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "i"]],
                    Editable->False], " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                        
                        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
                          Subsuperscript[ f, "", ""],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "D", "0"], 
                        f[ ]],
                      Editable->False], ")"}]}], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]], "-", 
              RowBox[{\(1\/2\), " ", 
                SuperscriptBox[
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False], "2"], " ", 
                InterpretationBox[
                  SubsuperscriptBox["\[Pi]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"i\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "i"],
                      Editable->False]],
                  Subsuperscript[ Pi, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "i"]],
                  Editable->False]}]}]}], "}"}], ",", 
        RowBox[{"{", 
          RowBox[{
            RowBox[{
              RowBox[{\(1\/2\), " ", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"]],
                  Editable->False], " ", 
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Gamma]", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"p\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "p"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ \[Gamma], 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "p"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "D", "0"], 
                      BaseMetricg[ la, "l$1"]],
                    Editable->False], ")"}]}], "-", 
              FractionBox[
                RowBox[{
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"i\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "i"],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "i"]],
                    Editable->False], " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                        
                        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
                          Subsuperscript[ f, "", ""],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "D", "0"], 
                        f[ ]],
                      Editable->False], ")"}]}], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]], "-", 
              RowBox[{\(1\/2\), " ", 
                SuperscriptBox[
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False], "2"], " ", 
                InterpretationBox[
                  SubsuperscriptBox["\[Pi]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"i\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "i"],
                      Editable->False]],
                  Subsuperscript[ Pi, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "i"]],
                  Editable->False]}]}], ",", 
            RowBox[{
              InterpretationBox[
                SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"i\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "i"],
                    Editable->False]],
                Subsuperscript[ 
                  OverTilde[ \[CapitalGamma]], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "i"]],
                Editable->False], "-", 
              FractionBox[
                RowBox[{
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"i\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "i"],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "i"]],
                    Editable->False], " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
                        
                        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
                          Subsuperscript[ f, "", ""],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "D", "a"], 
                        f[ ]],
                      Editable->False], ")"}]}], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]], "-", 
              FractionBox[
                RowBox[{
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"i\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "i"],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "i"]],
                    Editable->False], " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
                        
                        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
                          Subsuperscript[ f, "", ""],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "D", "b"], 
                        f[ ]],
                      Editable->False], ")"}]}], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]], "+", 
              FractionBox[
                RowBox[{
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False], " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\("D"\^"i"\),
                          Superscript[ "D", "i"],
                          Editable->False], "\[InvisibleSpace]", 
                        
                        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
                          Subsuperscript[ f, "", ""],
                          Editable->False]}],
                      SequenceForm[ 
                        Superscript[ "D", "i"], 
                        f[ ]],
                      Editable->False], ")"}]}], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]]}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[
      TableKK[AffineG[l\[Mu], l\[Nu], u0], {l\[Mu], l\[Nu]}]]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              FractionBox[
                InterpretationBox[
                  RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "0"], 
                    f[ ]],
                  Editable->False], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]], 
              FractionBox[
                InterpretationBox[
                  RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "a"], 
                    f[ ]],
                  Editable->False], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]]},
            {
              FractionBox[
                InterpretationBox[
                  RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "a"], 
                    f[ ]],
                  Editable->False], 
                RowBox[{"2", " ", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}]], 
              RowBox[{
                RowBox[{"-", 
                  FractionBox[
                    RowBox[{
                      InterpretationBox[
                        SubsuperscriptBox["\[Gamma]", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"b\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "b"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ \[Gamma], 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "b"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False], " ", 
                      RowBox[{"(", 
                        InterpretationBox[
                          RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                            
                            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\
\"\>"\)\),
                              Subsuperscript[ f, "", ""],
                              Editable->False]}],
                          SequenceForm[ 
                            Subscript[ "D", "0"], 
                            f[ ]],
                          Editable->False], ")"}]}], 
                    RowBox[{"2", " ", 
                      SuperscriptBox[
                        
                        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
                          Subsuperscript[ f, "", ""],
                          Editable->False], "3"]}]]}], "+", 
                FractionBox[
                  InterpretationBox[
                    RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Gamma]", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"b\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "b"],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ \[Gamma], 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "b"], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "D", "0"], 
                      BaseMetricg[ la, lb]],
                    Editable->False], 
                  RowBox[{"2", " ", 
                    SuperscriptBox[
                      
                      InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
                        Subsuperscript[ f, "", ""],
                        Editable->False], "2"]}]], "-", 
                RowBox[{\(1\/2\), " ", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Pi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ Pi, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}]}]}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Section"],

Cell[BoxData[
    \(ricciDn[la_, lb_] := 
      SumDumKK[RicciR[la, lb] // RiemannToAffine]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ricciDn[li, lj] //. BlockDiagonalRule // Absorb[#, BaseMetricg] &\) // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "j"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "j"], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"], "p"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "-", 
      InterpretationBox[
        RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "p"], 
          BaseAffineG[ li, lj, "u$1"]],
        Editable->False], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              GaugeF[ li, lj]],
            Editable->False], ")"}]}], "+", 
      InterpretationBox[
        RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "j"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "i"], 
          BaseAffineG[ "l$1", lj, "u$1"]],
        Editable->False], "-", 
      FractionBox[
        RowBox[{"3", " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          SuperscriptBox[
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  f[ ]],
                Editable->False], ")"}], "2"]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "4"]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "0"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BD[ l0, 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                    SequenceForm[ "", "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "j"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ li, "l$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"j\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "j"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ li, lj]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ lj, "l$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"j\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "j"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ li, lj]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ li, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ lj, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ li, "l$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ lj, "l$2"]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"j\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "j"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ li, lj]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "0"], 
            BD[ l0, 
              BaseMetricg[ li, lj]]],
          Editable->False], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("D"\^"p"\),
                      Superscript[ "D", "p"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "D", "p"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BD[ "u$1", 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\"i\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "j"], "i"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "j"], "i"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "p"], "j"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"j\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "j"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BaseMetricg[ li, lj]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"q"\),
                  Superscript[ "D", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "q"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BaseMetricg[ lj, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BaseMetricg[ lj, "u$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                    SequenceForm[ "", "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "j"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "p"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BD[ "l$1", 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                    SequenceForm[ "", "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "j"], "p"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "j"], "p"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("D"\^"p"\),
                      Superscript[ "D", "p"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "D", "p"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BD[ "u$1", 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "i"], 
            BD[ lj, 
              f[ ]]],
          Editable->False], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BaseMetricg[ lj, "u$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "j"], 
            BD[ li, 
              f[ ]]],
          Editable->False], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"i\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "i"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "p"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BaseMetricg[ li, "u$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      RowBox[{\(1\/4\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ li, lj]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Pi]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"],
                Editable->False]],
            Subsuperscript[ Pi, 
              SequenceForm[ 
                SequenceForm[ "", "p"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"]],
            Editable->False]}], 
        RowBox[{"4", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Pi]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ Pi, 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                  SequenceForm[ "", "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "j"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "j"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                  SequenceForm[ "", "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "j"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "j"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% /. StationaryRule // Dnup\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "j"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "j"], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"], "p"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "q"]],
          Editable->False]}], "-", 
      InterpretationBox[
        RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "p"], 
          BaseAffineG[ li, lj, "u$1"]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "j"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "D", "i"], 
          BaseAffineG[ "l$1", lj, "u$1"]],
        Editable->False], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("D"\^"p"\),
                      Superscript[ "D", "p"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "D", "p"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BD[ "u$1", 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\"i\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "j"], "i"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "j"], "i"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "p"], "j"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"j\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "j"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "j"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BaseMetricg[ li, lj]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], "j"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"q"\),
                  Superscript[ "D", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "q"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BaseMetricg[ lj, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BaseMetricg[ lj, "u$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                    SequenceForm[ "", "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "j"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "j"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "p"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BD[ "l$1", 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                    SequenceForm[ "", "j"],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "j"], "p"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "j"], "p"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("D"\^"p"\),
                      Superscript[ "D", "p"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "D", "p"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                BD[ "u$1", 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "i"], 
            BD[ lj, 
              f[ ]]],
          Editable->False], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\)\
,
                          SequenceForm[ "", "j"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "j"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "j"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BaseMetricg[ lj, "u$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "j"], 
            BD[ li, 
              f[ ]]],
          Editable->False], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"i\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "i"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "p"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BaseMetricg[ li, "u$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"j"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "j"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"j\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "j"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "j"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"j"\),
                  SequenceForm[ "", "j"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "j"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "j"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ricciDn[li, l0] //. BlockDiagonalRule // Absorb[#, BaseMetricg] &\) // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"i"\),
                            SequenceForm[ "", "i"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "i"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  BaseMetricg[ li, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              BD[ l0, 
                BaseMetricg[ li, "l$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"], "q"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                        SequenceForm[ "", "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "q"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "q"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$2", "l$1"]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ li, "l$2"]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "u$1", "u$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ li, "l$2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "i"], 
              BaseMetricg[ "u$1", "u$2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], "q"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  BaseMetricg[ "l$1", "l$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "i"], 
              BD[ l0, 
                BaseMetricg[ "l$1", "l$2"]]],
            Editable->False], ")"}]}], "+", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"q"\),
                  Superscript[ "D", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "q"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                    SequenceForm[ "", "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", "i"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"q"\),
                  Superscript[ "D", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "q"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\)\
,
                          SequenceForm[ "", "q"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "q"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$2", "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ li, "l$1"]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "0"], 
            BD[ li, 
              f[ ]]],
          Editable->False], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"i\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "i"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "i"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "i"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ 
              OverTilde[ \[CapitalGamma]], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "p"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\)\
,
                          SequenceForm[ "", "i"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "i"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "i"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                BaseMetricg[ li, "u$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "D", "i"], 
            BD[ l0, 
              f[ ]]],
          Editable->False], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "i"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              GaugeF[ li, "u$1"]],
            Editable->False], ")"}], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "i"], 
              GaugeF[ "l$1", "u$1"]],
            Editable->False], ")"}], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "i"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "q"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "q"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% /. StationaryRule // Dnup\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                        SequenceForm[ "", "i"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "i"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "i"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              GaugeF[ li, "u$1"]],
            Editable->False], ")"}], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "i"], 
              GaugeF[ "l$1", "u$1"]],
            Editable->False], ")"}], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"i"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "i"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                      SequenceForm[ "", "i"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "i"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "i"], "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "i"], "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"i"\),
                  SequenceForm[ "", "i"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "i"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "q"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "i"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "q"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ricciDn[l0, l0] //. BlockDiagonalRule // Absorb[#, BaseMetricg] &\) // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                  SequenceForm[ "", "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\"s\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "r"], "s"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "r"], "s"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"s\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "s"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "s"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$1", "l$4"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                        SequenceForm[ "", "r"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "r"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "r"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$3", "l$2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "u$1", "u$2"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Gamma]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], "q"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Gamma], 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  BaseMetricg[ "l$1", "l$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BD[ l0, 
                BaseMetricg[ "l$1", "l$2"]]],
            Editable->False], ")"}]}], "+", 
      FractionBox[
        RowBox[{"3", " ", 
          SuperscriptBox[
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  f[ ]],
                Editable->False], ")"}], "2"]}], 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "0"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BD[ l0, 
                  f[ ]]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{"3", " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("D"\^"p"\),
                    Superscript[ "D", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "D", "p"], 
                  f[ ]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BD[ "u$1", 
                f[ ]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              GaugeF[ "l$1", "u$1"]],
            Editable->False], ")"}], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"]}], "-", 
      RowBox[{\(1\/4\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "4"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% /. StationaryRule // Dnup\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("D"\^"p"\),
                    Superscript[ "D", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "D", "p"], 
                  f[ ]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BD[ "u$1", 
                f[ ]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "4"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(scalarR[] := 
      SumDumKK[RicciR[l\[Mu], u\[Mu]] // RiemannToAffine]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(scalarR[] //. BlockDiagonalRule // Absorb[#, BaseMetricg] &\) // 
      Can\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("D"\^"p"\),
                    Superscript[ "D", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "D", "p"], 
                  f[ ]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BD[ "u$1", 
                f[ ]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "p"], 
                  f[ ]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              BD[ "l$1", 
                f[ ]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                        SequenceForm[ "", "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "q"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "q"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              BaseMetricg[ "l$2", "u$1"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                        SequenceForm[ "", "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "q"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "q"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l$2", "u$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              f[ ]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                        SequenceForm[ "", "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "q"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "q"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              BaseMetricg[ "l$2", "l$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              f[ ]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"r"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "r"], 
              BaseMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"r"\),
                Superscript[ "D", "r"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "r"], 
              f[ ]],
            Editable->False], ")"}]}], "-", 
      FractionBox[
        RowBox[{"15", " ", 
          SuperscriptBox[
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "0"], 
                  f[ ]],
                Editable->False], ")"}], "2"]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\
\),
                      Subsuperscript[ f, "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "0"], 
                    f[ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BD[ l0, 
                  f[ ]]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"11", " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\)\
,
                          SequenceForm[ "", "q"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "q"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$2", "l$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          SuperscriptBox[
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False], "2"]}]], "+", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                    SequenceForm[ "", "r"],
                    Editable->False], "\[InvisibleSpace]", "\<\"s\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "r"], "s"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "r"], "s"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\)\
,
                          SequenceForm[ "", "r"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"s\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "r"], "s"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "r"], "s"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$3", "l$4"]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                    SequenceForm[ "", "r"],
                    Editable->False], "\[InvisibleSpace]", "\<\"s\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "r"], "s"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "r"], "s"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\)\
,
                          SequenceForm[ "", "r"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "r"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "r"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$3", "l$1"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"s"\)\
,
                          SequenceForm[ "", "s"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "s"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "s"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$4", "l$2"]],
              Editable->False], ")"}]}], 
        RowBox[{"4", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Gamma]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False]],
                  Subsuperscript[ \[Gamma], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BaseMetricg[ "u$1", "u$2"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Gamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"],
                Editable->False]],
            Subsuperscript[ \[Gamma], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["\[Gamma]", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                              SequenceForm[ "", "p"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"q\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "p"], "q"],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ \[Gamma], 
                        SequenceForm[ 
                          SequenceForm[ "", "p"], "q"], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "D", "0"], 
                    BaseMetricg[ "l$1", "l$2"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "0"], 
                BD[ l0, 
                  BaseMetricg[ "l$1", "l$2"]]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], " "], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                      SequenceForm[ "", "r"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "r"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "r"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"], "r"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "r"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"], "r"]],
          Editable->False], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], "r"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], "r"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                      SequenceForm[ "", "q"],
                      Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "q"], "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "q"], "r"], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "q"], "r"], "p"]],
          Editable->False], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"r"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "r"],
                    Editable->False]],
                Subsuperscript[ 
                  OverTilde[ \[CapitalGamma]], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "r"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "r"], 
              BaseAffineG[ "l$1", "l$2", "u$3"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "q"],
                    Editable->False]],
                Subsuperscript[ 
                  OverTilde[ \[CapitalGamma]], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "q"], "p"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              BaseAffineG[ "l$2", "l$1", "u$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              GaugeF[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Pi]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ Pi, 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              GaugeF[ "l$1", "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "3"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "q"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "q"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"0"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "0"], 
              BaseMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                  SequenceForm[ "", "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "r"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "r"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "r"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "r"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "3"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% /. StationaryRule\  // Dnup\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("D"\^"p"\),
                    Superscript[ "D", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "D", "p"], 
                  f[ ]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BD[ "u$1", 
                f[ ]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                    
                    Subsuperscript[ f, "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "D", "p"], 
                  f[ ]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              BD[ "l$1", 
                f[ ]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                        SequenceForm[ "", "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "q"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "q"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              BaseMetricg[ "l$2", "u$1"]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                        SequenceForm[ "", "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "q"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "q"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "p"], 
              BaseMetricg[ "l$2", "u$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              f[ ]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                        SequenceForm[ "", "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "q"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "q"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              BaseMetricg[ "l$2", "l$1"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"q"\),
                Superscript[ "D", "q"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "q"], 
              f[ ]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"r"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Gamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ \[Gamma], 
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "r"], 
              BaseMetricg[ "l$1", "l$2"]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"r"\),
                Superscript[ "D", "r"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "r"], 
              f[ ]],
            Editable->False], ")"}]}], "+", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("D"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("D"\^"p"\),
                  Superscript[ "D", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                  Subsuperscript[ f, "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "D", "p"], 
                f[ ]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False]}]], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], " "], "q"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "p"], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"r"\),
                      SequenceForm[ "", "r"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "r"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "r"], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"], "r"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "r"], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"], "r"]],
          Editable->False], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "q"], "r"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"], "r"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                      SequenceForm[ "", "q"],
                      Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "q"], "r"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "q"], "r"], "p"],
              Editable->False]],
          Subsuperscript[ 
            OverTilde[ \[CapitalGamma]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "q"], "r"], "p"]],
          Editable->False], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"r"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], "q"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"r\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "r"],
                    Editable->False]],
                Subsuperscript[ 
                  OverTilde[ \[CapitalGamma]], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "r"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "r"], 
              BaseAffineG[ "l$1", "l$2", "u$3"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("D"\^"p"\),
                Superscript[ "D", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[\(\[CapitalGamma]\&~\), 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "q"],
                    Editable->False]],
                Subsuperscript[ 
                  OverTilde[ \[CapitalGamma]], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "q"], "p"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "q"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "D", "p"], 
              BaseAffineG[ "l$2", "l$1", "u$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "3"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "p"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "q"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "q"], " "], 
            SequenceForm[ 
              SequenceForm[ "", " "], "q"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", 
        SuperscriptBox[
          InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
            Subsuperscript[ f, "", ""],
            Editable->False], "3"], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Pi]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                  SequenceForm[ "", "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "p"], "q"],
              Editable->False]],
          Subsuperscript[ Pi, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "p"], "q"]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{691, 694},
WindowMargins->{{11, Automatic}, {-4, Automatic}},
Visible->True,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"], 
    "Vdefine"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
            OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
            NotationBoxTag, TagStyle -> "NotationTemplateStyle"], " ", 
          "\[DoubleLongLeftRightArrow]", " ", 
          TagBox[ 
            RowBox[ {"Vector", "[", "\[Placeholder]", "]"}], NotationBoxTag, 
            TagStyle -> "NotationTemplateStyle"]}], "]"}], 
    "ivec"->OverscriptBox[ "\[DotlessI]", "^"], 
    "jvec"->OverscriptBox[ "\[DotlessJ]", "^"], 
    "vect"->OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
    "xvec"->OverscriptBox[ "x", "^"], "yvec"->OverscriptBox[ "y", "^"], 
    "zvec"->OverscriptBox[ "z", "^"], "rvec"->OverscriptBox[ "r", "^"], 
    "tvec"->OverscriptBox[ "\[Theta]", "^"], 
    "pvec"->OverscriptBox[ "\[Phi]", "^"]},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 100, 3, 43, "Title"],
Cell[1820, 54, 54, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1899, 59, 34, 0, 78, "Section"],

Cell[CellGroupData[{
Cell[1958, 63, 32, 0, 54, "Subsubsection"],

Cell[CellGroupData[{
Cell[2015, 67, 224, 3, 102, "Input"],
Cell[2242, 72, 8131, 199, 56, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10422, 277, 34, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[10481, 281, 186, 3, 102, "Input"],
Cell[10670, 286, 7754, 191, 57, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[18485, 484, 35, 0, 50, "Section"],
Cell[18523, 486, 238, 6, 41, "Text"],

Cell[CellGroupData[{
Cell[18786, 496, 79, 1, 56, "Input"],
Cell[18868, 499, 1606, 40, 56, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[20523, 545, 39, 0, 50, "Section"],

Cell[CellGroupData[{
Cell[20587, 549, 132, 3, 79, "Input"],
Cell[20722, 554, 1827, 44, 111, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22586, 603, 132, 3, 79, "Input"],
Cell[22721, 608, 1910, 47, 107, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[24668, 660, 132, 3, 79, "Input"],
Cell[24803, 665, 1302, 30, 91, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[26142, 700, 141, 3, 79, "Input"],
Cell[26286, 705, 1583, 38, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27906, 748, 141, 3, 79, "Input"],
Cell[28050, 753, 1789, 43, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[29888, 802, 39, 0, 50, "Section"],
Cell[29930, 804, 134, 2, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[30101, 811, 29, 0, 50, "Section"],
Cell[30133, 813, 119, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[30277, 819, 124, 2, 79, "Input"],
Cell[30404, 823, 11368, 271, 122, "Output"]
}, Open  ]],
Cell[41787, 1097, 387, 7, 194, "Input"],

Cell[CellGroupData[{
Cell[42199, 1108, 59, 1, 56, "Input"],
Cell[42261, 1111, 6644, 163, 80, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[48942, 1279, 83, 1, 56, "Input"],
Cell[49028, 1282, 4374, 106, 78, "Output"]
}, Open  ]],
Cell[53417, 1391, 336, 7, 150, "Input"],

Cell[CellGroupData[{
Cell[53778, 1402, 52, 1, 56, "Input"],
Cell[53833, 1405, 4374, 106, 78, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[58244, 1516, 124, 2, 79, "Input"],
Cell[58371, 1520, 5243, 126, 80, "Output"]
}, Open  ]],
Cell[63629, 1649, 334, 6, 172, "Input"],

Cell[CellGroupData[{
Cell[63988, 1659, 59, 1, 56, "Input"],
Cell[64050, 1662, 5243, 126, 80, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[69330, 1793, 76, 1, 56, "Input"],
Cell[69409, 1796, 524, 14, 76, "Output"]
}, Open  ]],
Cell[69948, 1813, 205, 3, 102, "Input"],

Cell[CellGroupData[{
Cell[70178, 1820, 52, 1, 56, "Input"],
Cell[70233, 1823, 524, 14, 76, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[70794, 1842, 124, 2, 79, "Input"],
Cell[70921, 1846, 5243, 126, 80, "Output"]
}, Open  ]],
Cell[76179, 1975, 323, 5, 172, "Input"],

Cell[CellGroupData[{
Cell[76527, 1984, 59, 1, 56, "Input"],
Cell[76589, 1987, 5243, 126, 80, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[81869, 2118, 76, 1, 56, "Input"],
Cell[81948, 2121, 524, 14, 76, "Output"]
}, Open  ]],
Cell[82487, 2138, 205, 3, 102, "Input"],

Cell[CellGroupData[{
Cell[82717, 2145, 52, 1, 56, "Input"],
Cell[82772, 2148, 524, 14, 76, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[83333, 2167, 124, 2, 79, "Input"],
Cell[83460, 2171, 656, 17, 69, "Output"]
}, Open  ]],
Cell[84131, 2191, 205, 3, 102, "Input"],

Cell[CellGroupData[{
Cell[84361, 2198, 59, 1, 56, "Input"],
Cell[84423, 2201, 656, 17, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[85116, 2223, 82, 1, 56, "Input"],
Cell[85201, 2226, 524, 14, 76, "Output"]
}, Open  ]],
Cell[85740, 2243, 184, 3, 102, "Input"],

Cell[CellGroupData[{
Cell[85949, 2250, 52, 1, 56, "Input"],
Cell[86004, 2253, 524, 14, 76, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[86565, 2272, 116, 2, 56, "Input"],
Cell[86684, 2276, 23112, 536, 182, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[109833, 2817, 106, 2, 56, "Input"],
Cell[109942, 2821, 8083, 188, 129, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[118074, 3015, 28, 0, 50, "Section"],
Cell[118105, 3017, 104, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[118234, 3023, 116, 2, 79, "Input"],
Cell[118353, 3027, 94960, 2304, 601, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[213350, 5336, 60, 1, 56, "Input"],
Cell[213413, 5339, 49726, 1210, 348, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[263176, 6554, 116, 2, 79, "Input"],
Cell[263295, 6558, 49861, 1194, 366, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[313193, 7757, 60, 1, 56, "Input"],
Cell[313256, 7760, 11847, 285, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[325140, 8050, 116, 2, 79, "Input"],
Cell[325259, 8054, 27306, 658, 214, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[352602, 8717, 60, 1, 56, "Input"],
Cell[352665, 8720, 6399, 160, 69, "Output"]
}, Open  ]],
Cell[359079, 8883, 104, 2, 56, "Input"],

Cell[CellGroupData[{
Cell[359208, 8889, 110, 2, 79, "Input"],
Cell[359321, 8893, 62841, 1514, 373, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[422199, 10412, 62, 1, 56, "Input"],
Cell[422264, 10415, 28610, 692, 201, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

