(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     72296,       1850]*)
(*NotebookOutlinePosition[     73064,       1877]*)
(*  CellTagsIndexPosition[     73020,       1873]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Cartan's Structure Equation", "Title",
  TextAlignment->Left,
  TextJustification->0,
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Tensorial forms", "Subsubsection"],

Cell["Basis one-form", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Fdefine[e, 1, "\<a\>"]; \ e[ua]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      SubsuperscriptBox[
        TagBox[
          StyleBox["e",
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
              SequenceForm[ "", "a"],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)]],
      Subsuperscript[ 
        StyleForm[ e, FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ "", " "], FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ "", "a"], FontWeight -> "Bold"]],
      Editable->False]], "Output"]
}, Open  ]],

Cell["Connection one-form", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Fdefine[\[Omega], 1, "\<-ba\>"]; \[Omega][la, ub]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      SubsuperscriptBox[
        TagBox[
          StyleBox["\[Omega]",
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], " "],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "b"],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)]],
      Subsuperscript[ 
        StyleForm[ \[Omega], FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ 
            SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ 
            SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
      Editable->False]], "Output"]
}, Open  ]],

Cell["Torsion two-form", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Fdefine[\[CapitalTheta], 2, "\<a\>"]; \[CapitalTheta][ua]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      SubsuperscriptBox[
        TagBox[
          StyleBox["\[CapitalTheta]",
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
              SequenceForm[ "", "a"],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)]],
      Subsuperscript[ 
        StyleForm[ \[CapitalTheta], FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ "", " "], FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ "", "a"], FontWeight -> "Bold"]],
      Editable->False]], "Output"]
}, Open  ]],

Cell["Curvature two-form", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Fdefine[\[CapitalOmega], 2, "\<-ba\>"]; \[CapitalOmega][la, 
      ub]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      SubsuperscriptBox[
        TagBox[
          StyleBox["\[CapitalOmega]",
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], " "],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)], 
        TagBox[
          StyleBox[
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], "b"],
              Editable->False],
            FontWeight->"Bold"],
          (StyleForm[ #, FontWeight -> "Bold"]&)]],
      Subsuperscript[ 
        StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ 
            SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
        StyleForm[ 
          SequenceForm[ 
            SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Structure equations", "Subsubsection"],

Cell["First structure equation", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[CapitalTheta][ua]\  == \ 
      XD[e[ua]]\  - \ XP[e[ub], \[Omega][lb, ua]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox[
          TagBox[
            StyleBox["\[CapitalTheta]",
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)]],
        Subsuperscript[ 
          StyleForm[ \[CapitalTheta], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", " "], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", "a"], FontWeight -> "Bold"]],
        Editable->False], "==", 
      RowBox[{
        InterpretationBox[
          RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["e",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ e, FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "a"], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ "d", 
            e[ ua]],
          Editable->False], "-", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["e",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ e, FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "b"], FontWeight -> "Bold"]],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
            "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Omega]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"a\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Omega], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "a"], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ 
            e[ ub], "\[Wedge]", 
            \[Omega][ lb, ua]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(firstRule = 
      Trule[XD[e[ua_]], \[CapitalTheta][ua]\  + \ 
          XP[e[ub], \[Omega][lb, ua]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["e",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]ua_\),
                    SequenceForm[ "", 
                      Pattern[ ua, 
                        Blank[ ]]],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ e, FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", 
                  Pattern[ ua, 
                    Blank[ ]]], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ "d", 
          e[ 
            Pattern[ ua, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["e",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ e, FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", " "], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", "p"], FontWeight -> "Bold"]],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
                "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Omega]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                              SequenceForm[ "", "p"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "p"], " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"a\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], "a"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Omega], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], " "], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "a"], FontWeight -> "Bold"]],
                  Editable->False]}],
              SequenceForm[ 
                e[ "u$1"], "\[Wedge]", 
                \[Omega][ "l$1", ua]],
              Editable->False], "+", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[CapitalTheta]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[CapitalTheta], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "a"], FontWeight -> "Bold"]],
              Editable->False]}], "]"}], "/;", "True"}]}]], "Output"]
}, Open  ]],

Cell["Second structure equation", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[CapitalOmega][la, ub]\  == \ 
      XD[\[Omega][la, ub]]\  - \ 
        XP[\[Omega][la, ud], \[Omega][ld, ub]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox[
          TagBox[
            StyleBox["\[CapitalOmega]",
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)]],
        Subsuperscript[ 
          StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ 
              SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
        Editable->False], "==", 
      RowBox[{
        InterpretationBox[
          RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Omega]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Omega], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ "d", 
            \[Omega][ la, ub]],
          Editable->False], "-", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Omega]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Omega], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "d"], FontWeight -> "Bold"]],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
            "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Omega]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\)\
,
                          SequenceForm[ "", "d"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "d"], " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Omega], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "d"], " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ 
            \[Omega][ la, ud], "\[Wedge]", 
            \[Omega][ ld, ub]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(secondRule = 
      Trule[XD[\[Omega][la_, ub_]], \[CapitalOmega][la, ub]\  + \ 
          XP[\[Omega][la, ud], \[Omega][ld, ub]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Omega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                        SequenceForm[ "", 
                          Pattern[ la, 
                            Blank[ ]]],
                        Editable->False], "\[InvisibleSpace]", "ub_"}],
                    SequenceForm[ 
                      SequenceForm[ "", 
                        Pattern[ la, 
                          Blank[ ]]], 
                      Pattern[ ub, 
                        Blank[ ]]],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Omega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ ub, 
                    Blank[ ]]], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ "d", 
          \[Omega][ 
            Pattern[ la, 
              Blank[ ]], 
            Pattern[ ub, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Omega]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                              SequenceForm[ "", "a"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "a"], " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"q\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], "q"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Omega], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"], FontWeight -> "Bold"]],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
                "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Omega]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                              SequenceForm[ "", "q"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "q"], " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"b\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], "b"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Omega], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
                  Editable->False]}],
              SequenceForm[ 
                \[Omega][ la, "u$2"], "\[Wedge]", 
                \[Omega][ "l$2", ub]],
              Editable->False], "+", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[CapitalOmega]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
              Editable->False]}], "]"}], "/;", "True"}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Bianchi identities", "Subsubsection"],

Cell["Take the exterior derivative to the first structure equation", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(XD[\[CapitalTheta][ua]\  - \ XD[e[ua]] + \ 
        XP[e[ub], \[Omega][lb, ua]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalTheta]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalTheta], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "a"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ "d", 
          \[CapitalTheta][ ua]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["e",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ e, FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "b"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Omega]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"a\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "a"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Omega], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "a"], FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Omega][ lb, ua]],
            Editable->False]}],
        SequenceForm[ 
          e[ ub], "\[Wedge]", 
          XD[ 
            \[Omega][ lb, ua]]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Omega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Omega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "a"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["e",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ e, FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", " "], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", "b"], FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              e[ ub]],
            Editable->False]}],
        SequenceForm[ 
          \[Omega][ lb, ua], "\[Wedge]", 
          XD[ 
            e[ ub]]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell["\<\
Using the structure equations we have the first Bianchi identity:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[%, {firstRule, secondRule}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalTheta]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalTheta], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "a"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ "d", 
          \[CapitalTheta][ ua]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["e",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ e, FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "p"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalOmega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "a"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ 
          e[ "u$1"], "\[Wedge]", 
          \[CapitalOmega][ "l$1", ua]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Omega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Omega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "a"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalTheta]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalTheta], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "p"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ 
          \[Omega][ "l$1", ua], "\[Wedge]", 
          \[CapitalTheta][ "u$1"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell["\<\
Take the exterior derivative to the second structure equation\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(XD[\[CapitalOmega][la, ub]\  - \ XD[\[Omega][la, ub]]\  + \ 
        XP[\[Omega][la, uc], \[Omega][lc, ub]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalOmega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ "d", 
          \[CapitalOmega][ la, ub]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Omega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Omega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "c"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Omega]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "b"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Omega], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Omega][ lc, ub]],
            Editable->False]}],
        SequenceForm[ 
          \[Omega][ la, uc], "\[Wedge]", 
          XD[ 
            \[Omega][ lc, ub]]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Omega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Omega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "c"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Omega]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "c"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Omega], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"], FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Omega][ la, uc]],
            Editable->False]}],
        SequenceForm[ 
          \[Omega][ lc, ub], "\[Wedge]", 
          XD[ 
            \[Omega][ la, uc]]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell["\<\
Using the structure equations we have the second Bianchi identity:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[%, {firstRule, secondRule}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalOmega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ "d", 
          \[CapitalOmega][ la, ub]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Omega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Omega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalOmega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ 
          \[Omega][ "l$1", ub], "\[Wedge]", 
          \[CapitalOmega][ la, "u$1"]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Omega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "p"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Omega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "p"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[CapitalOmega]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "p"], " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[CapitalOmega], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ 
          \[Omega][ la, "u$1"], "\[Wedge]", 
          \[CapitalOmega][ "l$1", ub]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{581, 682},
WindowMargins->{{10, Automatic}, {Automatic, 8}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 107, 3, 43, "Title"],
Cell[1827, 54, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1904, 59, 40, 0, 54, "Subsubsection"],
Cell[1947, 61, 30, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[2002, 65, 64, 1, 56, "Input"],
Cell[2069, 68, 964, 27, 54, "Output"]
}, Open  ]],
Cell[3048, 98, 35, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[3108, 102, 82, 1, 56, "Input"],
Cell[3193, 105, 1462, 39, 55, "Output"]
}, Open  ]],
Cell[4670, 147, 32, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[4727, 151, 90, 1, 56, "Input"],
Cell[4820, 154, 992, 27, 54, "Output"]
}, Open  ]],
Cell[5827, 184, 34, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[5886, 188, 103, 2, 56, "Input"],
Cell[5992, 192, 1476, 39, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[7517, 237, 44, 0, 32, "Subsubsection"],
Cell[7564, 239, 40, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[7629, 243, 110, 2, 56, "Input"],
Cell[7742, 247, 5660, 140, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13439, 392, 135, 3, 79, "Input"],
Cell[13577, 397, 6388, 152, 58, "Output"]
}, Open  ]],
Cell[19980, 552, 41, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[20046, 556, 145, 3, 79, "Input"],
Cell[20194, 561, 7586, 188, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27817, 754, 163, 3, 79, "Input"],
Cell[27983, 759, 8623, 204, 60, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[36655, 969, 43, 0, 32, "Subsubsection"],
Cell[36701, 971, 76, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[36802, 975, 113, 2, 56, "Input"],
Cell[36918, 979, 8115, 197, 55, "Output"]
}, Open  ]],
Cell[45048, 1179, 89, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[45162, 1185, 71, 1, 56, "Input"],
Cell[45236, 1188, 7425, 179, 56, "Output"]
}, Open  ]],
Cell[52676, 1370, 85, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[52786, 1376, 141, 2, 79, "Input"],
Cell[52930, 1380, 10008, 239, 55, "Output"]
}, Open  ]],
Cell[62953, 1622, 90, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[63068, 1628, 71, 1, 56, "Input"],
Cell[63142, 1631, 9126, 215, 57, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

