(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    490385,      12143]*)
(*NotebookOutlinePosition[    491153,      12170]*)
(*  CellTagsIndexPosition[    491109,      12166]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Conformal Transformation", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[BoxData[
    \(Tdefine[\[CapitalOmega], "\<\>"]\)], "Input"],

Cell[CellGroupData[{

Cell["Rules for conformal transformation", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(confRule1 = 
      Trule[Metricg[la_, lb_], \(\[CapitalOmega][]\^2\) Metricg[la, lb], 
        DnIndexQ[la]\  && \ DnIndexQ[lb]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "lb_"}],
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", 
              Pattern[ la, 
                Blank[ ]]], 
            Pattern[ lb, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox["g", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ g, 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False], " ", 
            SuperscriptBox[
              
              InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\
\"\>"\)\),
                Subsuperscript[ \[CapitalOmega], "", ""],
                Editable->False], "2"]}], "]"}], 
        "/;", \(DnIndexQ[la] && DnIndexQ[lb]\)}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(confRule2 = 
      Trule[Metricg[ua_, ub_], \(\[CapitalOmega][]\^\(-2\)\) Metricg[ua, ub], 
        UpIndexQ[ua]\  && \ UpIndexQ[ub]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]ua_\),
                SequenceForm[ "", 
                  Pattern[ ua, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "ub_"}],
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ ua, 
                  Blank[ ]]], 
              Pattern[ ub, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", 
              Pattern[ ua, 
                Blank[ ]]], 
            Pattern[ ub, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          FractionBox[
            InterpretationBox[
              SubsuperscriptBox["g", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False]],
              Subsuperscript[ g, 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"]],
              Editable->False], 
            SuperscriptBox[
              
              InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\
\"\>"\)\),
                Subsuperscript[ \[CapitalOmega], "", ""],
                Editable->False], "2"]], "]"}], 
        "/;", \(UpIndexQ[ua] && UpIndexQ[ub]\)}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(dimRule = Trule[Metricg[la_, ub_], Dimension, PairQ[la, ub]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "ub_"}],
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ ub, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", 
              Pattern[ la, 
                Blank[ ]]], 
            Pattern[ ub, 
              Blank[ ]]]],
        Editable->False], 
      "\[RuleDelayed]", \(DumFresh[Dimension] /; PairQ[la, ub]\)}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(confRule = {confRule1, confRule2, dimRule};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[Metricg[la, lb], confRule]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False], " ", 
      SuperscriptBox[
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False], "2"]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Affine connection", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[AffineG[la, lb, uc] // AffineToMetric, confRule]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"c"\),
                Superscript[ "\[PartialD]", "c"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "c"], 
              Metricg[ la, lb]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              Metricg[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              Metricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "c"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "c"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(rhs = 
      AffineG[la, lb, uc] + %[\([4]\)] + %[\([5]\)] + %[\([6]\)]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[CapitalGamma]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"],
            Editable->False]],
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], "c"]],
        Editable->False], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "c"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "c"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineConfRule = 
      Trule[AffineG[la_, lb_, uc_], rhs, 
        DnIndexQ[la] && DnIndexQ[lb] && UpIndexQ[uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[CapitalGamma]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lb_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "uc_"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]], 
              Pattern[ uc, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]], 
            Pattern[ uc, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox["\[CapitalGamma]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "c"],
                  Editable->False]],
              Subsuperscript[ \[CapitalGamma], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], " "], 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "c"]],
              Editable->False], "-", 
            FractionBox[
              RowBox[{
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\("\[PartialD]"\^"c"\),
                        Superscript[ "\[PartialD]", "c"],
                        Editable->False], "\[InvisibleSpace]", 
                      
                      InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\
\(\*"\<\"\"\>"\)\),
                        Subsuperscript[ \[CapitalOmega], "", ""],
                        Editable->False]}],
                    SequenceForm[ 
                      Superscript[ "\[PartialD]", "c"], 
                      \[CapitalOmega][ ]],
                    Editable->False], ")"}], " ", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}], 
              
              InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\
\"\>"\)\),
                Subsuperscript[ \[CapitalOmega], "", ""],
                Editable->False]], "+", 
            FractionBox[
              RowBox[{
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                      
                      InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\
\(\*"\<\"\"\>"\)\),
                        Subsuperscript[ \[CapitalOmega], "", ""],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "b"], 
                      \[CapitalOmega][ ]],
                    Editable->False], ")"}], " ", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "c"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"]],
                  Editable->False]}], 
              
              InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\
\"\>"\)\),
                Subsuperscript[ \[CapitalOmega], "", ""],
                Editable->False]], "+", 
            FractionBox[
              RowBox[{
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                      
                      InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\
\(\*"\<\"\"\>"\)\),
                        Subsuperscript[ \[CapitalOmega], "", ""],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "a"], 
                      \[CapitalOmega][ ]],
                    Editable->False], ")"}], " ", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "c"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"]],
                  Editable->False]}], 
              
              InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\
\"\>"\)\),
                Subsuperscript[ \[CapitalOmega], "", ""],
                Editable->False]]}], "]"}], 
        "/;", \(DnIndexQ[la] && DnIndexQ[lb] && UpIndexQ[uc]\)}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affine = ApplyRules[AffineG[la, lb, uc], affineConfRule]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[CapitalGamma]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"],
            Editable->False]],
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], "c"]],
        Editable->False], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "c"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "c"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Riemann tensor", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[RiemannR[la, lb, lc, ud] // RiemannToAffine, 
      affineConfRule]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"-", 
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "d"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "a"], 
          AffineG[ lb, lc, ud]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "d"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "b"], 
          AffineG[ la, lc, ud]],
        Editable->False], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ la, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"], "b"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], "b"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], "a"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], "a"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, lc]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ la, lc]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "d"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "p"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "d"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(rhs = AffineToMetric[Drop[%, 4]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        FractionBox[
          RowBox[{"2", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*\
"\<\"\"\>"\)\),
                    Subsuperscript[ \[CapitalOmega], "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "b"], 
                  \[CapitalOmega][ ]],
                Editable->False], ")"}], " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"d"\),
                    Superscript[ "\[PartialD]", "d"],
                    Editable->False], "\[InvisibleSpace]", 
                  
                  InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*\
"\<\"\"\>"\)\),
                    Subsuperscript[ \[CapitalOmega], "", ""],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "d"], 
                  \[CapitalOmega][ ]],
                Editable->False], ")"}], " ", 
            InterpretationBox[
              SubsuperscriptBox["g", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ g, 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}], 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "2"]]}], "+", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ la, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ lb, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ la, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ lb, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ la, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(riemannConfRule = 
        Trule[RiemannR[la_, lb_, lc_, ud_], 
          RiemannR[la, lb, lc, ud] + 
            rhs, \(DnIndexQ[la] && DnIndexQ[lb] && DnIndexQ[lc]\) && 
            UpIndexQ[ud]];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(riemann = 
      ApplyRules[RiemannR[la, lb, lc, ud], riemannConfRule]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["R", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "c"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "c"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "d"],
            Editable->False]],
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], "c"], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], "d"]],
        Editable->False], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ la, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ lb, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ la, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "d"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ lb, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ la, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "d"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Metricg[ub, ld]\ *\ riemann\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["R", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", "a"], "c"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False], "+", 
      FractionBox[
        RowBox[{"4", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"b"\),
                  Superscript[ "\[PartialD]", "b"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "b"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "a"], 
            BD[ lc, 
              \[CapitalOmega][ ]]],
          Editable->False], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}]
                        ,
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                Metricg[ la, ud]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"q"\),
                  Superscript[ "\[PartialD]", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "q"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"b"\),
                  Superscript[ "\[PartialD]", "b"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "b"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ lb, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ la, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "b"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ la, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "c"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "c"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "c"], "d"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "d"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ lb, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "b"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricci = Tsimplify[%\  /. \ dimRule]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["R", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ "", "a"], "c"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False], "+", 
      FractionBox[
        RowBox[{"4", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"2", " ", "Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "-", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ la, lc]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "a"], 
            BD[ lc, 
              \[CapitalOmega][ ]]],
          Editable->False], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                Metricg[ la, ud]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ lb, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ la, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "c"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "c"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "c"], "d"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "d"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\
\)\),
          Subsuperscript[ \[CapitalOmega], "", ""],
          Editable->False]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ lb, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "b"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False]}]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[CapitalOmega][]\^\(-2\)\ Metricg[ua, uc]\ *\ ricci\)], "Input"],

Cell[BoxData[
    RowBox[{
      FractionBox[
        RowBox[{"4", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "4"]], "-", 
      FractionBox[
        RowBox[{"2", " ", "Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "4"]], "-", 
      FractionBox[
        RowBox[{"3", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "a"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "a"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "4"]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "a"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "a"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "4"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"q"\),
                  Superscript[ "\[PartialD]", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "q"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "d"], 
            BD[ ud, 
              \[CapitalOmega][ ]]],
          Editable->False], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                Metricg[ la, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                Metricg[ la, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"b"\),
                  Superscript[ "\[PartialD]", "b"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "b"], 
                Metricg[ lb, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{
            InterpretationBox[\("\[PartialD]"\^"c"\),
              Superscript[ "\[PartialD]", "c"],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False]}],
          SequenceForm[ 
            Superscript[ "\[PartialD]", "c"], 
            BD[ lc, 
              \[CapitalOmega][ ]]],
          Editable->False], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"p\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "p"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                Metricg[ lc, "l$1"]],
              Editable->False], ")"}]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "a"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "a"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "d"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ la, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "a"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ la, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ lb, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "a"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "a"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "b"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ lb, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scalar = Tsimplify[% /. dimRule]\)], "Input"],

Cell[BoxData[
    RowBox[{
      FractionBox[
        RowBox[{"4", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "4"]], "-", 
      FractionBox[
        RowBox[{"5", " ", "Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "4"]], "+", 
      FractionBox[
        RowBox[{\(Dimension\^2\), " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "4"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"q"\),
                  Superscript[ "\[PartialD]", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "q"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"q"\),
                  Superscript[ "\[PartialD]", "q"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\)\
,
                          SequenceForm[ "", "p"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "p"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "p"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "q"], 
                Metricg[ "l$1", "l$2"]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"d"\),
                  Superscript[ "\[PartialD]", "d"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "d"], 
            BD[ ud, 
              \[CapitalOmega][ ]]],
          Editable->False], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"d"\),
                      Superscript[ "\[PartialD]", "d"],
                      Editable->False], "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "d"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                BD[ ud, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        RowBox[{"2", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                Metricg[ la, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        InterpretationBox[
          RowBox[{
            InterpretationBox[\("\[PartialD]"\^"c"\),
              Superscript[ "\[PartialD]", "c"],
              Editable->False], "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "c"], 
                \[CapitalOmega][ ]],
              Editable->False]}],
          SequenceForm[ 
            Superscript[ "\[PartialD]", "c"], 
            BD[ lc, 
              \[CapitalOmega][ ]]],
          Editable->False], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                    
                    InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\
\*"\<\"\"\>"\)\),
                      Subsuperscript[ \[CapitalOmega], "", ""],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "c"], 
                    \[CapitalOmega][ ]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                BD[ lc, 
                  \[CapitalOmega][ ]]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "d"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"c"\),
                  Superscript[ "\[PartialD]", "c"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\)\
,
                          SequenceForm[ "", "c"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "c"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "c"], 
                Metricg[ lc, ud]],
              Editable->False], ")"}]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "+", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "a"], 
                Metricg[ lb, ud]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"a"\),
                  Superscript[ "\[PartialD]", "a"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "a"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                    SequenceForm[ "", "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "d"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], "b"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "d"], " "], 
              SequenceForm[ 
                SequenceForm[ "", " "], "b"]],
            Editable->False]}], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "3"]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ la, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "a"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"q\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "q"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "q"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                Metricg[ la, "l$2"]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "q"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "a"], "q"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "-", 
      FractionBox[
        RowBox[{
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ la, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        RowBox[{"Dimension", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                
                InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\
\"\"\>"\)\),
                  Subsuperscript[ \[CapitalOmega], "", ""],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "p"], 
                \[CapitalOmega][ ]],
              Editable->False], ")"}], " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["g", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ g, 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                Metricg[ la, lc]],
              Editable->False], ")"}], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"]],
            Editable->False]}], 
        RowBox[{"2", " ", 
          SuperscriptBox[
            
            InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\
\>"\)\),
              Subsuperscript[ \[CapitalOmega], "", ""],
              Editable->False], "3"]}]], "+", 
      FractionBox[
        InterpretationBox[\(R\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ R, "", ""],
          Editable->False], 
        SuperscriptBox[
          
          InterpretationBox[\(\[CapitalOmega]\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>\
"\)\),
            Subsuperscript[ \[CapitalOmega], "", ""],
            Editable->False], "2"]]}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{703, 680},
WindowMargins->{{19, Automatic}, {Automatic, 5}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 57, 1, 43, "Title"],
Cell[1777, 52, 52, 1, 56, "Input"],
Cell[1832, 55, 65, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1922, 60, 59, 0, 54, "Subsubsection"],

Cell[CellGroupData[{
Cell[2006, 64, 161, 3, 81, "Input"],
Cell[2170, 69, 2647, 69, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4854, 143, 166, 3, 81, "Input"],
Cell[5023, 148, 2645, 69, 78, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7705, 222, 93, 1, 79, "Input"],
Cell[7801, 225, 1274, 36, 55, "Output"]
}, Open  ]],
Cell[9090, 264, 80, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[9195, 269, 70, 1, 56, "Input"],
Cell[9268, 272, 1149, 30, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10466, 308, 42, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[10533, 312, 92, 1, 79, "Input"],
Cell[10628, 315, 11557, 277, 117, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22222, 597, 104, 2, 56, "Input"],
Cell[22329, 601, 6812, 168, 76, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[29178, 774, 146, 3, 79, "Input"],
Cell[29327, 779, 10418, 255, 101, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[39782, 1039, 89, 1, 56, "Input"],
Cell[39874, 1042, 6812, 168, 76, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[46735, 1216, 39, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[46799, 1220, 111, 2, 79, "Input"],
Cell[46913, 1224, 67243, 1622, 397, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[114193, 2851, 65, 1, 56, "Input"],
Cell[114261, 2854, 78879, 1930, 565, "Output"]
}, Open  ]],
Cell[193155, 4787, 235, 5, 125, "Input"],

Cell[CellGroupData[{
Cell[193415, 4796, 103, 2, 79, "Input"],
Cell[193521, 4800, 80956, 1983, 565, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[274514, 6788, 60, 1, 56, "Input"],
Cell[274577, 6791, 70071, 1724, 504, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[344685, 8520, 68, 1, 56, "Input"],
Cell[344756, 8523, 54634, 1347, 443, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[399427, 9875, 85, 1, 58, "Input"],
Cell[399515, 9878, 54449, 1351, 444, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[454001, 11234, 65, 1, 56, "Input"],
Cell[454069, 11237, 36288, 902, 388, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

