(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    121513,       3224]*)
(*NotebookOutlinePosition[    124995,       3313]*)
(*  CellTagsIndexPosition[    124951,       3309]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Derivative Operators", "Title",
  TextAlignment->Left,
  TextJustification->0,
  FontSize->24],

Cell["Ref: General Relativity - R. M. Wald", "Text"],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[BoxData[
    \(Off[TorsionFreeFlag]\)], "Input"],

Cell[BoxData[
    \(Tdefine[\[Delta], "\<ba\>"]; Tdefine[f, "\<\>"]\)], "Input"],

Cell[CellGroupData[{

Cell["Ordinary Derivative", "Section"],

Cell[BoxData[
    \(DefineOperator[OD, \ "\<\[PartialD]\>", CD]\)], "Input"],

Cell[CellGroupData[{

Cell["Linearity", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, \ A_\  + \ B_]\  := \ OD[la, \ A]\  + OD[la, \ B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[la, RicciR[lb, lc] + MaxwellF[lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["F", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ F, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "a"], 
          MaxwellF[ lb, lc]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["R", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ R, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "a"], 
          RicciR[ lb, lc]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Leibnitz rule", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, \ A_\ B_]\  := \ OD[la, A]\ B\  + \ A\ OD[la, B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[la, \ RicciR[lb, lc] MaxwellF[uc, ud]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "d"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              RicciR[ lb, lc]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "d"],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "d"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              MaxwellF[ uc, ud]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutativity with contraction", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, \[Delta][lb_, uc_]] := 
      0\  /; \ \((DnIndexQ[lb]\  && \ 
              UpIndexQ[uc])\)\  || \ \((UpIndexQ[lb]\  && \ 
              DnIndexQ[uc])\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[ua, \[Delta][lb, uc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutativity with indices", "Subsubsection"],

Cell[BoxData[
    \(OD[la_, OD[lb_, expr_]]\  := 
      OD[lb, OD[la, expr]] /; \ \(! OrderedQ[{la, lb}]\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(OD[lb, OD[la, RicciR[ua, ub]]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox["R", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False]],
              Subsuperscript[ R, 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "b"], 
            RicciR[ ua, ub]],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\[PartialD]", "a"], 
        OD[ lb, 
          RicciR[ ua, ub]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Covariant Derivative", "Section"],

Cell[BoxData[
    \(DefineOperator[{CovD, \ 
        TildeCovD}, \ {"\<\[Del]\>", \*"\"\<\!\(\[Del]\&~\)\>\""}, 
      CD]\)], "Input"],

Cell[CellGroupData[{

Cell["Linearity", "Subsubsection"],

Cell[BoxData[
    \(CovD[la_, \ A_\  + \ B_]\  := \ 
      CovD[la, \ A]\  + \ CovD[la, \ B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[la, RicciR[lb, lc] + MaxwellF[lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["F", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ F, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          MaxwellF[ lb, lc]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["R", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ R, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          RicciR[ lb, lc]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCovD[la_, \ A_\  + \ B_]\  := \ 
      TildeCovD[la, \ A]\  + \ TildeCovD[la, \ B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[la, RicciR[lb, lc] + MaxwellF[lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["F", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ F, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\!\(\[Del]\&~\)", "a"], 
          MaxwellF[ lb, lc]],
        Editable->False], "+", 
      InterpretationBox[
        RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["R", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ R, 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\!\(\[Del]\&~\)", "a"], 
          RicciR[ lb, lc]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Leibnitz rule", "Subsubsection"],

Cell[BoxData[
    \(CovD[la_, \ A_\ B_]\  := \ 
      CovD[la, A]\ B\  + \ A\ CovD[la, B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[la, \ RicciR[lb, lc] MaxwellF[uc, ud]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              RicciR[ lb, lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "d"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "d"],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "d"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              MaxwellF[ uc, ud]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCovD[la_, \ A_\ B_]\  := \ 
      TildeCovD[la, A]\ B\  + \ A\ TildeCovD[la, B]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[la, \ RicciR[lb, lc] MaxwellF[uc, ud]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["F", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "c"], "d"],
                    Editable->False]],
                Subsuperscript[ F, 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "d"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              MaxwellF[ uc, ud]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["F", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"],
              Editable->False]],
          Subsuperscript[ F, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "d"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              RicciR[ lb, lc]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutativity with contraction", "Subsubsection"],

Cell[BoxData[
    \(CovD[la_, \[Delta][lb_, uc_]] := 
      0\  /; \ \((DnIndexQ[lb]\  && \ 
              UpIndexQ[uc])\)\  || \ \((UpIndexQ[lb]\  && \ 
              DnIndexQ[uc])\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[la, \[Delta][lb, uc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCovD[la_, \[Delta][lb_, uc_]] := 
      0\  /; \ \((DnIndexQ[lb]\  && \ 
              UpIndexQ[uc])\)\  || \ \((UpIndexQ[lb]\  && \ 
              DnIndexQ[uc])\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[ua, \[Delta][lb, uc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Consistency with directional derivative", "Subsubsection"],

Cell[BoxData[
    \(CovD[la_, f_[]]\  := \ OD[la, f[]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[la, f[]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\[PartialD]", "a"], 
        f[ ]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(TildeCovD[la_, f_[]]\  := \ OD[la, f[]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[la, f[]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\[PartialD]", "a"], 
        f[ ]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Torsionic and Torsion-free", "Subsubsection"],

Cell[TextData[{
  "Torsionic derivative operator: ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\_a\(\[Del]\_b 
            f\)\  = \ \[Del]\_b\(\[Del]\_a 
              f\)\  - \ \(t\_\(a\[InvisibleSpace]b\)\^p\) \[Del]\_p f\)]]
}], "Text"],

Cell[BoxData[
    \(CovD[la_, CovD[lb_, f_[]]]\  := \ 
      Module[{pair = DummyPair[]}, 
          CovD[lb, CovD[la, f[]]] - \ 
            Torsiont[la, lb, pair[\([2]\)]] 
              CovD[pair[\([1]\)], f[]]] /; \ \(! OrderedQ[{la, 
              lb}]\)\)], "Input"],

Cell[TextData[{
  "Because ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\_a f \[Equal] \[PartialD]\_a f\)]]
}], "Text"],

Cell[BoxData[
    \(CovD[la_, OD[lb_, f_[]]]\  := \ 
      Module[{pair = DummyPair[]}, 
          CovD[lb, OD[la, f[]]] - \ 
            Torsiont[la, lb, pair[\([2]\)]] 
              CovD[pair[\([1]\)], f[]]] /; \ \(! OrderedQ[{la, 
              lb}]\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[lb, CovD[la, f[]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              f[ ]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          OD[ lb, 
            f[ ]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Torsion-free derivative operator: ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\&~\_a\(\[Del]\&~\_b 
            f\)\  = \ \[Del]\&~\_b\(\[Del]\&~\_a f\)\)]]
}], "Text"],

Cell[BoxData[
    \(TildeCovD[la_, OD[lb_, f_[]]]\  := \ 
      TildeCovD[lb, OD[la, f[]]] /; \ \(! OrderedQ[{la, lb}]\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[lb, TildeCovD[la, f[]]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "b"], 
            f[ ]],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\!\(\[Del]\&~\)", "a"], 
        OD[ lb, 
          f[ ]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Relation between Derivative Operators", "Section"],

Cell[BoxData[
    \(Tdefine[{\[Mu], \[Nu], \[Xi], \[Eta]}, "\<a\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f[]\ \[Mu][lb]\  + \ \[Nu][lb]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "b"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      InterpretationBox[
        SubsuperscriptBox["\[Nu]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
            SequenceForm[ "", "b"],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False]],
        Subsuperscript[ \[Nu], 
          SequenceForm[ "", "b"], 
          SequenceForm[ "", " "]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Linearity and locality of ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Del]\_a\)\&~\ \(-\ \[Del]\_a\)\)]]
}], "Subsubsection"],

Cell[TextData[{
  "We show that ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\ \(-\ \[Del]\&~\_a\))\) \((f\ \[Mu]\_b\ \
 + \ \[Nu]\_b)\)\  = \ \(f(\[Del]\_a\ \(-\ \[Del]\&~\_a\))\) \[Mu]\_b\  + \ \
\((\[Del]\_a\ \(-\ \[Del]\&~\_a\))\) \[Nu]\_b\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[la, \ %]\  - \ TildeCovD[la, \ %]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Nu]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Nu], 
              SequenceForm[ "", "b"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          \[Nu][ lb]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              \[Mu][ lb]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              \[Mu][ lb]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Nu]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Nu], 
              SequenceForm[ "", "b"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\!\(\[Del]\&~\)", "a"], 
          \[Nu][ lb]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, there exists a tensor field ",
  Cell[BoxData[
      \(TraditionalForm\`\[Gamma]\_\(a\[InvisibleSpace]b\)\^c\)]],
  " such that ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\&~\_a \[Mu]\_b\  = \ \[Del]\_a \[Mu]\_b\  - \ \
\(\[Gamma]\_\(a\[InvisibleSpace]b\)\^p\) \[Mu]\_p\)]]
}], "Text"],

Cell[BoxData[
    \(Tdefine[\[Gamma], "\<abc\>"]\)], "Input"],

Cell[BoxData[
    \(TildeCovD[la_, \[Mu][lb_]] := 
      Module[{pair = DummyPair[]}, 
        CovD[la, \[Mu][lb]] - \[Gamma][la, lb, pair[\([2]\)]] \[Mu][
              pair[\([1]\)]]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[la, \[Mu][lb]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Mu]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Mu], 
              SequenceForm[ "", "b"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          \[Mu][ lb]],
        Editable->False], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Because ",
  Cell[BoxData[
      \(TraditionalForm\`\((\(\[Del]\_a\)\&~\ \(-\ \[Del]\_a\))\) \
\((\(\[Mu]\_b\) \[Xi]\^b)\)\  = \ 0\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Mu][lb] \[Xi][ub]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[Mu]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
            SequenceForm[ "", "b"],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False]],
        Subsuperscript[ \[Mu], 
          SequenceForm[ "", "b"], 
          SequenceForm[ "", " "]],
        Editable->False], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Xi]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
            SequenceForm[ "", "b"],
            Editable->False]],
        Subsuperscript[ \[Xi], 
          SequenceForm[ "", " "], 
          SequenceForm[ "", "b"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[la, %]\  - \ CovD[la, %] // Expand\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"-", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Xi]", 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False]],
                  Subsuperscript[ \[Xi], 
                    SequenceForm[ "", " "], 
                    SequenceForm[ "", "b"]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[Del]", "a"], 
                \[Xi][ ub]],
              Editable->False], ")"}]}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "b"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\!\(\[Del]\&~\)"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "b"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\!\(\[Del]\&~\)", "a"], 
              \[Xi][ ub]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "b"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
              SequenceForm[ "", "b"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "b"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "we have ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\&~\_a \[Xi]\^c\  = \ \[Del]\_a \[Xi]\^c\  + \ \
\(\[Gamma]\_\(a\[InvisibleSpace]p\)\^c\) \[Xi]\^p\)]]
}], "Text"],

Cell[BoxData[
    \(TildeCovD[la_, \[Xi][uc_]] := 
      Module[{pair = DummyPair[]}, 
        CovD[la, \[Xi][uc]] + \[Gamma][la, pair[\([1]\)], uc] \[Xi][
              pair[\([2]\)]]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TildeCovD[la, \[Xi][uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                SequenceForm[ "", "c"],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "c"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          \[Xi][ uc]],
        Editable->False], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[Gamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ \[Gamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Notice that ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Del]\_a\)\&~\(\[Del]\_b 
              f\)\  - \ \[Del]\_a\(\[Del]\_b 
              f\) = \(-\[Gamma]\_\(a\[InvisibleSpace]b\)\^c\) \[Del]\_c 
            f\)]],
  ". Using the definition for the torsion we find"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((TildeCovD[la, TildeCovD[lb, f[]]] - 
          CovD[la, CovD[lb, f[]]])\) - \((TildeCovD[lb, 
            TildeCovD[la, f[]]] - \ CovD[lb, CovD[la, f[]]])\)\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
            InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
              Subsuperscript[ f, "", ""],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "p"], 
            f[ ]],
          Editable->False], ")"}], " ", 
      InterpretationBox[
        SubsuperscriptBox["t", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"],
            Editable->False]],
        Subsuperscript[ t, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], "p"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, we obtain ",
  Cell[BoxData[
      \(TraditionalForm\`\[Gamma]\_\(a\[InvisibleSpace]b\)\^c\  - \ \
\[Gamma]\_\(b\[InvisibleSpace]a\)\^c\  = \ \(-t\_\(a\[InvisibleSpace]b\)\^c\)\
\)]]
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Section"],

Cell[CellGroupData[{

Cell[TextData[{
  "Linearity and locality of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \(+\
\ t\_\(a\[InvisibleSpace]b\)\^c\)\)\)\) \[Del]\_c\)]]
}], "Subsubsection"],

Cell["We show that", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \((f\ \[Mu]\_c\  + \ \
\[Nu]\_c)\)\  = \ \(f(\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \(+\ 
                          t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \
\[Mu]\_b\[IndentingNewLine]\ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \
\ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \
\ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \  + \ \((\[Del]\_a\(\[Del]\_b\ \(-\ \
\[Del]\_b\(\[Del]\_a\ \(+\ 
                          t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \
\[Nu]\_b\)]]], "Text",
  TextAlignment->Center],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[la, CovD[lb, f[]\ \[Mu][lc]\  + \ \[Nu][lc]]]\  - \ 
        CovD[lb, CovD[la, f[] \[Mu][lc]\  + \ \[Nu][lc]]]\  + \ 
        Torsiont[la, lb, ud] CovD[ld, f[] \[Mu][lc]\  + \ \[Nu][lc]] // 
      Expand\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CovD[ lb, 
            \[Nu][ lc]]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "b"], 
          CovD[ la, 
            \[Nu][ lc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CovD[ lb, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CovD[ la, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "d"], 
              \[Nu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "d"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
                Subsuperscript[ f, "", ""],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "d"], 
              f[ ]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dum[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CovD[ lb, 
            \[Nu][ lc]]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              \[Nu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "b"], 
          CovD[ la, 
            \[Nu][ lc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CovD[ lb, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Mu]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False]],
                    Subsuperscript[ \[Mu], 
                      SequenceForm[ "", "c"], 
                      SequenceForm[ "", " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Mu][ lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CovD[ la, 
                \[Mu][ lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Nu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Nu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Nu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[\(f\_\(\*"\<\"\"\>"\)\%\(\*"\<\"\"\>"\)\),
          Subsuperscript[ f, "", ""],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, there is a tensor ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\[InvisibleSpace]c\)\^d\)]],
  " such that ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \[Mu]\_c\  = \ \(-R\_\
\(a\[InvisibleSpace]b\[InvisibleSpace]c\)\^d\) \[Mu]\_d\)]],
  "."
}], "Text"],

Cell[BoxData[
    \(Tdefine[CurvatureR, "\<R\>", "\<-bacd-abdc\>"]\)], "Input"],

Cell[BoxData[
    \(CovD[la_, CovD[lb_, \[Mu][lc_]]]\  := \ 
      Module[{pair = DummyPair[]}, 
          CovD[lb, CovD[la, \[Mu][lc]]]\  - \ 
            Torsiont[la, lb, pair[\([2]\)]] CovD[pair[\([1]\)], \[Mu][lc]] - 
            CurvatureR[la, lb, lc, pair[\([2]\)]] \[Mu][
                pair[\([1]\)]]]\  /; \ \(! OrderedQ[{la, lb}]\)\  && \ 
          DnIndexQ[lc]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[lb, CovD[la, \[Mu][lc]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Mu][ lc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CovD[ lb, 
            \[Mu][ lc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "For scalar, ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \((\(\[Mu]\_c\) \[Xi]\
\^c)\) = 0\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[la, CovD[lb, \[Mu][lc] \[Xi][uc]]]\  - \ 
      CovD[lb, CovD[la, \[Mu][lc] \[Xi][uc]]]\  + \ 
      Torsiont[la, lb, ud] CovD[ld, \[Mu][lc] \[Xi][uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "c"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Xi][ uc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CovD[ lb, 
                \[Xi][ uc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                        SequenceForm[ "", "c"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "c"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Xi][ uc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CovD[ la, 
                \[Xi][ uc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "d"], 
              \[Xi][ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "c"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "c"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"d"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Mu]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Mu], 
                  SequenceForm[ "", "c"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "d"], 
              \[Mu][ lc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "d"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "d"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "c"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
              SequenceForm[ "", "c"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "c"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dum[%]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "b"], 
                  \[Xi][ "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              CovD[ lb, 
                \[Xi][ "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                        SequenceForm[ "", "p"],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[Del]", "a"], 
                  \[Xi][ "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              CovD[ la, 
                \[Xi][ "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "p"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                    SequenceForm[ "", "q"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "q"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Xi][ "u$2"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
              SequenceForm[ "", "q"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "q"], 
            SequenceForm[ "", " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"q\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "q"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "q"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Mu]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
              SequenceForm[ "", "q"],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False]],
          Subsuperscript[ \[Mu], 
            SequenceForm[ "", "q"], 
            SequenceForm[ "", " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, ",
  Cell[BoxData[
      \(TraditionalForm\`\((\[Del]\_a\(\[Del]\_b\ \(-\ \[Del]\_b\(\[Del]\_a\ \
\(+\ t\_\(a\[InvisibleSpace]b\)\^p\)\)\)\) \[Del]\_p)\) \[Xi]\^c\  = \ \
\(R\_\(a\[InvisibleSpace]b\[InvisibleSpace]p\)\^c\) \[Xi]\^p\)]]
}], "Text"],

Cell[BoxData[
    \(CovD[la_, CovD[lb_, \[Xi][uc_]]]\  := \ 
      Module[{pair = DummyPair[]}, 
          CovD[lb, CovD[la, \[Xi][uc]]]\  - \ 
            Torsiont[la, lb, pair[\([2]\)]] CovD[pair[\([1]\)], \[Xi][uc]] + 
            CurvatureR[la, lb, pair[\([1]\)], uc] \[Xi][
                pair[\([2]\)]]]\  /; \ \(! OrderedQ[{la, lb}]\)\  && \ 
          UpIndexQ[uc]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CovD[lb, CovD[la, \[Xi][uc]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              \[Xi][ uc]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[Del]", "a"], 
          CovD[ lb, 
            \[Xi][ uc]]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "p"], 
              \[Xi][ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["t", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ t, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "p"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "c"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "p"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "c"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Xi]", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
              SequenceForm[ "", "p"],
              Editable->False]],
          Subsuperscript[ \[Xi], 
            SequenceForm[ "", " "], 
            SequenceForm[ "", "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{645, 694},
WindowMargins->{{7, Automatic}, {3, Automatic}},
Visible->True,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"], 
    "Vdefine"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
            OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
            NotationBoxTag, TagStyle -> "NotationTemplateStyle"], " ", 
          "\[DoubleLongLeftRightArrow]", " ", 
          TagBox[ 
            RowBox[ {"Vector", "[", "\[Placeholder]", "]"}], NotationBoxTag, 
            TagStyle -> "NotationTemplateStyle"]}], "]"}], 
    "ivec"->OverscriptBox[ "\[DotlessI]", "^"], 
    "jvec"->OverscriptBox[ "\[DotlessJ]", "^"], 
    "vect"->OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
    "xvec"->OverscriptBox[ "x", "^"], "yvec"->OverscriptBox[ "y", "^"], 
    "zvec"->OverscriptBox[ "z", "^"], "rvec"->OverscriptBox[ "r", "^"], 
    "tvec"->OverscriptBox[ "\[Theta]", "^"], 
    "pvec"->OverscriptBox[ "\[Phi]", "^"]},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 100, 3, 43, "Title"],
Cell[1820, 54, 52, 0, 41, "Text"],
Cell[1875, 56, 52, 1, 56, "Input"],
Cell[1930, 59, 53, 1, 56, "Input"],
Cell[1986, 62, 80, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2091, 67, 38, 0, 78, "Section"],
Cell[2132, 69, 76, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2233, 74, 34, 0, 54, "Subsubsection"],
Cell[2270, 76, 90, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2385, 81, 74, 1, 56, "Input"],
Cell[2462, 84, 2479, 61, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[4990, 151, 38, 0, 32, "Subsubsection"],
Cell[5031, 153, 89, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[5145, 158, 74, 1, 56, "Input"],
Cell[5222, 161, 4741, 113, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10012, 280, 55, 0, 32, "Subsubsection"],
Cell[10070, 282, 194, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[10289, 290, 57, 1, 56, "Input"],
Cell[10349, 293, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[10433, 300, 51, 0, 32, "Subsubsection"],
Cell[10487, 302, 119, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[10631, 308, 63, 1, 56, "Input"],
Cell[10697, 311, 1517, 37, 55, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[12275, 355, 39, 0, 50, "Section"],
Cell[12317, 357, 135, 3, 79, "Input"],

Cell[CellGroupData[{
Cell[12477, 364, 34, 0, 54, "Subsubsection"],
Cell[12514, 366, 105, 2, 56, "Input"],

Cell[CellGroupData[{
Cell[12644, 372, 76, 1, 56, "Input"],
Cell[12723, 375, 2459, 61, 54, "Output"]
}, Open  ]],
Cell[15197, 439, 120, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[15342, 445, 81, 1, 56, "Input"],
Cell[15426, 448, 2495, 61, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[17970, 515, 38, 0, 32, "Subsubsection"],
Cell[18011, 517, 102, 2, 56, "Input"],

Cell[CellGroupData[{
Cell[18138, 523, 76, 1, 56, "Input"],
Cell[18217, 526, 4721, 113, 55, "Output"]
}, Open  ]],
Cell[22953, 642, 117, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[23095, 648, 81, 1, 56, "Input"],
Cell[23179, 651, 4757, 113, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[27985, 770, 55, 0, 32, "Subsubsection"],
Cell[28043, 772, 196, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[28264, 780, 59, 1, 56, "Input"],
Cell[28326, 783, 35, 1, 53, "Output"]
}, Open  ]],
Cell[28376, 787, 201, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[28602, 795, 64, 1, 56, "Input"],
Cell[28669, 798, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[28753, 805, 64, 0, 32, "Subsubsection"],
Cell[28820, 807, 67, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[28912, 812, 46, 1, 56, "Input"],
Cell[28961, 815, 345, 9, 54, "Output"]
}, Open  ]],
Cell[29321, 827, 72, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[29418, 832, 51, 1, 56, "Input"],
Cell[29472, 835, 345, 9, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[29866, 850, 51, 0, 32, "Subsubsection"],
Cell[29920, 852, 245, 6, 41, "Text"],
Cell[30168, 860, 272, 6, 148, "Input"],
Cell[30443, 868, 124, 4, 41, "Text"],
Cell[30570, 874, 268, 6, 148, "Input"],

Cell[CellGroupData[{
Cell[30863, 884, 56, 1, 56, "Input"],
Cell[30922, 887, 2626, 66, 54, "Output"]
}, Open  ]],
Cell[33563, 956, 193, 5, 41, "Text"],
Cell[33759, 963, 133, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[33917, 969, 66, 1, 56, "Input"],
Cell[33986, 972, 595, 16, 54, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[34642, 995, 56, 0, 50, "Section"],
Cell[34701, 997, 80, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[34806, 1002, 63, 1, 56, "Input"],
Cell[34872, 1005, 1137, 29, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36046, 1039, 148, 4, 56, "Subsubsection"],
Cell[36197, 1045, 270, 6, 41, "Text"],

Cell[CellGroupData[{
Cell[36492, 1055, 71, 1, 56, "Input"],
Cell[36566, 1058, 3399, 83, 54, "Output"]
}, Open  ]],
Cell[39980, 1144, 322, 8, 41, "Text"],
Cell[40305, 1154, 61, 1, 56, "Input"],
Cell[40369, 1157, 198, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[40592, 1165, 57, 1, 56, "Input"],
Cell[40652, 1168, 2827, 70, 56, "Output"]
}, Open  ]],
Cell[43494, 1241, 165, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[43684, 1250, 52, 1, 56, "Input"],
Cell[43739, 1253, 955, 25, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44731, 1283, 77, 1, 56, "Input"],
Cell[44811, 1286, 5301, 129, 57, "Output"]
}, Open  ]],
Cell[50127, 1418, 191, 5, 41, "Text"],
Cell[50321, 1425, 198, 4, 102, "Input"],

Cell[CellGroupData[{
Cell[50544, 1433, 57, 1, 56, "Input"],
Cell[50604, 1436, 2827, 70, 56, "Output"]
}, Open  ]],
Cell[53446, 1509, 304, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[53775, 1521, 192, 3, 125, "Input"],
Cell[53970, 1526, 1922, 49, 54, "Output"]
}, Open  ]],
Cell[55907, 1578, 224, 6, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[56180, 1590, 28, 0, 50, "Section"],

Cell[CellGroupData[{
Cell[56233, 1594, 217, 5, 54, "Subsubsection"],
Cell[56453, 1601, 28, 0, 41, "Text"],
Cell[56484, 1603, 722, 11, 63, "Text"],

Cell[CellGroupData[{
Cell[57231, 1618, 241, 4, 125, "Input"],
Cell[57475, 1624, 14486, 352, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[71998, 1981, 39, 1, 56, "Input"],
Cell[72040, 1984, 9512, 230, 77, "Output"]
}, Open  ]],
Cell[81567, 2217, 414, 10, 41, "Text"],
Cell[81984, 2229, 79, 1, 56, "Input"],
Cell[82066, 2232, 386, 7, 217, "Input"],

Cell[CellGroupData[{
Cell[82477, 2243, 62, 1, 56, "Input"],
Cell[82542, 2246, 6347, 157, 56, "Output"]
}, Open  ]],
Cell[88904, 2406, 234, 6, 41, "Text"],

Cell[CellGroupData[{
Cell[89163, 2416, 189, 3, 102, "Input"],
Cell[89355, 2421, 15334, 374, 81, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[104726, 2800, 39, 1, 56, "Input"],
Cell[104768, 2803, 9587, 234, 56, "Output"]
}, Open  ]],
Cell[114370, 3040, 277, 6, 41, "Text"],
Cell[114650, 3048, 386, 7, 217, "Input"],

Cell[CellGroupData[{
Cell[115061, 3059, 62, 1, 56, "Input"],
Cell[115126, 3062, 6347, 157, 56, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

