(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     63759,       1656]*)
(*NotebookOutlinePosition[     67237,       1745]*)
(*  CellTagsIndexPosition[     67193,       1741]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Differential Form Operators", "Subtitle",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Exterior Product", "Section"],

Cell[BoxData[
    \(Fdefine[{f, \[Alpha], \[Beta], \[Gamma]}, {0, 1, 2, 3}]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({XP[f[], f[]], XP[f[], \[Alpha][]], XP[\[Beta][], \[Alpha][]], 
      XP[\[Alpha][], \[Alpha][]], XP[\[Beta][], \[Beta][]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        SuperscriptBox[
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["f",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ f, FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"]],
            Editable->False], "2"], ",", 
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["f",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ f, FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Alpha]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Alpha], FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"]],
            Editable->False]}], ",", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Alpha]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"]],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
            "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Beta]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Beta], FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ 
            \[Alpha][ ], "\[Wedge]", 
            \[Beta][ ]],
          Editable->False], ",", "0", ",", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Beta]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Beta], FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"]],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
            "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Beta]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Beta], FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ 
            \[Beta][ ], "\[Wedge]", 
            \[Beta][ ]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(Dimension = 2;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({XP[\[Alpha][], f[]], XP[\[Alpha][], \[Beta][]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["f",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ f, FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Alpha]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Alpha], FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"]],
            Editable->False]}], ",", "0"}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(Dimension =. ;\)\)], "Input"],

Cell[BoxData[
    \(Fdefine[{f, \[Alpha], \[Beta], \[Gamma]}, {0, 1, 2, 
        3}, "\<a\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(XP[f[la], \[Alpha][la]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox[
          TagBox[
            StyleBox["f",
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)]],
        Subsuperscript[ 
          StyleForm[ f, FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", "a"], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", " "], FontWeight -> "Bold"]],
        Editable->False], " ", 
      InterpretationBox[
        SubsuperscriptBox[
          TagBox[
            StyleBox["\[Alpha]",
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)]],
        Subsuperscript[ 
          StyleForm[ \[Alpha], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", "a"], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", " "], FontWeight -> "Bold"]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \({XP[f[la], \[Alpha][lb]], XP[\[Beta][la], \[Alpha][lb]], 
      XP[\[Alpha][lb], \[Alpha][la]], 
      XP[\[Beta][la], \[Beta][lb]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["f",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ f, FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "a"], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Alpha]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Alpha], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "b"], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"]],
            Editable->False]}], ",", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Alpha]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "b"], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"]],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
            "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Beta]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Beta], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ 
            \[Alpha][ lb], "\[Wedge]", 
            \[Beta][ la]],
          Editable->False], ",", 
        RowBox[{"-", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Alpha]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", " "], FontWeight -> "Bold"]],
                  Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
                "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Alpha]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", "b"], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", " "], FontWeight -> "Bold"]],
                  Editable->False]}],
              SequenceForm[ 
                \[Alpha][ la], "\[Wedge]", 
                \[Alpha][ lb]],
              Editable->False], ")"}]}], ",", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Beta]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Beta], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"]],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
            "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Beta]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Beta], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "b"], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ 
            \[Beta][ la], "\[Wedge]", 
            \[Beta][ lb]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(XP[\[Alpha][ua], \[Alpha][la]]\)\(\ \ \)\( (*\ 
      should\ be\ zero\ *) \)\)\)], "Input"],

Cell[BoxData[
    RowBox[{"-", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Alpha]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"]],
              Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
            "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["\[Alpha]",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "a"], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ 
            \[Alpha][ la], "\[Wedge]", 
            \[Alpha][ ua]],
          Editable->False], ")"}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(Dimension = 2;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({XP[\[Alpha][la], f[lb]], XP[\ \[Alpha][lb], \[Beta][la]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["f",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ f, FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "b"], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Alpha]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Alpha], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "a"], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"]],
            Editable->False]}], ",", "0"}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(XP[\[Alpha][la] + f[la], \[Beta][ua]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox[
          TagBox[
            StyleBox["f",
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)]],
        Subsuperscript[ 
          StyleForm[ f, FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", "a"], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", " "], FontWeight -> "Bold"]],
        Editable->False], " ", 
      InterpretationBox[
        SubsuperscriptBox[
          TagBox[
            StyleBox["\[Beta]",
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)], 
          TagBox[
            StyleBox[
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False],
              FontWeight->"Bold"],
            (StyleForm[ #, FontWeight -> "Bold"]&)]],
        Subsuperscript[ 
          StyleForm[ \[Beta], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", " "], FontWeight -> "Bold"], 
          StyleForm[ 
            SequenceForm[ "", "a"], FontWeight -> "Bold"]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exterior Derivative", "Section"],

Cell[BoxData[
    \(\(Dimension =. ;\)\)], "Input"],

Cell[BoxData[
    \(Fdefine[{f, \[Alpha], \[Beta]}, {0, 1, 2}]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({XD[f[]], XD[\(-\[Alpha][]\)], XD[a\ \[Beta][]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["f",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox["\<\"\"\>",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ f, FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"], 
                StyleForm[ "", FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ "d", 
            f[ ]],
          Editable->False], ",", 
        RowBox[{"-", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Alpha]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox["\<\"\"\>",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox["\<\"\"\>",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                    StyleForm[ "", FontWeight -> "Bold"], 
                    StyleForm[ "", FontWeight -> "Bold"]],
                  Editable->False]}],
              SequenceForm[ "d", 
                \[Alpha][ ]],
              Editable->False], ")"}]}], ",", 
        RowBox[{"a", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Beta]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox["\<\"\"\>",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox["\<\"\"\>",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Beta], FontWeight -> "Bold"], 
                    StyleForm[ "", FontWeight -> "Bold"], 
                    StyleForm[ "", FontWeight -> "Bold"]],
                  Editable->False]}],
              SequenceForm[ "d", 
                \[Beta][ ]],
              Editable->False], ")"}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(XD[XP[\[Beta][], \[Alpha][]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Alpha]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Alpha][ ]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Beta]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Beta], FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ 
          XD[ 
            \[Alpha][ ]], "\[Wedge]", 
          \[Beta][ ]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Alpha]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox["\<\"\"\>",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Alpha], FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"], 
              StyleForm[ "", FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Beta]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Beta], FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Beta][ ]],
            Editable->False]}],
        SequenceForm[ 
          \[Alpha][ ], "\[Wedge]", 
          XD[ 
            \[Beta][ ]]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(XD[f[] XP[\[Beta][], \[Alpha][]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[
            TagBox[
              StyleBox["f",
                FontWeight->"Bold"],
              (StyleForm[ #, FontWeight -> "Bold"]&)], 
            TagBox[
              StyleBox["\<\"\"\>",
                FontWeight->"Bold"],
              (StyleForm[ #, FontWeight -> "Bold"]&)], 
            TagBox[
              StyleBox["\<\"\"\>",
                FontWeight->"Bold"],
              (StyleForm[ #, FontWeight -> "Bold"]&)]],
          Subsuperscript[ 
            StyleForm[ f, FontWeight -> "Bold"], 
            StyleForm[ "", FontWeight -> "Bold"], 
            StyleForm[ "", FontWeight -> "Bold"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox[
                      TagBox[
                        StyleBox["\[Alpha]",
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)], 
                      TagBox[
                        StyleBox["\<\"\"\>",
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)], 
                      TagBox[
                        StyleBox["\<\"\"\>",
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)]],
                    Subsuperscript[ 
                      StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                      StyleForm[ "", FontWeight -> "Bold"], 
                      StyleForm[ "", FontWeight -> "Bold"]],
                    Editable->False]}],
                SequenceForm[ "d", 
                  \[Alpha][ ]],
                Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
              "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Beta]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Beta], FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ 
              XD[ 
                \[Alpha][ ]], "\[Wedge]", 
              \[Beta][ ]],
            Editable->False], ")"}]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox[
            TagBox[
              StyleBox["f",
                FontWeight->"Bold"],
              (StyleForm[ #, FontWeight -> "Bold"]&)], 
            TagBox[
              StyleBox["\<\"\"\>",
                FontWeight->"Bold"],
              (StyleForm[ #, FontWeight -> "Bold"]&)], 
            TagBox[
              StyleBox["\<\"\"\>",
                FontWeight->"Bold"],
              (StyleForm[ #, FontWeight -> "Bold"]&)]],
          Subsuperscript[ 
            StyleForm[ f, FontWeight -> "Bold"], 
            StyleForm[ "", FontWeight -> "Bold"], 
            StyleForm[ "", FontWeight -> "Bold"]],
          Editable->False], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Alpha]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"]],
                Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
              "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox[
                      TagBox[
                        StyleBox["\[Beta]",
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)], 
                      TagBox[
                        StyleBox["\<\"\"\>",
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)], 
                      TagBox[
                        StyleBox["\<\"\"\>",
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)]],
                    Subsuperscript[ 
                      StyleForm[ \[Beta], FontWeight -> "Bold"], 
                      StyleForm[ "", FontWeight -> "Bold"], 
                      StyleForm[ "", FontWeight -> "Bold"]],
                    Editable->False]}],
                SequenceForm[ "d", 
                  \[Beta][ ]],
                Editable->False]}],
            SequenceForm[ 
              \[Alpha][ ], "\[Wedge]", 
              XD[ 
                \[Beta][ ]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["f",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ f, FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              f[ ]],
            Editable->False], ")"}], " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Alpha]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"]],
                Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
              "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Beta]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox["\<\"\"\>",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Beta], FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"], 
                  StyleForm[ "", FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ 
              \[Alpha][ ], "\[Wedge]", 
              \[Beta][ ]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Fdefine[{f, \[Alpha], \[Beta]}, {0, 1, 2}, "\<a\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({XD[f[la]], XD[\(-\[Alpha][la]\)], XD[a\ \[Beta][la]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
            InterpretationBox[
              SubsuperscriptBox[
                TagBox[
                  StyleBox["f",
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)], 
                TagBox[
                  StyleBox[
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False],
                    FontWeight->"Bold"],
                  (StyleForm[ #, FontWeight -> "Bold"]&)]],
              Subsuperscript[ 
                StyleForm[ f, FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                StyleForm[ 
                  SequenceForm[ "", " "], FontWeight -> "Bold"]],
              Editable->False]}],
          SequenceForm[ "d", 
            f[ la]],
          Editable->False], ",", 
        RowBox[{"-", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Alpha]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", " "], FontWeight -> "Bold"]],
                  Editable->False]}],
              SequenceForm[ "d", 
                \[Alpha][ la]],
              Editable->False], ")"}]}], ",", 
        RowBox[{"a", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox[
                    TagBox[
                      StyleBox["\[Beta]",
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)], 
                    TagBox[
                      StyleBox[
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False],
                        FontWeight->"Bold"],
                      (StyleForm[ #, FontWeight -> "Bold"]&)]],
                  Subsuperscript[ 
                    StyleForm[ \[Beta], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                    StyleForm[ 
                      SequenceForm[ "", " "], FontWeight -> "Bold"]],
                  Editable->False]}],
              SequenceForm[ "d", 
                \[Beta][ la]],
              Editable->False], ")"}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(XD[XP[\[Beta][la], \[Alpha][ua]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Alpha]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", " "], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", "a"], FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Alpha][ ua]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Beta]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Beta], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "a"], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"]],
            Editable->False]}],
        SequenceForm[ 
          XD[ 
            \[Alpha][ ua]], "\[Wedge]", 
          \[Beta][ la]],
        Editable->False], "-", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Alpha]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Alpha], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "a"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Beta]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Beta], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", " "], FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Beta][ la]],
            Editable->False]}],
        SequenceForm[ 
          \[Alpha][ ua], "\[Wedge]", 
          XD[ 
            \[Beta][ la]]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(XD[XP[\[Alpha][la], \[Alpha][ua]]]\)\(\ \ \)\( (*\ 
      should\ be\ zero\ *) \)\)\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Alpha]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", " "], FontWeight -> "Bold"]],
                Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
              "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox[
                      TagBox[
                        StyleBox["\[Alpha]",
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)], 
                      TagBox[
                        StyleBox[
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False],
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)], 
                      TagBox[
                        StyleBox[
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False],
                          FontWeight->"Bold"],
                        (StyleForm[ #, FontWeight -> "Bold"]&)]],
                    Subsuperscript[ 
                      StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                      StyleForm[ 
                        SequenceForm[ "", " "], FontWeight -> "Bold"], 
                      StyleForm[ 
                        SequenceForm[ "", "a"], FontWeight -> "Bold"]],
                    Editable->False]}],
                SequenceForm[ "d", 
                  \[Alpha][ ua]],
                Editable->False]}],
            SequenceForm[ 
              \[Alpha][ la], "\[Wedge]", 
              XD[ 
                \[Alpha][ ua]]],
            Editable->False], ")"}]}], "+", 
      InterpretationBox[
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox[
              TagBox[
                StyleBox["\[Alpha]",
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)], 
              TagBox[
                StyleBox[
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False],
                  FontWeight->"Bold"],
                (StyleForm[ #, FontWeight -> "Bold"]&)]],
            Subsuperscript[ 
              StyleForm[ \[Alpha], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", " "], FontWeight -> "Bold"], 
              StyleForm[ 
                SequenceForm[ "", "a"], FontWeight -> "Bold"]],
            Editable->False], "\[InvisibleSpace]", "\<\"\[Wedge]\"\>", 
          "\[InvisibleSpace]", 
          InterpretationBox[
            RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox[
                  TagBox[
                    StyleBox["\[Alpha]",
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)], 
                  TagBox[
                    StyleBox[
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False],
                      FontWeight->"Bold"],
                    (StyleForm[ #, FontWeight -> "Bold"]&)]],
                Subsuperscript[ 
                  StyleForm[ \[Alpha], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", "a"], FontWeight -> "Bold"], 
                  StyleForm[ 
                    SequenceForm[ "", " "], FontWeight -> "Bold"]],
                Editable->False]}],
            SequenceForm[ "d", 
              \[Alpha][ la]],
            Editable->False]}],
        SequenceForm[ 
          \[Alpha][ ua], "\[Wedge]", 
          XD[ 
            \[Alpha][ la]]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{664, 694},
WindowMargins->{{18, Automatic}, {Automatic, 0}},
Visible->True,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"], 
    "Vdefine"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
            OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
            NotationBoxTag, TagStyle -> "NotationTemplateStyle"], " ", 
          "\[DoubleLongLeftRightArrow]", " ", 
          TagBox[ 
            RowBox[ {"Vector", "[", "\[Placeholder]", "]"}], NotationBoxTag, 
            TagStyle -> "NotationTemplateStyle"]}], "]"}], 
    "ivec"->OverscriptBox[ "\[DotlessI]", "^"], 
    "jvec"->OverscriptBox[ "\[DotlessJ]", "^"], 
    "vect"->OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
    "xvec"->OverscriptBox[ "x", "^"], "yvec"->OverscriptBox[ "y", "^"], 
    "zvec"->OverscriptBox[ "z", "^"], "rvec"->OverscriptBox[ "r", "^"], 
    "tvec"->OverscriptBox[ "\[Theta]", "^"], 
    "pvec"->OverscriptBox[ "\[Phi]", "^"]},
Magnification->1,
StyleDefinitions -> "Demo.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 96, 2, 81, "Subtitle"],
Cell[1816, 53, 52, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[1893, 58, 35, 0, 77, "Section"],
Cell[1931, 60, 88, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[2044, 65, 156, 2, 77, "Input"],
Cell[2203, 69, 6010, 151, 56, "Output"]
}, Open  ]],
Cell[8228, 223, 51, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[8304, 228, 81, 1, 54, "Input"],
Cell[8388, 231, 1591, 41, 55, "Output"]
}, Open  ]],
Cell[9994, 275, 51, 1, 54, "Input"],
Cell[10048, 278, 106, 2, 54, "Input"],

Cell[CellGroupData[{
Cell[10179, 284, 56, 1, 54, "Input"],
Cell[10238, 287, 2047, 55, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[12322, 347, 165, 3, 77, "Input"],
Cell[12490, 352, 10480, 250, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23007, 607, 116, 2, 54, "Input"],
Cell[23126, 611, 2656, 63, 55, "Output"]
}, Open  ]],
Cell[25797, 677, 51, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[25873, 682, 91, 1, 54, "Input"],
Cell[25967, 685, 2319, 57, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[28323, 747, 70, 1, 54, "Input"],
Cell[28396, 750, 2045, 55, 55, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[30490, 811, 38, 0, 49, "Section"],
Cell[30531, 813, 51, 1, 54, "Input"],
Cell[30585, 816, 75, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[30685, 821, 81, 1, 54, "Input"],
Cell[30769, 824, 3229, 78, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34035, 907, 62, 1, 54, "Input"],
Cell[34100, 910, 4040, 103, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[38177, 1018, 66, 1, 54, "Input"],
Cell[38246, 1021, 8930, 217, 55, "Output"]
}, Open  ]],
Cell[47191, 1241, 84, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[47300, 1246, 87, 1, 54, "Input"],
Cell[47390, 1249, 4547, 110, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[51974, 1364, 66, 1, 54, "Input"],
Cell[52043, 1367, 5572, 135, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[57652, 1507, 120, 2, 54, "Input"],
Cell[57775, 1511, 5956, 141, 55, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

