(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    326293,       7977]*)
(*NotebookOutlinePosition[    327061,       8004]*)
(*  CellTagsIndexPosition[    327017,       8000]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Linearized Field Equation and Linealized Gravitational Waves", "Title",
  FontSize->24],

Cell["\<\
See: Introducing Einstein's Relativity, Ray D'Inverno, Chap 20, pp 271\
\>", "Text"],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Linealization", "Section"],

Cell[BoxData[
    \(\(Tdefine[{\[Eta], h}, "\<ba\>"];\)\)], "Input"],

Cell[TextData[{
  "Let ",
  Cell[BoxData[
      \(TraditionalForm\`g\_\(a\[InvisibleSpace]b\)\  = \ \[Eta]\_\(a\
\[InvisibleSpace]b\)\  + \ \[Epsilon]\ h\_\(a\[InvisibleSpace]b\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(metricRule\  = 
      Trule[Metricg[la_, 
          lb_], \[Eta][la, lb]\  + \ \[Epsilon]\ h[la, lb]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "lb_"}],
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", 
              Pattern[ la, 
                Blank[ ]]], 
            Pattern[ lb, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            RowBox[{"\[Epsilon]", " ", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}], "+", 
            InterpretationBox[
              SubsuperscriptBox["\[Eta]", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ \[Eta], 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}], "]"}], "/;", "True"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(etaRule\  = {Trule[BD[la_, \[Eta][lb_, lc_]], 0], \ 
        Trule[\[Eta][la_, ua_], 4, PairQ[{la, ua}]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^la_\),
                Superscript[ "\[PartialD]", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Eta]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]lb_\),
                        SequenceForm[ "", 
                          Pattern[ lb, 
                            Blank[ ]]],
                        Editable->False], "\[InvisibleSpace]", "lc_"}],
                    SequenceForm[ 
                      SequenceForm[ "", 
                        Pattern[ lb, 
                          Blank[ ]]], 
                      Pattern[ lc, 
                        Blank[ ]]],
                    Editable->False]],
                Subsuperscript[ \[Eta], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      Pattern[ lb, 
                        Blank[ ]]], 
                    Pattern[ lc, 
                      Blank[ ]]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", 
                Pattern[ la, 
                  Blank[ ]]], 
              \[Eta][ 
                Pattern[ lb, 
                  Blank[ ]], 
                Pattern[ lc, 
                  Blank[ ]]]],
            Editable->False], "\[RuleDelayed]", \(DumFresh[0] /; True\)}], 
        ",", 
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Eta]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "ua_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ ua, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ \[Eta], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ ua, 
                  Blank[ ]]]],
            Editable->False], 
          "\[RuleDelayed]", \(DumFresh[4] /; PairQ[{la, ua}]\)}]}], 
      "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "We shall neglect terms of second order or higher in ",
  Cell[BoxData[
      \(TraditionalForm\`\[Epsilon]\)]],
  " and raise and lower indices with ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta]\_\(a\[InvisibleSpace]b\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell["Connection", "Subsection"],

Cell[BoxData[
    \(\(\(SetAttributes[\[Epsilon], Constant]\)\(\ \)\( (*\ 
      for\ BD\ *) \)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(affine = AffineG[la, lb, uc]\  // \ AffineToMetric\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"c"\),
                Superscript[ "\[PartialD]", "c"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "c"], 
              Metricg[ la, lb]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              Metricg[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              Metricg[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[affine, Join[{metricRule}, etaRule]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", \(\[Epsilon]\^2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              h[ la, lb]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["h", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ h, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", \(\[Epsilon]\^2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              h[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["h", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ h, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", \(\[Epsilon]\^2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              h[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["h", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ h, 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              h[ la, lb]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              h[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              h[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Normal[Series[%, {\[Epsilon], 0, 1}]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              h[ la, lb]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              h[ lb, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "p"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              h[ la, "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affine = Absorb[%, \[Eta], {BD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              h[ lb, uc]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], "c"],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], " "], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              h[ la, uc]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"c"\),
                Superscript[ "\[PartialD]", "c"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["h", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ h, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "c"], 
              h[ la, lb]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(affineRule = Trule[AffineG[la_, lb_, uc_], affine]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[CapitalGamma]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lb_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "uc_"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]], 
              Pattern[ uc, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ \[CapitalGamma], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]], 
            Pattern[ uc, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                              SequenceForm[ "", "b"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "b"], " "],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"c\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], "c"],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], "c"]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "a"], 
                    h[ lb, uc]],
                  Editable->False], ")"}]}], "+", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                              SequenceForm[ "", "a"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "a"], " "],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"c\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], "c"],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], "c"]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "b"], 
                    h[ la, uc]],
                  Editable->False], ")"}]}], "-", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\("\[PartialD]"\^"c"\),
                      Superscript[ "\[PartialD]", "c"],
                      Editable->False], "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                              SequenceForm[ "", "a"],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\"b\"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "b"],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Superscript[ "\[PartialD]", "c"], 
                    h[ la, lb]],
                  Editable->False], ")"}]}]}], "]"}], "/;", 
        "True"}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RiemannR[la, lb, lc, ld]\  // \ RiemannToAffine\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"-", 
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "p"], "d"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "p"], "d"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], "d"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"], "d"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[CapitalGamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[CapitalGamma], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              AffineG[ lb, lc, "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[CapitalGamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[CapitalGamma], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              AffineG[ la, lc, "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Normal[
      Series[ApplyRules[%, 
          Join[{metricRule}, {affineRule}, etaRule]], {\[Epsilon], 0, 
          1}]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                h[ lb, lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ lb, "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              BD[ "u$1", 
                h[ la, lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              BD[ lc, 
                h[ la, "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(riemann = Absorb[%, \[Eta], {BD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ld]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ lb, ld]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ ld, 
                h[ lb, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              BD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              BD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Eta][ub, ud] riemann\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ld]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ lb, ld]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "d"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "b"], "d"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ ld, 
                h[ lb, lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "d"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "b"], "d"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              BD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "d"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "b"], "d"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              BD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "d"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "b"], "d"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricci = Absorb[%, \[Eta], {BD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "b"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ub]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ lb, ub]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"b"\),
                    Superscript[ "\[PartialD]", "b"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "b"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ ub, 
                h[ lb, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Eta][ua, uc] ricci\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "b"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ub]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ lb, ub]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"b"\),
                    Superscript[ "\[PartialD]", "b"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "b"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ ub, 
                h[ lb, lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(scalar = Dum[Absorb[%, \[Eta], {BD}]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{"\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  h[ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "u$2", 
                h[ "l$2", "l$1"]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricci\  - \ \(1\/2\) Metricg[la, lc] scalar\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "b"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ub]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ lb, ub]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"b"\),
                    Superscript[ "\[PartialD]", "b"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "b"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ ub, 
                h[ lb, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  h[ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "u$2", 
                h[ "l$2", "l$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(einstein = 
      Normal[Series[
          ApplyRules[%, Join[{metricRule}, etaRule]], {\[Epsilon], 0, 
            1}]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                h[ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ lc, 
                h[ la, "l$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ lc, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                h[ lc, "l$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ "l$1", "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ "l$1", "u$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  h[ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "u$2", 
                h[ "l$2", "l$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Bianchi identities", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(curvRule = Trule[RiemannR[la_, lb_, lc_, ld_], riemann]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["R", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                        SequenceForm[ "", 
                          Pattern[ la, 
                            Blank[ ]]],
                        Editable->False], "\[InvisibleSpace]", "lb_"}],
                    SequenceForm[ 
                      SequenceForm[ "", 
                        Pattern[ la, 
                          Blank[ ]]], 
                      Pattern[ lb, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lc_"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]], 
                    Pattern[ lb, 
                      Blank[ ]]], 
                  Pattern[ lc, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "ld_"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]], 
                Pattern[ lc, 
                  Blank[ ]]], 
              Pattern[ ld, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]], 
              Pattern[ lc, 
                Blank[ ]]], 
            Pattern[ ld, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                  SequenceForm[ "", "b"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"d\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "b"], "d"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "d"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "c"], 
                        h[ lb, ld]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "a"], 
                    BD[ lc, 
                      h[ lb, ld]]],
                  Editable->False], ")"}]}], "-", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                  SequenceForm[ "", "b"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"c\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "b"], "c"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "c"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "d"], 
                        h[ lb, lc]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "a"], 
                    BD[ ld, 
                      h[ lb, lc]]],
                  Editable->False], ")"}]}], "-", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                  SequenceForm[ "", "a"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"d\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "a"], "d"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "d"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "c"], 
                        h[ la, ld]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "b"], 
                    BD[ lc, 
                      h[ la, ld]]],
                  Editable->False], ")"}]}], "+", 
            RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                        InterpretationBox[
                          SubsuperscriptBox["h", 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                  SequenceForm[ "", "a"],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\"c\"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", "a"], "c"],
                              Editable->False], 
                            InterpretationBox[
                              RowBox[{
                                
                                InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                  SequenceForm[ "", " "],
                                  Editable->False], 
                                "\[InvisibleSpace]", "\<\" \"\>"}],
                              SequenceForm[ 
                                SequenceForm[ "", " "], " "],
                              Editable->False]],
                          Subsuperscript[ h, 
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "c"], 
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "]],
                          Editable->False]}],
                      SequenceForm[ 
                        Subscript[ "\[PartialD]", "d"], 
                        h[ la, lc]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "b"], 
                    BD[ ld, 
                      h[ la, lc]]],
                  Editable->False], ")"}]}]}], "]"}], "/;", 
        "True"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(bianchi1 = 
      Antisymmetrize[{la, lb, lc}, 
        BD[la, RiemannR[lb, lc, ld, le]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/3\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"b"\),
                                SequenceForm[ "", "b"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "b"], "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "b"], "c"], "d"],
                        Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"], "d"], "e"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], "d"], "e"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              RiemannR[ lb, lc, ld, le]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/3\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"c\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "c"], "d"],
                        Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"], "d"], "e"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], "d"], "e"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              RiemannR[ la, lc, ld, le]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/3\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["R", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]"a"\),
                                SequenceForm[ "", "a"],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\"b\"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", "a"], "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", "a"], "b"], "d"],
                        Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"], "d"], "e"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "], " "], " "],
                    Editable->False]],
                Subsuperscript[ R, 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], "d"], "e"], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "c"], 
              RiemannR[ la, lb, ld, le]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell["The Bianchi identy is satisfied", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[bianchi1, 
      Join[{curvRule}, Join[{metricRule}, etaRule]]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Gauge transformations", "Subsection"],

Cell[BoxData[
    \(Tdefine[\[Xi], "\<a\>"]\)], "Input"],

Cell[TextData[{
  "Under the coordinate transformation ",
  Cell[BoxData[
      \(TraditionalForm\`x'\^a\  = \ x\^a\  + \ \[Epsilon]\ \[Xi]\^a\)]],
  ", the metric transforms ",
  Cell[BoxData[
      \(TraditionalForm\`\(h\_\(a\[InvisibleSpace]b\)\%'\)\  = \ 
        h\_\(a\[InvisibleSpace]b\)\  - \ 
          2 \[PartialD]\_\((a\)\[Xi]\_\(\(b\)\()\)\)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(gaugeTr = 
      Trule[h[la_, lb_], 
        h[la, lb]\  - \ BD[la, \[Xi][lb]]\  - \ BD[lb, \[Xi][la]]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["h", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "lb_"}],
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ h, 
          SequenceForm[ 
            SequenceForm[ "", " "], " "], 
          SequenceForm[ 
            SequenceForm[ "", 
              Pattern[ la, 
                Blank[ ]]], 
            Pattern[ lb, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          RowBox[{
            RowBox[{"-", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["\[Xi]", 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False]],
                      Subsuperscript[ \[Xi], 
                        SequenceForm[ "", "b"], 
                        SequenceForm[ "", " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", "a"], 
                    \[Xi][ lb]],
                  Editable->False], ")"}]}], "-", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Xi]", 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], 
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False]],
                  Subsuperscript[ \[Xi], 
                    SequenceForm[ "", "a"], 
                    SequenceForm[ "", " "]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", "b"], 
                \[Xi][ la]],
              Editable->False], "+", 
            InterpretationBox[
              SubsuperscriptBox["h", 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], 
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False]],
              Subsuperscript[ h, 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "]],
              Editable->False]}], "]"}], "/;", "True"}]}]], "Output"]
}, Open  ]],

Cell["\<\
The Riemann and the Ricci tensors are invariant under the gauge \
transformation\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(riemann\  - \ ApplyRules[riemann, {gaugeTr}]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ricci\  - \ ApplyRules[ricci, {gaugeTr}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                h[ la, lc]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "c"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "p"], 
                      \[Xi][ lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  BD[ "l$1", 
                    \[Xi][ lc]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ la, 
                BD[ "l$1", 
                  \[Xi][ lc]]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "p"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      \[Xi][ "l$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "a"], 
                  BD[ lc, 
                    \[Xi][ "l$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ la, 
                BD[ lc, 
                  \[Xi][ "l$1"]]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ lc, 
                h[ la, "l$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "c"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "p"], 
                      \[Xi][ lc]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  BD[ "l$1", 
                    \[Xi][ lc]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                BD[ "l$1", 
                  \[Xi][ lc]]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "p"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "c"], 
                      \[Xi][ "l$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  BD[ lc, 
                    \[Xi][ "l$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                BD[ lc, 
                  \[Xi][ "l$1"]]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ lc, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                h[ lc, "l$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", " "], 
                          SequenceForm[ "", "p"]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", "p"], 
                      \[Xi][ "u$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  BD[ "l$1", 
                    \[Xi][ "u$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                BD[ "l$1", 
                  \[Xi][ "u$1"]]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\("\[PartialD]"\^"p"\),
                        Superscript[ "\[PartialD]", "p"],
                        Editable->False], "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", "p"], 
                          SequenceForm[ "", " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Superscript[ "\[PartialD]", "p"], 
                      \[Xi][ "l$1"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  BD[ "u$1", 
                    \[Xi][ "l$1"]]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                BD[ "u$1", 
                  \[Xi][ "l$1"]]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ "l$1", "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ "l$1", "u$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "b"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "b"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ lb, ub]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ lb, ub]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"b"\),
                    Superscript[ "\[PartialD]", "b"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "b"], 
                  h[ lb, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ ub, 
                h[ lb, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"d\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "d"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "d"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ la, ld]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ lc, 
                h[ la, ld]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"d"\),
                Superscript[ "\[PartialD]", "d"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"d"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "d"], 
                  h[ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "d"], 
              BD[ ld, 
                h[ la, lc]]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(% // Dum\) // Can[#, {BD}] &\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Let ",
  Cell[BoxData[
      \(TraditionalForm\`\[Psi]\_\(a\[InvisibleSpace]b\)\  = \ 
        h\_\(a\[InvisibleSpace]b\)\  - \ \(1\/2\) \(\[Eta]\_\(a\
\[InvisibleSpace]b\)\) h\)]]
}], "Text"],

Cell[BoxData[
    \(Tdefine[\[Psi], "\<ba\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(psiRule\  = \ {Trule[
          h[la_, lb_], \[Psi][la, lb]\  + \ \(1\/2\) \[Eta][la, lb] 
              h[lc, uc]], \ 
        Trule[\[Psi][la_, ua_], \(-h[la, ua]\), PairQ[{la, ua}]]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["h", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lb_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ h, 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]]],
            Editable->False], "\[RuleDelayed]", 
          RowBox[{
            RowBox[{"DumFresh", "[", 
              RowBox[{
                RowBox[{\(1\/2\), " ", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$\
24"\),
                            SequenceForm[ "", "$24"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"   \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "$24"], "   "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"  \
 "\),
                            SequenceForm[ "", "   "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"$24\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "   "], "$24"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "$24"], "   "], 
                      SequenceForm[ 
                        SequenceForm[ "", "   "], "$24"]],
                    Editable->False], " ", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Eta]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"b\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "b"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Eta], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}], "+", 
                InterpretationBox[
                  SubsuperscriptBox["\[Psi]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "b"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Psi], 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}], "]"}], "/;", "True"}]}], ",", 
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Psi]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "ua_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ ua, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ \[Psi], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ ua, 
                  Blank[ ]]]],
            Editable->False], "\[RuleDelayed]", 
          RowBox[{
            RowBox[{"DumFresh", "[", 
              RowBox[{"-", 
                InterpretationBox[
                  SubsuperscriptBox["h", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"$\
25"\),
                          SequenceForm[ "", "$25"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"   \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "$25"], "   "],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"   \
"\),
                          SequenceForm[ "", "   "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"$25\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "   "], "$25"],
                      Editable->False]],
                  Subsuperscript[ h, 
                    SequenceForm[ 
                      SequenceForm[ "", "$25"], "   "], 
                    SequenceForm[ 
                      SequenceForm[ "", "   "], "$25"]],
                  Editable->False]}], "]"}], "/;", \(PairQ[{la, ua}]\)}]}]}], 
      "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Under the gauge transformation ",
  Cell[BoxData[
      \(TraditionalForm\`\[Psi]\_\(a\[InvisibleSpace]b\)\)]],
  " transfoms as"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[h[la, lb]\  - \ \(1\/2\) \[Eta][la, lb] h[lc, uc], 
      Join[{gaugeTr}, psiRule, etaRule]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              \[Xi][ lb]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", "a"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "b"], 
          \[Xi][ la]],
        Editable->False], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              \[Xi][ "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", "p"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              \[Xi][ "l$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      InterpretationBox[
        SubsuperscriptBox["\[Psi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ \[Psi], 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% // Can[#, {BD}] &\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                    SequenceForm[ "", "b"],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", "b"], 
                  SequenceForm[ "", " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              \[Xi][ lb]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", "a"], 
              SequenceForm[ "", " "]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "b"], 
          \[Xi][ la]],
        Editable->False], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[Xi]", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                    SequenceForm[ "", "p"],
                    Editable->False]],
                Subsuperscript[ \[Xi], 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              \[Xi][ "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      InterpretationBox[
        SubsuperscriptBox["\[Psi]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ \[Psi], 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "from which we find ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\_a \[Psi]'\^\(a\[InvisibleSpace]b\)\  = \
\ \[PartialD]\_a \[Psi]\^\(a\[InvisibleSpace]b\)\  - \ \[PartialD]\^2 \
\[Xi]\^b\)]]
}], "Text"],

Cell["The Einstein tensor becomes", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[einstein, Join[psiRule, etaRule]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  \[Psi][ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                \[Psi][ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  \[Psi][ la, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ lc, 
                \[Psi][ la, "l$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  \[Psi][ lc, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                \[Psi][ lc, "l$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"p"\),
                            SequenceForm[ "", "p"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "p"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "p"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ "l$1", "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ lc, 
                h[ "l$1", "u$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ lc, 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  \[Psi][ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "u$2", 
                \[Psi][ "l$2", "l$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"r"\),
                            SequenceForm[ "", "r"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "r"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"r\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "r"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "r"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "r"]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  h[ "l$3", "u$3"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "u$2", 
                h[ "l$3", "u$3"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"q"\),
                  SequenceForm[ "", "q"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "q"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "q"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                  SequenceForm[ "", "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "c"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "c"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Absorb[%, \[Eta], {BD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  \[Psi][ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                \[Psi][ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "c"], 
                  \[Psi][ la, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ lc, 
                \[Psi][ la, "l$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  \[Psi][ lc, "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "u$1", 
                \[Psi][ lc, "l$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              BD[ "u$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/4\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"q\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "q"],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "q"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  h[ "l$2", "u$2"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "l$1", 
                h[ "l$2", "u$2"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^"p"\),
                Superscript[ "\[PartialD]", "p"],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"q"\),
                    Superscript[ "\[PartialD]", "q"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "q"], 
                  \[Psi][ "l$2", "l$1"]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", "p"], 
              BD[ "u$2", 
                \[Psi][ "l$2", "l$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq = Can[%, {BD}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^"p"\),
                    Superscript[ "\[PartialD]", "p"],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", "p"], 
                  \[Psi][ la, lc]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              BD[ "u$1", 
                \[Psi][ la, lc]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"c"\),
                            SequenceForm[ "", "c"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "c"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "c"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  \[Psi][ lc, "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              BD[ "l$1", 
                \[Psi][ lc, "u$1"]]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"c"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], "p"],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "p"], 
                  \[Psi][ la, "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "c"], 
              BD[ "l$1", 
                \[Psi][ la, "u$1"]]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_"q"\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["\[Psi]", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"q"\),
                            SequenceForm[ "", "q"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"p\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "q"], "p"],
                        Editable->False]],
                    Subsuperscript[ \[Psi], 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", "q"], "p"]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", "q"], 
                  \[Psi][ "u$2", "u$1"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "p"], 
              BD[ "l$2", 
                \[Psi][ "u$2", "u$1"]]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Eta]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ \[Eta], 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, if we choose the Einstein gauge ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\_a \[Psi]\^\(a\[InvisibleSpace]b\)\  = \ \
\(\[PartialD]\_a h\^\(a\[InvisibleSpace]b\)\  - \ \(1\/2\) \[PartialD]\^b 
                h = \ 0\)\)]],
  ", we have the wave equation"
}], "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`G\_\(a\[InvisibleSpace]b\)\  = \ \(\(\[Epsilon]\/2\) \
\[PartialD]\^2 \[Psi]\_\(a\[InvisibleSpace]b\)\  = \ \(-\[Kappa]\)\ T\_\(a\
\[InvisibleSpace]b\)\)\)]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell[BoxData[
    \(einsteinGauge = 
      Trule[BD[ua_, \[Psi][lb_, lc_]], 0, 
        PairQ[{ua, lb}] || PairQ[{ua, lc}]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{
          InterpretationBox[\("\[PartialD]"\^ua_\),
            Superscript[ "\[PartialD]", 
              Pattern[ ua, 
                Blank[ ]]],
            Editable->False], "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["\[Psi]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]lb_\),
                    SequenceForm[ "", 
                      Pattern[ lb, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lc_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ lb, 
                      Blank[ ]]], 
                  Pattern[ lc, 
                    Blank[ ]]],
                Editable->False]],
            Subsuperscript[ \[Psi], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ lb, 
                    Blank[ ]]], 
                Pattern[ lc, 
                  Blank[ ]]]],
            Editable->False]}],
        SequenceForm[ 
          Superscript[ "\[PartialD]", 
            Pattern[ ua, 
              Blank[ ]]], 
          \[Psi][ 
            Pattern[ lb, 
              Blank[ ]], 
            Pattern[ lc, 
              Blank[ ]]]],
        Editable->False], 
      "\[RuleDelayed]", \(DumFresh[0] /; 
        PairQ[{ua, lb}] || PairQ[{ua, lc}]\)}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[eq, einsteinGauge]\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
      RowBox[{"(", 
        InterpretationBox[
          RowBox[{\("\[PartialD]"\_"p"\), "\[InvisibleSpace]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\("\[PartialD]"\^"p"\),
                  Superscript[ "\[PartialD]", "p"],
                  Editable->False], "\[InvisibleSpace]", 
                InterpretationBox[
                  SubsuperscriptBox["\[Psi]", 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\)\
,
                          SequenceForm[ "", "a"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"],
                      Editable->False], 
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False]],
                  Subsuperscript[ \[Psi], 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "]],
                  Editable->False]}],
              SequenceForm[ 
                Superscript[ "\[PartialD]", "p"], 
                \[Psi][ la, lc]],
              Editable->False]}],
          SequenceForm[ 
            Subscript[ "\[PartialD]", "p"], 
            BD[ "u$1", 
              \[Psi][ la, lc]]],
          Editable->False], ")"}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "The gauge is not completely fixed because we can always carry out \
additional transformation with ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\^2 \[Xi]\^a\  = \ 0\)]],
  ", which leaves the Einstein gauge unaltered."
}], "Text"],

Cell[TextData[{
  "In vacuum, the Einstein equation becomes ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\^2 h\_\(a\[InvisibleSpace]b\)\  = \ 
        0\)]],
  " in the Einstein gauge ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\_a h\_b\^a\  - \ \(1\/2\) \[PartialD]\_b 
              h\  = \ 0\)]]
}], "Text"],

Cell[TextData[{
  "The vacuum equation implies ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]\^2 R\_\(a\[InvisibleSpace]b\
\[InvisibleSpace]c\[InvisibleSpace]d\)\  = \ 0\)]],
  " which obeys the wave equation. This implies that, in the linealized \
theory, gravitational effects propagate with the light velocity."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Linearized plane gravitational waves", "Section"],

Cell[TextData[{
  "We choose coordinates ",
  Cell[BoxData[
      \(TraditionalForm\`\((t, x, y, z)\)\)]],
  " and assume ",
  Cell[BoxData[
      \(TraditionalForm\`h\_\(a\[InvisibleSpace]b\)\  = \ \(h\_\(a\
\[InvisibleSpace]b\)\)(t, x)\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(assumeRule = {Trule[BD[ld_, BD[lc_, h[la_, lb_]]], 
          0, \((lc\  === \ \(-3\)\  || \ lc\  === \ \(-4\))\)], 
        Trule[BD[lc_, BD[ld_, h[la_, lb_]]], 
          0, \((lc\  === \ \(-3\)\  || \ lc\  === \ \(-4\))\)]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^ld_\),
                Superscript[ "\[PartialD]", 
                  Pattern[ ld, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^lc_\),
                    Superscript[ "\[PartialD]", 
                      Pattern[ lc, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\
\),
                            SequenceForm[ "", 
                              Pattern[ la, 
                                Blank[ ]]],
                            Editable->False], "\[InvisibleSpace]", "lb_"}],
                        SequenceForm[ 
                          SequenceForm[ "", 
                            Pattern[ la, 
                              Blank[ ]]], 
                          Pattern[ lb, 
                            Blank[ ]]],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 
                          Pattern[ la, 
                            Blank[ ]]], 
                        Pattern[ lb, 
                          Blank[ ]]]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", 
                    Pattern[ lc, 
                      Blank[ ]]], 
                  h[ 
                    Pattern[ la, 
                      Blank[ ]], 
                    Pattern[ lb, 
                      Blank[ ]]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", 
                Pattern[ ld, 
                  Blank[ ]]], 
              BD[ 
                Pattern[ lc, 
                  Blank[ ]], 
                h[ 
                  Pattern[ la, 
                    Blank[ ]], 
                  Pattern[ lb, 
                    Blank[ ]]]]],
            Editable->False], 
          "\[RuleDelayed]", \(DumFresh[0] /; 
            lc === \(-3\) || lc === \(-4\)\)}], ",", 
        RowBox[{
          InterpretationBox[
            RowBox[{
              InterpretationBox[\("\[PartialD]"\^lc_\),
                Superscript[ "\[PartialD]", 
                  Pattern[ lc, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\("\[PartialD]"\^ld_\),
                    Superscript[ "\[PartialD]", 
                      Pattern[ ld, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", 
                  InterpretationBox[
                    SubsuperscriptBox["h", 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], 
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\
\),
                            SequenceForm[ "", 
                              Pattern[ la, 
                                Blank[ ]]],
                            Editable->False], "\[InvisibleSpace]", "lb_"}],
                        SequenceForm[ 
                          SequenceForm[ "", 
                            Pattern[ la, 
                              Blank[ ]]], 
                          Pattern[ lb, 
                            Blank[ ]]],
                        Editable->False]],
                    Subsuperscript[ h, 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], 
                      SequenceForm[ 
                        SequenceForm[ "", 
                          Pattern[ la, 
                            Blank[ ]]], 
                        Pattern[ lb, 
                          Blank[ ]]]],
                    Editable->False]}],
                SequenceForm[ 
                  Superscript[ "\[PartialD]", 
                    Pattern[ ld, 
                      Blank[ ]]], 
                  h[ 
                    Pattern[ la, 
                      Blank[ ]], 
                    Pattern[ lb, 
                      Blank[ ]]]],
                Editable->False]}],
            SequenceForm[ 
              Superscript[ "\[PartialD]", 
                Pattern[ lc, 
                  Blank[ ]]], 
              BD[ 
                Pattern[ ld, 
                  Blank[ ]], 
                h[ 
                  Pattern[ la, 
                    Blank[ ]], 
                  Pattern[ lb, 
                    Blank[ ]]]]],
            Editable->False], 
          "\[RuleDelayed]", \(DumFresh[0] /; 
            lc === \(-3\) || lc === \(-4\)\)}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(riemannComp\  = \ 
        Table[riemann, {la, \(-1\), \(-4\), \(-1\)}, {lb, \(-1\), \(-4\), \
\(-1\)}, {lc, \(-1\), \(-4\), \(-1\)}, {ld, \(-1\), \(-4\), \(-1\)}];\)\)], \
"Input"],

Cell["The 20 component Riemann tensors fall into the three groups", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(group1 = 
      ApplyRules[{riemannComp[\([1, 2, 3, 4]\)], 
          riemannComp[\([1, 3, 3, 4]\)], riemannComp[\([1, 4, 3, 4]\)], 
          riemannComp[\([2, 3, 3, 4]\)], riemannComp[\([2, 4, 3, 4]\)], 
          riemannComp[\([3, 4, 3, 4]\)]}, assumeRule]\)], "Input"],

Cell[BoxData[
    \({0, 0, 0, 0, 0, 0}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(group2 = 
      ApplyRules[{riemannComp[\([1, 2, 1, 2]\)], 
          riemannComp[\([1, 2, 1, 3]\)], riemannComp[\([1, 2, 1, 4]\)], 
          riemannComp[\([1, 2, 2, 3]\)], riemannComp[\([1, 2, 2, 4]\)]}, 
        assumeRule]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "1"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 1],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 1], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -1]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 2], 
                  BD[ -2, 
                    h[ -1, -1]]],
                Editable->False], ")"}]}], "-", 
          RowBox[{"\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "2"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 2],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 2], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -2]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -2, 
                    h[ -1, -2]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "2"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 2],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 2], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 1], 
                      h[ -2, -2]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -1, 
                    h[ -2, -2]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -2, 
                    h[ -1, -3]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 1], 
                      h[ -2, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -1, 
                    h[ -2, -3]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -2, 
                    h[ -1, -4]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 1], 
                      h[ -2, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -1, 
                    h[ -2, -4]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 2], 
                  BD[ -2, 
                    h[ -1, -3]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "3"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 3],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 3], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -2, -3]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -2, 
                    h[ -2, -3]]],
                Editable->False], ")"}]}]}], ",", 
        RowBox[{
          RowBox[{\(-\(1\/2\)\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]1\),
                                SequenceForm[ "", 1],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 1], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 1], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -1, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 2], 
                  BD[ -2, 
                    h[ -1, -4]]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                  InterpretationBox[
                    RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                      InterpretationBox[
                        SubsuperscriptBox["h", 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]2\),
                                SequenceForm[ "", 2],
                                Editable->False], "\[InvisibleSpace]", "4"}],
                            SequenceForm[ 
                              SequenceForm[ "", 2], 4],
                            Editable->False], 
                          InterpretationBox[
                            RowBox[{
                              
                              InterpretationBox[\(\*"\<\"\"\>"\
\[InvisibleSpace]" "\),
                                SequenceForm[ "", " "],
                                Editable->False], 
                              "\[InvisibleSpace]", "\<\" \"\>"}],
                            SequenceForm[ 
                              SequenceForm[ "", " "], " "],
                            Editable->False]],
                        Subsuperscript[ h, 
                          SequenceForm[ 
                            SequenceForm[ "", 2], 4], 
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "]],
                        Editable->False]}],
                    SequenceForm[ 
                      Subscript[ "\[PartialD]", 2], 
                      h[ -2, -4]],
                    Editable->False]}],
                SequenceForm[ 
                  Subscript[ "\[PartialD]", 1], 
                  BD[ -2, 
                    h[ -2, -4]]],
                Editable->False], ")"}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(group3 = 
      ApplyRules[{riemannComp[\([1, 3, 1, 3]\)], 
          riemannComp[\([1, 3, 1, 4]\)], riemannComp[\([1, 3, 2, 3]\)], 
          riemannComp[\([1, 3, 2, 4]\)], riemannComp[\([1, 4, 1, 4]\)], 
          riemannComp[\([1, 4, 2, 4]\)], riemannComp[\([2, 3, 2, 3]\)], 
          riemannComp[\([2, 3, 2, 4]\)], riemannComp[\([2, 4, 2, 4]\)]}, 
        assumeRule]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "3"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 3],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 3], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 1], 
                    h[ -3, -3]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                BD[ -1, 
                  h[ -3, -3]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 1], 
                    h[ -3, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                BD[ -1, 
                  h[ -3, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "3"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 3],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 3], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -3]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                BD[ -2, 
                  h[ -3, -3]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                BD[ -2, 
                  h[ -3, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
4\),
                              SequenceForm[ "", 4],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 4], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 4], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 1], 
                    h[ -4, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                BD[ -1, 
                  h[ -4, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_1\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
4\),
                              SequenceForm[ "", 4],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 4], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 4], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -4, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 1], 
                BD[ -2, 
                  h[ -4, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "3"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 3],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 3], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -3]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 2], 
                BD[ -2, 
                  h[ -3, -3]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
3\),
                              SequenceForm[ "", 3],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 3], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 3], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -3, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 2], 
                BD[ -2, 
                  h[ -3, -4]]],
              Editable->False], ")"}]}], ",", 
        RowBox[{\(1\/2\), " ", "\[Epsilon]", " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                InterpretationBox[
                  RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
                    InterpretationBox[
                      SubsuperscriptBox["h", 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
4\),
                              SequenceForm[ "", 4],
                              Editable->False], "\[InvisibleSpace]", "4"}],
                          SequenceForm[ 
                            SequenceForm[ "", 4], 4],
                          Editable->False], 
                        InterpretationBox[
                          RowBox[{
                            
                            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
" "\),
                              SequenceForm[ "", " "],
                              Editable->False], 
                            "\[InvisibleSpace]", "\<\" \"\>"}],
                          SequenceForm[ 
                            SequenceForm[ "", " "], " "],
                          Editable->False]],
                      Subsuperscript[ h, 
                        SequenceForm[ 
                          SequenceForm[ "", 4], 4], 
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "]],
                      Editable->False]}],
                  SequenceForm[ 
                    Subscript[ "\[PartialD]", 2], 
                    h[ -4, -4]],
                  Editable->False]}],
              SequenceForm[ 
                Subscript[ "\[PartialD]", 2], 
                BD[ -2, 
                  h[ -4, -4]]],
              Editable->False], ")"}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "The vacuum equation ",
  Cell[BoxData[
      \(TraditionalForm\`R\_\(a\[InvisibleSpace]b\)\  = \ 0\)]],
  " and group 1 result in group2 vanishing. Thus, only the group3 are \
non-zero and these only involve the components ",
  Cell[BoxData[
      \(TraditionalForm\`h\_33\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`h\_34\)]],
  ", and ",
  Cell[BoxData[
      \(TraditionalForm\`h\_44\)]],
  "."
}], "Text"],

Cell[TextData[{
  "With the ansatz ",
  Cell[BoxData[
      \(TraditionalForm\`h\_\(a\[InvisibleSpace]b\)\  = \ \(h\_\(a\
\[InvisibleSpace]b\)\)(t - x)\)]],
  " and the asymptotic flatness the Einstein gauge condition gives"
}], "Text"],

Cell[TextData[Cell[BoxData[
    FormBox[
      RowBox[{\(h\_\(a\[InvisibleSpace]b\)\), " ", "=", " ", 
        RowBox[{"(", GridBox[{
              {"0", "0", "0", "0"},
              {"0", "0", "0", "0"},
              {"0", "0", \(h\_33\), \(h\_34\)},
              {"0", "0", \(h\_34\), \(-h\_33\)}
              }], ")"}]}], TraditionalForm]]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "That is, ",
  Cell[BoxData[
      \(TraditionalForm\`h\_\(a\[InvisibleSpace]b\)\)]],
  " only depends on two functions ",
  Cell[BoxData[
      \(TraditionalForm\`h\_33\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`h\_34\)]]
}], "Text"]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{695, 680},
WindowMargins->{{19, Automatic}, {Automatic, 5}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 93, 1, 43, "Title"],
Cell[1813, 52, 94, 2, 41, "Text"],
Cell[1910, 56, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1987, 61, 32, 0, 78, "Section"],
Cell[2022, 63, 68, 1, 56, "Input"],
Cell[2093, 66, 194, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[2312, 75, 134, 3, 79, "Input"],
Cell[2449, 80, 3535, 87, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6021, 172, 138, 2, 79, "Input"],
Cell[6162, 176, 3576, 92, 78, "Output"]
}, Open  ]],
Cell[9753, 271, 259, 7, 41, "Text"],

Cell[CellGroupData[{
Cell[10037, 282, 32, 0, 66, "Subsection"],
Cell[10072, 284, 110, 2, 56, "Input"],

Cell[CellGroupData[{
Cell[10207, 290, 83, 1, 56, "Input"],
Cell[10293, 293, 6305, 148, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16635, 446, 80, 1, 56, "Input"],
Cell[16718, 449, 14387, 337, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[31142, 791, 70, 1, 56, "Input"],
Cell[31215, 794, 7220, 169, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[38472, 968, 65, 1, 56, "Input"],
Cell[38540, 971, 4403, 100, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[42980, 1076, 83, 1, 56, "Input"],
Cell[43066, 1079, 7494, 178, 94, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[50609, 1263, 31, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[50665, 1267, 80, 1, 56, "Input"],
Cell[50748, 1270, 12969, 308, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[63754, 1583, 154, 4, 125, "Input"],
Cell[63911, 1589, 11830, 283, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[75778, 1877, 66, 1, 56, "Input"],
Cell[75847, 1880, 7645, 181, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[83529, 2066, 55, 1, 56, "Input"],
Cell[83587, 2069, 11545, 277, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[95169, 2351, 64, 1, 56, "Input"],
Cell[95236, 2354, 8014, 190, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[103287, 2549, 53, 1, 56, "Input"],
Cell[103343, 2552, 11914, 286, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[115294, 2843, 70, 1, 56, "Input"],
Cell[115367, 2846, 4208, 100, 56, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[119612, 2951, 76, 1, 72, "Input"],
Cell[119691, 2954, 14149, 337, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[133877, 3296, 153, 4, 125, "Input"],
Cell[134033, 3302, 14199, 337, 110, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[148281, 3645, 40, 0, 38, "Subsection"],

Cell[CellGroupData[{
Cell[148346, 3649, 88, 1, 56, "Input"],
Cell[148437, 3652, 11827, 267, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[160301, 3924, 122, 3, 79, "Input"],
Cell[160426, 3929, 9420, 211, 69, "Output"]
}, Open  ]],
Cell[169861, 4143, 47, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[169933, 4147, 107, 2, 79, "Input"],
Cell[170043, 4151, 35, 1, 53, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[170127, 4158, 43, 0, 38, "Subsection"],
Cell[170173, 4160, 56, 1, 56, "Input"],
Cell[170232, 4163, 370, 9, 43, "Text"],

Cell[CellGroupData[{
Cell[170627, 4176, 136, 3, 79, "Input"],
Cell[170766, 4181, 4205, 104, 55, "Output"]
}, Open  ]],
Cell[174986, 4288, 104, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[175115, 4295, 77, 1, 56, "Input"],
Cell[175195, 4298, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[175267, 4304, 73, 1, 56, "Input"],
Cell[175343, 4307, 26525, 628, 233, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[201905, 4940, 63, 1, 56, "Input"],
Cell[201971, 4943, 35, 1, 53, "Output"]
}, Open  ]],
Cell[202021, 4947, 211, 6, 44, "Text"],
Cell[202235, 4955, 58, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[202318, 4960, 218, 4, 123, "Input"],
Cell[202539, 4966, 8263, 200, 95, "Output"]
}, Open  ]],
Cell[210817, 5169, 160, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[211002, 5178, 136, 2, 99, "Input"],
Cell[211141, 5182, 6276, 154, 69, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[217454, 5341, 52, 1, 56, "Input"],
Cell[217509, 5344, 4318, 107, 54, "Output"]
}, Open  ]],
Cell[221842, 5454, 232, 6, 41, "Text"],
Cell[222077, 5462, 43, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[222145, 5466, 77, 1, 56, "Input"],
Cell[222225, 5469, 24430, 580, 151, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[246692, 6054, 56, 1, 56, "Input"],
Cell[246751, 6057, 15388, 364, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[262176, 6426, 50, 1, 56, "Input"],
Cell[262229, 6429, 8867, 208, 69, "Output"]
}, Open  ]],
Cell[271111, 6640, 310, 7, 44, "Text"],
Cell[271424, 6649, 267, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[271716, 6658, 136, 3, 79, "Input"],
Cell[271855, 6663, 1952, 54, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[273844, 6722, 62, 1, 56, "Input"],
Cell[273909, 6725, 1983, 48, 69, "Output"]
}, Open  ]],
Cell[275907, 6776, 259, 6, 66, "Text"],
Cell[276169, 6784, 331, 9, 44, "Text"],
Cell[276503, 6795, 335, 7, 66, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[276887, 6808, 55, 0, 50, "Section"],
Cell[276945, 6810, 253, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[277223, 6822, 259, 4, 148, "Input"],
Cell[277485, 6828, 6226, 160, 78, "Output"]
}, Open  ]],
Cell[283726, 6991, 203, 4, 79, "Input"],
Cell[283932, 6997, 75, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[284032, 7001, 292, 5, 148, "Input"],
Cell[284327, 7008, 52, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[284416, 7014, 259, 5, 148, "Input"],
Cell[284678, 7021, 22336, 491, 151, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[307051, 7517, 405, 7, 194, "Input"],
Cell[307459, 7526, 17448, 398, 151, "Output"]
}, Open  ]],
Cell[324922, 7927, 438, 15, 66, "Text"],
Cell[325363, 7944, 236, 6, 41, "Text"],
Cell[325602, 7952, 406, 10, 95, "Text"],
Cell[326011, 7964, 266, 10, 41, "Text"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

