(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     12491,        365]*)
(*NotebookOutlinePosition[     15969,        454]*)
(*  CellTagsIndexPosition[     15925,        450]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Misc", "Subtitle",
  TextAlignment->Center,
  TextJustification->0],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Status, SetCoordinates, and ClearCoordinates", "Section"],

Cell[BoxData[
    \(\(Dimension = 4;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Status[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"\<\"AutoCanFlag\"\>"},
                {"\<\"CoordinateBasisFlag\"\>"},
                {"\<\"EvaluateBDFlag\"\>"},
                {"\<\"MetricgFlag\"\>"},
                {"\<\"TorsionFreeFlag\"\>"},
                {"\<\"SyntaxCheckFlag\"\>"},
                {"\<\"InitCTensor\"\>"},
                {"\<\"Dimension\"\>"},
                {"\<\"Coordinates\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag",
             "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", 
            "InitCTensor", "Dimension", "Coordinates"}]], 
        "\[InvisibleSpace]", "\<\" == \"\>", "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"True"},
                {"True"},
                {"False"},
                {"True"},
                {"True"},
                {"False"},
                {"False"},
                {"4"},
                {"\<\"none\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {True, True, False, True, True, False, False, 4, 
            "none"}]]}],
      SequenceForm[ 
        MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag", 
          "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", "InitCTensor", 
          "Dimension", "Coordinates"}], " == ", 
        MatrixForm[ {True, True, False, True, True, False, False, 4, 
          "none"}]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(Dimension =. ;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Status[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"\<\"AutoCanFlag\"\>"},
                {"\<\"CoordinateBasisFlag\"\>"},
                {"\<\"EvaluateBDFlag\"\>"},
                {"\<\"MetricgFlag\"\>"},
                {"\<\"TorsionFreeFlag\"\>"},
                {"\<\"SyntaxCheckFlag\"\>"},
                {"\<\"InitCTensor\"\>"},
                {"\<\"Dimension\"\>"},
                {"\<\"Coordinates\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag",
             "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", 
            "InitCTensor", "Dimension", "Coordinates"}]], 
        "\[InvisibleSpace]", "\<\" == \"\>", "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"True"},
                {"True"},
                {"False"},
                {"True"},
                {"True"},
                {"False"},
                {"False"},
                {"\<\"Any\"\>"},
                {"\<\"none\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {True, True, False, True, True, False, False, "Any", 
            "none"}]]}],
      SequenceForm[ 
        MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag", 
          "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", "InitCTensor", 
          "Dimension", "Coordinates"}], " == ", 
        MatrixForm[ {True, True, False, True, True, False, False, "Any", 
          "none"}]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(SetCoordinates[{t, r, \[Theta], \[Phi]}]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Status[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"\<\"AutoCanFlag\"\>"},
                {"\<\"CoordinateBasisFlag\"\>"},
                {"\<\"EvaluateBDFlag\"\>"},
                {"\<\"MetricgFlag\"\>"},
                {"\<\"TorsionFreeFlag\"\>"},
                {"\<\"SyntaxCheckFlag\"\>"},
                {"\<\"InitCTensor\"\>"},
                {"\<\"Dimension\"\>"},
                {"\<\"Coordinates\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag",
             "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", 
            "InitCTensor", "Dimension", "Coordinates"}]], 
        "\[InvisibleSpace]", "\<\" == \"\>", "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"True"},
                {"True"},
                {"False"},
                {"True"},
                {"True"},
                {"False"},
                {"False"},
                {"4"},
                {\({t, r, \[Theta], \[Phi]}\)}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {True, True, False, True, True, False, False, 4, {t, 
            r, \[Theta], \[Phi]}}]]}],
      SequenceForm[ 
        MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag", 
          "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", "InitCTensor", 
          "Dimension", "Coordinates"}], " == ", 
        MatrixForm[ {True, True, False, True, True, False, False, 4, {t, 
          r, \[Theta], \[Phi]}}]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(ClearCoordinates[]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Status[]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"\<\"AutoCanFlag\"\>"},
                {"\<\"CoordinateBasisFlag\"\>"},
                {"\<\"EvaluateBDFlag\"\>"},
                {"\<\"MetricgFlag\"\>"},
                {"\<\"TorsionFreeFlag\"\>"},
                {"\<\"SyntaxCheckFlag\"\>"},
                {"\<\"InitCTensor\"\>"},
                {"\<\"Dimension\"\>"},
                {"\<\"Coordinates\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag",
             "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", 
            "InitCTensor", "Dimension", "Coordinates"}]], 
        "\[InvisibleSpace]", "\<\" == \"\>", "\[InvisibleSpace]", 
        InterpretationBox[
          RowBox[{"(", "\[NoBreak]", GridBox[{
                {"True"},
                {"True"},
                {"False"},
                {"True"},
                {"True"},
                {"False"},
                {"False"},
                {"\<\"Any\"\>"},
                {"\<\"none\"\>"}
                }], "\[NoBreak]", ")"}],
          
          MatrixForm[ {True, True, False, True, True, False, False, "Any", 
            "none"}]]}],
      SequenceForm[ 
        MatrixForm[ {"AutoCanFlag", "CoordinateBasisFlag", "EvaluateBDFlag", 
          "MetricgFlag", "TorsionFreeFlag", "SyntaxCheckFlag", "InitCTensor", 
          "Dimension", "Coordinates"}], " == ", 
        MatrixForm[ {True, True, False, True, True, False, False, "Any", 
          "none"}]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["LineElement", "Section"],

Cell[BoxData[
    \(\(coSys\  = \ {t, r, \[Theta], \[Phi]};\)\)], "Input"],

Cell[BoxData[
    \(\(metric\  = \ {{\(-E\^\(2  a[t, r]\)\), 0, 0, 0}, {0, 
            E\^\(2  b[t, r]\), 0, 0}, {0, 0, r\^2, 0}, {0, 0, 0, 
            r\^2\ Sin[\[Theta]]\^2}};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LineElement[coSys, metric]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\("s"\^2\),
          Superscript[ "s", 2],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", 
        RowBox[{
          RowBox[{\(\[ExponentialE]\^\(2\ b[t, r]\)\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(r\^2\),
                    Superscript[ r, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ r, 2]],
                Editable->False], ")"}]}], "-", 
          RowBox[{\(\[ExponentialE]\^\(2\ a[t, r]\)\), " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(t\^2\),
                    Superscript[ t, 2],
                    Editable->False]}],
                SequenceForm[ "d", 
                  Superscript[ t, 2]],
                Editable->False], ")"}]}], "+", 
          RowBox[{\(r\^2\), " ", 
            RowBox[{"(", 
              RowBox[{
                InterpretationBox[
                  RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                    InterpretationBox[\(\[Theta]\^2\),
                      Superscript[ \[Theta], 2],
                      Editable->False]}],
                  SequenceForm[ "d", 
                    Superscript[ \[Theta], 2]],
                  Editable->False], "+", 
                RowBox[{
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{"\<\"d\"\>", "\[InvisibleSpace]", 
                        InterpretationBox[\(\[Phi]\^2\),
                          Superscript[ \[Phi], 2],
                          Editable->False]}],
                      SequenceForm[ "d", 
                        Superscript[ \[Phi], 2]],
                      Editable->False], ")"}], " ", \(Sin[\[Theta]]\^2\)}]}], 
              ")"}]}]}]}],
      SequenceForm[ "d", 
        Superscript[ "s", 2], " = ", 
        Plus[ 
          Times[ 
            Power[ E, 
              Times[ 2, 
                b[ t, r]]], 
            SequenceForm[ "d", 
              Superscript[ r, 2]]], 
          Times[ -1, 
            Power[ E, 
              Times[ 2, 
                a[ t, r]]], 
            SequenceForm[ "d", 
              Superscript[ t, 2]]], 
          Times[ 
            Power[ r, 2], 
            Plus[ 
              SequenceForm[ "d", 
                Superscript[ \[Theta], 2]], 
              Times[ 
                SequenceForm[ "d", 
                  Superscript[ \[Phi], 2]], 
                Power[ 
                  Sin[ \[Theta]], 2]]]]]],
      Editable->False]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{644, 694},
WindowMargins->{{12, Automatic}, {0, Automatic}},
Visible->True,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"], 
    "Vdefine"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
            OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
            NotationBoxTag, TagStyle -> "NotationTemplateStyle"], " ", 
          "\[DoubleLongLeftRightArrow]", " ", 
          TagBox[ 
            RowBox[ {"Vector", "[", "\[Placeholder]", "]"}], NotationBoxTag, 
            TagStyle -> "NotationTemplateStyle"]}], "]"}], 
    "ivec"->OverscriptBox[ "\[DotlessI]", "^"], 
    "jvec"->OverscriptBox[ "\[DotlessJ]", "^"], 
    "vect"->OverscriptBox[ "\[Placeholder]", "\[RightVector]"], 
    "xvec"->OverscriptBox[ "x", "^"], "yvec"->OverscriptBox[ "y", "^"], 
    "zvec"->OverscriptBox[ "z", "^"], "rvec"->OverscriptBox[ "r", "^"], 
    "tvec"->OverscriptBox[ "\[Theta]", "^"], 
    "pvec"->OverscriptBox[ "\[Phi]", "^"]},
Magnification->1,
StyleDefinitions -> "Demo.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 73, 2, 81, "Subtitle"],
Cell[1793, 53, 52, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[1870, 58, 63, 0, 77, "Section"],
Cell[1936, 60, 51, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[2012, 65, 41, 1, 54, "Input"],
Cell[2056, 68, 1651, 41, 215, "Output"]
}, Open  ]],
Cell[3722, 112, 51, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[3798, 117, 41, 1, 54, "Input"],
Cell[3842, 120, 1669, 41, 215, "Output"]
}, Open  ]],
Cell[5526, 164, 73, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[5624, 169, 41, 1, 54, "Input"],
Cell[5668, 172, 1701, 41, 215, "Output"]
}, Open  ]],
Cell[7384, 216, 51, 1, 54, "Input"],

Cell[CellGroupData[{
Cell[7460, 221, 41, 1, 54, "Input"],
Cell[7504, 224, 1669, 41, 215, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[9222, 271, 30, 0, 49, "Section"],
Cell[9255, 273, 74, 1, 54, "Input"],
Cell[9332, 276, 194, 3, 80, "Input"],

Cell[CellGroupData[{
Cell[9551, 283, 59, 1, 54, "Input"],
Cell[9613, 286, 2850, 75, 78, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

