(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    119283,       3216]*)
(*NotebookOutlinePosition[    120344,       3252]*)
(*  CellTagsIndexPosition[    120300,       3248]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Static Monopoles in Einstein-Maxwell Theory", "Title",
  FontSize->24],

Cell[TextData[{
  "Ref: D. H. Park, ",
  StyleBox["General BPS Solution in Einstein-Maxwell-Dilaton Theory",
    FontSlant->"Italic"],
  ", J. Kor. Phy. Soc, 31, 894 (1997)"
}], "Text"],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[CellGroupData[{

Cell["Theory", "Section"],

Cell[CellGroupData[{

Cell["Action", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`I\  = \ \(-\(\[Integral]\(d\^4\) 
            x\ \(\@\(-g\)\)[\(1\/\[Kappa]\^2\) R\  + \ \(1\/4\) 
                  F\_\[Mu]\[Nu]\^2]\)\)\)]]], "Text",
  TextAlignment->Center]
}, Closed]],

Cell[CellGroupData[{

Cell["Static Metric", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]s\^2\  = \ \(-f\)\ \
d\[InvisibleSpace]t\^2\  + \ \(1\/f\) \(\[Gamma]\_\(i\[InvisibleSpace]j\)\) 
          d\[InvisibleSpace]\(x\^i\) d\[InvisibleSpace]x\^j\)]]], "Text",
  TextAlignment->Center]
}, Closed]],

Cell[CellGroupData[{

Cell["Effective Action", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`I\_eff\  = \ \(-\(1\/\[Kappa]\^2\)\) \
\(\[Integral]\(d\^3\) 
            x\ \(\@\[Gamma]\)[
              R\_3\  + \ \(1\/2\) \((\[PartialD]\_i ln\ f)\)\^2\  - \ \
\(1\/f\) \((\[PartialD]\_i v)\)\^2]\)\)]]], "Text",
  TextAlignment->Center],

Cell["where", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\(\[Kappa]\/\@2\) 
        F\_\(i\[InvisibleSpace]j\)\  = \ \(1\/f\) \(\[Epsilon]\_\(i\
\[InvisibleSpace]j\[InvisibleSpace]k\)\) \[PartialD]\^k v\)]]], "Text",
  TextAlignment->Center]
}, Closed]],

Cell[CellGroupData[{

Cell["Effective Metric in Potential Space", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`d\[InvisibleSpace]S\^2\  = \ \(\(G\_\(A\
\[InvisibleSpace]B\)\) d\[InvisibleSpace]\(\[CurlyPhi]\^A\) 
          d\[InvisibleSpace]\[CurlyPhi]\^B\  = \ \(-\(1\/f\)\) 
            d\[InvisibleSpace]v\^2\  + \ \(1\/\(2  f\^2\)\) 
            d\[InvisibleSpace]f\^2\)\)]]], "Text",
  TextAlignment->Center]
}, Closed]],

Cell[CellGroupData[{

Cell["Equation of Motion", "Subsubsection"],

Cell[TextData[{
  "Assume ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyPhi]\^A\  = \ \
\(\[CurlyPhi]\^A\)(\[Lambda])\)]],
  ". Then the equation of motion becomes"
}], "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`R\_\(i\[InvisibleSpace]j\)\  = \ \(-\[Epsilon]\)\ \
\[PartialD]\_i \[Lambda]\ \[PartialD]\_j \[Lambda]\)]]], "Text",
  TextAlignment->Center],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\(\[Xi]\^A\) \[Del]\_A \[Xi]\^B\  = \ 0\)]]], "Text",
  TextAlignment->Center],

Cell["where", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Xi]\^A\  = \ \
\(d\[InvisibleSpace]\[CurlyPhi]\^A\)\/\(d\[InvisibleSpace]\[Lambda]\), \ \ \ \
\ \ \[Epsilon]\  = \ \(G\_\(A\[InvisibleSpace]B\)\) \(\[Xi]\^A\) \
\[Xi]\^B\)]]], "Text",
  TextAlignment->Center],

Cell[TextData[{
  "Notice that for ",
  Cell[BoxData[
      \(TraditionalForm\`\[Epsilon]\  = \ 0\)]],
  " the three-dimensional Einstein equation simply gives ",
  Cell[BoxData[
      \(TraditionalForm\`\[Gamma]\_\(i\[InvisibleSpace]j\)\  = \ \
\[Delta]\_\(i\[InvisibleSpace]j\)\)]],
  "."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Killing Symmetries of Potential Space", "Section"],

Cell[BoxData[
    \(InitCTensor[{v, 
        f}, \ {{\(-\(1\/f\)\), 0}, {0, 1\/\(2\ f\^2\)}}]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Show[RiemannR]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Symmetry: "\[InvisibleSpace]"R"\_"[ab][cd]"\
\[InvisibleSpace]" = "\[InvisibleSpace]"R"\_"[cd][ab]"\),
      SequenceForm[ "Symmetry: ", 
        Subscript[ "R", "[ab][cd]"], " = ", 
        Subscript[ "R", "[cd][ab]"]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"v"\+\
_\),
                          SequenceForm[ "", 
                            dhPark`Tensor`Private`UnderBar[ "v"]],
                          Editable->False], "\[InvisibleSpace]", \("f"\+_\)}],
                      
                      SequenceForm[ 
                        SequenceForm[ "", 
                          dhPark`Tensor`Private`UnderBar[ "v"]], 
                        dhPark`Tensor`Private`UnderBar[ "f"]],
                      Editable->False], "\[InvisibleSpace]", \("v"\+_\)}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 
                        dhPark`Tensor`Private`UnderBar[ "v"]], 
                      dhPark`Tensor`Private`UnderBar[ "f"]], 
                    dhPark`Tensor`Private`UnderBar[ "v"]],
                  Editable->False], "\[InvisibleSpace]", \("f"\+_\)}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 
                      dhPark`Tensor`Private`UnderBar[ "v"]], 
                    dhPark`Tensor`Private`UnderBar[ "f"]], 
                  dhPark`Tensor`Private`UnderBar[ "v"]], 
                dhPark`Tensor`Private`UnderBar[ "f"]],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 
                    dhPark`Tensor`Private`UnderBar[ "v"]], 
                  dhPark`Tensor`Private`UnderBar[ "f"]], 
                dhPark`Tensor`Private`UnderBar[ "v"]], 
              dhPark`Tensor`Private`UnderBar[ "f"]], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], " "]],
          Editable->False], "\[InvisibleSpace]", "\<\" = \"\>", 
        "\[InvisibleSpace]", \(-\(1\/\(4\ f\^3\)\)\)}],
      SequenceForm[ 
        Subsuperscript[ R, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  dhPark`Tensor`Private`UnderBar[ "v"]], 
                dhPark`Tensor`Private`UnderBar[ "f"]], 
              dhPark`Tensor`Private`UnderBar[ "v"]], 
            dhPark`Tensor`Private`UnderBar[ "f"]], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "], " "]], " = ", 
        Times[ 
          Rational[ -1, 4], 
          Power[ f, -3]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(CD[la, RiemannR[lb, lc, ld, ue]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\)\
,
                          SequenceForm[ "", "b"],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], "d"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "e"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], "d"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "e"]],
          Editable->False]}],
      SequenceForm[ 
        Subscript[ "\[Del]", "a"], 
        RiemannR[ lb, lc, ld, ue]],
      Editable->False]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(cvR\  = SumDum[CDtoAffine[%]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
          InterpretationBox[
            SubsuperscriptBox["R", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], "d"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], "d"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "], "e"],
                Editable->False]],
            Subsuperscript[ R, 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], "d"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "e"]],
            Editable->False]}],
        SequenceForm[ 
          Subscript[ "\[PartialD]", "a"], 
          RiemannR[ lb, lc, ld, ue]],
        Editable->False], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                          SequenceForm[ "", 2],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 2], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 2], "b"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 2], "b"], "d"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "e"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "b"], "d"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "e"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                          SequenceForm[ "", 2],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 2], "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 2], "c"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 2], "c"], "d"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "e"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "c"], "d"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "e"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "d"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "d"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                          SequenceForm[ "", 2],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 2], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 2], " "], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 2], " "], "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"e\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "e"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "e"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "e"], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], " "], "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "e"], " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                          SequenceForm[ "", 1],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"b\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 1], "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 1], "b"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 1], "b"], "d"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "e"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "b"], "d"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "e"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                          SequenceForm[ "", 1],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"c\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 1], "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 1], "c"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 1], "c"], "d"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "], "e"],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "c"], "d"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "], "e"]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "d"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "d"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                          SequenceForm[ "", 1],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 1], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 1], " "], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 1], " "], "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"e\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "e"],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "e"], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "e"], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], " "], "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "e"], " "], " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "e"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "e"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "d"], "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                          SequenceForm[ "", 1],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 1], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 1], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 1], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"], "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], " "], " "], " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"e\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "e"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "e"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["R", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\"d\"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", " "], "d"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], "d"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], "d"], "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[
                      RowBox[{
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                          SequenceForm[ "", 2],
                          Editable->False], 
                        "\[InvisibleSpace]", "\<\" \"\>"}],
                      SequenceForm[ 
                        SequenceForm[ "", 2], " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", 2], " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", 2], " "], " "], " "],
              Editable->False]],
          Subsuperscript[ R, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], "d"], "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], " "], " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[
      cvR\  /. \ {la -> \(-i\), lb -> \(-j\), lc -> \(-k\), ld -> \(-l\), 
          ue -> m}, {i, 2}, {j, 2}, {k, 2}, {l, 2}, {m, 2}]\)], "Input"],

Cell[BoxData[
    \({{{{{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}}, {{{0, 0}, {0, 0}}, {{0, 0}, {0, 
              0}}}}, {{{{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}}, {{{0, 0}, {0, 
              0}}, {{0, 0}, {0, 0}}}}}\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Therefore, the curvature in the porential space is covariant constant: ",
  Cell[BoxData[
      \(TraditionalForm\`\[Del]\_a R\_\(b\[InvisibleSpace]c\[InvisibleSpace]d\
\)\^e\  = \ 0\)]]
}], "Text"],

Cell[CellGroupData[{

Cell["Killing Vectors", "Subsubsection"],

Cell[BoxData[
    \(Tdefine[K, "\<a\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(killingEq\  = \ 
      Metricg[la, lc]\ CD[lb, K[uc]]\  + \ 
        Metricg[lb, lc]\ CD[la, K[uc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "b"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                  SequenceForm[ "", "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "a"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[Del]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                    SequenceForm[ "", "c"],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", "c"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[Del]", "a"], 
              K[ uc]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                  SequenceForm[ "", "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(killingEq\  = SumDum[CDtoAffine[killingEq]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              K[ 2]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              K[ 2]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "2"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 2],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                  SequenceForm[ "", 2],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 2], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 2], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "b"], 
              K[ 1]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "a"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", "a"], 
              K[ 1]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                      SequenceForm[ "", 1],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 1], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 1], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 1], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
              SequenceForm[ "", 1],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                      SequenceForm[ "", 2],
                      Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", 2], "a"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", 2], "a"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "1"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], 1],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 2], "a"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], 1]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["K", 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
              SequenceForm[ "", " "],
              Editable->False], 
            InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
              SequenceForm[ "", 2],
              Editable->False]],
          Subsuperscript[ K, 
            SequenceForm[ "", " "], 
            SequenceForm[ "", 2]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                  SequenceForm[ "", 1],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", 1], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", 1], "b"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k22\  = \ 
      killingEq\ *\ f\^3 /. \ {la\  -> \ \(-2\), lb\  -> \ \(-2\)}\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"f", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                    SequenceForm[ "", 2],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 2]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", 2], 
              K[ 2]],
            Editable->False], ")"}]}], "-", 
      InterpretationBox[
        SubsuperscriptBox["K", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
            SequenceForm[ "", 2],
            Editable->False]],
        Subsuperscript[ K, 
          SequenceForm[ "", " "], 
          SequenceForm[ "", 2]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "=> ",
  Cell[BoxData[
      \(TraditionalForm\`K\^f\  = \ f\ AA[v]\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cset[K[2], \ f\ *\ AA[v]]; \ k22\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k12\  = \ 
      killingEq\ *\ f /. \ {la\  -> \ \(-1\), lb\  -> \ \(-2\)}\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("\[PartialD]"\_2\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["K", 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], 
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                    SequenceForm[ "", 1],
                    Editable->False]],
                Subsuperscript[ K, 
                  SequenceForm[ "", " "], 
                  SequenceForm[ "", 1]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "\[PartialD]", 2], 
              K[ 1]],
            Editable->False], ")"}]}], "+", 
      FractionBox[
        RowBox[{
          SuperscriptBox["AA", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "2"]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "=> ",
  Cell[BoxData[
      \(TraditionalForm\`K\^v\  = \ \(f\/2\) \(AA'\)[v]\  + \ BB[v]\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cset[K[1], \ \(f\/2\) D[AA[v], v]\  + \ BB[v]]; \ k12\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(k11\  = \ 
      killingEq\ *\ f /. \ {la\  -> \ \(-1\), lb\  -> \ \(-1\)}\)], "Input"],

Cell[BoxData[
    RowBox[{\(AA[v]\), "-", 
      RowBox[{"2", " ", 
        RowBox[{
          SuperscriptBox["BB", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}], "-", 
      RowBox[{"f", " ", 
        RowBox[{
          SuperscriptBox["AA", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "=> ",
  Cell[BoxData[
      \(TraditionalForm\`AA[v]\  = \ pp\ v\  + \ qq, \ 
      BB[v]\  = \ \(pp\/4\) v\^2\  + \ \(qq\/2\) v\ \  + \ rr\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AA[v_]\  = \ pp\ *\ v\  + \ qq; \ 
    BB[v_]\  = \ \(pp\/4\) v\^2\  + \ \(qq\/2\) v\  + \ rr; k11\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \({K[1], \ K[2]}\)], "Input"],

Cell[BoxData[
    \({\(f\ pp\)\/2 + rr + \(qq\ v\)\/2 + \(pp\ v\^2\)\/4, 
      f\ \((qq + pp\ v)\)}\)], "Output"]
}, Open  ]],

Cell["Therefore, we have three Killing vectors", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(2 \((K[1]\ \*"\"\<\!\(\[PartialD]\_v\%A\)\>\""\  + \ 
              K[2]\ \*"\"\<\!\(\[PartialD]\_f\%A\)\>\"")\) // 
        Simplify\)\  // \ Collect[#, {pp, qq, rr}] &\)], "Input"],

Cell[BoxData[
    \(2\ "\!\(\[PartialD]\_v\%A\)"\ rr + 
      qq\ \((2\ "\!\(\[PartialD]\_f\%A\)"\ f + 
            "\!\(\[PartialD]\_v\%A\)"\ v)\) + 
      pp\ \(("\!\(\[PartialD]\_v\%A\)"\ f + 
            2\ "\!\(\[PartialD]\_f\%A\)"\ f\ v + \("\!\(\[PartialD]\_v\%A\)"\ \
v\^2\)\/2)\)\)], "Output"]
}, Open  ]],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`K\_1\  = \ \[PartialD]\_v, \ \ \ \ \ K\_2\  = \ \((f\  \
+ \ v\^2\/2)\) \[PartialD]\_v\ \(+\ 2\) 
        f\ v\ \[PartialD]\_f, \ \ \ \ \ K\_3\  = \ 
      v\ \[PartialD]\_v\ \(+\ 2\) f\ \[PartialD]\_f\)]]], "Text",
  TextAlignment->Center]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutation Relations", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(killM\  = \ {{1, 0}, {f + v\^2\/2, 2  f\ v}, {v, 2  f}}; \ 
    MatrixForm[killM]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"1", "0"},
            {\(f + v\^2\/2\), \(2\ f\ v\)},
            {"v", \(2\ f\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Commutator[killM]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"[V\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_1\),
          Subscript[ 1],
          Editable->False], "\[InvisibleSpace]", "\<\", V\"\>", 
        "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_2\),
          Subscript[ 2],
          Editable->False], "\[InvisibleSpace]", "\<\"] = \"\>", 
        "\[InvisibleSpace]", \({v, 2\ f}\)}],
      SequenceForm[ "[V", 
        Subscript[ 1], ", V", 
        Subscript[ 2], "] = ", {v, 
        Times[ 2, f]}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"[V\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_1\),
          Subscript[ 1],
          Editable->False], "\[InvisibleSpace]", "\<\", V\"\>", 
        "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_3\),
          Subscript[ 3],
          Editable->False], "\[InvisibleSpace]", "\<\"] = \"\>", 
        "\[InvisibleSpace]", \({1, 0}\)}],
      SequenceForm[ "[V", 
        Subscript[ 1], ", V", 
        Subscript[ 3], "] = ", {1, 0}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"[V\"\>", "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_2\),
          Subscript[ 2],
          Editable->False], "\[InvisibleSpace]", "\<\", V\"\>", 
        "\[InvisibleSpace]", 
        InterpretationBox[\(\[Null]\_3\),
          Subscript[ 3],
          Editable->False], "\[InvisibleSpace]", "\<\"] = \"\>", 
        "\[InvisibleSpace]", \({1\/2\ \((\(-2\)\ f - 
                v\^2)\), \(-2\)\ f\ v}\)}],
      SequenceForm[ "[V", 
        Subscript[ 2], ", V", 
        Subscript[ 3], "] = ", {
        Times[ 
          Rational[ 1, 2], 
          Plus[ 
            Times[ -2, f], 
            Times[ -1, 
              Power[ v, 2]]]], 
        Times[ -2, f, v]}],
      Editable->False]], "Print"]
}, Open  ]],

Cell["Therefore", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\([K\_1, \ K\_2]\)\  = \ 
      K\_3, \ \ \ \ \ \([K\_1, \ K\_3]\)\  = \ 
      K\_1, \ \ \ \ \ \([K\_2, \ K\_3]\)\  = \ \(-K\_2\)\)]]], "Text",
  TextAlignment->Center]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Geodesic in Potential Space", "Section"],

Cell[BoxData[
    \(\(\(killM\  = \ {{1, 0}, {f + v\^2\/2, 2  f\ v}, {v, 
            2  f}};\)\(\ \ \)\( (*\ Killing\ vectors\ *) \)\)\)], "Input"],

Cell[BoxData[
    \(Format[aCharge[idx_]] := \ 
      SequenceForm["\<a\>", \ Subscript[idx]]\)], "Input"],

Cell["Asymptotic condition:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(asympRule\)\(\ \)\(=\)\(\ \)\({f\  -> \ 1\  - \ 2  M\ x, \ 
        v\  -> \ \(-\@2\) P\ x, \ \[Lambda]\  -> \ m\ x}\)\( (*\ 
        x\  = \ 1\/r\ *) \)\)\)], "Input"],

Cell[BoxData[
    \({f \[Rule] 1 - 2\ M\ x, 
      v \[Rule] \(-\@2\)\ P\ x, \[Lambda] \[Rule] m\ x}\)], "Output"]
}, Open  ]],

Cell["From the asymptotic values of the Killing vectors", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(killM\  /. \ asympRule\)\  /. \ x\  -> \ 0\)], "Input"],

Cell[BoxData[
    \({{1, 0}, {1, 0}, {0, 2}}\)], "Output"]
}, Open  ]],

Cell["We see that", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(aChargeRule\  = \ {aCharge[2]\  -> \ aCharge[1]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
            InterpretationBox[\(\[Null]\_2\),
              Subscript[ 2],
              Editable->False]}],
          SequenceForm[ "a", 
            Subscript[ 2]],
          Editable->False], "\[Rule]", 
        InterpretationBox[
          RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
            InterpretationBox[\(\[Null]\_1\),
              Subscript[ 1],
              Editable->False]}],
          SequenceForm[ "a", 
            Subscript[ 1]],
          Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We define a tangent vector to the geodesic and three Killing \
vectors:\
\>", "Text"],

Cell[BoxData[
    \(Tdefine[geoTanV, "\<\[Xi]\>", "\<a\>"]; \ 
    Tdefine[{K1, K2, K3}, "\<a\>"]\)], "Input"],

Cell[BoxData[
    \(Cset[K1[ua], {1, 0}]; \ Cset[K2[ua], {f\  + \ v\^2\/2, 2  f\ v}]; \ 
    Cset[K3[ua], {v, 2  f}]\)], "Input"],

Cell["Three conserved charges of the geodesic are given by", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({geoTanV[ua]\ Metricg[la, lb]\ K1[ub], 
      geoTanV[ua]\ Metricg[la, lb]\ K2[ub], 
      geoTanV[ua]\ Metricg[la, lb]\ K3[ub]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "a"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["K1", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False]],
            Subsuperscript[ K1, 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "b"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], ",", 
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "a"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["K2", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False]],
            Subsuperscript[ K2, 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "b"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}], ",", 
        RowBox[{
          InterpretationBox[
            SubsuperscriptBox["\[Xi]", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False]],
            Subsuperscript[ \[Xi], 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "a"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["K3", 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], 
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                SequenceForm[ "", "b"],
                Editable->False]],
            Subsuperscript[ K3, 
              SequenceForm[ "", " "], 
              SequenceForm[ "", "b"]],
            Editable->False], " ", 
          InterpretationBox[
            SubsuperscriptBox["g", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False]],
            Subsuperscript[ g, 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], 
              SequenceForm[ 
                SequenceForm[ "", " "], " "]],
            Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["Evaluation of the conserved charges:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(geoCV\  = SumDum\  /@ \ %; \ MatrixForm[geoCV]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              RowBox[{"-", 
                FractionBox[
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\),
                        SequenceForm[ "", 1],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", 1]],
                    Editable->False], "f"]}]},
            {
              RowBox[{
                RowBox[{"-", 
                  FractionBox[
                    RowBox[{\((f + v\^2\/2)\), " ", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", " "], 
                          SequenceForm[ "", 1]],
                        Editable->False]}], "f"]}], "+", 
                FractionBox[
                  RowBox[{"v", " ", 
                    InterpretationBox[
                      SubsuperscriptBox["\[Xi]", 
                        
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                          SequenceForm[ "", " "],
                          Editable->False], 
                        InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                          SequenceForm[ "", 2],
                          Editable->False]],
                      Subsuperscript[ \[Xi], 
                        SequenceForm[ "", " "], 
                        SequenceForm[ "", 2]],
                      Editable->False]}], "f"]}]},
            {
              RowBox[{
                RowBox[{"-", 
                  FractionBox[
                    RowBox[{"v", " ", 
                      InterpretationBox[
                        SubsuperscriptBox["\[Xi]", 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]1\)\
,
                            SequenceForm[ "", 1],
                            Editable->False]],
                        Subsuperscript[ \[Xi], 
                          SequenceForm[ "", " "], 
                          SequenceForm[ "", 1]],
                        Editable->False]}], "f"]}], "+", 
                FractionBox[
                  InterpretationBox[
                    SubsuperscriptBox["\[Xi]", 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], 
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]2\),
                        SequenceForm[ "", 2],
                        Editable->False]],
                    Subsuperscript[ \[Xi], 
                      SequenceForm[ "", " "], 
                      SequenceForm[ "", 2]],
                    Editable->False], "f"]}]}
            }], "\[NoBreak]", ")"}],
      MatrixForm[ {
        Times[ -1, 
          Power[ f, -1], 
          geoTanV[ 1]], 
        Plus[ 
          Times[ -1, 
            Power[ f, -1], 
            Plus[ f, 
              Times[ 
                Rational[ 1, 2], 
                Power[ v, 2]]], 
            geoTanV[ 1]], 
          Times[ 
            Power[ f, -1], v, 
            geoTanV[ 2]]], 
        Plus[ 
          Times[ -1, 
            Power[ f, -1], v, 
            geoTanV[ 1]], 
          Times[ 
            Power[ f, -1], 
            geoTanV[ 2]]]}]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "Express the tangent vector ",
  Cell[BoxData[
      \(TraditionalForm\`\[Xi]\^A\)]],
  " using the conserved charges ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\(1, 2, 3\)\)]],
  " of the geodesic:"
}], "Text"],

Cell[BoxData[
    \(\(trM\  = \ {{\(-\(1\/f\)\), 0}, {\(-\(v\/f\)\), 1\/f}};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(someV\  = \ Inverse[trM]\  . \ {aCharge[1], aCharge[3]}\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{\(-f\), " ", 
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                InterpretationBox[\(\[Null]\_1\),
                  Subscript[ 1],
                  Editable->False]}],
              SequenceForm[ "a", 
                Subscript[ 1]],
              Editable->False], ")"}]}], ",", 
        RowBox[{
          RowBox[{\(-f\), " ", "v", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Null]\_1\),
                    Subscript[ 1],
                    Editable->False]}],
                SequenceForm[ "a", 
                  Subscript[ 1]],
                Editable->False], ")"}]}], "+", 
          RowBox[{"f", " ", 
            RowBox[{"(", 
              InterpretationBox[
                RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Null]\_3\),
                    Subscript[ 3],
                    Editable->False]}],
                SequenceForm[ "a", 
                  Subscript[ 3]],
                Editable->False], ")"}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(Cset[geoTanV[ua], someV]\)], "Input"],

Cell["From the second equation of geoCV", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq2\  = \ 
      aCharge[2]\  - \ geoCV[\([2]\)]\  /. \ aChargeRule\  // \ 
        Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", \((2 - 2\ f + v\^2)\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
              InterpretationBox[\(\[Null]\_1\),
                Subscript[ 1],
                Editable->False]}],
            SequenceForm[ "a", 
              Subscript[ 1]],
            Editable->False], ")"}]}], "-", 
      RowBox[{"v", " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
              InterpretationBox[\(\[Null]\_3\),
                Subscript[ 3],
                Editable->False]}],
            SequenceForm[ "a", 
              Subscript[ 3]],
            Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell["we find", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol2Rule\  = \ 
      Flatten[Solve[eq2\  == \ 0, \ f]\  // \ Simplify]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{"f", "\[Rule]", 
        RowBox[{"1", "+", \(v\^2\/2\), "-", 
          FractionBox[
            RowBox[{"v", " ", 
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                    InterpretationBox[\(\[Null]\_3\),
                      Subscript[ 3],
                      Editable->False]}],
                  SequenceForm[ "a", 
                    Subscript[ 3]],
                  Editable->False], ")"}]}], 
            InterpretationBox[
              RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                InterpretationBox[\(\[Null]\_1\),
                  Subscript[ 1],
                  Editable->False]}],
              SequenceForm[ "a", 
                Subscript[ 1]],
              Editable->False]]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
  "From sol2Rule and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Xi]\^v\)]],
  " we can obtain ",
  Cell[BoxData[
      \(TraditionalForm\`\[Xi]\^f\)]],
  ". Thus, we finally have the following solution of the geodesic:"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[
      Flatten[{aChargeRule, \ \[Xi]\^\[Nu]\  == \ geoTanV[1], \ 
          sol2Rule}]]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              RowBox[{
                InterpretationBox[
                  RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                    InterpretationBox[\(\[Null]\_2\),
                      Subscript[ 2],
                      Editable->False]}],
                  SequenceForm[ "a", 
                    Subscript[ 2]],
                  Editable->False], "\[Rule]", 
                InterpretationBox[
                  RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                    InterpretationBox[\(\[Null]\_1\),
                      Subscript[ 1],
                      Editable->False]}],
                  SequenceForm[ "a", 
                    Subscript[ 1]],
                  Editable->False]}]},
            {
              RowBox[{\(\[Xi]\^\[Nu]\), "==", 
                RowBox[{\(-f\), " ", 
                  RowBox[{"(", 
                    InterpretationBox[
                      RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                        InterpretationBox[\(\[Null]\_1\),
                          Subscript[ 1],
                          Editable->False]}],
                      SequenceForm[ "a", 
                        Subscript[ 1]],
                      Editable->False], ")"}]}]}]},
            {
              RowBox[{"f", "\[Rule]", 
                RowBox[{"1", "+", \(v\^2\/2\), "-", 
                  FractionBox[
                    RowBox[{"v", " ", 
                      RowBox[{"(", 
                        InterpretationBox[
                          RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                            InterpretationBox[\(\[Null]\_3\),
                              Subscript[ 3],
                              Editable->False]}],
                          SequenceForm[ "a", 
                            Subscript[ 3]],
                          Editable->False], ")"}]}], 
                    InterpretationBox[
                      RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                        InterpretationBox[\(\[Null]\_1\),
                          Subscript[ 1],
                          Editable->False]}],
                      SequenceForm[ "a", 
                        Subscript[ 1]],
                      Editable->False]]}]}]}
            }], "\[NoBreak]", ")"}],
      MatrixForm[ {aCharge[ 2] -> aCharge[ 1], 
        Equal[ 
          Power[ \[Xi], \[Nu]], 
          Times[ -1, f, 
            aCharge[ 1]]], f -> Plus[ 1, 
          Times[ 
            Rational[ 1, 2], 
            Power[ v, 2]], 
          Times[ -1, v, 
            Power[ 
              aCharge[ 1], -1], 
            aCharge[ 3]]]}]]], "Output"]
}, Open  ]],

Cell["Signature of the geodesic:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Metricg[la, lb]\ geoTanV[ua]\ geoTanV[ub]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["\[Xi]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
            SequenceForm[ "", "a"],
            Editable->False]],
        Subsuperscript[ \[Xi], 
          SequenceForm[ "", " "], 
          SequenceForm[ "", "a"]],
        Editable->False], " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Xi]", 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
            SequenceForm[ "", " "],
            Editable->False], 
          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
            SequenceForm[ "", "b"],
            Editable->False]],
        Subsuperscript[ \[Xi], 
          SequenceForm[ "", " "], 
          SequenceForm[ "", "b"]],
        Editable->False], " ", 
      InterpretationBox[
        SubsuperscriptBox["g", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                SequenceForm[ "", "a"],
                Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                SequenceForm[ "", " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ "", " "], " "],
            Editable->False]],
        Subsuperscript[ g, 
          SequenceForm[ 
            SequenceForm[ "", "a"], "b"], 
          SequenceForm[ 
            SequenceForm[ "", " "], " "]],
        Editable->False]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Epsilon]\  = \ SumDum[%]\  // \ Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-f\), " ", 
        SuperscriptBox[
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                InterpretationBox[\(\[Null]\_1\),
                  Subscript[ 1],
                  Editable->False]}],
              SequenceForm[ "a", 
                Subscript[ 1]],
              Editable->False], ")"}], "2"]}], "+", 
      RowBox[{\(1\/2\), " ", 
        SuperscriptBox[
          RowBox[{"(", 
            RowBox[{
              RowBox[{\(-v\), " ", 
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                      InterpretationBox[\(\[Null]\_1\),
                        Subscript[ 1],
                        Editable->False]}],
                    SequenceForm[ "a", 
                      Subscript[ 1]],
                    Editable->False], ")"}]}], "+", 
              InterpretationBox[
                RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                  InterpretationBox[\(\[Null]\_3\),
                    Subscript[ 3],
                    Editable->False]}],
                SequenceForm[ "a", 
                  Subscript[ 3]],
                Editable->False]}], ")"}], "2"]}]}]], "Output"]
}, Open  ]],

Cell["\<\
Substituting the solution into the signature of the geodesic we \
find\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(%\  /. \ sol2Rule // Expand\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"-", 
        SuperscriptBox[
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                InterpretationBox[\(\[Null]\_1\),
                  Subscript[ 1],
                  Editable->False]}],
              SequenceForm[ "a", 
                Subscript[ 1]],
              Editable->False], ")"}], "2"]}], "+", 
      RowBox[{\(1\/2\), " ", 
        SuperscriptBox[
          RowBox[{"(", 
            InterpretationBox[
              RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                InterpretationBox[\(\[Null]\_3\),
                  Subscript[ 3],
                  Editable->False]}],
              SequenceForm[ "a", 
                Subscript[ 3]],
              Editable->False], ")"}], "2"]}]}]], "Output"]
}, Open  ]],

Cell["which is a manifestly constant.", "Text"],

Cell["At the event horizon the signature takes the following form:", "Text"],

Cell[BoxData[
    \(\(\(horizonRule\  = \ {f\  -> \ 0, \ v\  -> \ v\_h};\)\(\ \ \)\( (*\ 
      at\ the\ event\ horizon\ *) \)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Epsilon]\  /. \ horizonRule\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      SuperscriptBox[
        RowBox[{"(", 
          RowBox[{
            InterpretationBox[
              RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                InterpretationBox[\(\[Null]\_3\),
                  Subscript[ 3],
                  Editable->False]}],
              SequenceForm[ "a", 
                Subscript[ 3]],
              Editable->False], "-", 
            RowBox[{
              RowBox[{"(", 
                InterpretationBox[
                  RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                    InterpretationBox[\(\[Null]\_1\),
                      Subscript[ 1],
                      Editable->False]}],
                  SequenceForm[ "a", 
                    Subscript[ 1]],
                  Editable->False], ")"}], " ", \(v\_h\)}]}], ")"}], 
        "2"]}]], "Output"]
}, Open  ]],

Cell["From the asymptotic condition", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Series[eq2\  /. \ asympRule, \ {x, 0, \ 1}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        RowBox[{
          RowBox[{"(", 
            RowBox[{
              RowBox[{"2", " ", "M", " ", 
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                      InterpretationBox[\(\[Null]\_1\),
                        Subscript[ 1],
                        Editable->False]}],
                    SequenceForm[ "a", 
                      Subscript[ 1]],
                    Editable->False], ")"}]}], "+", 
              RowBox[{\(\@2\), " ", "P", " ", 
                RowBox[{"(", 
                  InterpretationBox[
                    RowBox[{"\<\"a\"\>", "\[InvisibleSpace]", 
                      InterpretationBox[\(\[Null]\_3\),
                        Subscript[ 3],
                        Editable->False]}],
                    SequenceForm[ "a", 
                      Subscript[ 3]],
                    Editable->False], ")"}]}]}], ")"}], " ", "x"}], "+", 
        InterpretationBox[\(O[x]\^2\),
          SeriesData[ x, 0, {}, 1, 2, 1]]}],
      SeriesData[ x, 0, {
        Plus[ 
          Times[ 2, M, 
            aCharge[ 1]], 
          Times[ 
            Power[ 2, 
              Rational[ 1, 2]], P, 
            aCharge[ 3]]]}, 1, 2, 1]]], "Output"]
}, Open  ]],

Cell[TextData[{
  "we find ",
  Cell[BoxData[
      \(TraditionalForm\`a\_3\/a\_1\  = \ \(-\@2\) M\/P\)]],
  "."
}], "Text"],

Cell["In summay, the geodesic of the potential space is given by", "Text"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\[Epsilon]\/a\_1\%2\  = \ \(\(-1\)\  + \ 
          M\^2\/P\^2\  = \ \((v\_h\/\@2\  + \ M\/P)\)\^2\)\)]]], "Text",
  TextAlignment->Center],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`f\  = \ 
      1\  - \ \(\@2\) \(M\/P\) v\  + \ 
        v\^2\/2, \ \ \ \ \ \(d\[InvisibleSpace]v\)\/\(\(a\_1\) d\
\[InvisibleSpace]\[Lambda]\)\  = \ \(-f\)\)]]], "Text",
  TextAlignment->Center],

Cell[TextData[{
  "Notice that if the horizon exists at which ",
  Cell[BoxData[
      \(TraditionalForm\`f\  \[Rule] 0\)]],
  ", the geodesic is always time-like (",
  Cell[BoxData[
      \(TraditionalForm\`\[Epsilon]\  > \ 0\)]],
  ") and satisfies the Bogomolnyi inequality ",
  Cell[BoxData[
      \(TraditionalForm\`M\^2\  \[GreaterEqual] \ P\^2\)]],
  "."
}], "Text"]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{717, 674},
WindowMargins->{{22, Automatic}, {Automatic, 3}},
PrintingCopies->1,
PrintingPageRange->{2, 100},
PrintingOptions->{"PrintingMargins"->{{54, 54}, {36, 36}},
"PaperSize"->{597.563, 842.375},
"PaperOrientation"->"Portrait",
"PrintCellBrackets"->False,
"PrintRegistrationMarks"->False,
"PrintMultipleHorizontalPages"->False,
"Magnification"->1},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 76, 1, 43, "Title"],
Cell[1796, 52, 185, 5, 66, "Text"],
Cell[1984, 59, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[2061, 64, 25, 0, 78, "Section"],

Cell[CellGroupData[{
Cell[2111, 68, 31, 0, 54, "Subsubsection"],
Cell[2145, 70, 228, 4, 43, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2410, 79, 38, 0, 32, "Subsubsection"],
Cell[2451, 81, 265, 4, 45, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2753, 90, 41, 0, 32, "Subsubsection"],
Cell[2797, 92, 292, 6, 51, "Text"],
Cell[3092, 100, 21, 0, 41, "Text"],
Cell[3116, 102, 235, 4, 45, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3388, 111, 60, 0, 32, "Subsubsection"],
Cell[3451, 113, 353, 6, 45, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3841, 124, 43, 0, 32, "Subsubsection"],
Cell[3887, 126, 181, 6, 41, "Text"],
Cell[4071, 134, 192, 3, 37, "Text"],
Cell[4266, 139, 129, 2, 35, "Text"],
Cell[4398, 143, 21, 0, 41, "Text"],
Cell[4422, 145, 261, 5, 48, "Text"],
Cell[4686, 152, 302, 9, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[5037, 167, 56, 0, 50, "Section"],
Cell[5096, 169, 106, 2, 72, "Input"],

Cell[CellGroupData[{
Cell[5227, 175, 47, 1, 56, "Input"],
Cell[5277, 178, 293, 6, 29, "Print"],
Cell[5573, 186, 3962, 95, 44, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9572, 286, 65, 1, 56, "Input"],
Cell[9640, 289, 2638, 66, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[12315, 360, 62, 1, 56, "Input"],
Cell[12380, 363, 34731, 843, 78, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[47148, 1211, 174, 3, 79, "Input"],
Cell[47325, 1216, 220, 3, 122, "Output"]
}, Open  ]],
Cell[47560, 1222, 217, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[47802, 1231, 40, 0, 54, "Subsubsection"],
Cell[47845, 1233, 52, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[47922, 1238, 132, 3, 79, "Input"],
Cell[48057, 1243, 3581, 89, 54, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[51675, 1337, 76, 1, 56, "Input"],
Cell[51754, 1340, 31317, 777, 101, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[83108, 2122, 110, 2, 58, "Input"],
Cell[83221, 2126, 1298, 33, 55, "Output"]
}, Open  ]],
Cell[84534, 2162, 101, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[84660, 2170, 65, 1, 56, "Input"],
Cell[84728, 2173, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[84800, 2179, 107, 2, 56, "Input"],
Cell[84910, 2183, 992, 25, 69, "Output"]
}, Open  ]],
Cell[85917, 2211, 124, 4, 44, "Text"],

Cell[CellGroupData[{
Cell[86066, 2219, 86, 1, 72, "Input"],
Cell[86155, 2222, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[86227, 2228, 107, 2, 56, "Input"],
Cell[86337, 2232, 353, 9, 53, "Output"]
}, Open  ]],
Cell[86705, 2244, 174, 5, 44, "Text"],

Cell[CellGroupData[{
Cell[86904, 2253, 131, 2, 92, "Input"],
Cell[87038, 2257, 35, 1, 53, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[87110, 2263, 47, 1, 56, "Input"],
Cell[87160, 2266, 114, 2, 72, "Output"]
}, Open  ]],
Cell[87289, 2271, 56, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[87370, 2275, 204, 3, 81, "Input"],
Cell[87577, 2280, 302, 6, 75, "Output"]
}, Open  ]],
Cell[87894, 2289, 291, 5, 47, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[88222, 2299, 46, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[88293, 2303, 114, 2, 99, "Input"],
Cell[88410, 2307, 238, 7, 121, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[88685, 2319, 50, 1, 56, "Input"],
Cell[88738, 2322, 579, 15, 28, "Print"],
Cell[89320, 2339, 556, 14, 28, "Print"],
Cell[89879, 2355, 777, 22, 44, "Print"]
}, Open  ]],
Cell[90671, 2380, 25, 0, 41, "Text"],
Cell[90699, 2382, 220, 4, 35, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[90968, 2392, 46, 0, 50, "Section"],
Cell[91017, 2394, 148, 2, 99, "Input"],
Cell[91168, 2398, 106, 2, 56, "Input"],
Cell[91277, 2402, 37, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[91339, 2406, 192, 3, 92, "Input"],
Cell[91534, 2411, 114, 2, 57, "Output"]
}, Open  ]],
Cell[91663, 2416, 65, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[91753, 2420, 77, 1, 56, "Input"],
Cell[91833, 2423, 58, 1, 53, "Output"]
}, Open  ]],
Cell[91906, 2427, 27, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[91958, 2431, 81, 1, 56, "Input"],
Cell[92042, 2434, 628, 18, 53, "Output"]
}, Open  ]],
Cell[92685, 2455, 95, 3, 41, "Text"],
Cell[92783, 2460, 110, 2, 79, "Input"],
Cell[92896, 2464, 129, 2, 99, "Input"],
Cell[93028, 2468, 68, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[93121, 2472, 161, 3, 102, "Input"],
Cell[93285, 2477, 6192, 149, 55, "Output"]
}, Open  ]],
Cell[99492, 2629, 52, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[99569, 2633, 79, 1, 56, "Input"],
Cell[99651, 2636, 4411, 111, 175, "Output"]
}, Open  ]],
Cell[104077, 2750, 233, 8, 41, "Text"],
Cell[104313, 2760, 91, 1, 72, "Input"],

Cell[CellGroupData[{
Cell[104429, 2765, 88, 1, 56, "Input"],
Cell[104520, 2768, 1254, 33, 53, "Output"]
}, Open  ]],
Cell[105789, 2804, 57, 1, 56, "Input"],
Cell[105849, 2807, 49, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[105923, 2811, 125, 3, 56, "Input"],
Cell[106051, 2816, 772, 21, 69, "Output"]
}, Open  ]],
Cell[106838, 2840, 23, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[106886, 2844, 104, 2, 56, "Input"],
Cell[106993, 2848, 868, 22, 75, "Output"]
}, Open  ]],
Cell[107876, 2873, 250, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[108151, 2885, 131, 3, 79, "Input"],
Cell[108285, 2890, 2720, 67, 127, "Output"]
}, Open  ]],
Cell[111020, 2960, 42, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[111087, 2964, 74, 1, 56, "Input"],
Cell[111164, 2967, 1872, 49, 55, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[113073, 3021, 75, 1, 56, "Input"],
Cell[113151, 3024, 1318, 34, 69, "Output"]
}, Open  ]],
Cell[114484, 3061, 94, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[114603, 3068, 60, 1, 56, "Input"],
Cell[114666, 3071, 835, 23, 69, "Output"]
}, Open  ]],
Cell[115516, 3097, 47, 0, 41, "Text"],
Cell[115566, 3099, 76, 0, 41, "Text"],
Cell[115645, 3101, 141, 2, 79, "Input"],

Cell[CellGroupData[{
Cell[115811, 3107, 62, 1, 56, "Input"],
Cell[115876, 3110, 866, 23, 69, "Output"]
}, Open  ]],
Cell[116757, 3136, 45, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[116827, 3140, 76, 1, 56, "Input"],
Cell[116906, 3143, 1327, 35, 57, "Output"]
}, Open  ]],
Cell[118248, 3181, 124, 5, 48, "Text"],
Cell[118375, 3188, 74, 0, 41, "Text"],
Cell[118452, 3190, 190, 3, 54, "Text"],
Cell[118645, 3195, 246, 5, 51, "Text"],
Cell[118894, 3202, 373, 11, 66, "Text"]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

