(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     44155,       1130]*)
(*NotebookOutlinePosition[     44923,       1157]*)
(*  CellTagsIndexPosition[     44879,       1153]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Curvature of SO(3) Group Space", "Title",
  FontSize->24],

Cell[BoxData[
    \(<< \ dhPark`Tensor`\)], "Input"],

Cell[BoxData[
    \(Off[CoordinateBasisFlag]\)], "Input"],

Cell[CellGroupData[{

Cell["Define totally anti-symmetric tensor", "Subsubsection"],

Cell[BoxData[
    \(Tdefine[\[Epsilon], "\<abc\>"]\)], "Input"],

Cell[BoxData[
    \(SetAntisymmetric[\[Epsilon]]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Rules", "Subsubsection"],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\([\[Xi]\_a, \ \[Xi]\_b]\)\  = \ \(\[Epsilon]\_\(a\
\[InvisibleSpace]b\)\^c\) \[Xi]\_c\)]]], "Text",
  TextAlignment->Center],

Cell[CellGroupData[{

Cell[BoxData[
    \(stRule = 
      Trule[Structuref[la_, lb_, lc_], \[Epsilon][la, lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        SubsuperscriptBox["f", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]la_\),
                    SequenceForm[ "", 
                      Pattern[ la, 
                        Blank[ ]]],
                    Editable->False], "\[InvisibleSpace]", "lb_"}],
                SequenceForm[ 
                  SequenceForm[ "", 
                    Pattern[ la, 
                      Blank[ ]]], 
                  Pattern[ lb, 
                    Blank[ ]]],
                Editable->False], "\[InvisibleSpace]", "lc_"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", 
                  Pattern[ la, 
                    Blank[ ]]], 
                Pattern[ lb, 
                  Blank[ ]]], 
              Pattern[ lc, 
                Blank[ ]]],
            Editable->False]],
        Subsuperscript[ f, 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], " "], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", 
                Pattern[ la, 
                  Blank[ ]]], 
              Pattern[ lb, 
                Blank[ ]]], 
            Pattern[ lc, 
              Blank[ ]]]],
        Editable->False], "\[RuleDelayed]", 
      RowBox[{
        RowBox[{"DumFresh", "[", 
          InterpretationBox[
            SubsuperscriptBox["\[Epsilon]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], "c"],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], " "],
                Editable->False]],
            Subsuperscript[ \[Epsilon], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "c"], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "]],
            Editable->False], "]"}], "/;", "True"}]}]], "Output"]
}, Open  ]],

Cell[TextData[Cell[BoxData[
    \(TraditionalForm\`\(D\_a\) 
        g\_\(b\[InvisibleSpace]c\)\  = \ \(\(D\_a\) \[CapitalGamma]\_\(b\
\[InvisibleSpace]c\)\^d\  = \ 0\)\)]]], "Text",
  TextAlignment->Center],

Cell[CellGroupData[{

Cell[BoxData[
    \(bdRule = Trule[BD[la_, expr_], 0]\)], "Input"],

Cell[BoxData[
    RowBox[{
      InterpretationBox[
        RowBox[{
          InterpretationBox[\("D"\^la_\),
            Superscript[ "D", 
              Pattern[ la, 
                Blank[ ]]],
            Editable->False], "\[InvisibleSpace]", "expr_"}],
        SequenceForm[ 
          Superscript[ "D", 
            Pattern[ la, 
              Blank[ ]]], 
          Pattern[ expr, 
            Blank[ ]]],
        Editable->False], 
      "\[RuleDelayed]", \(DumFresh[0] /; True\)}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Affine connection", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(AffineToMetric[AffineG[la, lb, lc]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                        SequenceForm[ "", "b"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              Metricg[ lb, lc]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              Metricg[ la, lc]],
            Editable->False], ")"}]}], "-", 
      RowBox[{\(1\/2\), " ", 
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"c"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["g", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "b"],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False]],
                Subsuperscript[ g, 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"], 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "c"], 
              Metricg[ la, lb]],
            Editable->False], ")"}]}], "+", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], "c"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], "c"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], "b"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], "b"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"a\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], "a"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], "a"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[%, {stRule, bdRule}]\)], "Input"],

Cell[BoxData[
    RowBox[{\(1\/2\), " ", 
      InterpretationBox[
        SubsuperscriptBox["\[Epsilon]", 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                    SequenceForm[ "", "a"],
                    Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"],
                Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], "c"],
            Editable->False], 
          InterpretationBox[
            RowBox[{
              InterpretationBox[
                RowBox[{
                  InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                    SequenceForm[ "", " "],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ "", " "], " "],
                Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "],
            Editable->False]],
        Subsuperscript[ \[Epsilon], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", "a"], "b"], "c"], 
          SequenceForm[ 
            SequenceForm[ 
              SequenceForm[ "", " "], " "], " "]],
        Editable->False]}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Curvature", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RiemannToAffine[RiemannR[la, lb, lc, ld]]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"-", 
          InterpretationBox[
            SubsuperscriptBox["\[CapitalGamma]", 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                        SequenceForm[ "", "a"],
                        Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"],
                    Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"], " "],
                Editable->False], 
              InterpretationBox[
                RowBox[{
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                        SequenceForm[ "", " "],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "],
                    Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                SequenceForm[ 
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "], "p"],
                Editable->False]],
            Subsuperscript[ \[CapitalGamma], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "], 
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"]],
            Editable->False]}], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "p"], "d"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "p"], "d"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "+", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "p"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "p"], "d"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "p"], "d"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"a"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[CapitalGamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"b"\),
                            SequenceForm[ "", "b"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "b"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "b"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[CapitalGamma], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "b"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "a"], 
              AffineG[ lb, lc, "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{
        RowBox[{"(", 
          InterpretationBox[
            RowBox[{\("D"\_"b"\), "\[InvisibleSpace]", 
              InterpretationBox[
                SubsuperscriptBox["\[CapitalGamma]", 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]\
"a"\),
                            SequenceForm[ "", "a"],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\"c\"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", "a"], "c"],
                        Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", "a"], "c"], " "],
                    Editable->False], 
                  InterpretationBox[
                    RowBox[{
                      InterpretationBox[
                        RowBox[{
                          
                          InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" \
"\),
                            SequenceForm[ "", " "],
                            Editable->False], 
                          "\[InvisibleSpace]", "\<\" \"\>"}],
                        SequenceForm[ 
                          SequenceForm[ "", " "], " "],
                        Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
                    SequenceForm[ 
                      SequenceForm[ 
                        SequenceForm[ "", " "], " "], "p"],
                    Editable->False]],
                Subsuperscript[ \[CapitalGamma], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", "a"], "c"], " "], 
                  SequenceForm[ 
                    SequenceForm[ 
                      SequenceForm[ "", " "], " "], "p"]],
                Editable->False]}],
            SequenceForm[ 
              Subscript[ "D", "b"], 
              AffineG[ la, lc, "u$1"]],
            Editable->False], ")"}], " ", 
        InterpretationBox[
          SubsuperscriptBox["g", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"d"\),
                  SequenceForm[ "", "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ "", "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                  SequenceForm[ "", " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ "", " "], " "],
              Editable->False]],
          Subsuperscript[ g, 
            SequenceForm[ 
              SequenceForm[ "", "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ "", " "], " "]],
          Editable->False]}], "-", 
      RowBox[{
        InterpretationBox[
          SubsuperscriptBox["\[CapitalGamma]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"p"\),
                      SequenceForm[ "", "p"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "p"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "p"], "c"], "d"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[CapitalGamma], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "p"], "c"], "d"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["f", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ f, 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(curvR = AffineToMetric[%];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ApplyRules[curvR, {stRule, bdRule}]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(-\(1\/4\)\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Epsilon]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[Epsilon], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Epsilon]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Epsilon], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False]}], "+", 
      RowBox[{\(1\/4\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Epsilon]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"c\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "c"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "c"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Epsilon], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "c"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Epsilon]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"b"\),
                      SequenceForm[ "", "b"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "b"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "b"], "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[Epsilon], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "b"], "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}], "-", 
      RowBox[{\(1\/2\), " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Epsilon]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"a"\),
                      SequenceForm[ "", "a"],
                      Editable->False], "\[InvisibleSpace]", "\<\"b\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "a"], "b"],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "a"], "b"], " "],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], "p"],
              Editable->False]],
          Subsuperscript[ \[Epsilon], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "a"], "b"], " "], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], "p"]],
          Editable->False], " ", 
        InterpretationBox[
          SubsuperscriptBox["\[Epsilon]", 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]"c"\),
                      SequenceForm[ "", "c"],
                      Editable->False], "\[InvisibleSpace]", "\<\"d\"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", "c"], "d"],
                  Editable->False], "\[InvisibleSpace]", "\<\"p\"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", "c"], "d"], "p"],
              Editable->False], 
            InterpretationBox[
              RowBox[{
                InterpretationBox[
                  RowBox[{
                    InterpretationBox[\(\*"\<\"\"\>"\[InvisibleSpace]" "\),
                      SequenceForm[ "", " "],
                      Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
                  SequenceForm[ 
                    SequenceForm[ "", " "], " "],
                  Editable->False], "\[InvisibleSpace]", "\<\" \"\>"}],
              SequenceForm[ 
                SequenceForm[ 
                  SequenceForm[ "", " "], " "], " "],
              Editable->False]],
          Subsuperscript[ \[Epsilon], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", "c"], "d"], "p"], 
            SequenceForm[ 
              SequenceForm[ 
                SequenceForm[ "", " "], " "], " "]],
          Editable->False]}]}]], "Output"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 723}},
ScreenStyleEnvironment->"Presentation",
WindowToolbars->{},
WindowSize->{567, 660},
WindowMargins->{{14, Automatic}, {Automatic, 6}},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 63, 1, 43, "Title"],
Cell[1783, 52, 52, 1, 56, "Input"],
Cell[1838, 55, 57, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[1920, 60, 61, 0, 54, "Subsubsection"],
Cell[1984, 62, 63, 1, 56, "Input"],
Cell[2050, 65, 61, 1, 56, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2148, 71, 30, 0, 32, "Subsubsection"],
Cell[2181, 73, 176, 3, 35, "Text"],

Cell[CellGroupData[{
Cell[2382, 80, 105, 2, 79, "Input"],
Cell[2490, 84, 3781, 97, 55, "Output"]
}, Open  ]],
Cell[6286, 184, 207, 4, 35, "Text"],

Cell[CellGroupData[{
Cell[6518, 192, 66, 1, 56, "Input"],
Cell[6587, 195, 503, 16, 54, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[7139, 217, 42, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[7206, 221, 68, 1, 56, "Input"],
Cell[7277, 224, 9006, 214, 110, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16320, 443, 64, 1, 56, "Input"],
Cell[16387, 446, 1565, 39, 69, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[18001, 491, 34, 0, 32, "Subsubsection"],

Cell[CellGroupData[{
Cell[18060, 495, 74, 1, 56, "Input"],
Cell[18137, 498, 16124, 385, 82, "Output"]
}, Open  ]],
Cell[34276, 886, 63, 1, 56, "Input"],

Cell[CellGroupData[{
Cell[34364, 891, 68, 1, 56, "Input"],
Cell[34435, 894, 9692, 232, 69, "Output"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

